{-----------------------------------------------------------------------------
 Unit Name: DiffEqSolver
 Project Name : Parametricas4
 Author:    Angel Montesinos Amilibia
 Dates: 2003-2009
 Purpose: Implements a Runge-Kutta embedded method for solving the differential
    equations that give curves of geometric significance.

-----------------------------------------------------------------------------}
{

Parametricas4, workshop for the geometry of surfaces in R^4.


MIT License

Copyright (c) [2003 - 2018] [Angel Montesinos-Amilibia], email montesin@uv.es

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


}

unit DiffEqSolver;

interface

uses
  Windows, Graphics,
  FParametrizationR2R4, VectorsAndUtil;

procedure DrawCurvesByPoint(can: TCanvas; parts, curveType: Integer; toPrint:
  Boolean;
  param: VectorsAndUtil.Vector2; foliations: Byte);

implementation

uses
  Forms, SysUtils, Controls, Math,
  FEvaluatorBytes_64_XE2, Fparam4, ResourceStrings, HigherDegreeLines;
const
  tol = 1.0E-4;
var
  angleTol: Double;
  rk1, rk2, rk3: VectorsAndUtil.Vector2;

function CalcElem(cl: Integer; var error, prevAngle: Double; u: VectorsAndUtil.Vector2;
  stepEstim
  : Double): boolean;
{computes the intermediate steps 'rk1, rk2, rk3' that are combined in the
 Runge-Kutta scheme to get the next point. 'prevAngle' serves to choose
 between the several directions that may satisfy the condition the nearest
 to that of the previous computed point}
var
  y, rk: VectorsAndUtil.Vector2;
  preDir, aux: Double;
  goOn: byte;
  i: integer;
begin
  rk:= rk3;
  preDir:= prevAngle;
  goOn:= MapR2R4.Directions(u, cl, rk1, prevAngle);
  if goOn > 0 then begin
    for i:= 1 to 2 do
      y[i]:= u[i] + stepEstim * rk1[i];
    goOn:= MapR2R4.Directions(y, cl, rk2, prevAngle);
    if goOn > 0 then begin
      for i:= 1 to 2 do
        y[i]:= u[i] + 0.25 * stepEstim * (rk1[i] + rk2[i]);
      goOn:= MapR2R4.Directions(y, cl, rk3, prevAngle);
      if goOn > 0 then
        error:= 0;
      begin
        for i:= 1 to 2 do
          error:= error + abs(stepEstim * (2.0 * rk3[i] - rk1[i] -
            rk2[i]));
        error:= (error + tol * 0.001) / param4.span2D;
        aux:= abs(prevAngle - preDir);
        error:= error + Min(aux, abs(2 * pi - aux)) * angleTol; {we don't permit
        that between a point and the next, the curve changes direction
        more than pi/90, that is more than two degrees}

        goOn:= ord(error < tol);
      end;
    end;
  end;
  CalcElem:= goOn > 0;
end;

function GetInitialStep(cl: Integer; alpha: Double; u: VectorsAndUtil.Vector2): Double;
const
  tol: Double = 1.0E-4;
var
  err, coef: Double;
  v: VectorsAndUtil.Vector2;
begin
  result:= kTheError;
  if (MapR2R4.Directions(u, cl, v, alpha) > 0) then begin
    result:= param4.radius2D / hypot(v[1], v[2]);
    CalcElem(cl, err, alpha, u, result);
    coef:= 0.8 * exp(ln(tol / err) / 3.0);
    result:= result * Max(0.1, Min(5, coef));
  end;
end;

function RunKut(var step, jump, guessStep, angle: Double; var x: VectorsAndUtil.Vector2;
  curvesOpt: Integer): boolean;
{'curvesOpt' gives the class of curves that the ODE solver must find.
 'angle' is the angle that the initial direction forms with the first axis.
 Given 'x' and an estimate 'guessStep' of the first step, RunKut looks for a
 a convenient step, 'step', and for  t+step an approximation 'x' to the solution,
 with local relative error less than 'tol'. Gives also the values of
 'rk1,rk2,rk3', that will be used for interpolating new points between  't'
 and 't+step'. In 'guessStep' it is returned a guess of the next step. The
 displacement of 'x' in the current step is returned in 'jump'. 'RunKut'
 returns True if all goes well in the computation of the directions that define
 the differential equation}
var
  err, coef: Double;
  shift: VectorsAndUtil.Vector2;
  keepGoing: boolean;
  i: integer;
begin
  repeat
    keepGoing:= CalcElem(curvesOpt, err, angle, x, guessStep);
    if keepGoing then begin
      coef:= 0.8 * exp(ln(tol / err) / 3.0);
      if coef < 0.5 then
        coef:= 0.5;
      if 2.0 < coef then
        coef:= 2.0;
      step:= guessStep;
      guessStep:= coef * step;
    end
  until (not keepGoing) or (err < tol);
  if keepGoing then begin
    for i:= 1 to 2 do begin
      shift[i]:= step * (rk1[i] + rk2[i] + 4.0 * rk3[i]) / 6.0;
      x[i]:= x[i] + shift[i]
    end;
    jump:= abs(shift[1]) + abs(shift[2]);
  end;
  RunKut:= keepGoing;
end;

function Interpolate(x, rk1, rk2, rk3: VectorsAndUtil.Vector2; h, theta: Double): VectorsAndUtil.Vector2;
{compute an interpolation for the solution at 't= prevT + theta*step', where
 'prevT' is the independent variable value at the last calculated point, 'prevT+step'
 is that at the next point and '0 <= theta <= 1'}

const
  threeHalves = 1.5;
  twoThirds = 0.666666666666666666;
  fourThirds = 1.33333333333333333;
var
  theta2, b1, b2, b3: Double;
  i: integer;
begin
  theta2:= sqr(theta);
  b1:= theta + (-threeHalves + twoThirds * theta) * theta2;
  b2:= (-0.5 + twoThirds * theta) * theta2;
  b3:= (2.0 - fourThirds * theta) * theta2;
  for i:= 1 to 2 do
    result[i]:= x[i] + h * (b1 * rk1[i] + b2 * rk2[i] + b3 * rk3[i]);
end;

procedure DrawCurvesByPoint(can: TCanvas; parts, curveType: Integer; toPrint:
  Boolean;
  param: VectorsAndUtil.Vector2; foliations: Byte);
{Computes, stores and draws all the curves of class 'curveType' starting at 'param'}
var
  notTooFar, firstStepsTaken, overtaken: boolean;
  deriv, x, y, xPrev: VectorsAndUtil.Vector2;
  aux, bb, bc, cc, inside, initialAngle, angle, step, guessStep, shift, theta,
    cut: Double;
  chunks, i, axis: integer;
  theSegment: Segment2;
  vec1, vec2: VectorsAndUtil.Vector2;
  counter: Longint;
  pen: TPen;
  goodPt: Byte;
begin
  angleTol:= 90 * tol / pi;
    {tolerance in direction change between a point and the next.
  In its application, this means that that change must be less than pi/90}
  counter:= 1;
  pen:= can.Pen;
  if toPrint then
    can.Pen.Width:= 2
  else
    can.Pen.Width:= 1;
  can.Pen.Mode:= pmCopy;
  if curveType in [kExtrNormDevLines, kExtrLatDevLines, kStrongPrincipalLines] then begin
    DrawHigherDegreeLines(curveType, can, param, foliations);
    Exit;
  end;
  cut:= 0.05 * param4.radius2D;
    {if the jump of a step is less than 'cut', we shall
  stor the computation in the understanding that probably we are approximating
  to a singular point}

  angle:= kTheError;
  try
    goodPt:= MapR2R4.Directions(param, curveType, deriv, angle);
    if goodPt > 0 then begin
      for axis:= 1 to parts do begin {draw all the curves by the initial point}
        if (curveType = kAsymptoticLines) and (axis = 3) then
          angle:= -kTheError; {a special key to be used in this case}
        overtaken:= FALSE;
        firstStepsTaken:= FALSE;
        goodPt:= MapR2R4.Directions(param, curveType, rk3, angle);
        initialAngle:= angle;
        if (goodPt and Foliations) > 0 then begin {draw only the curves
          corresponding to the foliatios chosen by user}
          param4.AddAndDrawLineSegment(param, param, 1);
          step:= 5*param4.radius2D/hypot(rk3[1],rk3[2]);
          guessStep:= step;
          x:= param;
          repeat
            xPrev:= x;
            counter:= counter + 1;
            if ((counter mod 30) = 0) then begin
              Application.ProcessMessages;
                {give user the chance of click the stop button}
              if param4.stop then begin
                {stop buttom was pressed}
                param4.stop:= False;
                can.Pen:= pen;
                Exit;
              end
            end;
            repeat
              x:= xPrev;
              notTooFar:= RunKut(step, shift, guessStep, angle, x, curveType);
              inside:= min(param4.halfEdges2D[1] - abs(x[1] -
                param4.center2D[1]),
                param4.halfEdges2D[2] - abs(x[2] - param4.center2D[2]));
              if not (notTooFar and (inside > 0)) then
                begin {reduce step and try again}
                guessStep:= 0.5 * step;
                step:= guessStep;
                shift:= abs(step) * (abs(rk1[1]) + abs(rk1[2]));
              end;
            until (notTooFar and (inside > 0)) or (shift < cut);
            if (shift >= cut) then begin
              {number of points that must be interpolated}
              chunks:= system.trunc(1 + shift / param4.maxSegLen);
              theta:= 1.0 / chunks;
              vec1:= xPrev;
              theSegment[1]:= xPrev;
              for i:= 1 to chunks do begin
                y:= Interpolate(xPrev, rk1, rk2, rk3, step, i * theta);
                theSegment[2]:= y;
                vec2:= y;
                param4.AddAndDrawLineSegment(vec1, vec2, 0);
                if firstStepsTaken then
                  begin {check whether the curve went again to
                  the initial point}
                  bc:= (param[1] - theSegment[1, 1]) * (theSegment[2, 1] -
                    theSegment[1, 1])
                    + (param[2] - theSegment[1, 2]) * (theSegment[2, 2] -
                      theSegment[1][2]);
                  if bc > 0 then begin
                    cc:= sqr(theSegment[2][1] - theSegment[1][1]) +
                      sqr(theSegment[2][2] - theSegment[1][2]);
                    if cc > bc then begin
                      bb:= sqr(param[1] - theSegment[1][1]) +
                        sqr(param[2] - theSegment[1][2]);
                      aux:= sqrt(abs(bb - sqr(bc) / cc));
                      if aux < 2 * param4.radius2D then
                        overtaken:= TRUE;
                    end;
                  end;
                end;
                theSegment[1]:= y;
                vec1:= vec2;
              end;
              firstStepsTaken:= TRUE;
            end;
          until (inside <= 0) or (shift < cut) or overtaken;
        end;

        {set the initial angle for drawing the next curve by the initial point}
        if curveType = kCurvatureLines then
          angle:= initialAngle + pi / 4
        else if curveType in [kMeanCurvLines, kPrincipalLines] then
          angle:= initialAngle + pi / 2
        else
          angle:= initialAngle + pi;
        if angle > 2 * pi then
          angle:= angle - 2 * pi
      end;
      param4.InfoBar.SimpleText:= '';
    end; {of  if goodPt...}
    if goodPt = 0 then begin
      if curveType = kAsymptoticLines then
        param4.InfoBar.SimpleText:= eNoAsymptoticLines
      else if curveType = kCurvatureLines then
        param4.InfoBar.SimpleText:= eUmbilicalPoint;
    end;
  except on exception do
      param4.InfoBar.SimpleText:= eComputationHalted;
  end;
  can.Pen:= pen;
end;

end.

