﻿unit FEvaluatorBytes_64_XE2;

{.$mode delphi}
{-----------------------------------------------------------------------------
 Unit Name: FEvaluator_64
 Author:    Angel Montesinos Amilibia
 October 15, 2011 -  November 29, 2018
 Purpose: Class TEvaluator_32. It is ancillary to TFunctionParser_32.
 From an instance of class TFunctionParser_32, and using its RPN
 equivalent, it creates evaluators of the function and its derivatives first and
 second. Provides also some useful applications of these evaluators

 For source code you may email me to:
     montesin@uv.es
-----------------------------------------------------------------------------}
{

Unit Name: FEvaluatorBytes_64
Evaluator of functions of several variables built from an instance of
TEvaluatorBytes_64



MIT License

Copyright (c) [2011 - 2018] [Angel Montesinos-Amilibia], email montesin@uv.es

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


}


{See OCFunctDDDoubleSafeIntegrals for documentation}



interface
uses
  InfByteArrays, FunctionParserBytes_64_XE2;

type
  Vector2= array[1..2] of Double;
  Matrix2= array[1..2] of Vector2;
  Vector3= array[1..3] of Double;
  Matrix3= array[1..3] of Vector3;

FirstDeriv= record
  fa0, dfa0: double;
end;

SecondDeriv= record
  fa0, dfa0, ddfa0: double;
end;

const

  kcvNum = 5; {number of special constants that the evaluator may
               insert in the object code}
  fpTypeSize = SizeOf(Double);
  eMaxNum: Double = 1.0e100;
  eMinNum: Double = 1.0e-100; {must be equal to 1/eMaxNum}
  eMaxExp: Double = 230;

    {error messages}
  eThereIsNoError = 'There is no error';
  eBadNumber = 'A number is miswritten.';
  eBadCharacters = 'You have used an illegal character.';
  eBadFunctionOrVariables = 'You mispelled a function or variable';
  eBadParentheses = 'Check the parentheses, please.';
  eTooMuchConstants = 'I cannot digest that much constants.';
  eSyntaxError = 'There is a sintax fault in the formula.';
  eErrorInParsing = 'I did fail in analysing the function.';
  eExpressionTooLarge = 'The expression is too long.';
  eNothingWritten = 'Some field may be empty';
  eBadVar = 'The allowed variables in this formula are: ';
  eBadVarSet = 'The allowed variables must be in the range ''p'' to ''z''.';
  eBadParamSet = 'Bad parameter set';
  eNoIntegralsInsideIntegrals = 'Integrals inside integrals are not allowed';
  eBadIntegralSyntax = 'Bad integral sintax';
  eAtMostTwoVarsInInt = 'One or at most two vars are allowed in integrals, besides ''r''.';
  eNaughtIntegral = 'The integral is naught because it doesn''t contain variables.';
  eTooMuchParts = 'Implicit function too long';
  eCouldnotCreateBlock= 'Could not create code block';
  eEmptyFormula= 'Formula reduces to nothing';
  eUndigestedPowerSymbol= 'Undigested power symbol';


type

  TEvaluatorBytes_64 = class(TObject)
  private
    data : array of Double; {stores the value of the
        parameters and constants that appear in the formula}
    emptyRegisters: Byte;
    minCorner, jump1, jump2, invJump1, invJump2: array of double;
    intList1: IntegralDataList1;
    intList2: IntegralDataList2;
    lastIntegral1, lastIntegral2, endFD, endVD: Integer;
    theArg:  PArray255D;
    currentJ: Integer; {Used for the derivatives to know where to read the
                         current position of 'u' and 'w'; Don't use for any else }
    lastPreserved: Byte; {last preserved item from the overflowing FPU stack}

    argNum, parNum, lengthActorList : Integer; {argNum=Length[allowedVars]}
    actorList : ByteArray;  {the RPN array of bytes that represents the function}
    ffBlock: PByte;
    ffCode: InfBArray; { 'ffCode.arr' shall be equal to ffBlock}
    vvBlock: PByte;
    vvCode: InfBArray; { 'vvCode.arr' shall be equal to vvBlock}
    aaBlock: PByte;
    aaCode: InfBArray; { 'aaCode.arr' shall be equal to aaBlock}
    ffBlockLength, vvBlockLength, aaBlockLength: Integer;

    procedure BuildFirstDerivative;
    procedure BuildSecondDerivative;
    procedure CodifyFunction;
    function ComputeFCodeLength: Integer;
    function ComputeDFCodeLength: Integer;
    function ComputeDDFCodeLength: Integer;
    function EvaluateBSplinesFunction(index: Integer): Double;
    function EvaluateBSplinesFunctionAndFirstDer(index: Integer): Double;
    function EvaluateBSplinesFunctionFirstAndSecondDer(index: Integer): Double;
    procedure LeaveRoomInFPUStack(neededPlaces: Integer; var cod: InfBArray);
    function LengthLeaveRoom(neededPlaces: Integer): Integer;
    function LengthPopFromStack: Integer;
    procedure LoadA(const j: Byte; var iba: InfBArray);
    procedure LoadCVPar(const j: Byte; var iba: InfBArray);
    procedure LoadData(const j: Word; var yarr: InfBArray);
    procedure LoadX(i: Word; var yarr: InfBArray);
    procedure OCRBytes(stub: ByteArray; var yarr: InfBArray);
    procedure LoadV(const j: Byte; var iba: InfBArray);
    procedure PopFromAppStack(var fCode: InfBArray);
    procedure WriteAcc(const j: Byte; var iba: InfBArray; pop: Boolean);
    procedure WriteVel(const j: Byte; var iba: InfBArray; pop: Boolean);
  public
    ff: RealFunction;
      {Gives the value of the funcion at the point X}
    vv: RealFunction;
      {Gives the value of the funcion at the point (XV[0],..., XV[argNum-1])
       and leaves in u[1] the value of the directional derivative of the
       function with respect to the vector (XV[argNum],..., XV[2*argNum-1])}
    aa: RealFunction;
      {Gives the value of the function at the point (XVA[0],..., XVA[argNum-1])
       and leaves in u[1] the value of the directional derivative of the
       function with respect to the vector (XVA[argNum],..., XVA[2*argNum-1]),
       and leaves in w[1] the value of the second derivative of the
       function with respect to the velocity (XVA[argNum],..., XVA[2*argNum-1]),
       and acceleration (XVA[2*argNum],..., XVA[3*argNum-1])}

    usedVarsAndParams: AnsiString;
    procedure Build(theF: TFunctionParserBytes_64);
    constructor Create;
    destructor Destroy; OVERRIDE;
    //function Dummy(index: Integer): Double;
    procedure DiffAndHessian(x: Vector3;  var df: Vector3; var hess: matrix3); overload;
         {Use this when ArgNum = 3}
    procedure DiffAndHessian(x: Vector2; var f: Double; var df: Vector2; var hess: matrix2); overload;
         {Use this when ArgNum = 2}
    function FirstD(const XV: array of Double): FirstDeriv;
    function gradF(const x: Vector3): Vector3; overload;
         {Use this when ArgNum = 3}
    function gradF(const x: Vector2): Vector2; overload;
         {Use this when ArgNum = 2}
    function SecondD(const XV: array of Double): SecondDeriv;
     // function TheF(const x: Vector3): Double;
     procedure UpdateParameters(parameters: array of Double);
  end;

  FourEvaluators= array[1..4] of TEvaluatorBytes_64;

  var
    theFs: FourEvaluators;
    theFsClones: array of FourEvaluators;

implementation
uses
  SysUtils, Math,
  {$IFDEF MSWINDOWS}
  Windows
  {$ELSE IF LINUX}
  BaseUnix
  {$ENDIF}
  ;

const
  c_sfr: Integer= $01A0; {room for storing locals and overflowings of fpu stack}
  c_LastReg = 7; {last register in the FPU}

  bA= $41 {'A'};
  bB= $42 {'B'};
  bC= $43 {'C'};
  bD= $44 {'D'};
  bE= $45 {'E'};
  bF= $46 {'F'};
  bG= $47 {'G'};
  bH= $48 {'H'};
  bI= $49 {'I'};
  bJ= $4A {'J'};
  bK= $4B {'K'};
  bL= $4C {'L'};
  bM= $4D {'M'};
  bN= $4E {'N'};
  bO= $4F {'O'};
  bP= $50 {'P'};
  bQ= $51 {'Q'};
  bR= $52 {'R'};
  bS= $53 {'S'};
  bT= $54 {'T'};
  bU= $55 {'U'};
  bV= $56 {'V'};
  bW= $57 {'W'};
  bX= $58 {'X'};
  bY= $59 {'Y'};
  bZ= $5A {'Z'};

  bPi = $70  {'p'};
  bSemiCol = $3B {';'};
  bQLow= $71{'q'};
  bZlow = $7A {'z'};

  b0= $30;
  b1= $31;
  b2= $32;
  b3= $33;
  b4= $34;
  b5= $35;
  b6= $36;
  b7= $37;
  b8= $38;
  b9= $39;

  bPlus= $2B;
  bMinus= $2D;
  bTimes= $2A;
  bDiv= $2F;
  bPow= $5E;
  bOpPar= $28;
  bClPar= $29;
  bNeg= $23;


  operators: ByteSet = [bPlus, bMinus, bTimes, bDiv, bPow];

  bLastByte= $FF;


var
  b_FtimesC: BArray3 = (bF, bTimes, bC);
  b_DOpPar: BArray2 = (bD, bOpPar);
  b_LoadPi: BArray2= ($D9, $EB);
  b_LoadHalfPi: array[0..9] of byte =($D9, $E8, //FLD1
                                       $D9 ,$E0, //FCHS
                                       $D9, $EB, //FLDPI
                                       $D9, $FD,  //FSCALE
                                       $DD, $D9); //FSTP(1)
  b_Load1: BArray2 = ($D9,$E8);
  b_Load0: BArray2 = ($D9,$EE);
  b_zero_comp: BArray4= ($D9, $EE, $DF, $F1);
    {4 bytes. ST(cLastReg) must be empty. Comparison of 0 with x=ST(0). After it,
     JA jumps if 0>x; JNA jumps if 0 <= x; JB, if 0 < x; JNB, if 0 >= x; JNE if 0<>x}
  b_1_comp: BArray4= ($D9, $E8, $DF, $F1);
    {4 bytes.  ST(cLastReg) must be empty. Comparison of 1 with x=ST(0). After it,
     JA jumps if 1>x; JNA jumps if 1 <= x; JB, if 1 < x; JNB, if 1 >= x; JNE if 1<>x}
  b_mod_pi: BArray8= ($D9, $EB, $D9, $C9, $D9, $F8, $DD, $D9);
    {8 bytes. Last ST must be empty}
  b_mod_2pi: array[0..9] of Byte= ($D9, $EB, $D8, $C0, $D9, $C9, $D9, $F8, $DD, $D9);
    {10 bytes. Last ST must be empty}
  b_ln: BArray6= ($D9, $ED, $D9, $C9, $D9, $F1);
    {6 bytes. Requires ST(0) > 0 and last ST empty}
  b_2toX: array[0..17] of Byte= ($D9, $C0, $D9, $FC, $D9, $C9, $D8, $E1, $D9,
                                 $F0, $D9, $E8, $DE, $C1, $D9, $FD, $DD, $D9);
    {18 bytes. Requires two last ST empty}
  b_div2: BArray6= ($D9, $E8, $D8, $C0, $DE, $F9);
    {6 bytes. Requires ST(cLastReg) empty}
  b_div_S: array[0..15] of Byte= ($D9, $EE, $DF, $F1, {zero compare}
             $75, $08, //if r_j <> 0, jump 8 bytes
             $DD, $D8,
             $DD, $D8, // nothing in FST
             $D9, $EE, $EB, $02, $DE, $F9);
    {Safe division. 16 bytes. Gives 0 if divisor is zero}

  b_times_Log2_e: BArray4= ($D9, $EA, $DE, $C9);
  b_FSTP_rec: BArray2 = ($DD,$9D); {FSTP fp64bit [RBP] plus 32 bit displacement}
  b_FLD_rec: BArray2 = ($DD,$85); {FLD fp64bit [RBP] plus 32 bit displacement}
  b_Ld_X1: BArray2 = ($DD, $02);  {FLD fp64bits [EDX]} {used only for fld the first indep var}
  b_Ld_X: BArray2  = ($DD, $42);  {FLD fp64bits [EDX] + disp8} {used for remaining
          indep vars and initial velocities and accelerations. Requires to fill it with
          its third entry, a byte, }
  b_Ld_Dt: BArray2 =  ($DD, $80);  {FLD from [EAX] + disp32}{used for
      fld formula consts, intermediate velocities and accelerations}
  b_Ld_Dt8: BArray2 = ($DD, $40);  {FLD from [EAX] + disp8}{used for freq const and params}
  b_StPdt: BArray2 = ($DD, $98); {FSTP to [EAX] + disp32} {used only for fstp intermediate
                                               velocities and accelerations}
  b_StDt: BArray2 =  ($DD, $90); {FST to [EAX] + disp32} {used only for fst intermediate
                                               velocities and accelerations}

  b_ln_S: Array[0..20] of Byte = ({b_zero_comp:}$D9, $EE, $DF, $F1, $72, $09, $DD, $D8,
      {LoadCVPar(5): next three bytes. Fill place 10} $DD, $40, 0,
      $D9, $E0, $EB, $06, {b_ln:} $D9, $ED, $D9, $C9, $D9, $F1);

  b_2toX_S: Array[0..48] of Byte = ($D9, $C0, $D9, $E1, {LoadCVPar(5), next 3 bytes} $DD, $40, 0,
      $DF, $F1, $DD, $D8, $72, $14, {b_2toX:} $D9, $C0, $D9, $FC, $D9, $C9, $D8, $E1, $D9,
      $F0, $D9, $E8, $DE, $C1, $D9, $FD, $DD, $D9 {b_2toX ended}, $EB, $10,
      {b_zero_comp:}$D9, $EE, $DF, $F1,  $DD, $D8, $72, $05,
      {LoadCVPar(3):} $DD, $40, 0 , $EB, $03, {LoadCVPar(1):} $DD, $40, 0);

  {These are the prolog and epilog when target is x86_64}
  b_Prolog: array[0..10] of Byte= ({push RBP: }$55, {RBP  <-- RSP:}
     {RSP  <--  RSP - $01A0: }$48, $81, $EC, {Int32Str(c_sfr):} $A0,$01,$00,$00,
     {mov %rsi,%edx:} $48, $89, $E5);


  b_Epilog: array[0..16] of Byte= ({fstp  [RBP+$10]. This clears ST(0):}  $DD, $5D, $10,
           {movsd  xmm0, [RBP-$08]. Result goes to xmm0: }$F2,$0F,$10,$45,$10,
           $48,$8D,$A5,$A0,$01,$00,$00,{Pop RBP} $5D, {ret:} $C3);

  b_plus_Code: BArray2= ($DE, $C1);
  b_minus_Code: BArray2= ($DE, $E9);
  b_times_Code: BArray2= ($DE, $C9);
  b_neg_code: BArray2= ($D9, $E0);
  b_cos_code: array[0..11] of Byte=  ({c_mod_2pi:} $D9, $EB, $D8, $C0, $D9, $C9, $D9, $F8, $DD, $D9,
                                      {cos:} $D9, $FF);
  b_sin_code: array[0..11] of Byte=  ({c_mod_2pi:} $D9, $EB, $D8, $C0, $D9, $C9, $D9, $F8, $DD,$D9,
                                      {sin:} $D9, $FE);
  b_Sqr: BArray2 = ($D8, $C8);
  b_Sqrt: array[0..13] of Byte= ($D9, $EE, $DF, $F1, $72, $06,
                    $DD, $D8, $D9, $EE, $EB, $02, $D9, $FA);
  b_Tan: array[0..11] of Byte = ($D9, $EB, $D9, $C9, $D9, $F8, $DD, $D9, $D9, $F2, $DD, $D8);


procedure AddInt32ToCode(n: Integer; var cod: InfBArray);
{Adds to 'cod' the code of the integer 'n', in the format that is
required by the CPU and increments the index of the last meanngful entry
of 'cod'. It is supposed that the length of 'cod' is not surpassed}
var
  c: PByte;
  theInt, i: Integer;
begin
  if cod.h+4 > cod.bigH then
    raise exception.create(eWrongAttemptToWrite);
  theInt:= n;
  c := @theInt;
  for i := 0 to 3 do begin
    inc(cod.h);
    cod.arr[cod.h] := c[i];
  end;
end;

function Int32Str(n: Integer): ByteArray;
var
  c: PByte;
  theInt, i: Integer;
begin
  theInt:= n;
  c := @theInt;
  SetLength(result, 4);
  for i := 0 to 3 do
    result[i] := c[i];
end;


{function Floor(x: Double): Integer;
begin
  result:= Trunc(x);
  if (result<>x) and (x < 0) then
    dec(result);
end; }



function disp32(n: Integer): BArray4;
{32 bit displacement from @data[0] to @data[n] written as ByteArray}
var
  bytes, i: Integer;
  c: PByte;
begin
  bytes:= fpTypeSize * n;
  c:= @bytes;
  for i := 0 to 3 do begin
    Result[i] := c[i];
  end;
end;


procedure sn(n: Pointer; var stub: array of Byte);
var
  c: PByte;
  i: Integer;
begin
  c:= PByte(@n);
  for i := 0 to SizeOf(Pointer) - 1 do
    stub[i]:= c[i];
end;

function EvaluateBSplineOnCurve(order, k1: Integer; a1: Double; dd: PArrayNodes1): Double;
var
  b0, a: Double;
  i,r: Integer;
  p: array[0..4,0..4] of Double; //4 is the expected maximal order
begin
  for i:= 0 to order do
    p[0,i]:= dd^[i+k1-order];
  for r:= 1 to order do begin
    b0:= 1/(order+1-r);
    for i:= r to order do begin
      a:= b0*(a1 - i - k1 + order);
      p[r,i]:= a*p[r-1,i]+(1-a)*p[r-1,i-1];
    end;
  end;
  result:= p[order,order];
end;

function EvaluateBSplineOnSurface(order1, order2, k1,k2: Integer; a1, a2: Double; dd: PArrayArrayNodes2): Double;
var
  b0, a: Double;
  i,j, r: Integer;
  d: array[0..4] of double;  //4 is the expected maximal order
  p: array[0..4, 0..4] of Double; //4 is the expected maximal order
begin
  for j:= k2-order2 to k2 do begin
    for i:= 0 to order1 do
      p[0,i]:=dd^[i+k1-order1,j];
    for r:= 1 to order1 do begin
      b0:= 1/(order1+1-r);
      for i:= r to order1 do begin
        a:= b0*(a1 - i - k1 + order1);
        p[r,i]:= a*p[r-1,i]+(1-a)*p[r-1,i-1];
      end;
    end;
    d[j-k2+order2]:= p[order1,order1];
  end;
  for i:= 0 to order2 do
    p[0,i]:= d[i];
  for r:=1 to order2 do begin
    b0:= 1/(order2+1-r);
    for i:= r to order2 do begin
      a:= b0*(a2 - i - k2 + order2);
      p[r,i]:= a*p[r-1,i]+(1-a)*p[r-1,i-1];
    end;
  end;
  result:= p[order2,order2];
end;

procedure TEvaluatorBytes_64.Build(theF: TFunctionParserBytes_64);
var
  i, j, k, a, b: Integer;
begin
  argNum:= theF.argNum;
  parNum:= theF.parNum;
  endFD:= theF.endFD;
  endVD:= theF.endVD;
  lastIntegral1:= theF.lastIntegral1;
  lastIntegral2:= theF.lastIntegral2;
  k:= High(theF.data);
  if k >= 0 then begin
    SetLength(data,k+1);
    for j:= 0 to k do
      data[j]:= theF.data[j];
  end;


  lengthActorList:= Length(theF.actorList);
  SetLength(actorList, lengthActorList);
  for i:= 0 to lengthActorList-1 do
    actorList[i]:= theF.actorList[i];

  k:= High(theF.minCorner);
  if k >= 0 then begin
    SetLength(minCorner,k+1);
    SetLength(jump1,k+1);
    SetLength(invJump1,k+1);
    SetLength(jump2,k+1);
    SetLength(invJump2,k+1);
    for i:= 0 to k do begin
      minCorner[i]:= theF.minCorner[i];
      jump1[i]:= theF.jump1[i];
      invJump1[i]:= theF.invJump1[i];
      jump2[i]:= theF.jump2[i];
      invJump2[i]:= theF.invJump2[i];
    end;
  end;

  if lastIntegral1 >= 0 then begin
    if high(intList1) < lastIntegral1 then
      SetLength(intList1, lastIntegral1+1);
    for i:= 0 to lastIntegral1 do begin
      intList1[i].var1:= theF.intList1[i].var1;
      for j:= 0 to 2 do
        for k:= 0 to cNodes1 do
         intList1[i].data[j,k]:= theF.intList1[i].data[j,k];
    end;
  end;
  if lastIntegral2 >= 0 then begin
    if high(intList2) < lastIntegral2 then
      SetLength(intList2, lastIntegral2+1);
    for i:= 0 to lastIntegral2 do begin
      intList2[i].var1:= theF.intList2[i].var1;
      intList2[i].var2:= theF.intList2[i].var2;
      for j:= 0 to 2 do
        for k:= 0 to 2 do
          for a:= 0 to cNodes2 do
            for b:= 0 to cNodes2 do
              IntList2[i].data[j,k,a,b]:= theF.IntList2[i].data[j,k,a,b];
    end;
  end;
  CodifyFunction;
  BuildFirstDerivative;
  BuildSecondDerivative;
end;

procedure TEvaluatorBytes_64.BuildFirstDerivative;
var
  i, j, n, a, pushes: Integer;
  ch: Byte;
  {$IFDEF MSWINDOWS}
    prot: Word;
  {$ELSE IF LINUX}
  prot: cInt;
  {$ENDIF}
  stub: ByteArray;
begin
  lastPreserved:= 1;
{$IFDEF MSWINDOWS}
  {if vvBlock = nil then begin
    vvBlockLength:= ComputeDFCodeLength;
    prot:= PAGE_EXECUTE_READWRITE;
    vvBlock:= VirtualAlloc(nil, vvBlockLength, MEM_COMMIT, prot);
  end
  else begin
    prot:= PAGE_EXECUTE_READ;
    VirtualProtect(vvBlock, vvBlockLength, PAGE_EXECUTE_READWRITE, @prot);
  end; }
  if vvBlock <> nil then begin
    VirtualFree(vvBlock, 0, MEM_RELEASE);
    vvBlock:= nil
  end;
  vvBlockLength:= ComputeDFCodeLength;
  prot:= PAGE_EXECUTE_READWRITE;
  vvBlock:= VirtualAlloc(nil, vvBlockLength, MEM_COMMIT, prot);
{$ELSE IF LINUX}
  if vvBlock <> nil then
    if Fpmunmap(vvBlock, vvBlockLength) <> 0 then
      raise exception.Create(eBadIntegralSyntax);
  prot:= PROT_EXEC OR PROT_READ OR PROT_WRITE;
  vvBlockLength:=  ComputeDFCodeLength;
  vvBlock:= FPmMAP(nil, vvBlockLength, prot, MAP_PRIVATE or MAP_ANONYMOUS, 0, 0);
{$ENDIF}
  vvCode.arr:= vvBlock;
  vvCode.h:= -1;
  vvCode.bigH:= vvBlockLength-1;
  IBAppend(vvCode, b_Prolog);{11 bytes. initiate 'vvBlock' with the prologue. }
  SetLength(stub, SizeOf(Pointer));
  if (lastIntegral1 < 0) and (lastIntegral2 < 0) then begin
    {Save the address of 'data' to rax}
    IBAppend(vvCode, [$48, $B8]); {2 bytes}
    sn(@data[0], stub);
    IBAppend(vvCode, stub); {SizeOf(Pointer)}
  end
  else begin
    {put the address of 'theArg[0]' to rax}
    IBAppend(vvCode, [$48, $B8]); {2 bytes}
    sn(@theArg, stub);
    IBAppend(vvCode, stub);

    {mov [rax], rdx. This makes  @theArg[0]:= @X[0]}
    IBAppend(vvCode,[$48, $89, $10]);

    {Save the address of 'data' to rax}
    IBAppend(vvCode, [$48, $B8]); {2 bytes}
    sn(@data[0], stub);
    IBAppend(vvCode, stub); {SizeOf(Pointer)}

    {mov [rbp+c_sfr+$10], rdx. This preserves rdx}
    IBAppend(vvCode, [$48,$89, $95]);
    IBAppend(vvCode,  Int32Str(c_sfr+$10));

    {mov [rbp+c_sfr+$18],rax. This preserves rax}
    IBAppend(vvCode, [$48,$89, $85]);
    IBAppend(vvCode, Int32Str(c_sfr+$18));

    {mov rcx, self. This puts @self in rcx}
    IBAppend(vvCode, [$48,$B9]);
    sn(self, stub);
    IBAppend(vvCode, stub);

    {mov [rbp+c_sfr+$20],rcx. This preserves rcx};
    IBAppend(vvCode, [$48,$89, $8D]);
    IBAppend(vvCode,  Int32Str(c_sfr+$20));
  end;

  emptyRegisters := c_LastReg + 1;
  i:= -1;
  j:= 0;
  while i < High(actorList) do
    begin
      inc(i);
      ch := actorList[i];
      if ch in  [bLastByte, bPi, b1] then begin
        LeaveRoomInFPUStack(1,vvCode);
        inc(j);
        case ch of
          bLastByte: begin
              ch:= actorList[i+1];
              i:= i+3;
              n:=  actorList[i];
              case ch of
                $01: begin{an independent variable}
                       LoadV(n, vvCode);
                       WriteVel(j, vvCode, True);
                       LoadX(n, vvCode);
                       dec(emptyRegisters);
                     end;
                $02: begin{a parameter}
                       IBAppend(vvCode, b_Load0);
                       WriteVel(j, vvCode, True);
                       LoadCVPar(n, vvCode);
                       dec(emptyRegisters);
                     end;
                $03: begin {a constant}
                       IBAppend(vvCode, b_Load0);
                       WriteVel(j, vvCode, True);
                       n:= (actorList[i-1] shl 8) + n;
                       LoadData(n, vvCode);
                       dec(emptyRegisters);
                    end;
                $04: {an integral}
                  begin
                    if n<0 then raise
                       exception.Create(eBadIntegralSyntax);
                    pushes:= c_LastReg + 1 - emptyRegisters;
                    LeaveRoomInFPUStack(c_LastReg+1, vvCode); {empty FPU}

                    {put in ESI, the order of the integral in 'intLis',
                     where 'EvaluateBSplinesFunction' will expect it}
                    IBAppend(vvCode,  [$C7,$C2]);
                    IBAppend(vvCode, Int32Str(n));  {mov edx, imm32}

                    {load in variable 'currentJ' the value of  'j'}
                    IBAppend(vvCode, [$48, $B8]);
                    SetLength(stub, SizeOf(Pointer));
                    sn(@currentJ, stub);
                    IBAppend(vvCode, stub); {mov rax, @currentJ}
                    IBAppend(vvCode, [$48, $C7, $00]);
                    IBAppend(vvCode, Int32Str(j)); {move to [rax]= @currentJ the value j}

                    {mov rax, @TEvaluator_64.EvaluateBSplinesFunction}
                    IBAppend(vvCode, [$48, $B8]);
                    sn(@TEvaluatorBytes_64.EvaluateBSplinesFunctionAndFirstDer, stub);
                    IBAppend(vvCode, stub);

                    IBAppend(vvCode, [$FF, $D0]); {call [rax]};

                     {it is supposed that the called function leaves its result in XMM0,
                     and we put that value into ST(0) of the FPU}
                    IBAppend(vvCode, [$f2, $0f, $11, $85]);
                    IBAppend(vvCode, Int32Str(c_sfr-$08));  {movsd [rbp-$08], xmm0}
                    IBAppend(vvCode, [$DD, $85]);
                    IBAppend(vvCode, Int32Str(c_sfr-$08));  {fld [rbp-$08]}

                    emptyRegisters := c_LastReg;
                    if pushes > 0 then
                       for  a:= 1 to pushes do
                         PopFromAppStack(vvCode);

                     {recover initial values of rcx, rdx and rax}
                    IBAppend(vvCode, [$48, $8B, $8D]);
                    IBAppend(vvCode, Int32Str(c_sfr+$20));
                    IBAppend(vvCode, [$48, $8B, $95]);
                    IBAppend(vvCode, Int32Str(c_sfr+$10));
                    IBAppend(vvCode, [$48, $8B, $85]);
                    IBAppend(vvCode, Int32Str(c_sfr-$18));
                   end;
                else
                  ;
              end;
            end;
          bPi: begin
                 IBAppend(vvCode, b_Load0);
                 WriteVel(j, vvCode, True);
                 IBAppend(vvCode, b_LoadPi);
                 dec(emptyRegisters);
               end;
          b1: begin
                 IBAppend(vvCode, b_Load0);
                 WriteVel(j, vvCode, True);
                 IBAppend(vvCode, b_Load1);
                 dec(emptyRegisters);
               end;
          else
            ;
        end;
      end
      else
        begin
          if (ch in [bPlus, bMinus, bTimes, bDiv, bPow]) then
            begin
              if (emptyRegisters = c_LastReg) then
                PopFromAppStack(vvCode);
              if ch = bPow then
                LeaveRoomInFPUStack(3, vvCode)
              else if ch in [bTimes, bDiv] then
                LeaveRoomInFPUStack(2, vvCode)
              else
                LeaveRoomInFPUStack(1, vvCode);
              case ch of
                bPlus: begin
                    IBAppend(vvCode, [$DE, $C1]);
                    LoadData(endFD + j-1, vvCode);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C1]);
                    WriteVel(j-1, vvCode, True);
                  end;
                bMinus : begin
                    IBAppend(vvCode, [$DE, $E9]);
                    LoadData(endFD + j-1, vvCode);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $E9]);
                    WriteVel(j-1, vvCode, True);
                  end;
                bTimes : begin
                    IBAppend(vvCode, [$D9, $C0, $D8, $CA, $D9, $CA]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    LoadData(endFD + j-1, vvCode);
                    IBAppend(vvCode, [$DE, $CA, $DE, $C1]);
                    WriteVel(j-1, vvCode, True);
                  end;
                bDiv : begin
                    IBAppend(vvCode, b_zero_comp);
                    IBAppend(vvCode,[$75, $0E, {if st0 <> 0, jump 14 bytes}
                                     $DD, $D8, // POP ST(0). r_{j-1} (numerator) in ST(0)
                                     $DD, $D8]); // POP ST(0). Nothing in  FST
                    IBAppend(vvCode, b_Load0);
                    WriteVel(j-1, vvCode, False);  //u[j-1]= 0, ST(0) = 0}
                    IBAppend(vvCode,[$EB, $28]);  {jump 40 bytes}
                    IBAppend(vvCode, b_Load1); {st0= 1, st1= y, st2= x}
                    IBAppend(vvCode, [$DE, $F1, {fdivrp, st0=1/y, st1=x}
                                      $D9, $C0, {fld st0, st0= 1/y, st1= 1/y, st2= x}
                                      $D8, $CA,$D9,$CA]);
                    LoadData(endFD + j-1, vvCode);
                    IBAppend(vvCode, [$D8, $CA, $D9, $CA, $D8, $C8, $DE, $C9]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9, $DE, $E9]);
                    WriteVel(j-1, vvCode, True);
                  end;
                bPow :  ;{This will not be used}
              end;
              inc(emptyRegisters);
              j:= j-1;
            end
          else {opCode of a function}
            begin
              if ch= bNeg then
                LeaveRoomInFPUStack(1, vvCode)
              else
                LeaveRoomInFPUStack(2, vvCode);
              case ch of
                bNeg : begin
                    IBAppend(vvCode, [$D9, $E0]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$D9, $E0]);
                    WriteVel(j, vvCode, True);
                  end;

                bA : begin
                    IBAppend(vvCode, b_mod_2pi);
                    IBAppend(vvCode, [$D9, $FB, $D9, $C9, $D9, $E0]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bB : begin
                    IBAppend(vvCode, b_mod_2pi);
                    IBAppend(vvCode, [$D9, $FB]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bC : begin
                    IBAppend(vvCode, [$D9, $EA, $DE, $C9]);
                    IBAppend(vvCode, b_2ToX_S);
                    IBAppend(vvCode, [$D9, $C0]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bD : begin
                    IBAppend(vvCode, [$D9, $C0, $D8, $C8, $D9, $C9, $D8, $C0]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bE : begin
                    IBAppend(vvCode, b_zero_comp);
                    IBAppend(vvCode, [$72, $08, $DD, $D8]); //nothing in FST
                    IBAppend(vvCode, b_Load0);  //ST(0)= 0
                    IBAppend(vvCode, b_Load0); //ST(0)= ST(1)= 0
                    IBAppend(vvCode, [$EB, $0E]); //jump 14 bytes
                    IBAppend(vvCode, [$D9, $FA]); //FSQRT, ST(0) = sqrt(r_j)
                    IBAppend(vvCode, [$D9, $C0]); //FLD ST(0), ST(0)= ST(1)= sqrt(r_j)
                    IBAppend(vvCode, [$D8, $C0]); //FADD ST(0), ST(0). ST(0)= 2*sqrt(r_j), ST(1)= sqrt(r_j)
                    LoadData(endFD + j, vvCode); //ST(0)= u_j, ST(1)= 2*sqrt(r_j), ST(2)= sqrt(r_j)
                    IBAppend(vvCode, [$DE, $F1]); //FDIVRP, ST(0)= u_j/(2*sqrt(r_j)), ST(1)= sqrt(r_j)
                    WriteVel(j, vvCode, True); // ST(0)= sqrt(r_j) or 0
                  end;

                bF : begin
                    IBAppend(vvCode, b_zero_comp);
                    IBAppend(vvCode, [$72, $0B, $DD, $D8]);
	                  LoadCVPar(5, vvCode);
                    IBAppend(vvCode, [$D9, $E0]);
	                  IBAppend(vvCode, b_Load0);
	                  IBAppend(vvCode, [$EB, $12, $D9, $C0]);
                    IBAppend(vvCode, b_ln);
                    IBAppend(vvCode, [$D9, $C9]);
	                  LoadData(endFD + j, vvCode);
	                  IBAppend(vvCode, [$DE, $F1]);
	                  WriteVel(j, vvCode, True);
                  end;

                bG : begin
                    IBAppend(vvCode, b_mod_pi);
                    IBAppend(vvCode, [$D9, $F2, $D9, $C1, $D8, $C8, $DE, $C1]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bH : begin
                    IBAppend(vvCode, [$D9, $EA, $DE, $C9]);
                    IBAppend(vvCode, b_2ToX_S);
	                  IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$D9, $E0, $D9, $C1, $D8, $F9, $D9, $FD, $D9,
                                      $CA, $D9, $FD, $DD, $D9, $D9, $C0, $D8, $C2,
                                      $D9, $CA, $DE, $E9]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
	                  WriteVel(j, vvCode, True);
                  end;

                bI : begin
                    IBAppend(vvCode, [$D9, $EA, $DE, $C9]);
                    IBAppend(vvCode, b_2ToX_S);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$D9, $E0, $D9, $C1, $D8, $F9, $D9, $FD, $D9, $CA,
                                      $D9, $FD, $DD, $D9, $D9, $C0, $D8, $C2, $D9, $C9,
                                      $DE, $E2]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode,[$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bJ : begin
                    IBAppend(vvCode, [$D9, $EA, $DE, $C9]);
                    IBAppend(vvCode, b_2ToX_S);
                    IBAppend(vvCode, [$D8, $C8]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$D8, $E9, $D9, $C9]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$DE, $C1, $DE, $F9, $D9, $C0, $D8, $C8]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$DE, $E1]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bK : begin
                    IBAppend(vvCode, [$D9, $C0, $D9, $E1]);
                    IBAppend(vvCode, b_1_comp);
                    IBAppend(vvCode, [$DD, $D8, $76, $18, $D9, $C0, $D8, $C8]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$DE, $E1, $D9, $FA]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$D8, $F1, $D9, $CA, $D9, $C9, $D9, $F3,
                                      $D9, $C9, $EB, $16]);
	                  IBAppend(vvCode, b_zero_comp);
                    IBAppend(vvCode, [$DD, $D8]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$D9, $E0]);
                    IBAppend(vvCode, b_LoadPi);
                    IBAppend(vvCode, [$D9, $FD, $DD, $D9, $72, $02, $D9,$E0]);
                    IBAppend(vvCode, b_Load0);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bL : begin
                    IBAppend(vvCode, [$D9, $C0, $D9, $E1]);
                    IBAppend(vvCode, b_1_comp);
                    IBAppend(vvCode, [$DD, $D8, $76, $18, $D9, $C0, $D8, $C8]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$DE, $E1, $D9, $FA]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$D8, $F1, $D9, $E0, $D9, $CA, $D9, $F3,
                                      $D9, $C9, $EB, $10]);
                    IBAppend(vvCode, b_zero_comp);
                    IBAppend(vvCode, [$DD, $D8, $72, $04]);
                    IBAppend(vvCode, b_LoadPi);
                    IBAppend(vvCode, [$EB, $02]);
                    IBAppend(vvCode, b_Load0);
                    IBAppend(vvCode, b_Load0);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bM : begin
                    IBAppend(vvCode, [$D9, $C0]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$D9, $F3, $D9, $C9, $D8, $C8]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$DC, $C1, $DE, $F1]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bN : begin
                    IBAppend(vvCode, b_zero_comp);
                    IBAppend(vvCode, [$76, $10, $D9, $E0]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$D9, $E0]);
                    WriteVel(j, vvCode, True);
                  end;

                bO : begin
                    IBAppend(vvCode, b_mod_pi);
                    IBAppend(vvCode, b_zero_comp);
                    IBAppend(vvCode, [$74, $12, $D9, $F2, $DC, $F1, $D9, $C9, $D9, $C0,
                                      $D8, $C8, $DE, $C2, $D9, $C9, $D9, $E0, $EB, $06,
                                      $DD, $D8]);
                    IBAppend(vvCode, b_Load0);
                    IBAppend(vvCode, b_Load0);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bP : begin
                    IBAppend(vvCode, b_1_comp);
                    IBAppend(vvCode, [$73, $1E, $D9, $C0, $D8, $C8]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$DE, $E9, $D9, $FA, $D9, $C0, $DE, $C2, $D9, $C9]);
                    IBAppend(vvCode, b_ln);
                    IBAppend(vvCode, [$D9, $C9]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$DE, $F1, $EB, $06, $DD, $D8]);
                    IBAppend(vvCode, b_Load0);
                    IBAppend(vvCode, b_Load0);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bQ : begin
                    IBAppend(vvCode, [$D9, $C0, $D9, $E1]);
                    IBAppend(vvCode, b_1_comp);
                    IBAppend(vvCode, [$DD, $D8, $76, $2A, $D9, $C0]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$DE, $C1, $D9, $C9]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$DE, $E1, $D9, $C0, $D8, $CA, $D9, $CA, $DE, $F1]);
                    IBAppend(vvCode, b_ln);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$D9, $E0, $D9, $C9, $D9, $FD, $D9, $CA,
                                      $DE, $F9, $D9, $E0, $EB, $12]);
                    IBAppend(vvCode, b_zero_comp);
                    IBAppend(vvCode, [$DD, $D8, $77, $05]);
                    LoadCVPar(1, vvCode);
                    IBAppend(vvCode, [$EB,$03]);
                    LoadCVPar(2, vvCode);
                    IBAppend(vvCode, b_Load0);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end;

                bR : begin
                    IBAppend(vvCode, [$D9, $C0, $D8, $C8]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$DE, $C1, $D9, $FA, $D9,
                                      $C0, $D9, $CA, $DE, $C1]);
                    IBAppend(vvCode, b_ln);
                    IBAppend(vvCode, [$D9, $C9]);
                    IBAppend(vvCode, b_Load1);
                    IBAppend(vvCode, [$DE, $F1]);
                    LoadData(endFD + j, vvCode);
                    IBAppend(vvCode, [$DE, $C9]);
                    WriteVel(j, vvCode, True);
                  end
              else
                 ;
              end;
            end;
        end;
    end; {del while}
  IBAppend(vvCode, b_epilog);   //ret
  @vv := @vvCode.arr[0];
  stub:= nil;
end;

procedure TEvaluatorBytes_64.BuildSecondDerivative;
var
  i, j, n, a, pushes: Integer;
  ch: Byte;
  {$IFDEF MSWINDOWS}
  prot: Word;
  {$ELSE IF LINUX}
  prot: cInt;
  {$ENDIF}
  stub: ByteArray;
begin
  lastPreserved:= 1;
  {$IFDEF MSWINDOWS}
  {if aaBlock = nil then begin
    aaBlockLength:= ComputeDDFCodeLength;
    prot:= PAGE_EXECUTE_READWRITE;
    aaBlock:= VirtualAlloc(nil, aaBlockLength, MEM_COMMIT, prot);
  end
  else begin
    prot:= PAGE_EXECUTE_READ;
    VirtualProtect(aaBlock, aaBlockLength, PAGE_EXECUTE_READWRITE, @prot); }
  if aaBlock <> nil then begin
    VirtualFree(aaBlock, 0, MEM_RELEASE);
    aaBlock:= nil;
  end;
  aaBlockLength:= ComputeDDFCodeLength;
  prot:= PAGE_EXECUTE_READWRITE;
  aaBlock:= VirtualAlloc(nil, aaBlockLength, MEM_COMMIT, prot);

  {$ELSE IF LINUX}
  if aaBlock <> nil then
    if Fpmunmap(aaBlock, aaBlockLength) <> 0 then
      raise exception.Create(eBadIntegralSyntax);
  prot:= PROT_EXEC OR PROT_READ OR PROT_WRITE;
  aaBlockLength:=  ComputeDDFCodeLength;
  aaBlock:= FPmMAP(nil, aaBlockLength, prot, MAP_PRIVATE or MAP_ANONYMOUS, 0, 0);
  {$ENDIF}
  aaCode.arr:= aaBlock;
  aaCode.h:= -1;
  aaCode.bigH:= aaBlockLength-1;

  IBAppend(aaCode, b_Prolog);{14 bytes. initiate 'vvBlock' with the prologue. }
  SetLength(stub, SizeOf(Pointer));
  if (lastIntegral1 < 0) and (lastIntegral2 < 0) then begin
    {Save the address of 'data' to rax}
    IBAppend(aaCode, [$48, $B8]); {2 bytes}
    sn(@data[0], stub);
    IBAppend(aaCode, stub); {SizeOf(Pointer)}
  end
  else begin
    {put the address of 'theArg[0]' to rax}
    IBAppend(aaCode, [$48, $B8]); {2 bytes}
    sn(@theArg, stub);
    IBAppend(aaCode, stub);

    {mov [rax], rdx. This makes  @theArg[0]:= @X[0]}
    IBAppend(aaCode,[$48, $89, $10]);

    {Save the address of 'data' to rax}
    IBAppend(aaCode, [$48, $B8]); {2 bytes}
    sn(@data[0], stub);
    IBAppend(aaCode, stub); {SizeOf(Pointer)}

    {mov [rbp+c_sfr+$10], rdx. This preserves rdx}
    IBAppend(aaCode, [$48,$89,$95]);
    IBAppend(aaCode, Int32Str(c_sfr+$10));

    {mov [rbp-$18],rax. This preserves rax, that is @data in [rbp-$18]}
    IBAppend(aaCode, [$48,$89,$85]);
    IBAppend(aaCode, Int32Str(c_sfr+$18));

    {mov rcx, self. This puts @self in rcx}
    IBAppend(aaCode, [$48,$B9]);
    sn(self, stub);
    IBAppend(aaCode, stub);

    {mov [rbp-$20],rcx. This preserves  @self  in [rbp-$20]}
    IBAppend(aaCode, [$48, $89,$8D]);
    IBAppend(aaCode, Int32Str(c_sfr+$20));
  end;

  emptyRegisters := c_LastReg + 1;
  i:= -1;
  j:= 0;
  while i < High(actorList) do  begin
    inc(i);
    ch := actorList[i];
    if ch in [bLastByte, bPi, b1] then begin
      LeaveRoomInFPUStack(1, aaCode);
      inc(j);
      case ch of
        bLastByte: begin
          ch:= actorList[i+1];
          i:= i + 3;
          n:= actorList[i];
          case ch of
            $01: begin
                LoadV(n, aaCode);
                WriteVel(j, aaCode, True);
                LoadA(n, aaCode);
                WriteAcc(j, aaCode, True);
                LoadX(n, aaCode);
                dec(emptyRegisters);
              end;
            $02, $03: begin
                n:= ord(actorList[i-1]) shl 8 + n;
                IBAppend(aaCode, b_Load0);  //PUSH  0 in ST(0)
                WriteVel(j, aaCode, False); // store ST(0) in u[j]
                WriteAcc(j, aaCode, True);  // guardar ST(0) in u[j], and pop
                LoadData(n, aaCode);
                dec(emptyRegisters);
              end;
            $04:  begin{an integral}
                if n<0 then raise
                  exception.Create(eBadIntegralSyntax);
                pushes:= c_LastReg + 1 - emptyRegisters;
                LeaveRoomInFPUStack(c_LastReg+1, aaCode); {empty FPU}

                {put in EDX, the order of the integral in 'intLis',
                     where 'EvaluateBSplinesFunctionFirstAndSecondDer' will expect it}
                IBAppend(aaCode, [$C7, $C2]); {mov edx, imm32}
                AddInt32ToCode(n, aaCode);

                {the four lines that follow mean this:
                 load in variable 'currentJ' the value of  'j'}
                IBAppend(aaCode, [$48, $B8]);
                SetLength(stub, SizeOf(Pointer));
                sn(@currentJ, stub);
                IBAppend(aaCode, stub); {mov rax, @currentJ}
                IBAppend(aaCode, [$48, $C7,$00]);
                AddInt32ToCode(j, aaCode); {mov [rax] imm 32 = j}

                IBAppend(aaCode, [$48, $B8]);
                sn(@TEvaluatorBytes_64.EvaluateBSplinesFunctionFirstAndSecondDer, stub);
                IBAppend(aaCode, stub);{mov rax, TEvaluator_64.EvaluateBSplinesFunction}

                IBAppend(aaCode, [$FF, $D0]); {call [RAX]}

                {it is supposed that the called function leaves its result in XMM0,
                and we put that value into ST(0) of the FPU}
                IBAppend(aaCode, [$f2, $0f, $11, $85]);
                IBAppend(aaCode, Int32Str(c_sfr-$08));  {movsd [rbp+c_sfr-$08], xmm0}
                IBAppend(aaCode, [$DD, $85]);
                IBAppend(aaCode, Int32Str(c_sfr-$08)); {fld [rbp+c_sfr-$08]. Thus ST(0) contains result of
                                             call to EvaluateBSplinesFunction}
                emptyRegisters := c_LastReg;
                if pushes > 0 then
                  for  a:= 1 to pushes do
                     PopFromAppStack(aaCode);
                IBAppend(aaCode, [$48, $8B, $8D]);
                IBAppend(aaCode, Int32Str(c_sfr+$20)); {Recover initial value of rcx}
                IBAppend(aaCode, [$48, $8B, $95]);
                IBAppend(aaCode, Int32Str(c_sfr+$10));  {Recover initial value of rdx}
                IBAppend(aaCode, [$48, $8B, $85]);
                IBAppend(aaCode, Int32Str(c_sfr+$18)); {Recover initial value of rax}
              end;
            else
              ;
          end;
        end;
        bPi: begin
            IBAppend(aaCode, b_Load0);
            WriteVel(j, aaCode, False);
            WriteAcc(j, aaCode, True);
            IBAppend(aaCode, b_LoadPi);
            dec(emptyRegisters);
          end;
        b1: begin
            IBAppend(aaCode, b_Load0);
            WriteVel(j, aaCode, False);
            WriteAcc(j, aaCode, True);
            IBAppend(aaCode, b_Load1);;
            dec(emptyRegisters);
          end;
        else
                  ;
      end;
    end
    else begin {opcode of a binary operator}
      if (ch in [bPlus, bMinus, bTimes, bDiv, bPow]) then begin
        if (emptyRegisters = c_LastReg) then
          PopFromAppStack(aaCode);
        if ch in [bPlus, bMinus]  then
          LeaveRoomInFPUStack(1, aaCode)
        else
          LeaveRoomInFPUStack(3, aaCode);
        case ch of
          bPlus: begin
              IBAppend(aaCode, [$DE, $C1]);  // FADDP r_j + r_{j-1} in ST(0)
              LoadData(endFD + j, aaCode);
              LoadData(endFD + j-1, aaCode);
              IBAppend(aaCode, [$DE, $C1]);  // FADDP. u_j + u_{j-1} in ST(0), r_j + r_{j-1} in ST(1)
              WriteVel(j-1, aaCode, True);
              LoadData(endVD+ j, aaCode);
              LoadData(endVD+ j-1, aaCode);
              IBAppend(aaCode, [$DE, $C1]); // FADDP. w_j + w_{j-1} in ST(0), r_j + r_{j-1} in ST(1)
              WriteAcc(j-1, aaCode, True);
            end;
          bMinus: begin
              IBAppend(aaCode, [$DE, $E9]); // FSUBP r_{j-1}-r_j in ST(0)
              LoadData(endFD + j-1, aaCode);
              LoadData(endFD + j, aaCode); // u_{j-1} in ST(0), u_j in ST(1), r_{j-1}-r_j in ST(2)
              IBAppend(aaCode, [$DE, $E9]); // FSUBP. u_{j-1} - u_j in ST(0), r_{j-1}-r_j in ST(1)
              WriteVel(j-1, aaCode, True);
              LoadData(endVD+ j-1, aaCode);
              LoadData(endVD+ j, aaCode); // w_{j-1} in ST(0), w_j in ST(1), r_{j-1}-r_j in ST(2)
              IBAppend(aaCode, [$DE, $E9]); // FSUBP. w_{j-1} - w_j in ST(0), r_{j-1}-r_j in ST(1)
              WriteAcc(j-1, aaCode, True);
            end;
          bTimes: begin
              LoadData(endFD + j, aaCode);
              LoadData(endFD + j-1, aaCode);
              IBAppend(aaCode, [$D9, $C0, $D8, $CA, $D8, $C0, $D9, $CA,
                                $D8, $CC, $D9, $C9, $D8, $CB, $DE, $C1]);
              WriteVel(j-1, aaCode, True);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$D8, $CB, $DE, $C1]);
              LoadData(endVD+ j-1, aaCode);
              IBAppend(aaCode, [$D8, $CA, $DE, $C1]);
              WriteAcc(j-1, aaCode, True);
              IBAppend(aaCode, [$DE, $C9]);
            end;
          bDiv: begin
              IBAppend(aaCode, b_zero_comp);
              IBAppend(aaCode, [$75, $14, $DD, $D8, $DD, $D8]);
              IBAppend(aaCode, b_Load0);
              WriteVel(j-1, aaCode, False);
              WriteAcc(j-1, aaCode, False);
              IBAppend(aaCode, [$EB, $48]);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$DE, $F1, $DC, $C9]);
              LoadData(endFD + j-1, aaCode);
              IBAppend(aaCode, [$D8, $C9]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D8, $CA, $D9, $C0, $D8, $CC, $D8, $EA]);
              WriteVel(j-1, aaCode, True);
              IBAppend(aaCode, [$DC, $C9, $D8, $C8, $D8, $CB, $DE, $E1, $D8, $C0]);
              LoadData(endVD+ j-1, aaCode);
              IBAppend(aaCode, [$D8, $CA, $DE, $C1]);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$D8, $CB, $DE, $CA, $DE, $E1]);
              WriteAcc(j-1, aaCode, True);
            end;
          else
             ;
        end;
        inc(emptyRegisters);
        j:= j-1;
      end
      else begin {opCode of a function}
        if ch in [bNeg,bN] then
          LeaveRoomInFPUStack(1, aaCode)
        else if ch= bC then
          LeaveRoomInFPUStack(2, aaCode)
        else
          LeaveRoomInFPUStack(3, aaCode);

        case ch of
          bNeg: begin
              IBAppend(aaCode, [$D9, $E0]);
              LoadData(endFD + j, aaCode); // endFD + j in ST(0), -r_j in ST(1)
              IBAppend(aaCode, [$D9, $E0]); //FCHS,  -endFD + j in ST(0), -r_j in ST(1)
              WriteVel(j, aaCode, True);
              LoadData(endVD+ j, aaCode); // w[j] in ST(0), -r_j in ST(1)
              IBAppend(aaCode, [$D9, $E0]); //FCHS,  -w[j] in ST(0), -r_j in ST(1)
              WriteAcc(j, aaCode, True);
            end;
          bA: begin
              IBAppend(aaCode, b_mod_2pi);
              IBAppend(aaCode, [$D9, $FB, $D9, $C9, $D9, $E0]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D8, $CB, $D9, $C9, $D8, $CA]);
              WriteVel(j, aaCode, True);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$DE, $CA, $DE, $E9]);
              WriteAcc(j, aaCode, True);
            end;
          bB: begin
              IBAppend(aaCode, b_mod_2pi);
              IBAppend(aaCode, [$D9, $FB]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D8, $CB, $D9, $C9, $D8, $CA]);
              WriteVel(j, aaCode, True);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$DE, $CA, $DE, $E9]);
              WriteAcc(j, aaCode, True);
            end;
          bC: begin
              IBAppend(aaCode, [$D9, $EA, $DE, $C9 ]);
              IBAppend(aaCode, b_2ToX_S);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D9, $C0, $D8, $CA]);
              WriteVel(j, aaCode,True);
              IBAppend(aaCode, [$D8,$C8]);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$DE, $C1]);
              IBAppend(aaCode, [$D8, $C9]);
              WriteAcc(j, aaCode, True);
            end;
          bD: begin
              IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D9, $C9, $D8, $C0]);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$D8, $C9]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$DC, $CA, $D9, $CA]);
              WriteVel(j, aaCode, True);
              IBAppend(aaCode, [$D9, $C9, $D8, $C8, $D8, $C0, $DE,$C1]);
              WriteAcc(j, aaCode, True);
            end;
          bE: begin
              IBAppend(aaCode, b_zero_comp);
              IBAppend(aaCode, [$72, $12, $DD, $D8]);
              IBAppend(aaCode, b_Load0);
              WriteVel(j, aaCode, False);
              WriteAcc(j, aaCode, False);
              IBAppend(aaCode, [$EB, $32, $D9, $FA, $D9, $C0, $D8, $C0]);
      		    IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$DE, $F1, $D9, $C0]);
      		    LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$DE, $C9, $D9, $C0]);
              WriteVel(j, aaCode, True);
              IBAppend(aaCode, [$D8, $C8, $D8, $C9, $D8, $C0]);
              LoadData(endVD+ j, aaCode);
        		  IBAppend(aaCode, [$DE, $CA, $DE, $E9]);
      		    WriteAcc(j, aaCode, True) ;
            end;
          bF: begin
              IBAppend(aaCode, b_zero_comp);
              IBAppend(aaCode, [$72, $17, $DD, $D8]);
              LoadCVPar(5, aaCode);
              IBAppend(aaCode, [$D9, $E0]);
              IBAppend(aaCode, b_Load0);
              WriteVel(j, aaCode, False);
              WriteAcc(j, aaCode, True);
              IBAppend(aaCode, [$EB, $30, $D9, $C0]);
              IBAppend(aaCode, b_ln);
              IBAppend(aaCode, [$D9,  $C9]);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$DE, $F1]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D8, $C9, $D9, $C0]);
              WriteVel(j, aaCode, True);
              IBAppend(aaCode, [$D8, $C8]);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$DE, $CA, $DE, $E9]);
              WriteAcc(j, aaCode, True);
            end;
          bG: begin
              IBAppend(aaCode, b_mod_pi);
              IBAppend(aaCode, [$D9, $F2, $D9, $C1, $D8, $C8, $DE, $C1]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D9, $C9, $D8, $CA]);
              WriteVel(j, aaCode, True);
              IBAppend(aaCode, [$D8, $C0, $D8, $CA, $D8, $C9]);
      		    LoadData(endVD+ j, aaCode);
      		    IBAppend(aaCode, [$DE, $CA, $DE, $C1]);
              WriteAcc(j, aaCode,True);
            end;
          bH: begin
              IBAppend(aaCode, [$D9, $EA, $DE, $C9]);
              IBAppend(aaCode, b_2ToX_S);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$D9, $E0, $D9, $C1, $D8, $F9, $D9, $FD, $D9, $CA,
                                $D9, $FD, $DD, $D9, $D9, $C0, $D8, $C2, $D9, $CA, $DE, $E9]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D8, $CB, $D9,$C9, $D8, $CA]);
              WriteVel(j, aaCode, True);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$DE, $CA, $DE, $C1]);
              WriteAcc(j, aaCode, True);
            end;
          bI: begin
              IBAppend(aaCode, [$D9, $EA, $DE, $C9]);
      		    IBAppend(aaCode, b_2ToX_S);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$D9, $E0, $D9, $C1, $D8, $F9, $D9, $FD, $D9, $CA, $D9, $FD,
                                $DD, $D9, $D9, $C0, $D8, $C2, $D9, $CA, $DE, $E9, $D9, $C9]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D8, $CB, $D9, $C9, $D8, $CA]);
              WriteVel(j, aaCode, True);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$DE, $CA, $DE, $C1]);
              WriteAcc(j, aaCode, True);
            end;
          bJ: begin
              IBAppend(aaCode, [$D9, $EA, $DE, $C9]);
              IBAppend(aaCode, b_2ToX_S);
              IBAppend(aaCode, [$D8, $C8]);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$D8, $E9, $D9, $C9]);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$DE, $C1, $DE, $F9, $D9, $C0, $D8, $C8]);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$DE, $E1]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D9, $C9, $D8, $CA]);
              WriteVel(j, aaCode, True);
              IBAppend(aaCode, [$D8, $C0, $D8, $CA, $D8, $C9]);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$DE, $CA, $DE, $E9]);
              WriteAcc(j, aaCode, True);
            end;
          bK: begin
              IBAppend(aaCode, [$D9, $C0, $D9, $E1]);
              IBAppend(aaCode, b_1_comp);
              IBAppend(aaCode, [$77, $28, $DD, $D8]);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$D9, $E0]);
              IBAppend(aaCode, b_LoadPi);
              IBAppend(aaCode, [$D9, $FD, $DD, $D9, $D9, $C9]);
              IBAppend(aaCode, b_zero_comp);
              IBAppend(aaCode, [$DD, $D8, $72, $02, $D9, $E0]);
              IBAppend(aaCode, b_Load0);
              WriteVel(j, aaCode, False);
              WriteAcc(j, aaCode, True);
              IBAppend(aaCode, [$EB, $30, $D8, $C8]);
              IBAppend(aaCode, b_load1);
              IBAppend(aaCode, [$DE, $E1, $D9, $FA]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D8, $F1, $D9, $C0]);
              WriteVel(j, aaCode, True);
              IBAppend(aaCode, [$D8, $C8, $D8, $CA, $D8, $F1]);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$D8, $F2, $DE, $C1]);
              WriteAcc(j, aaCode, True);
              IBAppend(aaCode, [$D9, $F3]);
            end;
          bL: begin
              IBAppend(aaCode, [$D9, $C0, $D9, $E1]);
              IBAppend(aaCode, b_1_comp);
              IBAppend(aaCode, [$77, $20, $DD, $D8]);
              IBAppend(aaCode, b_zero_comp);
              IBAppend(aaCode, [$DD, $D8, $72, $04]);
              IBAppend(aaCode, b_LoadPi);
              IBAppend(aaCode, [$EB, $02]);
              IBAppend(aaCode, b_Load0);
              IBAppend(aaCode, b_Load0);
              WriteVel(j, aaCode, False);
              WriteAcc(j, aaCode, True);
              IBAppend(aaCode, [$EB, $36, $D8, $C8]);
              IBAppend(aaCode, b_load1);
              IBAppend(aaCode, [$DE, $E1, $D9, $FA]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D8, $F1, $D9, $C0, $D9, $E0]);
              WriteVel(j, aaCode, True);
              IBAppend(aaCode, [$D8, $C8, $D8, $CA, $D8, $F1]);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$D8, $F2, $DE, $C1, $D9, $E0]);
              WriteAcc(j, aaCode, True);
              IBAppend(aaCode, [$D9, $C9, $D9, $F3]);
            end;
          bM: begin
              IBAppend(aaCode, [$D9, $C0, $D8, $C8]);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$DE, $C1]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D8, $F1]);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$DE, $F2, $D9, $C0]);
              WriteVel(j, aaCode, True);
              IBAppend(aaCode, [$D8, $C8, $D8, $CA, $D8, $C0, $DE, $E9]);
              WriteAcc(j, aaCode, True);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$D9, $F3]);
            end;
          bN: begin
              IBAppend(aaCode, b_zero_comp);
      		    IBAppend(aaCode, [$76, $1E, $D9, $E0]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D9, $E0]);
              WriteVel(j, aaCode, True);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$D9, $E0]);
              WriteAcc(j, aaCode, True);
            end;
          bO : begin
              IBAppend(aaCode, b_mod_pi);
              IBAppend(aaCode, b_zero_comp);
              IBAppend(aaCode, [$75, $12, $DD, $D8]);
              IBAppend(aaCode, b_Load0);
              WriteVel(j, aaCode, False);
              WriteAcc(j, aaCode, False);
              IBAppend(aaCode, [$EB, $36, $D9, $F2, $DC, $F1,
                                $D9, $C1, $D8, $C8, $DE, $C1]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D9, $C9, $D8, $CA, $D9, $E0]);
              WriteVel(j, aaCode, True);
              IBAppend(aaCode, [$D8, $C0, $D8, $CA, $D8, $C9]);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$DE, $CA, $DE, $E1]);
              WriteAcc(j, aaCode, True);
            end;
          bP: begin
              IBAppend(aaCode, b_1_comp);
              IBAppend(aaCode, [$72, $12, $DD, $D8]);
              IBAppend(aaCode, b_Load0);
              WriteVel(j, aaCode, False);
              WriteAcc(j, aaCode, False);
              IBAppend(aaCode, [$EB, $38, $D9, $C0, $D8, $C8]);
              IBAppend(aaCode, b_load1);
              IBAppend(aaCode, [$DE, $E9, $D9, $FA]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D8, $F1, $D9, $C0]);
              WriteVel(j, aaCode, True);
              IBAppend(aaCode, [$D8, $C8, $D8, $CA, $D8, $F1]);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$D8, $F2, $DE, $E1]);
              WriteAcc(j, aaCode, True);
              IBAppend(aaCode, [$DE, $C1]);
              IBAppend(aaCode, b_ln);
            end;
          bQ: begin
              IBAppend(aaCode, [$D9, $C0, $D9, $E1]);
              IBAppend(aaCode, b_1_comp);
              IBAppend(aaCode, [$77, $1F, $DD, $D8]);
              IBAppend(aaCode, b_zero_comp);
              IBAppend(aaCode, [$DD, $D8]);
              LoadCVPar(2, aaCode);
              IBAppend(aaCode, [$77, $08, $D9, $E0]);
              IBAppend(aaCode, b_Load0);
              WriteVel(j, aaCode, False);
              WriteAcc(j, aaCode, True);
              IBAppend(aaCode, [$EB, $4A, $D8, $C8]);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$DE, $E1]);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$DE, $F1]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D8, $C9, $D9, $C0]);
              WriteVel(j, aaCode, True);
              IBAppend(aaCode, [$D8, $C8, $D8, $C0, $D8, $CA]);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$D8, $CA, $DE, $C1]);
              WriteAcc(j, aaCode, True);
              IBAppend(aaCode, [$D9, $C9]);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$DE, $C1, $D8, $C8, $DE, $C9]);
              IBAppend(aaCode, b_ln);
              IBAppend(aaCode, b_Load1);
              IBAppend(aaCode, [$D9, $E0, $D9, $C9, $D9, $FD, $DD, $D9]);
            end;
          bR: begin
              IBAppend(aaCode, [$D9, $C0, $D8, $C8]);
              IBAppend(aaCode, b_load1);
              IBAppend(aaCode, [$DE, $C1, $D9, $FA]);
              LoadData(endFD + j, aaCode);
              IBAppend(aaCode, [$D8, $F1, $D9, $C0]);
              WriteVel(j, aaCode, True);
              IBAppend(aaCode, [$D8, $C8, $D8, $CA , $D8, $F1]);
              LoadData(endVD+ j, aaCode);
              IBAppend(aaCode, [$D8, $F2, $DE, $E1]);
              WriteAcc(j, aaCode, True);
              IBAppend(aaCode, [$DE, $C1]);
              IBAppend(aaCode, b_ln);
            end;
          else
                  ;
        end;
      end;
    end;
  end; {while}
  IBAppend(aaCode, b_epilog);
  @aa := @aaCode.arr[0];
  stub:= nil;
end;

procedure TEvaluatorBytes_64.DiffAndHessian(x: Vector3; var df: Vector3; var hess: matrix3);
begin
  aa([x[1],x[2],x[3],1,0,0,0,0,0]);
  df[1]:= data[endFd+1];
  hess[1,1]:= data[endVd+1];
  aa([x[1],x[2],x[3],0,1,0,0,0,0]);
  df[2]:= data[endFd+1];
  hess[2,2]:= data[endVd+1];
  aa([x[1],x[2],x[3],0,0,1,0,0,0]);
  df[3]:= data[endFd+1];
  hess[3,3]:= data[endVd+1];
  aa([x[1],x[2],x[3],1,1,0,0,0,0]);
  hess[1,2]:= (data[endVd+1] -  hess[1,1]- hess[2,2])/2;
  hess[2,1]:= hess[1,2];
  aa([x[1],x[2],x[3],0,1,1,0,0,0]);
  hess[2,3]:= (data[endVd+1] - hess[2,2]- hess[3,3])/2;
  hess[3,2]:= hess[2,3];
  aa([x[1],x[2],x[3],1,0,1,0,0,0]);
  hess[1,3]:= (data[endVd+1] - hess[3,3]- hess[1,1])/2;
  hess[3,1]:= hess[1,3];
end;

procedure TEvaluatorBytes_64.DiffAndHessian(x: Vector2; var f: Double; var df: Vector2; var hess: matrix2);
begin
  f:= aa([x[1],x[2],1,0,0,0]);
  df[1]:= data[endFd+1];
  hess[1,1]:= data[endVd+1];
  aa([x[1],x[2],0,1,0,0]);
  df[2]:= data[endFd+1];
  hess[2,2]:= data[endVd+1];
  aa([x[1],x[2],1,1,0,0]);
  hess[1,2]:= (data[endVd+1] -  hess[1,1]- hess[2,2])/2;
  hess[2,1]:= hess[1,2];
end;

function TEvaluatorBytes_64.FirstD(const XV: array of Double): FirstDeriv;
begin
  result.fa0:= vv(XV);
  result.dfa0:= data[endFd+1];
end;


function TEvaluatorBytes_64.EvaluateBSplinesFunction(index: Integer): Double;
  var
    u0, v0, a1, a2: Double;
    k1, k2, var1, var2, i: Integer;
begin
  if index <= $0F then begin{dimension = 1}
    i:= index-1;
    var1:= intList1[i].var1;
    u0:= minCorner[var1]-(cQ+2)*Jump1[var1];
    a1:= (theArg[var1]-u0)*invJump1[var1];
    k1:= Floor(a1);
    k1:= Min(k1, cN1+cQ+2);
    k1:= max(k1,cQ+2);
    result:= EvaluateBSplineOnCurve(4,k1,a1,@intList1[i].Data[0]);
  end
  else begin
    i:= index shr 4 - 1;
    var1:= intList2[i].var1;
    var2:= intList2[i].var2;
    u0:= minCorner[var1]-(cQ+2)*Jump2[var1];
    v0:= minCorner[var2]-(cQ+2)*Jump2[var2];
    a1:= (theArg[var1]-u0)*invJump2[var1];
    a2:= (theArg[var2]-v0)*invJump2[var2];
    k1:= Floor(a1);
    k1:= Min(k1, cN2+cQ+2);
    k1:= max(k1,cQ+2);
    k2:= Floor(a2);
    k2:= Min(k2, cN2+cQ+2);
    k2:= max(k2,cQ+2);
    result:= EvaluateBSplineOnSurface(4,4,k1,k2,a1,a2,@intList2[i].Data[0,0]);
  end
end;

function TEvaluatorBytes_64.EvaluateBSplinesFunctionAndFirstDer(index: Integer):Double;
  var
    u0, v0, a1, a2, aux: Double;
    k1, k2, var1, var2: Integer;
begin
  if index <= $0F then begin{dimension = 1}
    dec(index);
    var1:= intList1[index].var1;
    u0:= minCorner[var1]-(cQ+2)*Jump1[var1];
    a1:= (theArg^[var1]-u0)*invJump1[var1];
    k1:= Floor(a1);
    k1:= Min(k1, cN1+cQ+2);
    k1:= max(k1,cQ+2);
    {evaluate the integral value with a quartic B-spline}
    result:= EvaluateBSplineOnCurve(4,k1,a1,@intList1[index].Data[0]);
    data[endFd + currentJ]:= theArg^[var1+ argNum]*EvaluateBSplineOnCurve(3,k1,a1,@intList1[index].Data[1]);
  end
  else begin
    index:= index shr 4 - 1;
    var1:= intList2[index].var1;
    var2:= intList2[index].var2;
    u0:= minCorner[var1]-(cQ+2)*Jump2[var1];
    v0:= minCorner[var2]-(cQ+2)*Jump2[var2];
    a1:= (theArg^[var1]-u0)*invJump2[var1];
    a2:= (theArg^[var2]-v0)*invJump2[var2];
    k1:= Floor(a1);
    k1:= Min(k1, cN2+cQ+2);
    k1:= max(k1,cQ+2);
    k2:= Floor(a2);
    k2:= Min(k2, cN2+cQ+2);
    k2:= max(k2,cQ+2);
    {evaluate integral}
    result:= EvaluateBSplineOnSurface(4,4,k1,k2,a1,a2,@intList2[index].Data[0,0]);
     {evaluate partial derivative of the integral with respect to each of
      its two variables}
    data[endFd + currentJ]:= 0;
    aux:= theArg^[var1+ argNum];
    if aux <> 0 then
      data[endFd + currentJ]:= aux*EvaluateBSplineOnSurface(3,4,k1,k2,a1,a2,@intList2[index].Data[1,0]);
    aux:= theArg^[var2+ argNum];
    if aux <> 0 then
      data[endFd + currentJ]:= data[endFd + currentJ] + aux*EvaluateBSplineOnSurface(4,3,k1,k2,a1,a2,@intList2[index].Data[0,1]);
  end
end;

function TEvaluatorBytes_64.EvaluateBSplinesFunctionFirstAndSecondDer(index: Integer):Double;
var
    u0, v0, a1, a2, d1,d2,d3: Double;
    k1, k2, var1, var2: Integer;
begin
  if index <= $0F then begin{dimension = 1}
    dec(index);
    var1:= intList1[index].var1;
    u0:= minCorner[var1]-(cQ+2)*Jump1[var1];
    a1:= (theArg^[var1]-u0)*invJump1[var1];
    k1:= Floor(a1);
    k1:= Min(k1, cN1+cQ+2);
    k1:= max(k1,cQ+2);
    result:= EvaluateBSplineOnCurve(4,k1,a1,@intList1[index].Data[0]);
    d1:= EvaluateBSplineOnCurve(3,k1,a1,@intList1[index].Data[1]);
    data[endFd + currentJ]:= d1 * theArg^[var1+ argNum];
    data[endVd + currentJ]:= d1 * theArg^[var1 + 2*argNum];
    data[endVd + currentJ]:= data[endVd + currentJ] +
       sqr(theArg^[var1+ argNum])*EvaluateBSplineOnCurve(2,k1,a1,@intList1[index].Data[2]);
  end
  else begin
    index:= index shr 4 - 1;
    var1:= intList2[index].var1;
    var2:= intList2[index].var2;
    u0:= minCorner[var1]-(cQ+2)*Jump2[var1];
    v0:= minCorner[var2]-(cQ+2)*Jump2[var2];
    a1:= (theArg^[var1]-u0)*invJump2[var1];
    a2:= (theArg^[var2]-v0)*invJump2[var2];
    k1:= Floor(a1);
    k1:= Min(k1, cN2+cQ+2);
    k1:= max(k1,cQ+2);
    k2:= Floor(a2);
    k2:= Min(k2, cN2+cQ+2);
    k2:= max(k2,cQ+2);
    result:= EvaluateBSplineOnSurface(4,4,k1,k2,a1,a2,@intList2[index].Data[0,0]);
    d1:= EvaluateBSplineOnSurface(3,4,k1,k2,a1,a2,@intList2[index].Data[1,0]);
    d2:= EvaluateBSplineOnSurface(4,3,k1,k2,a1,a2,@intList2[index].Data[0,1]);
    data[endFd + currentJ]:= d1*theArg^[var1+ argNum]+ d2*theArg[var2 + argNum];
    data[endVd + currentJ]:= d1*theArg^[var1 + 2*argNum]+ d2*theArg[var2 + 2*argNum];
    d1:= EvaluateBSplineOnSurface(2,4,k1,k2,a1,a2,@intList2[index].Data[2,0]);
    d2:= EvaluateBSplineOnSurface(4,2,k1,k2,a1,a2,@intList2[index].Data[0,2]);
    d3:= EvaluateBSplineOnSurface(3,3,k1,k2,a1,a2,@intList2[index].Data[1,1]);
    data[endVd + currentJ]:= data[endVd + currentJ] + d1*sqr(theArg^[var1+ argNum])+ d2*sqr(theArg[var2 + argNum])+
             2*d3*theArg^[var1 + argNum]*theArg^[var2 + argNum] ;
  end
end;

procedure TEvaluatorBytes_64.CodifyFunction;
var
  i, n, a, pushes: Integer;
  ch: Byte;
  {$IFDEF MSWINDOWS}
    prot: Word;
  {$ELSE IF LINUX}
  prot: cInt;
  {$ENDIF}
  stub: ByteArray;
begin
  lastPreserved:= 1;
{$IFDEF MSWINDOWS}
  {if ffBlock = nil then begin
    ffBlockLength:= ComputeFCodeLength;
    prot:= PAGE_EXECUTE_READWRITE;
    ffBlock:= VirtualAlloc(nil, ffBlockLength, MEM_COMMIT, prot);
  end
  else begin
    prot:= PAGE_EXECUTE_READ;
    VirtualProtect(ffBlock, ffBlockLength, PAGE_EXECUTE_READWRITE, @prot);
  end;}
  if ffBlock <> nil then begin
    VirtualFree(ffBlock, 0, MEM_RELEASE);
    ffBlock:= nil
  end;
  ffBlockLength:= ComputeFCodeLength;
  prot:= PAGE_EXECUTE_READWRITE;
  ffBlock:= VirtualAlloc(nil, ffBlockLength, MEM_COMMIT, prot);
{$ELSE IF LINUX}
  if ffBlock <> nil then
    if Fpmunmap(ffBlock, ffBlockLength) <> 0 then
      raise exception.Create(eBadIntegralSyntax);
  prot:= PROT_EXEC OR PROT_READ OR PROT_WRITE;
  ffBlockLength:=  ComputeFCodeLength;
  ffBlock:= FPmMAP(nil, ffBlockLength, prot, MAP_PRIVATE or MAP_ANONYMOUS, 0, 0);
{$ENDIF}
  ffCode.arr:= ffBlock;
  ffCode.h:= -1;
  ffCode.bigH:= ffBlockLength-1;
  IBAppend(ffCode, b_Prolog);{11 bytes. initiate 'FBlock' with the prologue. }
  SetLength(stub, SizeOf(Pointer));
  if (lastIntegral1 < 0) and (lastIntegral2 < 0) then begin
    {Save the address of 'data' to rax}
    IBAppend(ffCode, [$48, $B8]); {2 bytes}
    sn(@data[0], stub);
    IBAppend(ffCode, stub); {SizeOf(Pointer)}
  end
  else begin
    {put the address of 'theArg[0]' to rax}
    IBAppend(ffCode, [$48, $B8]); {2 bytes}
    sn(@theArg, stub);
    IBAppend(ffCode, stub);

    {mov [rax], rdx. This makes  @theArg[0]:= @X[0]}
    IBAppend(ffCode,[$48, $89, $10]);

    {Save the address of 'data' to rax}
    IBAppend(ffCode, [$48, $B8]); {2 bytes}
    sn(@data[0], stub);
    IBAppend(ffCode, stub); {SizeOf(Pointer)}

    IBAppend(ffCode, [$48,$89, $95]); {mov [rbp+c_sfr+$10], rdx.}
    IBAppend(ffCode, Int32Str(c_sfr+$10));

    {mov [rbp+c_sfr+$$18],rax. This preserves rax}
    IBAppend(ffCode, [$48,$89, $85]);
    IBAppend(ffCode, Int32Str(c_sfr+$18));

    {mov rcx, self}
    IBAppend(ffCode, [$48,$B9]);
    sn(self, stub);
    IBAppend(ffCode, stub); {SizeOf(Pointer)}

    {mov [rbp+c_sfr+$20],rcx. This preserves rcx}
    IBAppend(ffCode, [$48,$89, $8D]);
    IBAppend(ffCode, Int32Str(c_sfr+$20));
  end;
  emptyRegisters := c_LastReg + 1;
  i:= -1;
  while i < High(ActorList) do
    begin
      inc(i);
      ch := actorList[i];
      if ch in [bLastByte, bPi, b1] then begin
        if ch <> bLastByte then
          LeaveRoomInFPUStack(1, ffCode);
        case ch of
          bLastByte: if actorList[i+1] < $04 then begin
              LeaveRoomInFPUStack(1, ffCode);
              SetLength(stub, 4);
              stub:= ExtractBA(@actorList[0], i, 4);
              OCRBytes(stub, ffCode);
              i:= i+3;
              dec(emptyRegisters);
            end
            else begin
              inc(i, 3);
              n:= actorList[i];
              if n < 0 then
                exception.Create(eBadIntegralSyntax);
              pushes:= c_LastReg + 1 - emptyRegisters;
              LeaveRoomInFPUStack(c_LastReg+1, ffCode);  {empts FPU}

              {put in EDX, the order of the integral in 'intLis',
               where 'EvaluateBSplinesFunction' will expect it}
              IBAppend(ffCode,  [$C7,$C2]);
              IBAppend(ffCode, Int32Str(n));

              {mov rax, TEvaluator_64.EvaluateBSplinesFunction}
              IBAppend(ffCode, [$48, $B8]);
              SetLength(stub, SizeOf(Pointer));
              sn(@TEvaluatorBytes_64.EvaluateBSplinesFunction, stub);
              IBAppend(ffCode, stub);
              IBAppend(ffCode, [$FF, $D0]); {call [rax]};

              {it is supposed that the called function leaves its result in XMM0,
               and we put that value into ST(0) of the FPU}
              IBAppend(ffCode, [$f2, $0f, $11, $85]);
              IBAppend(ffCode, Int32Str(c_sfr-$08));  {movsd [rbp+c_sfr-$08], xmm0}
              IBAppend(ffCode, [$DD, $85]);
              IBAppend(ffCode, Int32Str(c_sfr-$08));  {fld [rbp+c_sfr-$08]. Thus ST(0) contains result of
                                       call to EvaluateBSplinesFunction}

              emptyRegisters := c_LastReg;
              if pushes > 0 then
                for a:= 1 to pushes do
                  PopFromAppStack(ffCode);

              {recover initial values of rcx, rdx and rax}
              IBAppend(ffCode, [$48, $8B, $8D]);
              IBAppend(ffCode, Int32Str(c_sfr-$20));
              IBAppend(ffCode, [$48, $8B, $95]);
              IBAppend(ffCode, Int32Str(c_sfr+$10));
              IBAppend(ffCode, [$48, $8B, $85]);
              IBAppend(ffCode, Int32Str(c_sfr+$18));
            end;
          bPi: begin
                 IBAppend(ffCode, b_LoadPi);
                 dec(emptyRegisters);
               end;
          b1: begin
                 IBAppend(ffCode, b_Load1);
                 dec(emptyRegisters);
               end
          else
            ;
        end;
      end
      else
        begin
          if (ch in [bPlus, bMinus, bTimes, bDiv, bPow]) then
            begin
              if (emptyRegisters = c_LastReg) then
                PopFromAppStack(ffCode);
              case ch of
                bPlus : IBAppend(ffCode, b_plus_code) ;
                bMinus : IBAppend(ffCode, b_minus_code) ;
                bTimes : IBAppend(ffCode, b_times_code) ;
                bDiv :
                  begin
                    LeaveRoomInFPUStack(1, ffCode);
                    IBAppend(ffCode, b_div_S) ;
                  end;
                bPow :  ;{This shall not be used. Used exp(#2 * ln#1) <-- #1^#2}
              end;
              inc(emptyRegisters);
            end
          else {opCode of a function}
            begin
              if (ch in [bA, bB, bC, bE, bF, bG, bM]) then
                LeaveRoomInFPUStack(1, ffCode)
              else if (ch in [bH, bI, bJ, bK, bL, bO, bP, bQ, bR]) then
                LeaveRoomInFPUStack(2, ffCode);
              case ch of
                bNeg : IBAppend(ffCode, b_neg_code);
                bA : IBAppend(ffCode, b_cos_code);
                bB : IBAppend(ffCode, b_sin_code);
                bC : begin
                      IBAppend(ffCode, b_times_Log2_e );
                      IBAppend(ffCode,  b_2ToX_S);
                     end;
                bD : IBAppend(ffCode, b_Sqr);
                bE : IBAppend(ffCode, b_Sqrt);
                bF : IBAppend(ffCode, b_ln_S);
                bG : IBAppend(ffCode, b_tan);
                bH : begin
                       IBAppend(ffCode, [$D9, $EA, $DE, $C9]);
                       IBAppend(ffCode, b_2ToX_S);
                       IBAppend(ffCode, [$D9, $C0, $D9, $E8, $DE, $F1, $DE, $E9]);
                       IBAppend(ffCode, b_div2);
                     end;
                bI : begin
                       IBAppend(ffCode, [$D9, $EA, $DE, $C9]);
                       IBAppend(ffCode, b_2ToX_S);
                       IBAppend(ffCode, [$D9, $C0, $D9, $E8, $DE, $F1, $DE, $C1]);
                       IBAppend(ffCode, b_div2);
                     end;
                bJ : begin
                       IBAppend(ffCode, [$D9, $EA, $DE, $C9]);
                       IBAppend(ffCode, b_2ToX_S);
                       IBAppend(ffCode, [$D8, $C8, $D9, $E8, $D8, $E9, $D9, $C9,
                                        $D9, $E8, $DE, $C1, $DE, $F9]);
                     end;
                bK : begin
                       IBAppend(ffCode, [$D9, $C0, $D9, $E1]);
                       IBAppend(ffCode, b_1_comp);
                       IBAppend(ffCode, [$DD, $D8, $72, $0E, $D9, $C0, $D8, $C8, $D9, $E8, $DE,
                               $E1, $D9, $FA, $D9, $F3, $EB, $14]);
                       IBAppend(ffCode, b_zero_comp);
                       IBAppend(ffCode, [$DD, $D8]);
                       IBAppend(ffCode, b_LoadHalfPi);
                       IBAppend(ffCode, [$72, $02, $D9, $E0]);
                     end;
                bL : begin
                       IBAppend(ffCode, [$D9, $C0, $D9, $E1]);
                       IBAppend(ffCode, b_1_comp);
                       IBAppend(ffCode, [$DD, $D8, $72, $10, $D9, $C0, $D8, $C8, $D9, $E8, $DE,
                               $E1, $D9, $FA, $D9, $C9, $D9, $F3, $EB, $0E]);
                       IBAppend(ffCode, b_zero_comp);
                       IBAppend(ffCode, [$DD, $D8, $72, $04, $D9, $EB, $EB, $02, $D9, $EE]);
                     end;
                bM : IBAppend(ffCode, [$D9, $E8, $D9, $F3]);
                bN : IBAppend(ffCode, [$D9, $E1]);
                bO : begin
                       IBAppend(ffCode, b_mod_pi);
                       IBAppend(ffCode, [$D9, $FB, $D9, $C9]);
                       IBAppend(ffCode, b_div_S);
                     end;
                bP : begin
                       IBAppend(ffCode, b_1_comp);
                       IBAppend(ffCode,[$77, $14, $D9, $C0, $D8, $C8, $D9, $E8, $DE, $E9,
                               $D9, $FA, $DE, $C1]);
                       IBAppend(ffCode, b_ln);
                       IBAppend(ffCode,[$EB, $04, $DD, $D8, $D9, $EE]);
                     end;
                bQ : begin
                       IBAppend(ffCode,[$D9, $C0, {fld(0), r in ST(0) and in ST(1)}
                           $D9, $E1 {abs, |r| in ST(0), r in ST(1)}]);
                       IBAppend(ffCode, b_1_comp);
                       IBAppend(ffCode,[$DD, $D8,{pop(0), r in ST(0)}
                           $76, $1C, {if |r| >=1, jump $1C= 28 bytes}
                           $D9, $C0, $D9, $E8, $DE, $C1, $D9, $C9,
                           $D9, $E8, $DE, $E1, $DE, $F9 {(1+r)/(1-r) in ST(0)}]);
                       IBAppend(ffCode, b_ln {2 atanh(r) in ST(0)});
                       IBAppend(ffCode, b_div2 {atanh(r) in ST(0)});
                       IBAppend(ffCode,[$EB, $0F {jump 15 bytes}]);
                       IBAppend(ffCode, b_zero_comp);
                       IBAppend(ffCode,[$DD, $D8,  {pop(0), nothing in ST(0)}
                              $77, $05] {if r < 0, jump 5 bytes} );
                       LoadCVPar(5, ffCode);
                       IBAppend(ffCode,[$EB, $02 {jump 2 bytes}]);
                       LoadCVPar(0, ffCode);
                     end;
                bR : begin
                       IBAppend(ffCode, b_zero_comp);
                       IBAppend(ffCode,[$D9, $E1, $D9, $C0, $D8, $C8, $D9, $E8, $DE, $C1,
                                $D9, $FA, $DE, $C1]);
                       IBAppend(ffCode, b_ln);
                       IBAppend(ffCode,[$76, $02, $D9, $E0]);
                     end;
                else
                  raise Exception.Create(eErrorInParsing);
                end;
            end;
        end;
    end; {del while}
  IBAppend(ffCode, b_Epilog);
  @ff := @ffCode.arr[0];
  stub:= nil;
end;

function TEvaluatorBytes_64.ComputeDDFCodeLength: Integer;
  var
    i, j, a, n, pushes: Integer;
    ch: Byte;
  begin
  result:= 0;
  inc(result, 11); {IBAppend(aaCode, b_Prolog)}
  if (lastIntegral1 < 0) and (lastIntegral2 < 0) then begin
    Inc(result, 2); {IBAppend(aaCode, [$48, $B8])}
    Inc(result, SizeOf(Pointer));
  end
  else begin
  (*Inc(result,2); {IBAppend(aaCode, [$48, $B8])}
    Inc(result, SizeOf(Pointer)); {IBAppend(aaCode, stub)}
    Inc(result, 3); {IBAppend(aaCode,[$48, $89, $10])}
    Inc(result,2); {IBAppend(aaCode, [$48, $B8])}
    Inc(result, SizeOf(Pointer)); {IBAppend(aaCode, stub)
    Inc(result, 3); {IBAppend(aaCode, [$48,$89, $95])}
    Inc(result, 4); {IBAppend(aaCode, Int32Str(c_sfr+$10))}
    Inc(result, 3); {IBAppend(aaCode, [$48,$89, $85])}
    Inc(result, 4); {IBAppend(aaCode, Int32Str(c_sfr+$18))}
    Inc(result, 2); {IBAppend(aaCode, [$48,$B9])}
    Inc(result, SizeOf(Pointer)); {IBAppend(aaCode, stub)}
    Inc(result, 3); {IBAppend(aaCode, [$48, $89, $8D])}
    Inc(result, 4); {IBAppend(aaCode, Int32Str(c_sfr+$20))} *)

    Inc(result, 3*SizeOf(Pointer) + 30);
  end;

  emptyRegisters := c_LastReg + 1;
  i:= -1;
  j:= 0;
  while i < High(actorList) do  begin
    inc(i);
    ch := actorList[i];
    if ch in [bLastByte, bPi, b1] then begin
      Inc(result, LengthLeaveRoom(1)); {LeaveRoomInFPUStack(1, aaCode)}
      inc(j);
      case ch of
        bLastByte: begin
          ch:= actorList[i+1];
          i:= i + 3;
          n:= actorList[i];
          case ch of
            $01: begin
              (*Inc(result, 3); {LoadV(n, aaCode)}
                Inc(result, 6); {WriteVel(j, aaCode, True)}
                Inc(result, 3); {LoadA(n, aaCode)}
                Inc(result, 6); {WriteAcc(j, aaCode, True)}
                Inc(result, 3); {LoadX(n, aaCode)} *)

                Inc(result, 21);
                dec(emptyRegisters);
              end;
            $02, $03: begin
                n:= ord(actorList[i-1]) shl 8 + n;
              (*Inc(result, 2); {IBAppend(aaCode, b_Load0)}
                Inc(result, 6); {WriteVel(j, aaCode, False)}
                Inc(result, 6); {WriteAcc(j, aaCode, True)}
                Inc(result, 6); {LoadData(n, aaCode)} *)

                Inc(result, 20);
                dec(emptyRegisters);
              end;
            $04:  begin{an integral}
                pushes:= c_LastReg + 1 - emptyRegisters;

                Inc(result, LengthLeaveRoom(c_LastReg+1)); {LeaveRoomInFPUStack(c_LastReg+1, aaCode)}

              (*Inc(result, 2); {IBAppend(aaCode, [$C7, $C2])}
                Inc(result, 4); {AddInt32ToCode(n, aaCode)}

                Inc(result, 2); {IBAppend(aaCode, [$48, $B8])}
                Inc(result, SizeOf(Pointer)); {IBAppend(aaCode, stub)}
                Inc(result, 3); {IBAppend(aaCode, [$48, $C7, $00])}
                Inc(result, 4); {AddInt32ToCode(j, aaCode)}

                Inc(result, 2); {IBAppend(aaCode, [$48, $B8]);}
                Inc(result, SizeOf(Pointer)); {sn}

                Inc(result, 2); {IBAppend(aaCode, [$FF, $D0])}

  
                Inc(result, 4); {IBAppend(aaCode, [$f2, $0f, $11, $85])}
                Inc(result, 4); {IBAppend(aaCode, Int32Str(c_sfr-$08))}
                Inc(result, 2); {IBAppend(aaCode, [$DD, $85])}
                Inc(result, 4); {IBAppend(aaCode, Int32Str(c_sfr-$08))} *)

                Inc(result, 2*SizeOf(Pointer) + 33);
                                             
                emptyRegisters := c_LastReg;
                if pushes > 0 then
                  for  a:= 1 to pushes do
                     Inc(result, LengthPopFromStack);

              (*Inc(result, 3); {IBAppend(aaCode, [$48, $8B, $8D])}
                Inc(result, 4); {IBAppend(aaCode, Int32Str(-$20))}
                Inc(result, 3); {IBAppend(aaCode, [$48, $8B, $95])}
                Inc(result, 4); {IBAppend(aaCode, Int32Str(-$10))}
                Inc(result, 3); {IBAppend(aaCode, [$48, $8B, $85])}
                Inc(result, 4); {IBAppend(aaCode, Int32Str(-$18))} *)

                Inc(result, 21);
              end;
            else
              ;
          end;
        end;
        bPi: begin
          (*Inc(result, 2); {IBAppend(aaCode, b_Load0)}
            Inc(result, 6); {WriteVel(j, aaCode, False)}
            Inc(result, 6); {WriteAcc(j, aaCode, True)}
            Inc(result, 2); {IBAppend(aaCode, b_LoadPi)} *)

            Inc(result, 16);
            dec(emptyRegisters);
          end;
        b1: begin
          (*Inc(result, 2); {IBAppend(aaCode, b_Load0)}
            Inc(result, 6); {WriteVel(j, aaCode, False)}
            Inc(result, 6); {WriteAcc(j, aaCode, True)}
            Inc(result, 2); {IBAppend(aaCode, b_Load1)}*)

            Inc(result, 16);
            dec(emptyRegisters);
          end;
        else
                  ;
      end;
    end
    else begin {opcode of a binary operator}
      if (ch in [bPlus, bMinus, bTimes, bDiv, bPow]) then begin
        if (emptyRegisters = c_LastReg) then
          Inc(result, LengthPopFromStack); {PopFromAppStack(aaCode)}
        if ch in [bPlus, bMinus]  then
          Inc(result, LengthLeaveRoom(1))
        else
          Inc(result, LengthLeaveRoom(3));
        case ch of
          bPlus: begin
            (*Inc(result, 2); {IBAppend(aaCode, [$DE, $C1])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 6); {LoadData(endFD + j-1, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $C1])}
              Inc(result, 6); {WriteVel(j-1, aaCode, True)}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 6); {LoadData(endVD+ j-1, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $C1])}
              Inc(result, 6); {WriteAcc(j-1, aaCode, True)}*)

              Inc(result, 42);
            end;
          bMinus: begin
            (*Inc(result, 2); {IBAppend(aaCode, [$DE, $E9])}
              Inc(result, 6); {LoadData(endFD + j-1, aaCode)}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $E9])}
              Inc(result, 6); {WriteVel(j-1, aaCode, True)}
              Inc(result, 6); {LoadData(endVD+ j-1, aaCode)}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $E9])}
              Inc(result, 6); {WriteAcc(j-1, aaCode, True)} *)

              Inc(result, 42);
            end;
          bTimes: begin
              (* Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 6); {LoadData(endFD + j-1, aaCode)}
              Inc(result, 16);{IBAppend(aaCode, [$D9, $C0, $D8, $CA, $D8, $C0, $D9, $CA,
                                                 $D8, $CC, $D9, $C9, $D8, $CB, $DE, $C1])}
              Inc(result, 6); {WriteVel(j-1, aaCode, True)}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $CB, $DE, $C1])}
              Inc(result, 6); {LoadData(endVD+ j-1, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $CA, $DE, $C1])}
              Inc(result, 6); {WriteAcc(j-1, aaCode, True)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $C9])} *)

              Inc(result, 62);
            end;
          bDiv: begin
            (*Inc(result, 4); {IBAppend(aaCode, b_zero_comp)}
              Inc(result, 6); {IBAppend(aaCode, [$75, $14, $DD, $D8, $DD, $D8])}
              Inc(result, 2); {IBAppend(aaCode, b_Load0)}
              Inc(result, 6); {WriteVel(j-1, aaCode, False)}
              Inc(result, 6); {WriteAcc(j-1, aaCode, False)}
              Inc(result, 2); {IBAppend(aaCode, [$EB, $48])}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $F1, $DC, $C9])}
              Inc(result, 6); {LoadData(endFD + j-1, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$D8, $C9])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 8); {IBAppend(aaCode, [$D8, $CA, $D9, $C0, $D8, $CC, $D8, $EA])}
              Inc(result, 6); {WriteVel(j-1, aaCode, True)}
              Inc(result, 10); {IBAppend(aaCode, [$DC, $C9, $D8, $C8, $D8, $CB, $DE, $E1, $D8, $C0])}
              Inc(result, 6); {LoadData(endVD+ j-1, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $CA, $DE, $C1])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 6); {IBAppend(aaCode, [$D8, $CB, $DE, $CA, $DE, $E1])}
              Inc(result, 6); {WriteAcc(j-1, aaCode, True)} *)

              Inc(result, 98);
            end;
          else
             ;
        end;
        inc(emptyRegisters);
        j:= j-1;
      end
      else begin {opCode of a function}
        if ch in [bNeg,bN] then
          Inc(result, LengthLeaveRoom(1))
        else if ch= bC then
          Inc(result, LengthLeaveRoom(2))
        else
          Inc(result, LengthLeaveRoom(3));

        case ch of
          bNeg: begin
           (* Inc(result, 2); {IBAppend(aaCode, [$D9, $E0])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$D9, $E0])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$D9, $E0])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)} *)

              Inc(result, 30);
            end;
          bA: begin
           (* Inc(result, 10); {IBAppend(aaCode, b_mod_2pi)}
              Inc(result, 6); {IBAppend(aaCode, [$D9, $FB, $D9, $C9, $D9, $E0])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 10); {IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D8, $CB, $D9, $C9, $D8, $CA])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $CA, $DE, $E9])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)} *)

              Inc(result, 54);
            end;
          bB: begin
            (*Inc(result, 10); {IBAppend(aaCode, b_mod_2pi)}
              Inc(result, 2); {IBAppend(aaCode, [$D9, $FB])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 10); {IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D8, $CB, $D9, $C9, $D8, $CA])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $CA, $DE, $E9])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)} *)

              Inc(result, 50);
            end;
          bC: begin
           (* Inc(result, 4); {IBAppend(aaCode, [$D9, $EA, $DE, $C9 ])}
              Inc(result, 49); {IBAppend(aaCode, b_2ToX_S)}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D9, $C0, $D8, $CA])}
              Inc(result, 6); {WriteVel(j, aaCode,True)}
              Inc(result, 2); {IBAppend(aaCode, [$D8,$C8])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $C1])}
              Inc(result, 2); {IBAppend(aaCode, [$D8, $C9])}
              Inc(result, 6); {WriteAcc(j, aaCode, True);} *)

              Inc(result, 87);
            end;
          bD: begin
           (* Inc(result, 8); {IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D9, $C9, $D8, $C0])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$D8, $C9])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$DC, $CA, $D9, $CA])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 8); {IBAppend(aaCode, [$D9, $C9, $D8, $C8, $D8, $C0, $DE,$C1])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)} *)

              Inc(result, 46);
            end;
          bE: begin
            (*Inc(result, 4); {IBAppend(aaCode, b_zero_comp)}
              Inc(result, 4); {IBAppend(aaCode, [$72, $12, $DD, $D8])}
              Inc(result, 2); {IBAppend(aaCode, b_Load0)}
              Inc(result, 6); {WriteVel(j, aaCode, False)}
              Inc(result, 6); {WriteAcc(j, aaCode, False)}
              Inc(result, 8); {IBAppend(aaCode, [$EB, $32, $D9, $FA, $D9, $C0, $D8, $C0])}
      		    Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $F1, $D9, $C0])}
      		    Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $C9, $D9, $C0])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {IBAppend(aaCode, [$D8, $C8, $D8, $C9, $D8, $C0])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
        		  Inc(result, 4); {IBAppend(aaCode, [$DE, $CA, $DE, $E9])}
      		    Inc(result, 6); {WriteAcc(j, aaCode, True)} *)

              Inc(result, 74);
            end;
          bF: begin
            (*Inc(result, 4); {IBAppend(aaCode, b_zero_comp)}
              Inc(result, 4); {IBAppend(aaCode, [$72, $17, $DD, $D8])}
              Inc(result, 3); {LoadCVPar(5, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$D9, $E0])}
              Inc(result, 2); {IBAppend(aaCode, b_Load0)}
              Inc(result, 6); {WriteVel(j, aaCode, False)}
              Inc(result, 6); {WriteAcc(j, aaCode, True)}
              Inc(result, 4); {IBAppend(aaCode, [$EB, $30, $D9, $C0])}
              Inc(result, 6); {IBAppend(aaCode, b_ln)}
              Inc(result, 2); {IBAppend(aaCode, [$D9, $C9])}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $F1])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $C9, $D9, $C0])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 2); {IBAppend(aaCode, [$D8, $C8])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $CA, $DE, $E9])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)} *)

              Inc(result, 77);
            end;
          bG: begin
            (*Inc(result, 8); {IBAppend(aaCode, b_mod_pi)}
              Inc(result, 8); {IBAppend(aaCode, [$D9, $F2, $D9, $C1, $D8, $C8, $DE, $C1])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 8); {IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D9, $C9, $D8, $CA])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {IBAppend(aaCode, [$D8, $C0, $D8, $CA, $D8, $C9])}
      		    Inc(result, 6); {LoadData(endVD+ j, aaCode)}
      		    Inc(result, 4); {IBAppend(aaCode, [$DE, $CA, $DE, $C1])}
              Inc(result, 6); {WriteAcc(j, aaCode,True)} *)

              Inc(result, 58);
            end;
          bH: begin
              (* Inc(result, 4); {IBAppend(aaCode, [$D9, $EA, $DE, $C9])}
              Inc(result, 49); {IBAppend(aaCode, b_2ToX_S)}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 22); {IBAppend(aaCode, [$D9, $E0, $D9, $C1, $D8, $F9, $D9, $FD, $D9, $CA,
                                $D9, $FD, $DD, $D9, $D9, $C0, $D8, $C2, $D9, $CA, $DE, $E9])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 10); {IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D8, $CB, $D9,$C9, $D8, $CA]);}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $CA, $DE, $C1])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)} *)

              Inc(result, 115);
            end;
          bI: begin
            (*Inc(result, 4); {IBAppend(aaCode, [$D9, $EA, $DE, $C9])}
      		    Inc(result, 49); {IBAppend(aaCode, b_2ToX_S)}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 24); {IBAppend(aaCode, [$D9, $E0, $D9, $C1, $D8, $F9, $D9, $FD, $D9, $CA, $D9, $FD,
                                            $DD, $D9, $D9, $C0, $D8, $C2, $D9, $CA, $DE, $E9, $D9, $C9])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 10); {IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D8, $CB, $D9, $C9, $D8, $CA])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $CA, $DE, $C1])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)} *)

              Inc(result, 117);
            end;
          bJ: begin
              (*Inc(result, 4); {IBAppend(aaCode, [$D9, $EA, $DE, $C9])}
              Inc(result, 49); {IBAppend(aaCode, b_2ToX_S)}
              Inc(result, 2); {IBAppend(aaCode, [$D8, $C8])}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $E9, $D9, $C9])}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 8); {IBAppend(aaCode, [$DE, $C1, $DE, $F9, $D9, $C0, $D8, $C8])}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $E1])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 8); {IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D9, $C9, $D8, $CA])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {IBAppend(aaCode, [$D8, $C0, $D8, $CA, $D8, $C9])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $CA, $DE, $E9])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)} *)

              Inc(result, 117);
            end;
          bK: begin
            (*Inc(result, 4); {IBAppend(aaCode, [$D9, $C0, $D9, $E1])}
              Inc(result, 4); {IBAppend(aaCode, b_1_comp)}
              Inc(result, 4); {IBAppend(aaCode, [$77, $28, $DD, $D8])}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 2); {IBAppend(aaCode, [$D9, $E0])}
              Inc(result, 2); {IBAppend(aaCode, b_LoadPi)}
              Inc(result, 6); {IBAppend(aaCode, [$D9, $FD, $DD, $D9, $D9, $C9])}
              Inc(result, 4); {IBAppend(aaCode, b_zero_comp)}
              Inc(result, 6); {IBAppend(aaCode, [$DD, $D8, $72, $02, $D9, $E0])}
              Inc(result, 2); {IBAppend(aaCode, b_Load0)}
              Inc(result, 6); {WriteVel(j, aaCode, False)}
              Inc(result, 6); {WriteAcc(j, aaCode, True)}
              Inc(result, 4); {IBAppend(aaCode, [$EB, $30, $D8, $C8])}
              Inc(result, 2); {IBAppend(aaCode, b_load1)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $E1, $D9, $FA])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $F1, $D9, $C0])}
              Inc(result, 6); {WriteVel(j, aaCode, True);}
              Inc(result, 6); {IBAppend(aaCode, [$D8, $C8, $D8, $CA, $D8, $F1])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $F2, $DE, $C1])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)}
              Inc(result, 2); {IBAppend(aaCode, [$D9, $F3])} *)

              Inc(result, 98);
            end;
          bL: begin
            (*Inc(result, 4); {IBAppend(aaCode, [$D9, $C0, $D9, $E1])}
              Inc(result, 4); {IBAppend(aaCode, b_1_comp)}
              Inc(result, 4); {IBAppend(aaCode, [$77, $20, $DD, $D8])}
              Inc(result, 4); { IBAppend(aaCode, b_zero_comp)}
              Inc(result, 4); {IBAppend(aaCode, [$DD, $D8, $72, $04])}
              Inc(result, 2); {IBAppend(aaCode, b_LoadPi)}
              Inc(result, 2); {IBAppend(aaCode, [$EB, $02])}
              Inc(result, 2); {IBAppend(aaCode, b_Load0)}
              Inc(result, 2); {IBAppend(aaCode, b_Load0)}
              Inc(result, 6); {WriteVel(j, aaCode, False)}
              Inc(result, 6); {WriteAcc(j, aaCode, True)}
              Inc(result, 4); {IBAppend(aaCode, [$EB, $36, $D8, $C8])}
              Inc(result, 2); {IBAppend(aaCode, b_load1)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $E1, $D9, $FA])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 6); {IBAppend(aaCode, [$D8, $F1, $D9, $C0, $D9, $E0])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {IBAppend(aaCode, [$D8, $C8, $D8, $CA, $D8, $F1])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 6); {IBAppend(aaCode, [$D8, $F2, $DE, $C1, $D9, $E0])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)}
              Inc(result, 4); {IBAppend(aaCode, [$D9, $C9, $D9, $F3])} *)

              Inc(result, 96);
            end;
          bM: begin
            (*Inc(result, 4); {IBAppend(aaCode, [$D9, $C0, $D8, $C8])}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $C1])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$D8, $F1])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $F2, $D9, $C0])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 8); {IBAppend(aaCode, [$D8, $C8, $D8, $CA, $D8, $C0, $DE, $E9])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 2); {IBAppend(aaCode, [$D9, $F3])} *)

              Inc(result, 50);
            end;
          bN: begin
           (* Inc(result, 4); {IBAppend(aaCode, b_zero_comp)}
      		    Inc(result, 4); {IBAppend(aaCode, [$76, $1E, $D9, $E0])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$D9, $E0])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 2); {IBAppend(aaCode, [$D9, $E0])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)} *)

              Inc(result, 36);
            end;
          bO : begin
           (* Inc(result, 8); {IBAppend(aaCode, b_mod_pi)}
              Inc(result, 4); {IBAppend(aaCode, b_zero_comp)}
              Inc(result, 4); {IBAppend(aaCode, [$75, $12, $DD, $D8])}
              Inc(result, 2); {IBAppend(aaCode, b_Load0)}
              Inc(result, 6); {WriteVel(j, aaCode, False)}
              Inc(result, 6); {WriteAcc(j, aaCode, False)}
              Inc(result, 12); {IBAppend(aaCode, [$EB, $36, $D9, $F2, $DC, $F1,
                                $D9, $C1, $D8, $C8, $DE, $C1])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 10); {IBAppend(aaCode, [$D9, $C0, $D8, $C8, $D9, $C9, $D8, $C, $D9, $E0])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {IBAppend(aaCode, [$D8, $C0, $D8, $CA, $D8, $C9])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $CA, $DE, $E1])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)} *)

              Inc(result, 86);
            end;
          bP: begin
            (*Inc(result, 4); {IBAppend(aaCode, b_1_comp)}
              Inc(result, 4); {IBAppend(aaCode, [$72, $12, $DD, $D8])}
              Inc(result, 2); {IBAppend(aaCode, b_Load0)}
              Inc(result, 6); {WriteVel(j, aaCode, False)}
              Inc(result, 6); {WriteAcc(j, aaCode, False)}
              Inc(result, 6); {IBAppend(aaCode, [$EB, $38, $D9, $C0, $D8, $C8])}
              Inc(result, 2); {IBAppend(aaCode, b_load1)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $E9, $D9, $FA])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $F1, $D9, $C0])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {IBAppend(aaCode, [$D8, $C8, $D8, $CA, $D8, $F1])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $F2, $DE, $E1])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $C1])}
              Inc(result, 6); {IBAppend(aaCode, b_ln)} *)

              Inc(result, 80);
            end;
          bQ: begin
            (*Inc(result, 4); {IBAppend(aaCode, [$D9, $C0, $D9, $E1])}
              Inc(result, 4); {IBAppend(aaCode, b_1_comp)}
              Inc(result, 4); {IBAppend(aaCode, [$77, $1F, $DD, $D8])}
              Inc(result, 4); {IBAppend(aaCode, b_zero_comp)}
              Inc(result, 2); {IBAppend(aaCode, [$DD, $D8])}
              Inc(result, 3); {LoadCVPar(2, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$77, $08, $D9, $E0])}
              Inc(result, 2); {IBAppend(aaCode, b_Load0)}
              Inc(result, 6); {WriteVel(j, aaCode, False)}
              Inc(result, 6); {WriteAcc(j, aaCode, True)}
              Inc(result, 4); {IBAppend(aaCode, [$EB, $4A, $D8, $C8])}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $E1])}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $F1])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $C9, $D9, $C0])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {IBAppend(aaCode, [$D8, $C8, $D8, $C0, $D8, $CA])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $CA, $DE, $C1])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)}
              Inc(result, 2); {IBAppend(aaCode, [$D9, $C9])}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 6); {IBAppend(aaCode, [$DE, $C1, $D8, $C8, $DE, $C9])}
              Inc(result, 6); {IBAppend(aaCode, b_ln)}
              Inc(result, 2); {IBAppend(aaCode, b_Load1)}
              Inc(result, 8); {IBAppend(aaCode, [$D9, $E0, $D9, $C9, $D9, $FD, $DD, $D9])} *)

              Inc(result, 115);
            end;
          bR: begin
           (* Inc(result, 4); {IBAppend(aaCode, [$D9, $C0, $D8, $C8])}
              Inc(result, 2); {IBAppend(aaCode, b_load1)}
              Inc(result, 4); {IBAppend(aaCode, [$DE, $C1, $D9, $FA])}
              Inc(result, 6); {LoadData(endFD + j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $F1, $D9, $C0])}
              Inc(result, 6); {WriteVel(j, aaCode, True)}
              Inc(result, 6); {IBAppend(aaCode, [$D8, $C8, $D8, $CA , $D8, $F1])}
              Inc(result, 6); {LoadData(endVD+ j, aaCode)}
              Inc(result, 4); {IBAppend(aaCode, [$D8, $F2, $DE, $E1])}
              Inc(result, 6); {WriteAcc(j, aaCode, True)}
              Inc(result, 2); {IBAppend(aaCode, [$DE, $C1])}
              Inc(result, 6); {IBAppend(aaCode, b_ln)} *)

              Inc(result, 56);
            end;
          else
                  ;
        end;
      end;
    end;
  end; {while}
  Inc(result, 17); {IBAppend(aaCode, b_epilog)}
 end;
  

function TEvaluatorBytes_64.ComputeDFCodeLength: Integer;
var
  i, j, a, pushes: Integer;
  ch: Byte;
begin
  result:= 0;
  inc(result, 11); {b_Prolog, 14 bytes}
  inc(result, 2); {IBAppend(ffCode, [$48, $B8]);}
  inc(result, SizeOf(Pointer)); {IBAppend(ffCode, stub);}
  if (lastIntegral1 >= 0) or (lastIntegral2 >= 0) then begin
    (*inc(result, 2); {IBAppend(ffCode, [$48, $B8]);}
    inc(result, SizeOf(pointer)); {sn(@theArg,, stub);}

    inc(result, 3); {IBAppend(ffCode,[$48, $89, $10])}

    inc(result, 3); {IBAppend(ffCode, [$48,$89, $95])}
    inc(result, 4 ); {IBAppend(ffCode, Int32Str(-$10));}

    inc(result, 3); {IBAppend(ffCode, [$48,$89, $85])}
    inc(result, 4 ); {IBAppend(ffCode, Int32Str(-$18)); }

    {mov rcx, self}
    inc(result, 2); {IBAppend(ffCode, [$48,$89]);}
    inc(result, SizeOf(Pointer)); {IBAppend(ffCode,sn(self, stub);}

    inc(result, 3); {IBAppend(ffCode, [$48,$89, $8D]);}
    inc(result, 4); {IBAppend(ffCode, Int32Str(-$20));}*)

    inc(result, 28 + 2*sizeOf(Pointer));
  end;
  emptyRegisters := c_LastReg + 1;
  i:= -1;
  j:= 0;
  while i < High(ActorList) do
    begin
      inc(i);
      ch := actorList[i];
      if ch in  [bLastByte, bPi, b1] then begin
        inc(result, LengthLeaveRoom(1));
        inc(j);
        case ch of
          bLastByte: begin
              ch:= actorList[i+1];
              i:= i+3;
              case ch of
                $01: begin{an independent variable}
                       inc(result,3); {LoadV(n, vvCode);}
                       inc(result,6); {WriteVel(j, vvCode, True);}
                       inc(result, 3); {max length of LoadX}
                       dec(emptyRegisters);
                     end;
                $02: begin{a parameter}
                       inc(result,2 ); {IBAppend(vvCode, b_Load0)}
                       inc(result,6); {WriteVel(j, vvCode, True)}
                       inc(result, 3); {max length of LoadCVPar}
                       dec(emptyRegisters);
                     end;
                $03: begin {a constant}
                       inc(result,2 ); {IBAppend(vvCode, b_Load0)}
                       inc(result,6); {WriteVel(j, vvCode, True)}
                       inc(result, 6); {LoadData(n, vvCode)}
                       dec(emptyRegisters);
                    end;
                $04: {an integral}
                  begin
                    pushes:= c_LastReg + 1 - emptyRegisters;
                    inc(result, LengthLeaveRoom(c_LastReg+1));

                    inc(result,2); {IBAppend(vvCode,  [$C7,$C6]);}
                    inc(result,4); {IBAppend(vvCode, Int32Str(n))}

                    inc(result,2); {IBAppend(vvCode, [$48, $B8])}
                    inc(result, SizeOf(Pointer));
                    inc(result, 3); {IBAppend(vvCode, [$48, $C7, $00])}
                    inc(result, 4); {IBAppend(vvCode, Int32Str(j))}

                    inc(result,2); {IBAppend(vvCode, [$48, $B8])}
                    inc(result, SizeOf(Pointer));
                    inc(result,2); {IBAppend(vvCode, [$FF, $D0])}

                    inc(result,4); {IBAppend(vvCode, [$f2, $0f, $11, $85])}
                    inc(result,4); {IBAppend(vvCode, Int32Str(-$08))}
                    inc(result,2); {IBAppend(vvCode, [$DD, $85])}
                    inc(result,4); {IBAppend(vvCode, Int32Str(-$08))}

                    emptyRegisters := c_LastReg;
                    if pushes > 0 then
                       for  a:= 1 to pushes do
                         inc(result, LengthPopFromStack);

                    inc(result, 21);
                    {IBAppend(vvCode, [$48, $8B, $8D]);
                    IBAppend(vvCode, Int32Str(-$20));
                    IBAppend(vvCode, [$48, $8B, $95]);
                    IBAppend(vvCode, Int32Str(-$10));
                    IBAppend(vvCode, [$48, $8B, $85]);
                    IBAppend(vvCode, Int32Str(-$18));}
                   end;
                else
                  ;
              end;
            end;
          bPi: begin
                 inc(result, 2); {IBAppend(vvCode, b_Load0)}
                 inc(result, 6); {WriteVel(j, vvCode, True)}
                 inc(result, 2); {IBAppend(vvCode, b_LoadPi)}
                 dec(emptyRegisters);
               end;
          b1: begin
                 inc(result, 2); {IBAppend(vvCode, b_Load0)}
                 inc(result, 6); {WriteVel(j, vvCode, True)}
                 inc(result, 2); {IBAppend(vvCode, b_Load1)}
                 dec(emptyRegisters);
               end;
          else
            ;
        end;
      end
      else begin
        if (ch in [bPlus, bMinus, bTimes, bDiv, bPow]) then
            begin
             if (emptyRegisters = c_LastReg) then
                inc(result, LengthPopFromStack);
              if ch = bPow then
                inc(result, LengthLeaveRoom(3))
              else if ch in [bTimes, bDiv] then
                inc(result, LengthLeaveRoom(2))
              else
                inc(result, LengthLeaveRoom(1));

              case ch of
                bPlus: begin
                 (* inc(result, 2); {IBAppend(vvCode, [$DE, $C1])}
                    inc(result, 6); {LoadData(endFD + j-1, vvCode)}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C1])}
                    inc(result, 6); {WriteVel(j-1, vvCode, True)}*)

                    inc(result, 22);
                  end;
                bMinus : begin
                 (* inc(result, 2); {IBAppend(vvCode, [$DE, $E9])}
                    inc(result, 6); {LoadData(endFD + j-1, vvCode)}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $E9])}
                    inc(result, 6); {WriteVel(j-1, vvCode, True);}*)

                    inc(result, 22);
                  end;
                bTimes : begin
                 (* inc(result, 6); {IBAppend(vvCode, [$D9, $C0, $D8, $CA, $D9, $CA])}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6); {LoadData(endFD + j-1, vvCode)}
                    inc(result, 4); {IBAppend(vvCode, [$DE, $CA, $DE, $C1])}
                    inc(result, 6); {WriteVel(j-1, vvCode, True)}*)

                    inc(result, 30);
                  end;
                bDiv : begin
                 (* inc(result, 4); {IBAppend(vvCode, b_zero_comp)}
                    inc(result, 6); {IBAppend(vvCode,[$75, $0E,$DD, $D8, $DD, $D8])}
                    inc(result, 2); {IBAppend(vvCode, b_Load0)}
                    inc(result, 6); {WriteVel(j-1, vvCode, False)}
                    inc(result, 2); {IBAppend(vvCode,[$EB, $28])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 8); {IBAppend(vvCode, [$DE, $F1, $D9, $C0, $D8, $CA, $D9, $CA])}
                    inc(result, 6); {LoadData(endFD + j-1, vvCode)}
                    inc(result, 8); {IBAppend(vvCode, [$D8, $CA, $D9, $CA, $D8, $C8, $DE, $C9])}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 4); {IBAppend(vvCode, [$DE, $C9, $DE, $E9])}
                    inc(result, 6); {WriteVel(j-1, vvCode, True);} *)

                    inc(result, 60);
                  end;
                bPow :  ;{This will not be used}
              end;
              inc(emptyRegisters);
              j:= j-1;
            end
          else {opCode of a function}
            begin
              if ch= bNeg then
                inc(result, LengthLeaveRoom(1))
              else
                inc(result, LengthLeaveRoom(2));

              case ch of
                bNeg : begin
                 (* inc(result, 2); {IBAppend(vvCode, [$D9, $E0])}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$D9, $E0])}
                    inc(result, 6); {WriteVel(j, vvCode, True)} *)

                    inc(result, 16);
                  end;

                bA : begin
                 (* inc(result, 10); {IBAppend(vvCode, b_mod_2pi)}
                    inc(result, 6); {IBAppend(vvCode, [$D9, $FB, $D9, $C9, $D9, $E0])}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6); {WriteVel(j, vvCode, True)}*)

                    inc(result, 30);
                  end;

                bB : begin
                 (* inc(result, 10); {IBAppend(vvCode, b_mod_2pi)}
                    inc(result, 2);  {IBAppend(vvCode, [$D9, $FB])}
                    inc(result, 6);  {LoadData(endFD + j, vvCode)}
                    inc(result, 2);  {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6);  {WriteVel(j, vvCode, True)} *)

                    inc(result, 26);
                  end;

                bC : begin
                  (*inc(result, 4); {IBAppend(vvCode, [$D9, $EA, $DE, $C9])}
                    inc(result, 49); {IBAppend(vvCode, b_2ToX_S)}
                    inc(result, 2); {IBAppend(vvCode, [$D9, $C0])}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6); {WriteVel(j, vvCode, True)} *)

                    inc(result, 69);
                  end;

                bD : begin
                   (* inc(result, 8); {IBAppend(vvCode, [$D9, $C0, $D8, $C8, $D9, $C9, $D8, $C0]);}
                    inc(result, 6); {LoadData(endFD + j, vvCode);}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9]);}
                    inc(result, 6); {WriteVel(j, vvCode, True);} *)

                    inc(result, 22);
                  end;

                bE : begin
                 (* inc(result, 4); {IBAppend(vvCode, b_zero_comp)}
                    inc(result, 4); {IBAppend(vvCode, [$72, $08, $DD, $D8])}
                    inc(result, 2); {IBAppend(vvCode, b_Load0)}
                    inc(result, 2); {IBAppend(vvCode, b_Load0)}
                    inc(result, 2); {IBAppend(vvCode, [$EB, $0E])}
                    inc(result, 2); {IBAppend(vvCode, [$D9, $FA])}
                    inc(result, 2); {IBAppend(vvCode, [$D9, $C0])}
                    inc(result, 2); {IBAppend(vvCode, [$D8, $C0])}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $F1])}
                    inc(result, 6); {WriteVel(j, vvCode, True)} *)

                    inc(result, 34);
                  end;

                bF : begin
                  (*inc(result, 4); {IBAppend(vvCode, b_zero_comp)}
                    inc(result, 4); {IBAppend(vvCode, [$72, $0B, $DD, $D8])}
	                  inc(result, 3); {LoadCVPar(5, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$D9, $E0])}
	                  inc(result, 4); {IBAppend(vvCode, b_Load0)}
	                  inc(result, 4); {IBAppend(vvCode, [$EB, $12, $D9, $C0])}
                    inc(result, 6); {IBAppend(vvCode, b_ln)}
                    inc(result, 2); {IBAppend(vvCode, [$D9, $C9])}
	                  inc(result, 6); {LoadData(endFD + j, vvCode)}
	                  inc(result, 2); {IBAppend(vvCode, [$DE, $F1])}
	                  inc(result, 6); {WriteVel(j, vvCode, True); } *)

                    inc(result, 43);
                  end;

                bG : begin
                  (*inc(result, 8); {IBAppend(vvCode, b_mod_pi)}
                    inc(result, 8); {IBAppend(vvCode, [$D9, $F2, $D9, $C1, $D8, $C8, $DE, $C1])}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6); {WriteVel(j, vvCode, True)} *)

                    inc(result, 30);
                  end;

                bH : begin
                  (*inc(result, 4); {IBAppend(vvCode, [$D9, $EA, $DE, $C9])}
                    inc(result, 49) {IBAppend(vvCode, b_2ToX_S)}
	                  inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 22); {IBAppend(vvCode, [$D9, $E0, $D9, $C1, $D8, $F9, $D9, $FD, $D9,
                                                  $CA, $D9, $FD, $DD, $D9, $D9, $C0, $D8, $C2,
                                                  $D9, $CA, $DE, $E9]);}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
	                  inc(result, 6); {WriteVel(j, vvCode, True)} *)

                    inc(result, 91);
                  end;

                bI : begin
                    (*inc(result, 4); {IBAppend(vvCode, [$D9, $EA, $DE, $C9])}
                    inc(result, 49); {IBAppend(vvCode, b_2ToX_S)}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 22); {IBAppend(vvCode, [$D9, $E0, $D9, $C1, $D8, $F9, $D9, $FD, $D9, $CA,
                                        $D9, $FD, $DD, $D9, $D9, $C0, $D8, $C2, $D9, $C9,
                                        $DE, $E2])}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode,[$DE, $C9])}
                    inc(result, 6); {WriteVel(j, vvCode, True)} *)

                    inc(result, 91);
                  end;

                bJ : begin
                   (*inc(result, 4); {IBAppend(vvCode, [$D9, $EA, $DE, $C9]);}
                    inc(result, 49); {IBAppend(vvCode, b_2ToX_S)}
                    inc(result, 2); {IBAppend(vvCode, [$D8, $C8])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 4); {IBAppend(vvCode, [$D8, $E9, $D9, $C9])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 8); {IBAppend(vvCode, [$DE, $C1, $DE, $F9, $D9, $C0, $D8, $C8])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $E1])}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6); {WriteVel(j, vvCode, True)}  *)

                    inc(result, 89);
                  end;

                bK : begin
                  (*inc(result, 4); {IBAppend(vvCode, [$D9, $C0, $D9, $E1]);}
                    inc(result, 4); {IBAppend(vvCode, b_1_comp);}
                    inc(result, 8); {IBAppend(vvCode, [$DD, $D8, $76, $18, $D9, $C0, $D8, $C8])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 4); {IBAppend(vvCode, [$DE, $E1, $D9, $FA])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 12);{IBAppend(vvCode, [$D8, $F1, $D9, $CA, $D9, $C9, $D9, $F3,
                                      $D9, $C9, $EB, $16])}
	                  inc(result, 4); {IBAppend(vvCode, b_zero_comp)}
                    inc(result, 2); {IBAppend(vvCode, [$DD, $D8])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 2); {IBAppend(vvCode, [$D9, $E0])}
                    inc(result, 2); {IBAppend(vvCode, b_LoadPi)}
                    inc(result, 8); {IBAppend(vvCode, [$D9, $FD, $DD, $D9, $72, $02, $D9,$E0])}
                    inc(result, 2); {IBAppend(vvCode, b_Load0)}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6); {WriteVel(j, vvCode, True)} *)

                    inc(result, 72);
                  end;

                bL : begin
                  (*inc(result, 4); {IBAppend(vvCode, [$D9, $C0, $D9, $E1])}
                    inc(result, 4); {IBAppend(vvCode, b_1_comp)}
                    inc(result, 8); {IBAppend(vvCode, [$DD, $D8, $76, $18, $D9, $C0, $D8, $C8])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 4); {IBAppend(vvCode, [$DE, $E1, $D9, $FA])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 12); {IBAppend(vvCode, [$D8, $F1, $D9, $E0, $D9, $CA, $D9, $F3,
                                      $D9, $C9, $EB, $10])}
                    inc(result, 4); {IBAppend(vvCode, b_zero_comp)}
                    inc(result, 4); {IBAppend(vvCode, [$DD, $D8, $72, $04])}
                    inc(result, 2); {IBAppend(vvCode, b_LoadPi)}
                    inc(result, 2); {IBAppend(vvCode, [$EB, $02])}
                    inc(result, 2); {IBAppend(vvCode, b_Load0)}
                    inc(result, 2); {IBAppend(vvCode, b_Load0)}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6); {WriteVel(j, vvCode, True)} *)

                    inc(result, 66);
                  end;

                bM : begin
                  (*inc(result, 2); {IBAppend(vvCode, [$D9, $C0])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 6); {IBAppend(vvCode, [$D9, $F3, $D9, $C9, $D8, $C8])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 4); {IBAppend(vvCode, [$DC, $C1, $DE, $F1])}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6); {WriteVel(j, vvCode, True)} *)

                    inc(result,30);
                  end;

                bN : begin
                 (* inc(result, 4); {IBAppend(vvCode, b_zero_comp);}
                    inc(result, 4); {IBAppend(vvCode, [$76, $10, $D9, $E0]);}
                    inc(result, 6); {LoadData(endFD + j, vvCode);}
                    inc(result, 2); {IBAppend(vvCode, [$D9, $E0]); }
                    inc(result, 6); {WriteVel(j, vvCode, True);} *)

                    inc(result, 22);
                  end;

                bO : begin
                 (* inc(result, 8); {IBAppend(vvCode, b_mod_pi)}
                    inc(result, 4); {IBAppend(vvCode, b_zero_comp)}
                    inc(result, 22); {IBAppend(vvCode, [$74, $12, $D9, $F2, $DC, $F1, $D9, $C9, $D9, $C0,
                                         $D8, $C8, $DE, $C2, $D9, $C9, $D9, $E0, $EB, $06, $DD, $D8])}
                    inc(result, 2); {IBAppend(vvCode, b_Load0)}
                    inc(result, 2); {IBAppend(vvCode, b_Load0)}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6); {WriteVel(j, vvCode, True)}*)

                    inc(result, 52);
                  end;

                bP : begin
                 (* inc(result, 4); {IBAppend(vvCode, b_1_comp)}
                    inc(result, 6); {IBAppend(vvCode, [$73, $1E, $D9, $C0, $D8, $C8])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 10); {IBAppend(vvCode, [$DE, $E9, $D9, $FA, $D9, $C0, $DE, $C2, $D9, $C9])}
                    inc(result, 6); {IBAppend(vvCode, b_ln)}
                    inc(result, 2); {IBAppend(vvCode, [$D9, $C9])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 6); {IBAppend(vvCode, [$DE, $F1, $EB, $06, $DD, $D8])}
                    inc(result, 2); {IBAppend(vvCode, b_Load0)}
                    inc(result, 2); {IBAppend(vvCode, b_Load0)}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6); {WriteVel(j, vvCode, True)} *)

                    inc(result, 56);
                  end;

                bQ : begin
                 (* inc(result, 4); {IBAppend(vvCode, [$D9, $C0, $D9, $E1])}
                    inc(result, 4); {IBAppend(vvCode, b_1_comp)}
                    inc(result, 6); {IBAppend(vvCode, [$DD, $D8, $76, $2A, $D9, $C0])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 4); {IBAppend(vvCode, [$DE, $C1, $D9, $C9])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 10); {IBAppend(vvCode, [$DE, $E1, $D9, $C0, $D8, $CA, $D9, $CA, $DE, $F1])}
                    inc(result, 6); {IBAppend(vvCode, b_ln)}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 14); {IBAppend(vvCode, [$D9, $E0, $D9, $C9, $D9, $FD, $D9, $CA,
                                      $DE, $F9, $D9, $E0, $EB, $12])}
                    inc(result, 4); {IBAppend(vvCode, b_zero_comp)}
                    inc(result, 4); {IBAppend(vvCode, [$DD, $D8, $77, $05])}
                    inc(result, 6); { LoadCVPar(1, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$EB,$03])}
                    inc(result, 3); {LoadCVPar(2, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, b_Load0)}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6); {WriteVel(j, vvCode, True)} *)

                    inc(result, 89);
                  end;

                bR : begin
                 (* inc(result, 4); {IBAppend(vvCode, [$D9, $C0, $D8, $C8])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 10); {IBAppend(vvCode, [$DE, $C1, $D9, $FA, $D9,
                                      $C0, $D9, $CA, $DE, $C1])}
                    inc(result, 6); {IBAppend(vvCode, b_ln)}
                    inc(result, 2); {IBAppend(vvCode, [$D9, $C9])}
                    inc(result, 2); {IBAppend(vvCode, b_Load1)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $F1])}
                    inc(result, 6); {LoadData(endFD + j, vvCode)}
                    inc(result, 2); {IBAppend(vvCode, [$DE, $C9])}
                    inc(result, 6); {WriteVel(j, vvCode, True)} *)

                    inc(result, 42);
                  end
              else
                 ;
              end;
            end;
        end;
      end;
   inc(result, 17); {epilog}
end;

function TEvaluatorBytes_64.ComputeFCodeLength: Integer;
var
  i, a, pushes: Integer;
  ch: Byte;
begin
  result:= 0;
  lastPreserved:= 1;
  inc(result, 11); {b_Prolog, 11 bytes}
  {Save the address of 'data' to rax}
  inc(result, 2); {IBAppend(ffCode, [$48, $B8]);}
  inc(result, SizeOf(Pointer)); {IBAppend(ffCode, stub);}
  if (lastIntegral1 >= 0) or (lastIntegral2 >= 0) then begin
    (*inc(result, 2); IBAppend(ffCode, [$48, $B8]);
                      sn(@theArg, stub);
      inc(result,SizeOf(Pointer)); IBAppend(ffCode, stub);
      inc(result, 3); IBAppend(ffCode,[$48, $89, $10]);
      inc(result, 3); IBAppend(ffCode, [$48,$89, $95]);
      inc(result, 4); IBAppend(ffCode, Int32Str(c_sfr+$10));

      inc(result, 3); IBAppend(ffCode, [$48,$89, $85]);
      inc(result, 4); IBAppend(ffCode, Int32Str(c_sfr+$18));

      inc(result, 2); IBAppend(ffCode, [$48,$B9]);
                      sn(self, stub);
      inc(result,SizeOf(Pointer)); IBAppend(ffCode, stub)}

      inc(result, 3); IBAppend(ffCode, [$48,$89, $8D]);
      inc(result, 4); IBAppend(ffCode, Int32Str(c_sfr+$20));
      *)

    inc(result, 28 + 2*sizeOf(Pointer));
  end;

  emptyRegisters := c_LastReg + 1;
  i:= -1;
  while i < High(ActorList) do
    begin
      inc(i);
      ch := actorList[i];
      if ch in [bLastByte, bPi, b1] then begin
        if ch <> bLastByte then
          inc(result, LengthLeaveRoom(1));
        case ch of
          bLastByte: if actorList[i+1] < $04 then begin
              inc(result, LengthLeaveRoom(1));
              inc(result, 6); {OCRBytes adds at most 6 bytes}
              inc(i, 3);
              dec(emptyRegisters);
            end
            else begin
              inc(i, 3);
              pushes:= c_LastReg + 1 - emptyRegisters;
              inc(result, LengthLeaveRoom(c_LastReg+1)); {LeaveRoomInFPUStack(c_LastReg+1, ffCode)}
              inc(result,2); //IBAppend(ffCode,  [$C7,$C2]);
              inc(result, 4); //IBAppend(ffCode, Int32Str(n));

              inc(result,2); //IBAppend(ffCode, [$48, $B8]);
              inc(result, SizeOf(Pointer)); //IBAppend(ffCode, stub);
              inc(result, 2); //IBAppend(ffCode, [$FF, $D0]);

              inc(result, 4); //IBAppend(ffCode, [$f2, $0f, $11, $85]);
              inc(result, 4); //IBAppend(ffCode, Int32Str(c_sfr-$08));
              inc(result, 2); //IBAppend(ffCode, [$DD, $85]);
              inc(result, 4); //IBAppend(ffCode, Int32Str(c_sfr-$08));

              emptyRegisters := c_LastReg;
              if pushes > 0 then
                for a:= 1 to pushes do
                  inc(result, LengthPopFromStack);

              {IBAppend(ffCode, [$48, $8B, $8D]);
               IBAppend(ffCode, Int32Str(c_sfr-$20));
               IBAppend(ffCode, [$48, $8B, $95]);
               IBAppend(ffCode, Int32Str(c_sfr+$10));
               IBAppend(ffCode, [$48, $8B, $85]);
               IBAppend(ffCode, Int32Str(c_sfr+$18));}
              inc(result, 21);
            end;
          bPi: begin
                 inc(result, 2); {b_LoadPi};
                 dec(emptyRegisters);
               end;
          b1: begin
                 inc(result, 2); {b_Load1};
                 dec(emptyRegisters);
               end
          else
            ;
        end;
      end
      else
        begin
          if (ch in [bPlus, bMinus, bTimes, bDiv, bPow]) then
            begin
              if (emptyRegisters = c_LastReg) then
                inc(result, LengthPopFromStack);
              case ch of
                bPlus, bMinus, bTimes : inc(result, 2);
                bDiv :
                  begin
                    inc(result, LengthLeaveRoom(1));
                    inc(result, 16); { b_div_S};
                  end;
                bPow :
                  raise
                    exception.Create(eUndigestedPowerSymbol); {why not exp(#2 * ln#1) <-- #1^#2 ?}
              end;
              inc(emptyRegisters);
            end
          else {opCode of a function}
            begin
              if (ch in [bA, bB, bC, bE, bF, bG, bM]) then
                inc(result, LengthLeaveRoom(1))
              else if (ch in [bH, bI, bJ, bK, bL, bO, bP, bQ, bR]) then
                inc(result, LengthLeaveRoom(2));
              case ch of
                bNeg : inc(result, 2);
                bA, bB: inc(result, 12); {b_cos_code, b_sin_code}
                bC : begin
                   (* inc(result, 4); {b_times_Log2_e}
                      inc(result, 49);  {b_2ToX_S}  *)
                       inc(result, 53);
                     end;
                bD : inc(result, 2);  {b_Sqr}
                bE : inc(result, 14); {b_Sqrt}
                bF : inc(result, 21);  {b_ln_S}
                bG : inc(result, 12);  {b_tan}
                bH, bI : begin
                       (*inc(result, 4);
                       inc(result, 49);  {b_2ToX_S}
                       inc(result, 8);
                       inc(result, 6);  {b_div2}*)

                       inc(result, 67);
                     end;
                bJ : begin
                      (* inc(result, 4);
                       inc(result, 49);  {b_2ToX_S}
                       inc(result, 14); *)

                       inc(result, 67);
                     end;
                bK : begin
                    (* inc(result, 4);
                       inc(result, 4); {b_1_comp}
                       inc(result, 18);
                       inc(result, 4); {b_zero_comp}
                       inc(result, 2);
                       inc(result, 10); {b_LoadHalfPi}
                       inc(result, 4); *)

                       inc(result, 46);
                     end;
                bL : begin
                    (* inc(result, 4);
                       inc(result, 4); {b_1_comp}
                       inc(result, 20);
                       inc(result, 4); {b_zero_comp}
                       inc(result, 10); *)

                       inc(result, 42);
                     end;
                bM : inc(result, 4);
                bN : inc(result, 2);
                bO : begin
                    (* inc(result, 8); {b_mod_pi}
                       inc(result, 4);
                       inc(result, 16); {b_div_S} *)

                       inc(result, 28);
                     end;
                bP : begin
                    (* inc(result, 4); {b_1_comp}
                       inc(result, 14);
                       inc(result, 6); {b_ln}
                       inc(result, 6); *)

                       inc(result, 30);
                     end;
                bQ : begin
                    (* inc(result, 4);
                       inc(result, 4); {b_1_comp}
                       inc(result, 18); {(1+r)/(1-r)]}
                       inc(result, 6); {b_ln}
                       inc(result, 6);  {b_div2}
                       inc(result, 2);
                       inc(result, 4); {b_zero_comp}
                       inc(result, 4);
                       inc(result, 3); {LoadCVPar(5, fCode)}
                       inc(result, 2);
                       inc(result, 2); {LoadCVPar(0, fCode)} *)

                       inc(result, 55);
                     end;
                bR : begin
                    (* inc(result, 4); {b_zero_comp}
                       inc(result, 14);
                       inc(result, 6); {b_ln}
                       inc(result, 4); *)

                       inc(result, 28);
                     end;
                else
                  raise Exception.Create(eErrorInParsing);
                end;
            end;
        end;
    end; {del while}
  inc(result, 17); {b_Epilog};
end;




constructor TEvaluatorBytes_64.Create;
begin
  inherited Create;
  b_ln_s[10]:= fpTypeSize * 5;
  b_2toX_S[6]:= fpTypeSize * 5;
  b_2toX_S[43]:= fpTypeSize * 3;
  b_2toX_S[48]:= fpTypeSize;
  ffBlockLength:= 0;
  vvBlockLength:= 0;
  aaBlockLength:= 0;
end;

destructor TEvaluatorBytes_64.Destroy;
begin
  data := nil;
  minCorner:= nil;
  jump1:= nil;
  invJump1:= nil;
  jump2:= nil;
  invJump2:= nil;
  intList1:= nil;
  intList2:= nil;
  {$ifdef mswindows}
  if ffBlock <> nil then begin
    VirtualFree(ffBlock, 0, MEM_RELEASE);
    ffBlock:= nil
  end;
  if vvBlock <> nil then begin
    VirtualFree(vvBlock, 0, MEM_RELEASE);
    vvBlock:= nil
  end;
  if aaBlock <> nil then begin
    VirtualFree(aaBlock, 0, MEM_RELEASE);
  aaBlock:= nil
  end;
  {$endif}
  {$ifdef Linux}
  if ffBlock <> nil then begin
    fpmunmap(ffBlock, ffBlockLength);
    ffBlock:= nil;
  end;
  if vvBlock <> nil then begin
    fpmunmap(vvBlock, vvBlockLength);
    vvBlock:= nil;
  end;
  if aaBlock <> nil then begin
    fpmunmap(aaBlock, aaBlockLength);
    aaBlock:= nil;
  end;
  {$endif}
  actorList:= nil;
  inherited;
end;

{function TEvaluatorBytes_64.Dummy(index: Integer): Double;
begin
  result:= sqrt(1+sqr(index));
end;}

function TEvaluatorBytes_64.gradF(const x: Vector3): Vector3;
         {Use this when ArgNum = 3}
begin
  vv([x[1],x[2],x[3],1,0,0]);
  result[1]:= data[endFd+1];
  vv([x[1],x[2],x[3],0,1,0]);
  result[2]:= data[endFd+1];
  vv([x[1],x[2],x[3],0,0,1]);
  result[3]:= data[endFd+1];
end;


function TEvaluatorBytes_64.gradF(const x: Vector2): Vector2;
         {Use this when ArgNum = 2}
begin
  vv([x[1],x[2],1,0]);
  result[1]:= data[endFd+1];
  vv([x[1],x[2],0,1]);
  result[2]:= data[endFd+1];
end;


PROCEDURE TEvaluatorBytes_64.LeaveRoomInFPUStack(neededPlaces: Integer; var cod: InfBArray);
var
  stub: ByteArray;
begin
  IF emptyRegisters >= neededPlaces THEN
    Exit;
  SetLength(stub, 4);
  IF emptyRegisters = 0 THEN BEGIN
    inc(lastPreserved);
    inc(cod.h);
    cod.arr[cod.h] := $D9;
    inc(cod.h);
    cod.arr[cod.h] := $F6;  {fdecstp: ST(cLastReg] becomes ST(0)}
    IBAppend(cod, b_FSTP_rec);
    stub:= Int32Str(c_sfr-8*lastPreserved);  {fstp [RBP - 8*lastPreserved]}
    IBAppend(cod, stub);
    inc(emptyRegisters);
  END;
  IF emptyRegisters < neededPlaces then begin
    if emptyRegisters = c_LastReg then begin
      inc(lastPreserved);
      IBAppend(cod, b_FSTP_rec);
      stub:= Int32Str(c_sfr-8*lastPreserved);  {fstp [RBP - 8*lastPreserved]}
      IBAppend(cod, stub);
      inc(emptyRegisters);
    end
    else REPEAT
      inc(lastPreserved);
      inc(cod.H);
      cod.arr[cod.H]:= $D9;
      inc(cod.H);
      cod.arr[cod.H]:= $C0 + c_LastReg - emptyRegisters; {push last occupied ST to ST(0)}
      IBAppend(cod, b_FSTP_rec);
      stub:= Int32Str(c_sfr-8*lastPreserved);  {fstp [RBP - 8*lastPreserved]}
      IBAppend(cod, stub);
      inc(cod.H);
      cod.arr[cod.H]:= $DD;
      inc(cod.H);
      cod.arr[cod.H]:= $C0 + c_LastReg - emptyRegisters;
         {last occupied place contains the former value: declare it free}
      inc(emptyRegisters);
    UNTIL emptyRegisters = neededPlaces;
  end;
  stub:= nil;
END;

function TEvaluatorBytes_64.LengthLeaveRoom(neededPlaces: Integer): Integer;
begin
  result:= 0;
  if emptyRegisters >= neededPlaces then
    Exit;
  if emptyRegisters = 0 then begin
    inc(lastPreserved);
    (*inc(result, 2); {fdecstp: ST(cLastReg] becomes ST(0)}
    inc(result, 6); {b_FSTP_rec + disp32} *)
    inc(result, 8);
    inc(emptyRegisters);
  end;
  if emptyRegisters < neededPlaces then begin
    if emptyRegisters = c_LastReg then begin
      inc(lastPreserved);
      inc(result, 6); {b_FSTP_rec + disp32}
      inc(emptyRegisters);
    end
    else repeat
      inc(lastPreserved);
      (*inc(result, 2); {push last occupied ST to ST(0)}
      inc(result, 6); {b_FSTP_rec + disp32}
      inc(result, 2); {delare freelast occupied place} *)
      inc(result, 10);
      inc(emptyRegisters);
    until emptyRegisters = neededPlaces;
  end;
end;

function  TEvaluatorBytes_64.LengthPopFromStack: Integer;
VAR
  m : byte;
BEGIN
  result:= 0;
  m := c_LastReg + 1 - emptyRegisters; {ST(m) is the first empty register}
  if m = 1 then
    inc(result, 8)
  else
    inc(result, 10);
  dec(lastPreserved);
  dec(emptyregisters);
END;

procedure TEvaluatorBytes_64.LoadA(const j: Byte; var iba: InfBArray);
{Reads from the function stack the acceleration 'i'
 and stores it in ST(0)}
begin
  IBAppend(iba, b_Ld_X);
  inc(iba.h);
  iba.arr[iba.h]:= fpTypeSize * (2*argNum + j -1);
end;

procedure TEvaluatorBytes_64.LoadCVPar(const j: Byte; var iba: InfBArray);
begin
  if j=0 then
    IBAppend(iba, [$DD,$00])
  else begin
    IBAppend(iba, b_Ld_Dt8);
    inc(iba.h);
    iba.arr[iba.h]:= fpTypeSize*j;
  end;
end;

procedure TEvaluatorBytes_64.LoadData(const j: word; var yarr: InfBArray);
{Reads from the address of the var data[j] as a Double and stores it in ST(0)}
var
  bytes, i: Integer;
  c: PByte;
begin
  if j= 0 then
    IBAppend(yarr, [$DD,$00])
  else begin
    IBAppend(yarr, b_Ld_Dt);
    bytes:= fpTypeSize * j;
    c:= @bytes;
    for i := 0 to 3 do begin
      inc(yarr.h);
      yarr.arr[yarr.h] := c[i];
    end;
  end;
end;

procedure TEvaluatorBytes_64.LoadV(const j: Byte; var iba: InfBArray);
{Reads from the function stack the acceleration 'j'
 and stores it in ST(0)}
begin
  IBAppend(iba, b_Ld_X);
  inc(iba.h);
  iba.arr[iba.h]:= fpTypeSize * (argNum + j -1);
end;

procedure TEvaluatorBytes_64.LoadX(i: Word; var yarr: InfBArray);
{Reads from the function stack the independent variable 'i' of FPType
type and stores it in ST(0)}

begin
  if i = 1 then
    IBAppend(yarr, b_Ld_X1)
  else begin
    IBAppend(yarr, b_Ld_X);
    inc(yarr.h);
    yarr.arr[yarr.h]:= fpTypeSize * (i-1);
  end;
end;

procedure TEvaluatorBytes_64.OCRBytes(stub: ByteArray; var yarr: InfBArray);
{'stub' is a string of four bytes beginning with $FF that represent the loading
 of an argument, a parameter or a constant. Gives the bytes to be added to an
 object code string}
begin
  case stub[1] of
    $01: LoadX(stub[3], yarr);
    $02: LoadCVPar(stub[3], yarr);
    $03: LoadData(stub[2] shl 8 + stub[3], yarr)
    else
    ;
  end;
end;




PROCEDURE TEvaluatorBytes_64.PopFromAppStack(var fCode: InfBArray);
VAR
  m : byte;
  stub: ByteArray;
BEGIN
  m := c_LastReg + 1 - emptyRegisters; {ST(m) is the first empty register}
  if m = 1 then begin
    IBAppend(fCode,  b_FLD_rec);
    SetLength(stub,4);
    stub:= Int32Str(c_sfr- 8*lastPreserved);
    IBAppend(fCode, stub);
    IBAppend(fCode, $D9);
    inc(fCode.h);
    fCode.arr[fCode.h]:= $C8 + m;
  end
  else begin
    IBAppend(fCode, $DD);
    inc(fCode.h);
    fCode.arr[fCode.h]:= $D8 + m;
    IBAppend(fCode, b_FLD_rec);
    stub:= Int32Str(c_sfr- 8*lastPreserved);
    IBAppend(fCode, stub);
    IBAppend(fCode, $D9);
    inc(fCode.h);
    fCode.arr[fCode.h]:= $C8 + m;
  end;
  dec(lastPreserved);
  dec(emptyregisters);
  stub:= nil;
END;

function TEvaluatorBytes_64.SecondD(const XV: array of Double): SecondDeriv;
begin
  result.fa0:= aa(XV);
  result.dfa0:= data[endFd+1];
  result.ddfa0:= data[endVd+1];
end;

procedure TEvaluatorBytes_64.UpdateParameters(parameters:
          array of Double);
{Changes the value of the parameters stored in 'data', so that they
 will have the new values in the following evaluations}
  var
    h,i: Integer;
    //worthless: Double;
begin
  h:= high(parameters);
  if (h >-1) and (h < parNum) then for i:=0 to high(parameters) do
    data[i+6]:= parameters[i];
  {worthless:= dummy(argNum);
  if worthless < 0 then
    beep(100,100);}
end;


procedure TEvaluatorBytes_64.WriteAcc(const j: Byte; var iba: InfBArray; pop: Boolean);
{Writes ST(0), as an FPType, into  'data[endVd+i]', and pops ST(0)}
var
  stb: BArray4;
begin
  stb:= disp32(endVd + j);
  if pop then begin
    IBAppend(iba, b_StPDt);
    IBAppend(iba, stb);
  end
  else begin
    IBAppend(iba, b_StDt);
    IBAppend(iba, stb);
  end;
end;

procedure TEvaluatorBytes_64.WriteVel(const j: Byte; var iba: InfBArray; pop: Boolean);
{Writes ST(0), as an FPType, into  'data[endFd+i]', and pops ST(0)}
var
  stb: BArray4;
begin
  stb:= disp32(endFd + j);
  if pop then begin
    IBAppend(iba, b_StPDt);
    IBAppend(iba, stb);
  end
  else begin
    IBAppend(iba, b_StDt);
    IBAppend(iba, stb);
  end;
end;


end.

