
{-----------------------------------------------------------------------------
 Unit Name: FLinePalette
 Project Name : Parametricas4
 Author:    Angel Montesinos Amilibia
 Dates: 2003-2009
 Purpose: Simple color picker for curves

-----------------------------------------------------------------------------}
{Parametricas4, workshop for the geometry of surfaces in R^4.


MIT License

Copyright (c) [2003 - 2018] [Angel Montesinos-Amilibia], email montesin@uv.es

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


}


unit FLinePalette;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, ComCtrls,
  VectorsAndUtil;

const
  kLastPltColor = 511; {must be an odd number}
  kMedPltIndex = (kLastPltColor + 1) div 2;

type
  TColorArray = array[0..65535] of TColor;
  pTColorArray = ^TColorArray;

  Plt = array[0..kLastPltColor] of TColor;

type
  TLinePalette = class(TForm)
    HSBox: TPaintBox;
    meterBox: TPaintBox;
    procedure DrawHSPoint;
    procedure DrawMeterPt;
    procedure MouseMoveInHSBox(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure MouseMoveInMeter(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure PaintHSBox(Sender: TObject);
    procedure OnPaintPicker(Sender: TObject);
    procedure OnPaintMeter(Sender: TObject);
  private
    own: TComponent;
    HSBm, meterBm: TBitmap;
    HSRect, meterRect: TRect;
    HSctr, HSPt: Vector2Int;
    meterPos: Integer;
    function PtInWheel(var r: Double; h, v: Integer): Boolean;
    procedure DrawWheel;
    procedure FillMeter;
    function FromWheelToHS(x, y: Double): Vector2;
    function FromHSLToRGB(hs: Vector2; b: Double): TColor;

  public
    curveColor: TColor;
    procedure Initiate;
    procedure BuildPlt(var pal: Plt; h1, h2: Double);
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
  end;

var
  linePalette: TLinePalette;
  bckgPal: Plt;

implementation
{$R *.dfm}
uses math,
  ResourceStrings, FParam4;

const
  cTwoPi: Double = 6.283185307179586476925;
  cWheelRadius = 80;
  cMeterLast = 164;
  cHalfMeterLast = cMeterLast div 2;
  cMaxComp = 255;

procedure TLinePalette.BuildPlt(var pal: Plt; h1, h2: Double);
{build a RGB color palette 'pal' that has its brightest value at the center,
and goes downward towards black with hue and saturation 'h2' and upward
towards black with hue and saturation 'h1'}
const
  cMargin = 0.05;
var
  hs1, hs2: Vector2;
  step: Double;
  i: Integer;
begin
  hs1[1]:= h1; {hue of first color}
  hs1[2]:= 0.9; {saturation of first color}
  hs2[1]:= h2; {hue of last color}
  hs2[2]:= 0.9; {saturation of last color}
  step:= (1-2*cMargin)/(kMedPltIndex-1);
  for i:= 0 to kMedPltIndex-1 do begin
    {with this choice of 'i' and 'step', all the entries of the palette are
     assigned, the least light is 'cMargin' and the greatest '1-cMargin'}
    pal[i]:= FromHSLToRGB(hs1,cMargin+i*step);
    pal[kLastPltColor-i]:= FromHSLToRGB(hs2,cMargin+i*step);
  end;

end;

constructor TLinePalette.Create(aOwner: TComponent);
begin
  inherited create(aOwner);
  own:= owner;
  caption:= ePalette;
  HSBox.canvas.Brush.Color:= clWhite;
  HSBox.canvas.Pen.Color:= clWhite;
  HSBox.Canvas.Brush.Style:= bsSolid;
  HSBm:= TBitmap.Create;
  HSBm.Width:= HSBox.Width;
  HSBm.Height:= HSBox.Height;
  HSBm.PixelFormat:= pf32bit;
  HSBm.Canvas.Brush.Style:= bsSolid;
  HSBm.Canvas.Brush.Color:= clWhite;
  SetRect(HSRect, 0, 0, HSBox.Width, HSBox.Height);

  HSctr[1]:= HSBox.Width div 2;
  HSctr[2]:= HSBox.Height div 2;
  HSPt[1]:= 40;
  HsPt[2]:= 40;
  DrawWheel;

  meterBox.canvas.Brush.Color:= clWhite;
  meterBm:= TBitmap.Create;
  meterBm.Width:= meterBox.Width;
  meterBm.Height:= meterBox.Height;
  meterBm.PixelFormat:= pf32bit;
  meterBm.Canvas.Brush.Style:= bsSolid;
  meterBm.Canvas.Brush.Color:= clWhite;
  SetRect(meterRect, 0, 0, meterBox.Width, meterBox.Height);
  meterPos:= cHalfMeterLast;
  FillMeter;
  DrawMeterPt;

  {color to be used in the curves of foliations}
  curveColor:= meterBm.canvas.Pixels[meterPos, 6];

  {build the backgrounds palette, with colors given by hues 0 and 0.5}
  BuildPlt(bckgPal, 0, 0.5);
end;

destructor TLinePalette.Destroy;
begin
  HSBm.Free;
  meterBm.Free;
  inherited destroy;
end;

procedure TLinePalette.DrawHSPoint;
{draw spot in the hue-saturation circle}
begin
  HSBox.Canvas.Ellipse(HSPt[1] + 3, HSPt[2] + 3, HSPt[1] - 3, HSPt[2] - 3);
end;

procedure TLinePalette.DrawMeterPt;
{draw spot in the color meter}
begin
  meterBox.Canvas.Ellipse(meterPos + 3, 3, meterPos - 3, 9);
end;

procedure TLinePalette.DrawWheel;
{draw the hue-saturation circle with light = 0.6}
var
  i, j, k: Integer;
  r: Double;
  hs: Vector2;
  line: pTColorArray;
begin
  for j:= 0 to HSRect.Bottom - 1 do begin
    line:= HSBm.ScanLine[j];
    for i:= 0 to HSRect.right - 1 do begin
      if PtInWheel(r, i, j) then begin
        if r = 0 then begin {draw the central pixel}
          k:= round(0.6 * cMaxComp) and $FF;
          line[i]:= k + k shl 8 + k shl 16;
        end else begin
          hs:= FromWheelToHS(i, j); {obtain the hue and saturation}
          line[i]:= FromHSLToRGB(hs, 0.6); {convert to RGB and draw pixel}
        end
      end;
    end;
  end;
end;

procedure TLinePalette.FillMeter;
{fill meter with colors from black to white with hue and saturation given
 by the position of the spot in the color circle}
var
  hs: Vector2;
  step: Double;
  cc: array[0..cMeterLast] of TColor;
  i, j: Integer;
  line: pTColorArray;
begin
  {from the position of the spot in the color circle, 'HSPt', find the
  hue and saturation of the color, 'hs'}
  hs:= FromWheelToHS(HSPt[1], HSPt[2]);

  {fill the meter }
  step:= 1 / cMeterLast;
  for i:= 0 to cMeterLast do
    cc[i]:= FromHSLToRGB(hs, i * step);
  for j:= 0 to 11 do begin
    line:= meterBm.Scanline[j];
    for i:= 0 to cMeterLast do
      line^[i]:= cc[i];
  end;
end;

function TLinePalette.FromWheelToHS(x, y: Double): Vector2;
{from a point in the color circle, obtain the corresponding values
of hue and saturation}
var
  a, b: Double;
begin
  a:= HSctr[1] - x;
  b:= y - HSctr[2];
  result[2]:= hypot(a, b) / cWheelRadius; {saturation: value between 0 and 1}
  if result[2] > 0 then begin
    result[1]:= 0.5 + myArctan2(b, a) / cTwoPi; {arctan2 gives values between -pi and pi.
    Therefore arctan2/cTwoPi gives values between -0.5 and 0.5. So result[1] goes
    from 0 to 1}
    result[1]:= result[1] - int(result[1]);
    if result[1] > 1 then
      result[1]:= result[1] - 1
    else if result[1] < 0 then
      result[1]:= result[1] + 1;
  end else {the point is in the center of the circle. Give any value to its hue,
        because its saturation will be zero and the final color will be grey}
    result[1]:= 0.5;
end;

function TLinePalette.FromHSLToRGB(hs: Vector2; b: Double): TColor;
{'hs[1] means hue; 'hs[2]', saturation; and 'b', light. Applies
 a standard formula for conversion from HSL to RGB, where the RGB components
 go from 0 to 'cMaxComp'}
const
  oneover3 = 1 / 3;
  oneOver6 = 1 / 6;
  twoOver3 = 2 / 3;
var
  k: Integer;
  p, q: Double;
  t: vector3;
begin
  if (b <= 0.5) then
    q:= b * (1 + hs[2]) * cMaxComp
  else
    q:= (b + hs[2] * (1 - b)) * cMaxComp;
  p:= 2 * b * cMaxComp - q;
  t[2]:= hs[1];
  t[1]:= hs[1] + oneover3;
  if t[1] > 1 then
    t[1]:= t[1] - 1;
  t[3]:= hs[1] - oneover3;
  if t[3] < 0 then
    t[3]:= t[3] + 1;
  for k:= 1 to 3 do
    if t[k] < oneOver6 then
      t[k]:= p + 6 * (q - p) * t[k]
    else if t[k] < 0.5 then
      t[k]:= q
    else if t[k] < twoOver3 then
      t[k]:= p + 6 * (q - p) * (twoOver3 - t[k])
    else
      t[k]:= p;
  result:= (round(t[1]) and $FF) +
    ((round(t[2]) and $FF) shl 8) +
    ((round(t[3]) and $FF) shl 16);
end;

procedure TLinePalette.Initiate;
begin
  FillMeter;
end;

procedure TLinePalette.MouseMoveInMeter(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if (ssLeft in Shift) and (x <> meterPos)
    and PtInRect(meterRect, Point(x, y)) then begin
    meterPos:= x;
    curveColor:= meterBm.Canvas.Pixels[x, 5];
    MeterBox.Canvas.Draw(0, 0, meterBm);
    DrawMeterPt;
  end;
end;

procedure TLinePalette.MouseMoveInHSBox(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  a: Double;
begin
  if (ssLeft in Shift) and PtInWheel(a, x, y) then begin
    HSBox.Canvas.Draw(0, 0, HSBm);
    HSPt[1]:= x;
    HSPt[2]:= y;
    DrawHSPoint;
    FillMeter;
    curveColor:= meterBm.Canvas.Pixels[meterPos, 5]; {new color for curves of foliations}
    OnPaintMeter(nil);
  end;
end;

procedure TLinePalette.PaintHSBox(Sender: TObject);
begin
  HSBox.Canvas.Draw(0, 0, HSBm);
  HSBox.Canvas.Pen.Mode:= pmCopy;
  HSBox.Canvas.Pen.Color:= clWhite;
  HSBox.Canvas.Brush.Color:= clBlack;
  DrawHSPoint;
end;

procedure TLinePalette.OnPaintPicker(Sender: TObject);
begin
  Initiate;
end;

function TLinePalette.PtInWheel(var r: Double; h, v: Integer): Boolean;
begin
  r:= hypot(h - HSCtr[1], v - HSCtr[2]); {distance of point from the center}
  result:= (r <= cWheelRadius); {it must be less or equal than radius}
end;

procedure TLinePalette.OnPaintMeter(Sender: TObject);
{splash the bitmap on the panel and draw the spot}
begin
  meterBox.canvas.Draw(0, 0, meterBm);
  DrawMeterPt;
end;

end.

