unit FunctionAndPartDerivs_64;

{.$mode delphi}
{.$mode object pascal}


{Unit Name: FunctionAndPartDerivs_64
parser  for real functions of several variables of Double type. For X86-64 machines
and Linux systems.

Author:    Angel Montesinos Amilibia
 November 2, 2018 -
 Purpose: Class TFunctionParser_64. Parses a text for a "Realfunction" type.
This text may contain a mathematical formula with at most five
independent variables, with names chosen among 'p q r s t u v w x y z',
and at several parameters with names chosen among same characters
not already used as independent variables.

Then it builds an array of executable functions for the nitial input
function and all of its partial derivatives of order no greater than
three. That array isd declared as
   'ff : array of array of array of RealFunction',
and it is such that if X is an array of Double, with at least so many
entrees as independent variables used in the input, the value of ff[i,j,k](X) is
as follows, where  a  denotes the number of used independen vars:
   ff[0,0,0](X)  : the Value of the input function at X.
   ff[1..a,0,0](X) : the values of its first order partial derivatives at X
   ff[1..a, 1..a, 0](X) : the values of its second order partial derivatives at X
   ff[1..a, 1..a, 1..a](X): : the values of its third order partial derivatives at X.

 History: I am indebted to Dmitry Kaschenko and Vladimir Safin for the general
    idea of translating an evaluation tree to direct X86 opcode that I learned
    by studying their component "FormulaCompiler".

Copyright (C) 2018 by Angel Montesinos Amilibia

This unit is free software; you can redistribute it and/or modify it under
the terms of the MIT license, that is:

 ..................................................................

MIT License

Copyright (c) [2018] [Angel Montesinos-Amilibia]

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

....................................................................

}



interface
uses
  InfByteArrays;

const
  kcvNum = 6; {number of special constants that the evaluator may
               insert in the object code}
  kInitParams = 12;
  kInitIntegers = 23;
  kEndIntegers = 32;
  kEndPrevConst = 40; {initial value for EndFD}


  fpTypeSize = SizeOf(Double);
  eMaxNum: Double = 1.0e100;
  eMinNum: Double = 1.0e-100; {must be equal to 1/eMaxNum}
  eMaxExp: Double = 230;
  eHalfPi: Double = pi/2;

    {error messages}
  eThereIsNoError = 'There is no error';
  eBadNumber = 'A number is miswritten.';
  ebadOpSeq = 'Bad sequence of + - * / ^';
  eBadCharacters = 'You have used an illegal character.';
  eBadFunctionOrVariables = 'You mispelled a function or variable';
  eBadParentheses = 'Check the parentheses, please.';
  eTooMuchConstants = 'I cannot digest that much constants.';
  eSyntaxError = 'There is a syntax fault in the formula.';
  eErrorInParsing = 'I did fail in analysing the function.';
  eExpressionTooLarge = 'The expression is too long.';
  eNothingWritten = 'Some field may be empty';
  eBadVar = 'The allowed variables in this formula are: ';
  eBadVarSet = 'The allowed variables must be in the range ''p'' to ''z''.';
  eBadParamSet = 'Bad parameter set';
  eTooMuchParts = 'Implicit function too long';
  eCouldnotCreateBlock= 'Could not create code block';
  eEmptyFormula= 'Formula reduces to nothing';
  eErrorInPowerConversion = 'Power symbols conversion failed';
  eNoIntegrals = 'No integrals allowed';

type
  BArray13= array[1..3] of Byte;
  BArray2= array[0..1] of Byte;
  BArray3= array[0..2] of Byte;
  BArray4= array[0..3] of Byte;
  BArray6= array[0..5] of Byte;
  BArray8= array[0..7] of Byte;
  BArray10= array[0..9] of Byte;


  {We shall store the pointers to the opcodes, whose type is 'PByte' in a unique array}

  Array253D = array[0..253] of Double;
  RealFunction = function (const X: array of Double): Double of object;
  AnsiCharSet= Set of AnsiChar;
  string2 = string[2];
  goodStrPairs = array[0..21] of string2;

const
  goodStrings: goodStrPairs=('++', '+', '+-', '-', '-+', '-', '--', '+', '*+', '*', '*-', '*#',
                             '**', '^', '/+', '/', '/-', '/#', '^+', '^', '^-', '^#');


type
  WrittenFunction = record
    t : AnsiString;
    by : Byte;
  end;

  const
    bA= $41 {'A'};
    bB= $42 {'B'};
    bC= $43 {'C'};
    bD= $44 {'D'};
    bE= $45 {'E'};
    bF= $46 {'F'};
    bG= $47 {'G'};
    bH= $48 {'H'};
    bI= $49 {'I'};
    bJ= $4A {'J'};
    bK= $4B {'K'};
    bL= $4C {'L'};
    bM= $4D {'M'};
    bN= $4E {'N'};
    bO= $4F {'O'};
    bP= $50 {'P'};
    bQ= $51 {'Q'};
    bR= $52 {'R'};
    bS= $53 {'S'};
    bT= $54 {'T'};
    bU= $55 {'U'};
    bV= $56 {'V'};
    bW= $57 {'W'};
    bX= $58 {'X'};
    bY= $59 {'Y'};
    bZ= $5A {'Z'};

    bOpPar= ord('(');
    bClPar= ord(')');

    bQuestion = ord('?');

    kMaxNumOfNames = 50; {The number of strings of length less}
                       {than 8 that the parser recognizes as}
                       {function names. For instance, asin  }
                       {and arcsin are both recognized.     }

    knownFuncs : array[1..kMaxNumOfNames] of
      WrittenFunction=
           ((t:'COS'; by: bA {A}),
            (t:'SIN'; by: bB {B}),
            (t:'SEN'; by: bB {B}),
            (t:'EXP'; by: bC {C}),
            (t:'SQR'; by: bD {D}),
            (t:'SQRT'; by: bE {E}),
            (t:'LN'; by: bF {F}),
            (t:'LOG'; by: bF {F}),
            (t:'TG'; by: bG {G}),
 {10}       (t:'TAN'; by: bG {G}),
            (t:'SH'; by: bH {H}),
            (t:'SINH'; by: bH {H}),
            (t:'CH'; by:  bI {I}),
            (t:'COSH'; by: bI {I}),
            (t:'TH'; by: bJ {J}),
            (t:'TANH'; by: bJ {J}),
            (t:'ARCSEN'; by: bK {K}),
            (t:'ASEN';  by: bK {K}),
            (t:'ARCSIN';  by: bK {K}),
 {20}       (t:'ASIN';  by: bK {K}),
            (t:'ARCCOS'; by: bL {L}),
            (t:'ACOS'; by: bL {L}),
            (t:'ARCOS'; by: bL {L}),
            (t:'ARCTG'; by: bM {M}),
            (t:'ARCTAN'; by: bM {M}),
            (t:'ATAN'; by: bM {M}),
            (t:'ABS'; by: bN {N}),
            (t:'COTG'; by: bO {O}),
            (t:'COTAN'; by: bO {O}),
 {30}       (t:'COT'; by: bO {O}),
            (t:'ARGCH'; by: bP {P}),
            (t:'ACOSH'; by: bP {P}),
            (t:'ARGCOSH'; by: bP {P}),
            (t:'ARCCOSH'; by: bP {P}),
            (t:'ARCTANH'; by: bQ {Q}),
            (t:'ARGTANH'; by: bQ {Q}),
            (t:'ARGTH'; by: bQ {Q}),
            (t:'ATANH'; by: bQ {Q}),
            (t:'ARCSINH'; by: bR {R}),
 {40}       (t:'ASINH'; by: bR {R}),
            (t:'ARGSH'; by: bR {R}),
            (t:'ARGSINH'; by: bR {R}),
            (t: 'INT'; by: bS {S}),
            (t: 'INTEGRAL'; by: bS {S}), {not used here}
            (t: 'SGN'; by: bT {T}),
            (t: 'SIGN'; by: bT {T}),
            (t: 'INV' ; by: bU {U}),
            (t: 'CHS' ; by: bV {V}),
            (t: 'CUBE'; by: bW {W}), {x -> x^3}
            (t: 'QUAD'; by: bX {W}) {x -> x^4}
           );
  type

  capitals = 'A'..'Z';
  capitalSet= Set of Capitals;


  TFunctionAndPartDerivs_64 = class(TObject)
  {It is supposed that the number of independent variables is less or equal to 5. They must
  be named by a subset of (p,q,r,s,t,u,v,w,x,y,z). The parameters must be a subset
  of the set, but with empty intersection to the subset of independent variables. The
  call to 'ParseNewText' leads to the preparation of the input function and all
  its partial derivatives of order 1 to 3.}
  private
    rawText : AnsiString; {the text of the function}
    uFormula: AnsiString;
    allowedVars, allowedPars, allVars: AnsiString; {variable names that may appear in 'rawText'.
                 must be a subset of [p,q,r,s,t,u,v,w,x,y,z]}
    reduceConsts: Boolean; {whether in 'ParseNewText' must apply  'ReduceConstants'}
    lastPreserved: Byte; {last preserved item from the overflowing FPU stack}
    {------------------------------------------------------}

    emptyRegisters: Byte;

    FBlocks: array of PByte; {executable DEP block where the object codes of the function and
       its partial derivatives are stored}
    FBlockLengths: array of Integer; {likst of the lengths of FBloks}
    lastBlock: Integer;

    auxBlock: PByte;
    auxBlockLength: Integer;


    FuncAssigner: array of array of array of Integer; {The partial deritative f[i,j,k]  of f is
      the function whose FBlock is  FBlocks[FuncAssigner[i,j,k]]. This table is built by the
      procedure 'AssignFunctions'. The initial function has as FBlock equal to FuncAssigner[0,0,0],
      its first derivatives have as FBlock  FuncAssigner[1..argNum,0,0]. The second derivatives
      have FuncAssigner[1..argNum, 1..argNum,0] and the third derivatives
      FuncAssigner[1..argnum, 1..ArgNum, 1..ArgNum]}

    fs: array of RealFunction;
      {The functin and its pds. Use it as  ff[2,1,0](X), where '2' denotes the second variable
      independent , the '1' the first one, and 'ff[2,1,0](X)' is the second partial derivative
      of the input function with respect to the second and first independent variables. Of
      course, that value is the same as that of 'ff[1,2,0](X)'}

    fCode: InfBArray;
    depth: Integer;

    procedure AssignFunctions;
    function AtomBA(theCode: AnsiSTring): AnsiSTring;
    function BuildPD(blockInd, varInd: Integer; source: AnsiString): AnsiString;
    procedure ChangeMinusAndDiv(var t: AnsiString);
    function CheckFunction : String;
    procedure CodifyFunction(var Block: PByte; var theF: RealFunction; var bLength: Integer;
             action: AnsiString); {builds 'fC' from 'actorList'}
    function ComputeCodeLength(str: AnsiString): Integer;
    procedure ConvertPowers(var t: AnsiString);
    function ExpressionBA(theCode: AnsiSTring) : AnsiSTring;
    function FactorBA(theCode: AnsiString): AnsiString;
    function InNumber(targetStr : AnsiString; var startChar : AnsiChar; var startPos : Integer) : AnsiString;
    function InText(targetStr: AnsiString; var startChar : AnsiChar; var startPos : Integer) : AnsiString;
    procedure LeaveRoomInFPUStack(neededPlaces: Integer; var cod: InfBArray);
    function  LengthLeaveRoom(neededPlaces: Integer): Integer;
    function LengthPopFromStack: Integer;
    procedure LoadData(const j: Word; var yarr: InfBArray);
    procedure LoadX(i: Byte; var yarr: InfBArray);
    procedure OCStr(stub: ByteArray; var yarr: InfBArray);
    procedure PopFromAppStack;
    procedure ReduceConstants(var str: ansiString);
    procedure ReduceSymbols(source: AnsiString; var dest: AnsiString);
    function TermBA(theCode: AnsiString) : AnsiString;
    procedure UnifyChars;
    function UseTilde(t: AnsiString; var xp: Double) : Boolean;
    function APD(t: AnsiString;  ind: Integer):  AnsiString;
    function TPD(t: AnsiString;  ind: Integer):  AnsiString;
    function FPD(t: AnsiString;  ind: Integer):  AnsiString;
    function FunctionPD(t: AnsiString; b: AnsiChar; ind: Integer):  AnsiString;
    function XPD(t: AnsiString;  ind: Integer):  AnsiString;


  //protected

  public
    uReduced : AnsiString;
    actorList : AnsiString; {the RPN string that represents the function}
    argNum, parNum : Integer; {Length[vars], Length[params] as called in "ParseNewText"}
    endFD, endVD: Byte;
    data : Array253D; {stores the values of general constants, independent variables,
       parameters, constants that appear in the formula, and the
       intermediate values of velocities and accelerations as follows:
          useful consts:    data[0]:= 0;  (unused)
                            data[1] := eMaxNum;
                            data[2] := -eMaxNum;
                            data[3] := eMinNum;
                            data[4] := -eMinNum;
                            data[5] := eMaxExp;
                            data[6] := eHalfPi;
                            data[7]..data[11],  unused and MUST remain unused
                            data[12]..data[22], places reserved for parameters;
                            data[23]..data[32],  values 1..5, -1..-5;
                            data[33] .. data[40], values 1/2..1/5, -1/2..-1/5;
                            data[41 .. endFD],   other constants;
          kcvNum:            6
          kinitParams:      12
          kendParams:       12 + parNum-1
          knitIntegers:     23
          kendInteger:      32
          kEndPrevConst:    40  (initial value for endFD)
          other consts:     41 .. endFD
       where
          kcvNum = 6; //data[6] is the last basic number used make more robust the code
          parNum:= theF.parNum;
        }

    usedVarsAndParams: Set of Capitals;
    f: RealFunction; {f= ff[0,0,0]}
    ff : array of array of array of RealFunction;//function(const X : array of Double) : Double;
     {'X' must be an array of Double whose length is 'argNum'.
      Gives as result the value of the function at 'X'.}
    tirar: Int64;
    procedure BuildPDs;
    constructor Create;
    destructor Destroy; override;
    procedure ParseNewText(t, vars, pars: AnsiString; codify, withPDs: Boolean);
    {'t' is the text of the function (capitalization is unimportant); 'vars' is
     a string with some of the characters in 'PQRSTUVWXYZ'. ' pars '
     is a string with some of the characters in 'PQRSTUVWXYZ' that do not
     appear in 'vars'. The characters in 'vars' will be treated as independent
     variables, that is, if vars= 'xyw', the function will be called as
     f(3, 7.21, 2.4) and this will mean that the value of the variable x is 3,
     that of y is 7.21 and that of w is 2.4. However, the function f may have
     a text as  t = 'cos(p x) + q y^2 w'. Then 'pars' would be 'pq'. The value
     of the parameters p and q will be passed to the function using the
           procedure UpdateParameters;}
    function EvaluateActorList(str: AnsiString): Double;
    procedure UpdateParameters(parameters: array of Double);

  end;
FourFPD= array[1..4] of TFunctionAndPartDerivs_64;

procedure ReadConstantField(theStr: AnsiString; var val: Double); overload;
function ReadConstantField(theStr: AnsiString): Double; overload;

var
  theFPD: FourFPD;
  theFPDClones: array of FourFPD;


implementation
uses
  SysUtils, math,
  {$IFDEF MSWINDOWS}
  Windows,
  {$ELSE IF LINUX}
  BaseUnix,
  {$ENDIF}
  Forms;
type

  pieceClass = (openParent, closeParent, operationSymb,
    negation, constantSymb, varSymb, functionSymb, badThing);


const
  c_sfr: Integer= $01A0; {room for storing locals and overflowings of fpu stack}
  c_LastReg = 7; {last register in the FPU}

  upCaseLetters = ['A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z'];
  variables = ['P','Q','R','S','T','U','V','W','X','Y','Z'];
  digits = ['0','1','2','3','4','5','6','7','8','9'];
  operatorsChar = [';','+','-','*','/','^', #$7E];
  properOpChars = ['+','-','*','/','^', #$7E];


  {AnsiChar constants}
  kReturn = AnsiChar(13);
  kLineFeed = AnsiChar(10);
  kTab = AnsiChar($09);
  kLastChar = #$FF;
  kTilde= #$7E;
  kFE= #$FE;

var
  uAttribution : array['#'..'z'] of pieceClass;

  {Code chunks. Don't change them when you change "Double"
   Complete information about the instruction sets x86
   x87, etc. may be downloaded from the sites of Intel and AMD. We use
   the instruction names that appear there and the format of hexadecimal
   characters. A byte is shown in hexadecimal manner}

  b_LoadPi: BArray2= ($D9, $EB);
  b_LoadHalfPi: array[0..9] of byte =($D9, $E8, //FLD1
                                       $D9 ,$E0, //FCHS
                                       $D9, $EB, //FLDPI
                                       $D9, $FD,  //FSCALE
                                       $DD, $D9); //FSTP(1)
  b_Load1: BArray2 = ($D9,$E8);
  b_Load0: BArray2 = ($D9,$EE);
  b_zero_comp: BArray4= ($D9, $EE, $DF, $F1);
    {4 bytes. ST(cLastReg) must be empty. Comparison of 0 with x=ST(0). After it,
     JA=$77 jumps if 0>x; JNA= $76 jumps if 0 <= x; JB= $72, if 0 < x; JNB= $73, if 0 >= x; JNE= $75 if 0<>x}
  b_1_comp: BArray4= ($D9, $E8, $DF, $F1);
    {4 bytes.  ST(cLastReg) must be empty. Comparison of 1 with x=ST(0). After it,
     JA= $77 jumps if 1>x; JNA=$76 jumps if 1 <= x; JB= $72, if 1 < x; JNB= $73, if 1 >= x; JNE=$75 if 1<>x}
  b_mod_pi: BArray8= ($D9, $EB, $D9, $C9, $D9, $F8, $DD, $D9);
    {8 bytes. Last ST must be empty}
  b_ln: BArray6= ($D9, $ED, $D9, $C9, $D9, $F1);
    {6 bytes. Requires ST(0) > 0 and last ST empty}
  b_div2: BArray6= ($D9, $E8, $D8, $C0, $DE, $F9);
    {6 bytes. Requires ST(cLastReg) empty}
  b_div_S: array[0..15] of Byte= ($D9, $EE, $DF, $F1, {zero compare}
             $75, $08, //if r_j <> 0, jump 8 bytes
             $DD, $D8,
             $DD, $D8, // nothing in FST
             $D9, $EE, $EB, $02, $DE, $F9);
    {Safe division. 16 bytes. Gives 0 if divisor is zero}
  b_inv_S: array[0..9] of Byte= ({Compare 0 st(0)}  $D9, $EE, $DF, $F1, {jump if st(0)=0}
       $74, $04, {Load 1} $D9, $E8, {Div st(0) by st(1) and pop} $DE, $F1);
     {Safe inversion 18 bytes. Gives 0 if argument is zero}
  b_times_Log2_e: BArray4= ($D9, $EA, $DE, $C9);
  b_FSTP_rec: BArray2 = ($DD,$9D); {FSTP fp64bit [RBP] plus 32 bit displacement}
  b_FLD_rec: BArray2 = ($DD,$85); {FLD fp64bit [RBP] plus 32 bit displacement}
  b_Ld_X1: BArray2 = ($DD, $02);  {FLD fp64bits [EDX]} {used only for fld the first indep var}
  b_Ld_X: BArray2  = ($DD, $42);  {FLD fp64bits [EDX] + disp8} {used for remaining
          indep vars and initial velocities and accelerations. Requires to fill it with
          its third entry, a byte, }
  b_Ld_Dt: BArray2 =  ($DD, $80);  {FLD from [EAX] + disp32}{used for
      fld formula consts, intermediate velocities and accelerations}
  b_Ld_Dt8: BArray2 = ($DD, $40);  {FLD from [EAX] + disp8}{used for freq const and params}

  b_ln_S: Array[0..20] of Byte = ({b_zero_comp:}$D9, $EE, $DF, $F1, $72, $09, $DD, $D8,
      {LoadCVPar(5): next three bytes. Fill place 10} $DD, $40, 0,
      $D9, $E0, $EB, $06, {b_ln:} $D9, $ED, $D9, $C9, $D9, $F1);

  b_2toX_S: Array[0..48] of Byte = ($D9, $C0, $D9, $E1, {LoadCVPar(5), next 3 bytes} $DD, $40, 0,
      $DF, $F1, $DD, $D8, $72, $14, {b_2toX:} $D9, $C0, $D9, $FC, $D9, $C9, $D8, $E1, $D9,
      $F0, $D9, $E8, $DE, $C1, $D9, $FD, $DD, $D9 {b_2toX ended}, $EB, $10,
      {b_zero_comp:}$D9, $EE, $DF, $F1,  $DD, $D8, $72, $05,
      {LoadCVPar(3):} $DD, $40,0 , $EB, $03, {LoadCVPar(1):} $DD, $40, 0);

  {These are the prolog and epilog when target is x86_64}

  {$IFDEF MSWINDOWS}
  b_Prolog: array[0..10] of Byte= ({push RBP: }$55,
     {RSP  <--  RSP - $01A0: }$48, $81, $EC, {Int32Str(c_sfr):} $A0,$01,$00,$00,
     {RBP  <-- RSP:} $48, $89, $E5);
  b_Epilog: array[0..16] of Byte= ({fstp  [RBP-$08]. This clears ST(0):}  $DD, $5D, $10,
           {movsd  xmm0, [RBP-$08]. Result goes to xmm0: }$F2,$0F,$10,$45,$10,
           $48,$8D,$A5, $A0,$01,$00,$00, {lea rsp,[rbp+c_sfr]. Recover initial value of RSP}
           {pop RBP:}  $5D,{ret:} $C3);
  {$ELSE IF LINUX}
  b_Prolog: array[0..13] of Byte= ({push RBP: }$55, {RBP  <-- RSP:} $48, $89, $E5,
     {RSP  <--  RSP - $01A0: }$48, $81, $EC, {Int32Str(c_sfr):} $A0,$01,$00,$00,
     {mov %rsi,%edx:} $48, $89, $F2);
   b_Epilog: array[0..9] of Byte= ({fstp  [RBP-$08]. This clears ST(0):}  $DD, $5D, $F8,
           {movsd  xmm0, [RBP-$08]. Result goes to xmm0: }$F2,$0F,$10,$45,$F8,
           {leaveq:}  $C9,{ret:} $C3);
  {$ENDIF}

  b_plus_Code: BArray2= ($DE, $C1);
  b_minus_Code: BArray2= ($DE, $E9);
  b_times_Code: BArray2= ($DE, $C9);
  b_neg_code: BArray2= ($D9, $E0);
  b_cos_code: array[0..11] of Byte=  ({c_mod_2pi:} $D9, $EB, $D8, $C0, $D9, $C9, $D9, $F8, $DD, $D9,
                                      {cos:} $D9, $FF);
  b_sin_code: array[0..11] of Byte=  ({c_mod_2pi:} $D9, $EB, $D8, $C0, $D9, $C9, $D9, $F8, $DD, $D9,
                                      {sin:} $D9, $FE);
  b_Sqr: BArray2 = ($D8, $C8);
  b_Sqrt: array[0..13] of Byte= ($D9, $EE, $DF, $F1, $72, $06,
                    $DD, $D8, $D9, $EE, $EB, $02, $D9, $FA);
  b_Tan: array[0..11] of Byte = ($D9, $EB, $D9, $C9, $D9, $F8, $DD, $D9, $D9, $F2, $DD, $D8);

  b_exp : array[0..21] of Byte = ({FLD log_2e} $D9, $EA, {FMULP st(0) st(1)} $DE, $C9, {FLD st(0)} $D9, $C0,
        {FRNDINT} $D9, $FC, {FXCH} $D9, $C9, {FSUB} $D8, $E1, {FX2M1} $D9, $F0, {ADD 1} $D9, $E8,
        $DE, $C1, {Scale} $D9, $FD, {FSTP} $DD, $D9);
  b_sinh : array[0..35] of Byte = ( {b_exp}{FLD log_2e} $D9, $EA, {FMULP st(0) st(1)} $DE, $C9, {FLD st(0)} $D9, $C0,
        {FRNDINT} $D9, $FC, {FXCH} $D9, $C9, {FSUB} $D8, $E1, {FX2M1} $D9, $F0, {ADD 1} $D9, $E8,
        $DE, $C1, {Scale} $D9, $FD, {FSTP} $DD, $D9, {FLD st(0)} $D9, $C0, {invert st(0)} $D9, $E8, $DE, $F1,
        {FSUBP st(1)-st(0)} $DE, $E9, {Div by 2} $D9, $E8, $D8, $C0, $DE, $F9);
  b_cosh :  array[0..35] of Byte = ( {b_exp}{FLD log_2e} $D9, $EA, {FMULP st(0) st(1)} $DE, $C9, {FLD st(0)} $D9, $C0,
        {FRNDINT} $D9, $FC, {FXCH} $D9, $C9, {FSUB} $D8, $E1, {FX2M1} $D9, $F0, {ADD 1} $D9, $E8,
        $DE, $C1, {Scale} $D9, $FD, {FSTP} $DD, $D9, {FLD st(0)} $D9, $C0, {invert st(0)} $D9, $E8, $DE, $F1,
        {FSUBP st(1)-st(0)} $DE, $C1, {Div by 2} $D9, $E8, $D8, $C0, $DE, $F9);
  b_tanh : array[0..33] of Byte = ({ADD st(0)+st(0)} $D8, $C0,{b_exp} {FLD log_2e} $D9, $EA, {FMULP st(0) st(1)} $DE, $C9, {FLD st(0)} $D9, $C0,
        {FRNDINT} $D9, $FC, {FXCH} $D9, $C9, {FSUB} $D8, $E1, {FX2M1} $D9, $F0, {ADD 1} $D9, $E8,
        $DE, $C1, {Scale} $D9, $FD, {FSTP} $DD, $D9, {FLD st(0)} $D9, $C0, {FLD 1} $D9, $E8, {FADD st(0)+st(1)} $DC, $C1,
        {FSUBP st(2)-st(0)} $DE, $EA, {FDIVP st(1)/(st(0)} $DE, $F9);
  b_atanh: array[0..41] of byte=({FLD st(0)} $D9, $C0, {FABS st(0)} $D9, $E1, {FCOMI 1 to st(0)} $D9, $E8,
        $DF, $F1, {JUMP if |x|<1} $77, $06, {Discard st(1)} $DD, $D9, {FSUB st(0)-st(0)} $D8, $E0, {Jump to end} $EB, $1A,
        {FXCH st(0) st(1)} $D9, $C9, {Discard st(1)} $DD, $D9, {FLD st(0)} $D9, $C0, {FLD 1} $D9, $E8,
        {FADD st(0)+st(1)} $DC, $C1, {FSUBP st(0)-st(2)} $DE, $E2,{FDIVP  st(0)/st(1)} $DE, $F1,
        {b_ln} $D9, $ED, $D9, $C9, $D9, $F1, {b_div2} $D9, $E8, $D8, $C0, $DE, $F9);


 (* b_power: array[0..77] of Byte = ({Comp 0 st(0)} $D9, $EE, $DF, $F1, {Jump if st(0)<>0} $75, $08,
       {Discard st(1)} $DD, $D9, {Discard st(0)} $DD, $D8, {Load 0} $D9, $EE, {Jump to end} $EB, $40,
       {FCOMI} $D9, $EE, $DF, $F1, {JUMP if x>0} $72, $26, {FABS} $D9, $E1, {FL{D} $D9, $C1,
       {MUL by Pi} $D9, $EB, $DE, $C9, {FCOS} $D9, $FF, {FXCH}$D9, $CA, {FXCH}$D9, $C9, {FYL2X} $D9, $F1,
       {FLD st(0)} $D9, $C0,
       {FRNDINT} $D9, $FC, {FXCH}$D9, $C9, {FSUB} $D8, $E1, {F2XM1} $D9, $F0, {Add 1} $D9, $E8, $DE, $C1,
       {FSCALE} $D9, $FD,  {FMULP} $DE, $CA, {FSTP} $DD, $D8, {JUMP  to end} $EB, $14,  {FYL2X} $D9, $F1,
       {FLD} $D9, $C0, {FRNDINT} $D9, $FC, {FXCH} $D9, $C9, {FSUB} $D8, $E1, {F2XM1} $D9, $F0,
       {ADD 1} $D9, $E8, $DE, $C1, {SCALE} $d9, $FD, {FSTP} $DD, $D9);
 *)

 b_power: array[0..29] of Byte =({Comp 0 st(0)} $D9, $EE, $DF, $F1, {Jump if st(0)<>0} $75, $04,
       {Discard st(1)} $DD, $D9, {JUMP to end} $EB, $14, {FYL2X} $D9, $F1, {FLD st(0)} $D9, $C0,
       {FRNDINT} $D9, $FC, {FXCH} $D9, $C9, {FSUB} $D8, $E1, {exp_2 st(0)-1} $D9, $F0, {Add 1}
       $D9, $E8, $DE, $C1, {Scale} $D9, $FD, {Discard st(1)} $DD, $D9);

  b_Tilde: array [0..45] of Byte = ({Comp 0 st(0)} $D9, $EE, $DF, $F1, {Jump if st(0)<>0} $75, $04,
       {Discard st(1)} $DD, $D9, {Jump to end} $EB, $24,
       {Compare 0 ST(1)} $D9, $EE, $DF, $F2, {Jump if y >= 0} $76, $0A, {Div 1 by x} $D9, $E8, $DE, $F1,
       {FXCH} $D9, $C9, {FABS} $D9, $E1, {FXCH} $D9, $C9, {Load 1} $D9, $E8, {Compare 1 st(2)}
       $D9, $E8, $DF, $F3, {Jump if y = 0} $77, $08, {FMUL} $D8, $C9, {y:= y-1} $D9, $E8, $DE, $EB,
       {JUMP -$0E} $EB, $F2, {Discard st(2)} $DD, $D9, {Discard st(1)} $DD, $D9);

  b_sgn: array[0..17] of Byte = ({Comp 0 st(0)} $D9, $EE, $DB, $F1,  {Discard st(1)} $DD, $D9,
       {Jump if st(0)<>0} $75, $02, {JUMP to end} $EB, $08, {Discard st(0)} $DD, $D8, {Load 1} $D9, $E8,
       {Jump if x >0} $72, $02, {FCHS} $D9, $E0);


procedure AddInt32ToCode(n: Integer; var cod: InfBArray);
{Adds to 'cod' the code of the integer 'n', in the format that is
required by the CPU and increments the index of the last meanngful entry
of 'cod'. It is supposed that the length of 'cod' is not surpassed}
var
  c: PByte;
  theInt, i: Integer;
begin
  if cod.h+4 > cod.bigH then
    raise exception.create(eWrongAttemptToWrite);
  theInt:= n;
  c := @theInt;
  for i := 0 to 3 do begin
    inc(cod.h);
    cod.arr[cod.h] := c[i];
  end;
end;

function CorrectPair(st : string2): string2;
var
  i, h: Integer;
begin
  i:= 0;
  result:= '';
  h:= High(goodStrings);
  while i < h do begin
    if st = goodStrings[i] then begin
      result:= goodStrings[i+1];
      break;
    end;
    inc(i,2);
  end;
end;

function disp32(n: Integer): BArray4;
{32 bit displacement from @data[0] to @data[n] written as ByteArray}
var
  bytes, i: Integer;
  c: PByte;
begin
  bytes:= fpTypeSize * n;
  c:= @bytes;
  for i := 0 to 3 do begin
    Result[i] := c[i];
  end;
end;

function DropBlanksEtc(tx: AnsiString): AnsiString;
var
  h, i: Integer;
  ch: AnsiChar;
  pst: string2;
  rslt: AnsiString;
begin
  rslt:= '';
  h:= Length(tx);
  i:= 1;
  if (h > 0) then
    while i <= h do begin
      ch:= tx[i];
      if ch in  [' ', kReturn, kTab, kLineFeed] then begin {Eat all these characters, or
                    convert them to a '*'}
        if ( (i>1) and  (tx[i-1] in digits) ) and ( (i < h) and (tx[i+1] in digits) ) then
          rslt:= rslt + '*'
      end
      else if (ch = '-') and   ((i = 1) or (tx[i-1]= '(')) then
        rslt:= rslt + '#'
      else if (ch in ['{', '[']) then
        rslt:= rslt+ '('
      else if (ch in  ['}', ']']) then begin
        if rslt[Length(rslt)] <> '(' then
          rslt:= rslt+ ')'
        else
          Delete(rslt, Length(rslt),1);
      end
      else if ch = '\' then
        rslt:= rslt + '/'
      else if ch in properOpChars then begin
        if (i < h) and (tx[i+1] in properOpChars) then begin
          if (i < h-1) and (tx[i+2] in properOpChars) then
              raise Exception.Create(ebadOpSeq);
          pst:= CorrectPair(string2(ch + tx[i+1]));
          if pst <>'' then begin
             rslt:= rslt + pst;
             inc(i);
          end
            else raise Exception.Create(ebadOpSeq);
        end
        else
          rslt:= rslt + ch;
      end
      else
        rslt:= rslt + ch;
      inc(i);
    end;
  result:= rslt;
end;


procedure InitializeParser;
var
  character : AnsiChar;
begin
  for character := '#' to 'z' do
    case character of
      '(':
        uAttribution[character] := openParent;
      '#' :
        uAttribution[character] := negation;
      ')' :
        uAttribution[character] := closeParent;
      '+', '-', '*', '/', '^', ';':
        uAttribution[character] := operationSymb;
      '0', '1', 'p' :
        uAttribution[character] := constantSymb;
      'q'..'z' :
        uAttribution[character] := varSymb;
      'A'..'Z' :
        uAttribution[character] := functionSymb;
    else
      uAttribution[character] := badThing;
    end; {of case}
end; {InitializeParser}


function IdentifyFunction(theFunctionText : AnsiString) : Byte;
{If 'theFunctionText' is among the known functions, IdentifyFunction }
{returns the capital letter that corresponds to it. else, it    }
{returns '?'.                                                        }
var
  i : Integer;
  foundFunction : Boolean;
begin
  i := 0;
  repeat
    i := i + 1;
    foundFunction := (theFunctionText = knownFuncs[i].t);
  until (i = kMaxNumOfNames) or foundFunction;
  if foundFunction then
    IdentifyFunction := knownFuncs[i].by
  else
    IdentifyFunction := bQuestion;
end;


function Int32Str(n: Integer): ByteArray;
var
  c: PByte;
  theInt, i: Integer;
begin
  theInt:= n;
  c := @theInt;
  SetLength(result, 4);
  for i := 0 to 3 do
    result[i] := c[i];
end;

function PeelExpression(f: AnsiString): AnsiString;
var
  i, depth, minDepth, len: Integer;
  ch: AnsiChar;
begin
  result:= '';
  minDepth:= maxInt;
  depth:= 0;
  i:= 0;
  len:= Length(f);
  while (mindepth > 0) and (i<len)  do begin
    inc(i);
    ch:= f[i];
    if ch = '(' then inc(depth)
    else if ch = ')' then dec(depth)
    else if ch in  [kLastChar, kFE] then begin
       minDepth:= min(mindepth, depth);
       inc(i);
    end
    else
      minDepth:= min(mindepth, depth);
  end;
  if mindepth > 0 then begin
    result:= Copy(f, minDepth +1, len-2*minDepth);
  end;
end;

function PieceBeginning(t: AnsiString; const last : Integer; soughtChars : AnsicharSet) :
  Integer;
{This function acts upon 't'. It is assumed that 't[last]'
is not in the string 'soughtChars'. Then, it scans backwards
't' until in reaches one of the symbols in 'soughtChars' while being at
<=zero parentheses depth. Result is 1 (if not found) or the index following that symbol.}
var
  i, depth : Integer;
  ch : AnsiChar;
  found : Boolean;
begin
  i := last + 1;
  depth := 0;
  found := False;
  while (i > 1) and not found do
    begin
      dec(i);
      ch := t[i];
      if (i > 2) and (t[i-1] in [kLastChar, kFE]) then
        i:= i-1
      else if (ch = ')') then
        inc(depth)
      else if (ch = '(') then
        dec(depth);
      if ((ch in soughtChars) and (depth = 0)) or (depth < 0) then
        found := true;
    end;
  if found then
    result := i + 1
  else
    result := 1;
end;

function PieceEndStr(const first : Integer; const theCode: AnsiString; h: Integer;
   soughtChars: AnsicharSet; var symbol : AnsiChar) : Integer;
{This function acts upon 'theCode^', from which only the chars 'theCode[0]..
theCode[h]' are meaningful. It assumes that 'theCode[first]''
it is not in 'soughtChars'. It scans forward 'theCode' until in reaches one
of the chars in 'soughtChars' while being at non-negative parentheses depth.
Result is the index previous to that byte, or 'h' otherwise . In the first
case, 'symbol' is the found char in 'soughtChars', and it is bQuestion in
the other case}
var
  i, dep: Integer;
  ch : AnsiChar;
  found : Boolean;
begin
  symbol := '?';
  i := first-1;
  dep := 0;
  found := False;
  while (i < h) and not found do
    begin
      inc(i);
      ch := theCode[i];
      if (ch = '(') then
        inc(dep)
      else if (ch = ')') then
        dec(dep);
      if dep < 0 then begin
        found:= True;
        break;
      end;
      if (ch in [kFE, kLastChar]) then
          i:= i+1
      else if ((ch in soughtChars) and (dep = 0)) or (dep < 0) then
        found := true;
    end;
  if found then
    begin
      result := i - 1;
      symbol := ch;
    end
  else
    result := h;
end;



procedure ReadConstantField(theStr: AnsiString; var val: Double);
var
  forConstFunct: TFunctionAndPartDerivs_64;
  mc: array of Double;
begin
  val := 0;
  mc:= nil;
  forConstFunct:= TFunctionAndPartDerivs_64.create;
  try
    forConstFunct.ParseNewText(theStr,'','', True, False);
    val := forConstFunct.fs[0](mc);
  finally
    forConstFunct.Free;
  end;
end;

function ReadConstantField(theStr: AnsiString): Double; overload;
var
  forConstFunct: TFunctionAndPartDerivs_64;
  mc: array of Double;
begin
  result := 0;
  mc:= nil;
  forConstFunct:= TFunctionAndPartDerivs_64.create;
  try
    forConstFunct.ParseNewText(theStr,'','', True, False);
    try
      result := forConstFunct.fs[0](mc);
    except on Exception do
      beep(100,100);
    end;
  finally
    forConstFunct.Free;
  end;
end;

procedure sn(n: Pointer; var stub: array of Byte);
var
  c: PByte;
  i, sz: Integer;
begin
  c:= PByte(@n);
  sz:= SizeOf(Pointer);
  for i := 0 to sz - 1 do
    stub[i]:= c[i];
end;


procedure TFunctionAndPartDerivs_64.AssignFunctions;
var
  i, j, k, lst: Integer;
begin
  lst:= -1;
  inc(lst);
  SetLength(FuncAssigner, argNum+1);
  for i:= 0 to argnum do begin
    SetLength(FuncAssigner[i], argNum+1);
    for j:= 0 to argNum do
      SetLength(FuncAssigner[i,j], argNum+1);
  end;

  FuncAssigner[0,0,0]:= lst;
  for i:= 1 to argNum do begin
    inc(lst);
    FuncAssigner[i,0,0] := lst;
  end;
  for i:= 1 to argNum do
    for j:= i to argNum do begin
      inc(lst);
      FuncAssigner[i,j,0]:= lst;
      if i<>j then
        FuncAssigner[j,i,0]:= lst;
    end;
  for i:= 1 to argNum do
    for j:= i to argNum do
      for k:= j to argNum do begin
        inc(lst);
        FuncAssigner[i,j,k]:= lst;   {I know that there will be repetitions but}
        FuncAssigner[i,k,j]:= lst;   {probably the penalty is unnoticeable}
        FuncAssigner[j,i,k]:= lst;
        FuncAssigner[j,k,i]:= lst;
        FuncAssigner[k,i,j]:= lst;
        FuncAssigner[k,j,i]:= lst;
      end;
  lastBlock:= lst;
end;

function TFunctionAndPartDerivs_64.AtomBA(theCode: AnsiSTring): AnsiSTring;
var
  ini, fini: Integer;
  base: AnsiSTring;
  ch: AnsiChar;
begin
  result:= '';
  ini:= 1;
  fini:= Length(theCode);
  if theCode  <> '' then begin
    base:= PeelExpression(theCode);
    if base <> '' then begin
      result:= ExpressionBA(base);
    end
    else begin
      ch:= theCode[1];
      case ch of
        '0', '1', 'p' : Result:= ch;
         kLastChar, kFE: begin
            result:= ch + theCode[2];
         end;
        'A'..'Z' :
            begin
              Assert((theCode[2] = '(') and (theCode[fini]= ')'));
              ini:= ini+2;
              base:= Copy(theCode, ini, fini-ini);
              Assert(ch <>  'S', 'AtomString');
              result:= ExpressionBA(base);
              result:= result + ch;
            end
          else
             beep(100,100);
        end;
        ReduceConstants(result);
      end;
    end;
  if result = '' then
    result:= '0';
end; {AtomString}

function TFunctionAndPartDerivs_64.BuildPD(blockInd, varInd: Integer; source: AnsiString): AnsiString;
{'Source' is the text  of the function 'g' from which we need to obtain its partial
derivative g[varInd], and Blocks[blockInd] shall be the Fblock that must contain the machine
code of g[varInd]}
var
  aCode: AnsiString;
begin
  if not (allVars[varInd] in usedVarsAndParams) then
    aCode:= '0'
  else begin
    FBlockLengths[blockInd]:= 0;
    aCode:= XPD(source, varInd);
    //ConvertPowers(aCode);
  end;
  actorList:= '';
  actorList := ExpressionBA(aCode);
  CodifyFunction(FBlocks[blockInd], fs[blockInd], FBlockLengths[blockInd], actorList);
  result:= aCode;
end;

procedure TFunctionAndPartDerivs_64.BuildPDs;
var
  source0, source1, source2, source3: AnsiString;
  i, j, k, count: Integer;
begin
  source0:= uReduced;
  for i:= 1 to argNum do begin
    count:= FuncAssigner[i, 0, 0];
    source1:= BuildPD(count,i,source0);
    ff[i,0,0]:= fs[count];
    for j:= i to argNum do begin
      count:= FuncAssigner[i, j, 0];
      source2:= BuildPD(count,j,source1);
      ff[i,j,0]:= fs[count];
      ff[j,i,0]:= fs[count];
      for k:= j to argNum do begin
        count:= FuncAssigner[i,j,k];
        {source3:= }BuildPD(count,k,source2);
        ff[i,j,k]:= fs[count];
        ff[i,k,j]:= fs[count];
        ff[j,k,i]:= fs[count];
        ff[j,i,k]:= fs[count];
        ff[k,j,i]:= fs[count];
        ff[k,i,j]:= fs[count];
      end;
    end;
  end;
  //TestOCFunc.atext:= source3;
end;


procedure TFunctionAndPartDerivs_64.ChangeMinusAndDiv(var t: AnsiString);
const
  op= ['+', '-', '*', '/'];
var
  i, k, lst: Integer;
  ch: AnsiChar;
begin
  if t = '' then
    Exit;
  lst:= Length(t);
  i:= 0;
  while i < lst do begin
    inc(i);
    ch:= t[i];
    if ch in [kLastChar, kFE] then
      inc(i)
    else if ch = '#' then begin
      k:= PieceEndStr(i+1,t,lst, op, ch);
      if k = lst then
        t:= t+')'
      else
        Insert(')', t, k+1);
      Delete(t, i, 1);
      Insert('V(', t, i);
      inc(i, 1);
      inc(lst, 2);
    end
    else if ch = '-' then begin
      k:= PieceEndStr(i+1,t,lst, op, ch);
      if k = lst then
        t:= t+')'
      else
        Insert(')',t,k+1);
      Delete(t, i, 1);
      Insert('+V(',t,i);
      inc(i,2);
      inc(lst,3);
    end
    else if ch= '/' then begin
      k:= PieceEndStr(i+1,t,lst, op, ch);
      if k = lst then
        t:= t+')'
      else
        Insert(')',t,k+1);
      Delete(t, i, 1);
      Insert('*U(',t,i);
      inc(i,2);
      inc(lst,3);
    end
  end;
end;




function TFunctionAndPartDerivs_64.CheckFunction : String;
{Checks the syntax of the function}
var
  species, speciesAnt : pieceClass;
  tok : AnsiChar;
  lastRead, lst: Integer;
  stopHere : boolean;

  function hasError : boolean;
  var
    foundError, usual : boolean;
  begin
    foundError := false;
    if species = badThing then
      foundError := true
    else
      begin
        usual := (species = operationSymb) or (species = closeParent);
        case speciesAnt of
          openParent :
            foundError := usual;
          closeParent, constantSymb, varSymb :
            foundError := not usual;
          negation, operationSymb :
            foundError := usual or (species = negation);
          functionSymb :
            foundError := (species <> openParent);
        end; {of case}
      end; {of else}
    hasError := foundError;
  end; {of hasError}

begin {CheckFunction}
  result:= '';
  if uReduced = '' then begin
    result:= eEmptyFormula;
    Exit;
  end;
  result := '';
  lst:= Length(uReduced);
  lastRead := 0;
  speciesAnt := openParent;
  repeat
    inc(lastRead);
    tok := uReduced[lastRead];
    if (tok in [kFE, kLastChar]) then begin
      lastRead:= lastRead + 1;
      species := constantSymb;
    end
    else
      species := uAttribution[tok];
    stopHere := hasError;
    speciesAnt := species;
  until (lastRead = lst) or stopHere;
  if (not stopHere) then
    begin
      species := closeParent;
      stopHere := hasError;
    end;
  if stopHere then
    result := eSyntaxError;
end; {CheckFunction}

procedure TFunctionAndPartDerivs_64.CodifyFunction(var Block: PByte;
    var thef: RealFunction; var bLength: Integer; action: AnsiString);
var
  i: Integer;
  ch: AnsiChar;
  stub: ByteArray;

  {$IFDEF MSWINDOWS}
  prot: Word;
  {$ELSE IF LINUX}
  prot: cInt;
  {$ENDIF}
begin
  lastPreserved:= 4;
  {$IFDEF MSWINDOWS}
  {if Block = nil then begin
    bLength:= ComputeCodeLength;
    prot:= PAGE_EXECUTE_READWRITE;
    Block:= VirtualAlloc(nil,bLength, MEM_COMMIT, prot);
  end
  else begin
    prot:= PAGE_EXECUTE_READ;
    VirtualProtect(Block, bLength, PAGE_EXECUTE_READWRITE, @prot);
  end;}
  if Block <> nil then begin
    VirtualFree(Block, 0, MEM_RELEASE);
    Block:= nil
  end;
  bLength:= ComputeCodeLength(action);
  prot:= PAGE_EXECUTE_READWRITE;
  Block:= VirtualAlloc(nil, bLength, MEM_COMMIT, prot);
  {$ELSE IF LINUX}
   if Block <> nil then
    if Fpmunmap(Block, bLength) <> 0 then
      raise exception.Create(eCouldnotCreateBlock);
   prot:= PROT_EXEC OR PROT_READ OR PROT_WRITE;
   bLength:= ComputeCodeLength(action);
   Block:= FPmMAP(nil, bLength+1, prot, MAP_PRIVATE or MAP_ANONYMOUS, 0, 0);
   if Block = nil then
      i:= fpgeterrno;
  {$ENDIF}
  fCode.arr:= Block;
  fCode.h:= -1;
  fCode.bigH:= bLength-1;
  IBAppend(fCode, b_Prolog);{In MSWindows, 11 bytes. In Linux, 14 bytes.
                             initiate 'fCode' with the prologue. }
  SetLength(stub, SizeOf(Pointer));
  {Save the address of 'data[0]' to rax}
  IBAppend(fCode, [$48, $B8]); {2 bytes}
  SetLength(stub, SizeOf(Pointer));
  sn(@data[0], stub);
  IBAppend(fCode, stub); {SizeOf(Pointer)}

  emptyRegisters := c_LastReg + 1;
  i:= 0;
  while i < Length(action) do
    begin
      inc(i);
      ch := action[i];
      if ch in [kLastChar, 'p' , '0',  '1'] then begin
        if ch <> kLastChar then
          LeaveRoomInFPUStack(1, fCode);
        case ch of
          kLastChar:  begin
              LeaveRoomInFPUStack(1, fCode);
              stub[0]:= ord(action[i]);
              stub[1]:= ord(action[i+1]);
              OCStr(stub, fCode);
              i:= i+1;
              dec(emptyRegisters);
            end;
          'p': begin
              IBAppend(fCode, b_LoadPi);
              dec(emptyRegisters);
            end;
          '0': begin
              IBAppend(fCode, b_Load0);
              dec(emptyRegisters);
            end;
          '1': begin
              IBAppend(fCode, b_Load1);
              dec(emptyRegisters);
            end
          else
            ;
        end;
      end
      else
        begin
          if (ch in ['+', '-', '*', '/', '^', kTilde]) then
            begin
              if (emptyRegisters = c_LastReg) then
                PopFromAppStack;
              case ch of
                '+' : IBAppend(fCode, b_plus_code) ;  //fCode:= fCode + b_plus_code ;
                '-' : IBAppend(fCode, b_minus_code) ;//fCode:= fCode + b_minus_code;
                '*' : IBAppend(fCode, b_times_code) ;//fCode:= fCode + b_times_code;
                '/' :  begin
                     LeaveRoomInFPUStack(1, fCode);
                     IBAppend(fCode, b_div_S) ;//fCode:= fCode + b_div_S;
                  end;
                '^' : begin
                    LeaveRoomInFPUStack(1, fCode);
                    IBAppend(fCode, b_Power);
                  end;
                kTilde : begin
                    LeaveRoomInFPUStack(2, fCode);
                    IBAppend(fCode, b_Tilde);
                  end;
              end;
              inc(emptyRegisters);
            end
          else {opCode of a function}
            begin
              if (ch in ['A', 'B',  'E', 'F', 'G', 'M', 'T', 'U', 'W']) then
                LeaveRoomInFPUStack(1, fCode)
              else if (ch in ['C', 'H', 'I', 'J', 'K', 'L', 'O', 'P', 'Q', 'R']) then
                LeaveRoomInFPUStack(2, fCode);
              case ch of
                '#' : IBAppend(fCode, b_neg_code);
                'A' : IBAppend(fCode, b_cos_code);
                'B' : IBAppend(fCode, b_sin_code);
                'C' : begin
                      {IBAppend(fCode, b_times_Log2_e );
                      IBAppend(fCode,  b_2ToX_S);}
                      IBAppend(fCode, b_exp);
                     end;
                'D' : IBAppend(fCode, b_Sqr);
                'E' : IBAppend(fCode, b_Sqrt);
                'F' : IBAppend(fCode, b_ln_S);
                'G' : IBAppend(fCode, b_tan);
                'H' : begin
                       (*IBAppend(fCode, [$D9, $EA, $DE, $C9]);
                       IBAppend(fCode, b_2ToX_S);
                       IBAppend(fCode, [$D9, $C0, $D9, $E8, $DE, $F1, $DE, $E9]);
                       IBAppend(fCode, b_div2);*)
                       IBAppend(fCode, b_sinh);
                     end;
                'I' : begin
                       (*IBAppend(fCode, [$D9, $EA, $DE, $C9]);
                       IBAppend(fCode, b_2ToX_S);
                       IBAppend(fCode, [$D9, $C0, $D9, $E8, $DE, $F1, $DE, $C1]);
                       IBAppend(fCode, b_div2); *)
                       IBAppend(fCode, b_cosh);
                     end;
                'J' : begin
                       {IBAppend(fCode, [$D9, $EA, $DE, $C9]);
                       IBAppend(fCode, b_2ToX_S);
                       IBAppend(fCode, [$D8, $C8, $D9, $E8, $D8, $E9, $D9, $C9,
                                        $D9, $E8, $DE, $C1, $DE, $F9]);}
                       IBAppend(fCode, b_tanh);
                     end;
                'K' : begin
                       IBAppend(fCode, [$D9, $C0, $D9, $E1]);
                       IBAppend(fCode, b_1_comp);
                       IBAppend(fCode, [$DD, $D8, $72, $0E, $D9, $C0, $D8, $C8, $D9, $E8, $DE,
                               $E1, $D9, $FA, $D9, $F3, $EB, $14]);
                       IBAppend(fCode, b_zero_comp);
                       IBAppend(fCode, [$DD, $D8]);
                       IBAppend(fCode, b_LoadHalfPi);
                       IBAppend(fCode, [$72, $02, $D9, $E0]);
                     end;
                'L' : begin
                       IBAppend(fCode, [$D9, $C0, $D9, $E1]);
                       IBAppend(fCode, b_1_comp);
                       IBAppend(fCode, [$DD, $D8, $72, $10, $D9, $C0, $D8, $C8, $D9, $E8, $DE,
                               $E1, $D9, $FA, $D9, $C9, $D9, $F3, $EB, $0E]);
                       IBAppend(fCode, b_zero_comp);
                       IBAppend(fCode, [$DD, $D8, $72, $04, $D9, $EB, $EB, $02, $D9, $EE]);
                     end;
                'M' : IBAppend(fCode, [$D9, $E8, $D9, $F3]);
                'N' : IBAppend(fCode, [$D9, $E1]);
                'O' : begin
                       IBAppend(fCode, b_mod_pi);
                       IBAppend(fCode, [$D9, $FB, $D9, $C9]);
                       IBAppend(fCode, b_div_S);
                     end;
                'P' : begin
                       IBAppend(fCode, b_1_comp);
                       IBAppend(fCode,[$77, $14, $D9, $C0, $D8, $C8, $D9, $E8, $DE, $E9,
                               $D9, $FA, $DE, $C1]);
                       IBAppend(fCode, b_ln);
                       IBAppend(fCode,[$EB, $04, $DD, $D8, $D9, $EE]);
                     end;
                'Q' : begin
                       (*IBAppend(fCode,[$D9, $C0, {fld(0), r in ST(0) and in ST(1)}
                           $D9, $E1 {abs, |r| in ST(0), r in ST(1)}]);
                       IBAppend(fCode, b_1_comp);
                       IBAppend(fCode,[$DD, $D8,{pop(0), r in ST(0)}
                           $76, $1C, {if |r| >=1, jump $1C= 28 bytes}
                           $D9, $C0, $D9, $E8, $DE, $C1, $D9, $C9,
                           $D9, $E8, $DE, $E1, $DE, $F9 {(1+r)/(1-r) in ST(0)}]);
                       IBAppend(fCode, b_ln {2 atanh(r) in ST(0)});
                       IBAppend(fCode, b_div2 {atanh(r) in ST(0)});
                       IBAppend(fCode,[$EB, $0F {jump 15 bytes}]);
                       IBAppend(fCode, b_zero_comp);
                       IBAppend(fCode,[$DD, $D8,  {pop(0), nothing in ST(0)}
                              $77, $05] {if r < 0, jump 5 bytes} );
                       LoadCVPar(5, fCode);
                       IBAppend(fCode,[$EB, $02 {jump 2 bytes}]);
                       LoadCVPar(0, fCode);*)
                       IBAppend(fcode, b_atanH);
                     end;
                'R' : begin
                       IBAppend(fCode, b_zero_comp);
                       IBAppend(fCode,[$D9, $E1, $D9, $C0, $D8, $C8, $D9, $E8, $DE, $C1,
                                $D9, $FA, $DE, $C1]);
                       IBAppend(fCode, b_ln);
                       IBAppend(fCode,[$76, $02, $D9, $E0]);
                     end;
                'T' : IBAppend(fCode, b_sgn);
                'U' : IBAppend(fCode, b_inv_S);
                'V' : IBAPPend(fCode, [$D9, $E0]);
                'W' : IBAPPend (fCode, [$D9, $C0, $D8, $C8, $DE, $C9]);
                'X' : IBAPPend (fCode, [$D8, $C8,$D8, $C8]);
                else
                  raise Exception.Create(eErrorInParsing);
                end;
            end;
        end;
    end; {del while}
  IBAppend(fCode, b_Epilog);
  //if fCode.h > 2000 then  begin
    if fcode.h > bLength then
      beep(100,100);
  //end;
  {$IFDEF MSWINDOWS}
    @theF := @fCode.arr[0];
  {$ENDIF}
  {$IFDEF linux}
    @theF :=  Block;
  {$ENDIF}
  stub:= nil;
end;

function TFunctionAndPartDerivs_64.ComputeCodeLength(str: AnsiString): Integer;
var
  i: Integer;
  ch: AnsiChar;
begin
  result:= 0;
  lastPreserved:= 4;
  {$IFDEF MSWINDOWS}
  inc(result, 11); {b_Prolog, 11 bytes}
  {$ELSE IF LINUX}
  inc(result, 14); {b_Prolog, 14 bytes}
  {$ENDIF}
  {Save the address of 'data' to rax}
  inc(result, 2); {IBAppend(ffCode, [$48, $B8]);}
  inc(result, SizeOf(Pointer)); {IBAppend(ffCode, stub);}

  //inc(result, 2+ SizeOf(Pointer));

  emptyRegisters := c_LastReg + 1;
  i:= 0;
  while i < Length(str) do
    begin
      inc(i);
      ch := str[i];
      if ch in [kLastChar, 'p' , '0',  '1'] then begin
        if ch <> kLastChar then
          inc(result, LengthLeaveRoom(1));
        case ch of
          kLastChar: begin
              inc(result, LengthLeaveRoom(1));
              inc(result, 6); {OCRBytes adds at most 6 bytes}
              i:= i+1;
              dec(emptyRegisters);
            end;
          'p': begin
                 inc(result, 2); {b_LoadPi};
                 dec(emptyRegisters);
               end;
          '0', '1': begin
                 inc(result, 2); {b_Load1 or b_load0};
                 dec(emptyRegisters);
               end
          else
            ;
        end;
      end
      else
        begin
          if (ch in ['+', '-', '*', '/', '^', kTilde]) then
            begin
              if (emptyRegisters = c_LastReg) then
                inc(result, LengthPopFromStack);
              case ch of
                '+', '-', '*' : inc(result, 2);
                '/' :
                  begin
                    inc(result, LengthLeaveRoom(1));
                    inc(result, 16); { b_div_S};
                  end;
                '^' : begin
                    inc(result,  LengthLeaveRoom(1));
                    inc(result, 30);
                  end;
                kTilde : begin
                    inc(result,  LengthLeaveRoom(2));
                    inc(result, 46);
                  end;
              end;
              inc(emptyRegisters);
            end
          else {opCode of a function}
            begin
              if (ch in ['A', 'B', 'C', 'E', 'F', 'G', 'M', 'T', 'U', 'W']) then
                inc(result, LengthLeaveRoom(1))
              else if (ch in ['H', 'I', 'J', 'K', 'L', 'O', 'P', 'Q', 'R']) then
                inc(result, LengthLeaveRoom(2));
              case ch of
                '#' : inc(result, 2);
                'A', 'B': inc(result, 12); {b_cos_code, b_sin_code}
                'C' : begin
                      inc(result, 22); {b_exp}
                     end;
                'D': inc(result, 2);  {b_Sqr}
                'E' : inc(result, 14); {b_Sqrt}
                'F' : inc(result, 21);  {b_ln_S}
                'G' : inc(result, 12);  {b_tan}
                'H', 'I' : begin
                       inc(result, 36) ; {b_sinh}
                     end;
                'J' : begin
                       inc(result, 34); {b_tanh}
                     end;
                'K' : begin
                       inc(result, 4);
                       inc(result, 4); {b_1_comp}
                       inc(result, 18);
                       inc(result, 4); {b_zero_comp}
                       inc(result, 2);
                       inc(result, 10); {b_LoadHalfPi}
                       inc(result, 4);
                     end;
                'L' : begin
                       inc(result, 4);
                       inc(result, 4); {b_1_comp}
                       inc(result, 20);
                       inc(result, 4); {b_zero_comp}
                       inc(result, 10);
                     end;
                'M' : inc(result, 4);
                'N' : inc(result, 2);
                'O' : begin
                       inc(result, 8); {b_mod_pi}
                       inc(result, 4);
                       inc(result, 16); {b_div_S}
                     end;
                'P' : begin
                       inc(result, 4); {b_1_comp}
                       inc(result, 14);
                       inc(result, 6); {b_ln}
                       inc(result, 6);
                     end;
                'Q' : begin
                       inc(result, 42); {b_atanh}
                     end;
                'R' : begin
                       inc(result, 4); {b_zero_comp}
                       inc(result, 14);
                       inc(result, 6); {b_ln}
                       inc(result, 4);
                     end;
                'T' : inc(result, 18); {b_sgn}
                'U' : inc(result, 10); {b_inv_S}
                'V':  Inc(result, 2); {change sign}
                'W':  Inc(result, 6); {cube}
                'X' : Inc(result, 4);  {x^4}
                else
                  raise Exception.Create(eErrorInParsing);
                end;
            end;
        end;
    end; {del while}
  {$IFDEF MSWINDOWS}
  inc(result, 17); {b_Epilog};
  {$ELSE IF LINUX}
  inc(result, 10); {b_Epilog};
  {$ENDIF}
end;



procedure TFunctionAndPartDerivs_64.ConvertPowers(var t: AnsiString);
{It is assumed that 'lst' is the last meaningful value in the initial 'target'. Also,
 that 'target' is a valid expression of a mathematical formula in the language used in
 TFunctionAndPartDerivs_64.ReduceSymbols. Then, this function eliminates some ambiguities.
 That is, let a^b be smallest meaningful chunck of 'target' that contains the power
 symbol '^' which appears at some place of 'target'. If 'b = b1' represents the '1',
 then 'a^b' is changed to 'a'. This procedure assumes also  the following convention:
                  a^b^c = a^(b*c), a^b^c^d = a* (b*c*d), etc.
 and then changes a^b^c, a^b^c^d to a^(b*c), a^(b*c*d), etc.}
const
  pwrs = ['^', kTilde];
var
   i, k, count, expIni, expFini, powPos, lst: Integer;
   c : AnsiChar;
   pw: Double;
   expnt, aux: AnsiString;
   good : Boolean;
begin
   lst:= Length(t);
   i:= 2; {a meaningful formula 't' cannot begin with a '^'}
   while i < lst do begin
     c := t[i];
     if (c in pwrs) then begin
       powPos:= i;
       expnt:= '';
       expFini:= powPos-1;
       count:= 0;
       repeat
         inc(count);
         expIni := expFini + 2;
         expFini := PieceEndStr(expIni, t, lst, properOpChars, c);
         expnt:= expnt + Copy(t, expIni, expFini-expIni+1);
         if c in pwrs then
           expnt:= expnt + '*';
       until not (c in pwrs) or (expFini >= lst);
       k:= PieceBeginning(t, powPos-1, ['#', '+', '-', '*', '/', '^', kTilde]);
       if UseTilde(expnt, pw) then begin
         Assert((k =1) or not(t[k-1] in pwrs), 'In convert powers');
         if pw = 0 then begin
           Delete(t, k, expfini-k+1);   {lst := lst - expfini+k-1 }
           Insert('1', t, k);    {lst := lst +1. Hence:}
           lst:= Length(t);
           i:= k+1;
         end
         else if pw = 1 then begin
           Delete(t, powPos, expFini-powPos+1);
           lst:= Length(t);
           i:= powPos;
         end
         else if pw = 2 then begin
           Delete(t, powPos, expFini-powPos+1);
           Insert('D(',t, k);
           Insert(')', t, powPos+2);
           lst:= Length(t);
           i:= powPos + 2;
         end
         else if pw = 3 then begin
           Delete(t, powPos, expFini-powPos+1);
           Insert('W(',t, k);
           Insert(')', t, powPos+2);
           lst:= length(t);
           i:= powPos+2;
         end
         else if pw = 4 then  begin
           Delete(t, powPos, expFini-powPos+1);
           Insert('X(',t, k);
           Insert(')', t, powPos+2);
           lst:= Length(t);
           i:= powPos+2;
         end
         else begin
           k:= kInitIntegers-1;
           good:= False;
           while k <= endFD do begin
             inc(k);
             good:= (pw = data[k]);
             if good then
               break
           end;
           if good then
             expnt:= kLastChar + chr(k)
           else  begin
             inc(endFD);
             data[endFD]:= pw;
             expnt:= kLastChar + chr(endFD)
           end;
           Delete(t, powpos, expFini-powPos+1);
           Insert(kTilde + expnt, t, powpos);
           lst:= Length(t);
           i:= powPos+ length(expnt) + 1;
         end
       end
       else begin
          aux:= Copy(t, k, powPos-k);
          aux:= '*A(' + kLastChar + char(6)+ '*(1+V(T('+ aux + ')))*' + expnt+')';
          Insert('N(',t, k);
          Insert(')', t, powPos+2);
          if count > 1  then
            expnt:= '(' + expnt + ')';
          Delete(t, powPos+4 , expFini-powPos);
          Insert(expnt, t, powPos+4);
          Insert(aux, t, powPos+4+Length(expnt));
          lst:= Length(t);
          i:= powPos+3+Length(expnt)+ length(aux);
       end;
     end
     else {'c' is not a power symbol}
       inc(i);
   end;
 end;


constructor TFunctionAndPartDerivs_64.Create;
var
  i: Integer;
begin
  inherited Create;
  rawText := '';
  data[0]:= 0; {used for [EAX]}
  data[1] := eMaxNum;
  data[2] := -eMaxNum;
  data[3] := eMinNum;
  data[4] := -eMinNum;
  data[5] := eMaxExp;
  data[6] := eHalfPi;
  for i:= kcvNum {=6} + 1 to kInitIntegers-1 do
    data[i]:= 0;
  for i:= 0 to 4 do begin
    data[kInitIntegers+i]:= i+1;
    data[kInitIntegers+ 5 +i]:= -(i+1);
  end;
  for i:= 1 to 4 do begin
    data[kEndIntegers+i]:= 1/(i+1);
    data[kEndIntegers+ 4+ i]:= -1/(i+1);
  end;
  endFD:= kEndPrevConst;
  b_ln_s[10]:= fpTypeSize * 5;
  b_2toX_S[6]:= fpTypeSize * 5;
  b_2toX_S[43]:= fpTypeSize * 3;
  b_2toX_S[48]:= fpTypeSize;
end;

destructor TFunctionAndPartDerivs_64.Destroy;
var
  i, j, h: Integer;
begin
  rawText := '';
  h:= High(FBlocks);
  if h >= 0 then begin
    for i:= 0 to h do begin
      {$IFDEF MSWINDOWS}
      VirtualFree(FBlocks[i], 0, MEM_RELEASE);
      {$ELSE IF LINUX}
      fpmunmap(FBlocks[i], FBlockLengths[i]);
      {$ENDIF}
      FBlocks[i]:= nil;
    end;
    FBlockLengths:= nil;
  end;

  fs:= nil;

  h:= high(ff);
  if h >= 0 then begin
    for i:= 0 to h do begin
      for j:= 0 to h do
         ff[i,j]:= nil;
      ff[i]:= nil;
    end;
    ff:= nil;
  end;


  uReduced:= '';
  actorList:= '';

  h:= high(FuncAssigner);
  if h >= 0 then begin
    for i:= 0 to h do begin
      for j:= 0 to h do
         FuncAssigner[i,j]:= nil;
      FuncAssigner[i]:= nil;
    end;
    FuncAssigner:= nil;
  end;

  inherited;
end;

function TFunctionAndPartDerivs_64.EvaluateActorList(str: AnsiString): Double;
var
  lf: RealFunction;
  mc: array of Double;
begin
  {$IFDEF MSWINDOWS}
  VirtualFree(auxBlock, 0, MEM_RELEASE);
  {$ELSE IF LINUX}
  fpmunmap(auxBlock, auxBlockLength);
  {$ENDIF}
  auxBlock:= nil;
  auxBlockLength:= 0;
  CodifyFunction(auxBlock, lf, auxBlockLength, str);
  mc:= nil;
  result:= lf(mc);
end;


function TFunctionAndPartDerivs_64.ExpressionBA(theCode : AnsiString) : AnsiString;
const
  plus =['+'];
var
  aux, aPD: AnsiString;
  first, last, initLast, lst: Integer;
  symbol: AnsiChar;
begin
  result:= '';
  if theCode <> '' then begin
    aux:= PeelExpression(theCode);
    if aux= '' then
      aux:= theCode;
    initLast:= Length(aux);
    lst := PieceEndStr(1, aux, initLast, plus, symbol);
    if lst = initLast then
      Result := TermBA(aux)
    else begin
      aPD := TermBA(Copy(aux, 1, lst));
      if aPD <> '0' then
        result:= aPD;
      if lst < initLast then
        repeat
          first:=lst+2;
          last:= initLast;
          lst := PieceEndStr(first, aux, last, plus, symbol);
          aPD:= TermBA(Copy(aux, first, lst-first+1));
          if aPD <> '0' then begin
            if result <>'' then
              result:= result + aPD + '+'
            else
              result:= aPD;
          end;
        until lst >= initLast;
      ReduceConstants(result);
    end;
  end;
  if result = '' then
    result:= '0';
end;


function TFunctionAndPartDerivs_64.FactorBA(theCode: AnsiString): AnsiString;
const
  power= ['^', kTilde];
var
  base, expnt, aux : AnsiString;
  initLast, first, lst: Integer;
  ch, symbol: ansichar;
begin
  result:= '';
  if theCode <> '' then begin
    aux:= PeelExpression(theCode);
    if aux <> '' then
      result:= ExpressionBA(aux)
    else begin
      initLast:= length(theCode);
      lst := PieceEndStr(1, theCode, initLast, power, symbol);
      if lst = initLast then
        Result := AtomBA(theCode)
      else begin
        base := AtomBA(Copy(theCode, 1, lst));
        if base  = '0' then begin
          result:= '0';
          Exit;
        end;
        expnt:= '';
        first := lst + 2;
        if first < initLast-1 then begin
          lst := PieceEndStr(first, theCode, initLast, power, ch);
          Assert(lst = initLast, 'In FactorBA');
          if ch = '?' then begin
            expnt:= ExpressionBA(Copy(theCode, first, initLast-first+1));
            if expnt = '0' then
              result:= '1'
            else if expnt = '1' then
              result:= base
            else
              result:=  expnt +  base + symbol;
          end
          else
            beep(100,100);
        end
        else begin
          expnt:= AtomBA(Copy(theCode, first, initLast-first+1));
          if base = '1' then
            result:= '1'
          else if expnt = '0' then
            result:=  '1'
          else if expnt = '1' then
            result:= base
          else
            result:= expnt + base + symbol;
        end;
      end;
    end;
  end;
  if result = '' then
    result:= '0';
end;



function TFunctionAndPartDerivs_64.InNumber(targetStr : AnsiString;
  var startChar : AnsiChar;
  var startPos : Integer) : AnsiString;
var
  valid, alreadyBlank, alreadyDot: boolean;
  charPos, stringLength: Integer;
  theChar: AnsiChar;
  a: Double;
  res: AnsiString;

  procedure PutNewNumber;
    var
      i: Integer;
      fnd: Boolean;
  begin
    fnd:= False;
    a:=StrToFloat(res);
    if a = 0 then
      res:= '0'
    else if a= 1 then
      res:= '1'
    else begin
      for i:= kInitIntegers to kEndPrevConst do
        if a = data[i] then begin
          res:= kLastChar + chr(i);
          fnd:= True;
          break;
        end;
      if not fnd then begin
         inc(endFD);
         data[endFD]:= a;
         res:= kLastChar + chr(endFD);
      end;
    end;
  end;

begin
  stringLength := length(targetStr);
  alreadyDot := false;
  valid := true;
  alreadyBlank := false;
  result := '';
  res:= '';
  charPos := startPos;
  theChar := startChar;
  if theChar = '#' then begin
     res:= '-';
     inc(charPos);
  end;
  repeat
    theChar := targetStr[charPos];
    case theChar of
      '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' : begin
          {if (charPos > 0) and targetStgr[charPos-1] = '#' then
            res:= res + '-';}
          res := res + theChar;
          if charPos = stringLength  then begin
            PutNewNumber;
          end;
        end;
      '.', ',', '''' :
          if alreadyDot or (res = '') then
            valid := False
          else begin
            alreadyDot:= True;
            res := res + '.';
          end;
      ' ' :
          if res[length(res)] = '.' then
            valid := False
          else if (charPos < stringLength) and (targetStr[charPos+1] in digits) then begin
            PutNewNumber;
            res:= res+ '*';
            alreadyBlank := true;
            startChar:= '*';
          end;
      else begin
        PutNewNumber;
        startChar:= '1';
        dec(charPos);
        alreadyBlank := true;
      end;
    end; {case}
    inc(charPos);
  until (charPos > stringLength) or (not valid) or alreadyBlank;
  dec(charPos);
  result:= res;
  if valid then
    startPos := charPos
  else begin
    startChar := '?';
    raise Exception.Create(eBadNumber);
  end;
end;

function TFunctionAndPartDerivs_64.InText(targetStr: AnsiString; var startChar : AnsiChar;
         var startPos : Integer) : AnsiString;
var
  valid: boolean;
  charPos, stringLength: Integer;
  theChar: AnsiChar;
  rslt: AnsiString;

  function PutTimes(fini: Integer): AnsiString;
  var
    k: Integer;
    ch: AnsiChar;
  begin
    valid:= True;
    result:= '';
    if (fini = startPos) then begin
       ch:= targetStr[startPos];
       if Pos(ch, allVars) > 0 then begin
         usedVarsAndParams:= usedVarsAndParams + [ch];
         result:= ch;
       end
       else
         valid:= False;
    end
    else if fini > startPos then
      ch:= targetStr[startPos];
      result:= ch;
      if Pos(ch, allVars) > 0 then
         usedVarsAndParams:= usedVarsAndParams + [ch];
      for k:= startPos+1 to fini do begin
        ch:= targetStr[k];
        if Pos(ch,allVars) > 0 then begin
          result:= result + '*'+ ch;
          usedVarsAndParams:= usedVarsAndParams + [ch];
        end
        else begin
          valid:= False;
          break
        end
      end;
  end;

  function FoundP: AnsiString;
  var
    ch: AnsiChar;
  begin
    result:= '';
    if (charPos  = stringLength) then
      result:= rslt+'P'
    else if (charPos  = stringLength-1) then begin
       ch:= targetStr[charPos+1];
       if (ch = 'I') then begin
         if charPos = startPos then
           result:= 'p'
         else
           result:= PutTimes(charPos-1) + '*p';
         inc(charPos);
       end
       else
         result:= rslt + 'P';
    end
    else begin
      ch:= targetStr[charPos+1];
      if (ch = 'I') then begin
        if (targetStr[charPos+2]= 'N') then begin
          rslt:= rslt + 'PIN';
          inc(charPos, 2);
        end
        else begin
          if charPos = startPos then
            result:= 'p'
          else
           result:= PutTimes(charPos-1) + '*p';
          inc(charPos);
        end;
      end
      else
        result:= rslt + 'P';
    end;
  end;

  function FoundParenthesis: AnsiString;
  var
    i, b, len: Integer;
    sh: AnsiString;
  begin
    len:= charPos-startPos;
    b:= bQuestion;
    if len < 2 then begin
      result:= PutTimes(charPos-1) + '*(';
      startPos:= charPos;
    end
    else begin
      for i:= 0  to len-2 do begin
        sh:= Copy(targetStr, startPos+i,  len-i);
        b:= IdentifyFunction(sh);
        if b <> bQuestion then begin
          if i > 0 then
            result:= PutTimes(startPos+i-1) + '*'+ kFE+ ansichar(b)+ '('
          else
            result:= kFE+ char(b)+ '(';
          if char(b) = 'S' then
            raise exception.Create(eNoIntegrals);
          break;
        end;
      end;
      if b = bQuestion then
        result:= PutTimes(charPos-1) + '*(';
    end;
    startPos:= charPos;
    inc(depth);
  end;

begin
  rslt:= '';
  stringLength := length(targetStr);
  valid := true;
  charPos := startPos-1;
  while (charPos < stringLength) and  valid do begin
    inc(charPos);
    theChar:= targetStr[charPos];
    case theChar of
      'A'..'O', 'Q'..'Z': begin
           rslt:= rslt + theChar;
           if charPos = stringLength then
             rslt:= PutTimes(charPos);
        end;
      'P': begin
            rslt:= FoundP;
            if rslt[Length(rslt)]= 'p' then
              break;
         end;
      '(', '[', '{': begin
           rslt:= FoundParenthesis;
           break;
        end;
      else begin
        dec(charPos);
        rslt:= PutTimes(charPos);
        break;
      end;
    end;
  end;
  if not valid then
    raise exception.Create(eBadCharacters);
  result:= rslt;
  startPos:= charPos;
  startChar:= rslt[length(rslt)];
end;

PROCEDURE TFunctionAndPartDerivs_64.LeaveRoomInFPUStack(neededPlaces: Integer; var cod: InfBArray);
var
  stub: ByteArray;
BEGIN
  IF emptyRegisters >= neededPlaces THEN
    Exit;
  SetLength(stub, 4);
  IF emptyRegisters = 0 THEN BEGIN
    inc(lastPreserved);
    inc(cod.h);
    cod.arr[cod.h] := $D9;
    inc(cod.h);
    cod.arr[cod.h] := $F6;  {fdecstp: ST(cLastReg] becomes ST(0)}
    IBAppend(cod, b_FSTP_rec);
    stub:= Int32Str(c_sfr-8*lastPreserved);  {fstp [RBP - 8*lastPreserved]}
    IBAppend(cod, stub);
    inc(emptyRegisters);
  END;
  IF emptyRegisters < neededPlaces then begin
    if emptyRegisters = c_LastReg then begin
      inc(lastPreserved);
      IBAppend(cod, b_FSTP_rec);
      stub:= Int32Str(c_sfr-8*lastPreserved);  {fstp [RBP - 8*lastPreserved]}
      IBAppend(cod, stub);
      inc(emptyRegisters);
    end
    else REPEAT
      inc(lastPreserved);
      inc(cod.H);
      cod.arr[cod.H]:= $D9;
      inc(cod.H);
      cod.arr[cod.H]:= $C0 + c_LastReg - emptyRegisters; {push last occupied ST to ST(0)}
      IBAppend(cod, b_FSTP_rec);
      stub:= Int32Str(c_sfr-8*lastPreserved);  {fstp [RBP - 8*lastPreserved]}
      IBAppend(cod, stub);
      inc(cod.H);
      cod.arr[cod.H]:= $DD;
      inc(cod.H);
      cod.arr[cod.H]:= $C0 + c_LastReg - emptyRegisters;
         {last occupied place contains the former value: declare it free}
      inc(emptyRegisters);
    UNTIL emptyRegisters = neededPlaces;
  end;
  stub:= nil;
END;

function  TFunctionAndPartDerivs_64.LengthLeaveRoom(neededPlaces: Integer): Integer;
begin
  result:= 0;
  if emptyRegisters >= neededPlaces then
    Exit;
  if emptyRegisters = 0 then begin
    (*inc(result, 2); {fdecstp: ST(cLastReg] becomes ST(0)}
    inc(result, 6); {b_FSTP_rec + disp32} *)
    inc(result, 8);
    inc(emptyRegisters);
  end;
  if emptyRegisters < neededPlaces then begin
    if emptyRegisters = c_LastReg then begin
      inc(result, 6); {b_FSTP_rec + disp32}
      inc(emptyRegisters);
    end
    else repeat
      (*inc(result, 2); {push last occupied ST to ST(0)}
      inc(result, 6); {b_FSTP_rec + disp32}
      inc(result, 2); {delare freelast occupied place} *)
      inc(result, 10);
      inc(emptyRegisters);
    until emptyRegisters = neededPlaces;
  end;
end;


function  TFunctionAndPartDerivs_64.LengthPopFromStack: Integer;
VAR
  m : byte;
BEGIN
  result:= 0;
  m := c_LastReg + 1 - emptyRegisters; {ST(m) is the first empty register}
  if m = 1 then
    inc(result, 8)
  else
    inc(result, 10);
  dec(emptyregisters);
END;

procedure TFunctionAndPartDerivs_64.LoadData(const j: word; var yarr: InfBArray);
{Reads from the address of the var data[j] as a Double and stores it in ST(0)}
var
  bytes, i: Integer;
  c: PByte;
begin
  if j= 0 then
    IBAppend(yarr, [$DD,$00])
  else begin
    IBAppend(yarr, b_Ld_Dt);
    bytes:= fpTypeSize * j;
    c:= @bytes;
    for i := 0 to 3 do begin
      inc(yarr.h);
      yarr.arr[yarr.h] := c[i];
    end;
  end;
end;


procedure TFunctionAndPartDerivs_64.LoadX(i: Byte; var yarr: InfBArray);
{Reads from the function stack the independent variable 'i' of FPType
type and stores it in ST(0)}
begin
  if i = 1 then
    IBAppend(yarr, b_Ld_X1)
  else begin
    IBAppend(yarr, b_Ld_X);
    inc(yarr.h);
    yarr.arr[yarr.h]:= fpTypeSize * (i-1);
  end;
end;

procedure TFunctionAndPartDerivs_64.OCStr(stub: ByteArray; var yarr: InfBArray);
{'stub' is a string of two bytes beginning with bLastByte that represent the loading
 of an argument, a parameter or a constant. Gives the bytes to be added to an
 object code string}
begin
  if stub[1] <= argNum then
    LoadX(stub[1], yarr)
  else
    LoadData(stub[1], yarr);
end;

procedure TFunctionAndPartDerivs_64.ParseNewText(t, vars, pars: AnsiString;
        codify, withPDs: Boolean);
{Parses and builds the function given in the string 't'. Each
 character in the string 'vars' identifies one of the possible arguments
 on which the function may depend. Each character in the string 'pars'
 identifies one of the possible parameters that may appear in the function.}
var
  i, j, len, last : Integer;
  ch : AnsiChar;
begin
  rawText := t;
  usedVarsAndParams:= [];
  last:= High(FBlocks);
   if last >= 0 then begin
    for i:= 0 to last do begin
      {$IFDEF MSWINDOWS}
      VirtualFree(FBlocks[i], 0, MEM_RELEASE);
      {$ELSE IF LINUX}
      fpmunmap(FBlocks[i], FBlockLengths[i]);
      {$ENDIF}
      FBlocks[i]:= nil;
    end;
    FBlockLengths:= nil;
  end;
  allowedVars:= '';
  if vars = '' then
    argNum := 0
  else begin
    len := Length(vars);
    for i := 1 to len do begin
      ch := AnsiChar(AnsiUpperCase(vars[i])[1]);
      if not (ch in variables) then
        raise Exception.Create(eBadVarSet);
      if pos(ch,allowedVars) = 0 then
        allowedVars:= allowedVars + ch;
    end;
    argNum:= Length(allowedVars);
  end;

  allowedPars:= '';
  if pars = '' then
    parNum := 0
  else begin
    len := Length(pars);
    for i := 1 to len do begin
      ch := AnsiChar(AnsiUpperCase(pars[i])[1]);
      if not (ch in variables) or (Pos(ch, allowedVars)<>0) then
        raise Exception.Create(eBadVarSet);
      if pos(ch,allowedPars) = 0 then
        allowedPars:= allowedPars + ch;
    end;
    parNum:= Length(allowedPars);
  end;

  allVars:= allowedVars + allowedPars;
  reduceConsts := (allVars <> '');
  for i:= kInitParams to kInitParams + parNum - 1 do
    data[i]:= 0;
  endFD:= kEndPrevConst;
  AssignFunctions;
  SetLength(FBlocks, lastBlock+1);
  SetLength(FBlockLengths, lastBlock + 1);
  SetLength(fs, lastBlock+1);
  SetLength(ff, argNum+1);
  for i:= 0 to argnum do begin
    SetLength(ff[i], argNum+1);
    for j:= 0 to argNum do
      SetLength(ff[i,j], argNum+1);
  end;

  UnifyChars;
  uReduced:= '';

  ReduceSymbols(uFormula, uReduced);
  if CheckFunction <> '' then
    raise Exception.Create(eSyntaxError);
  ConvertPowers(uReduced);
  ChangeMinusAndDiv(uReduced);
  actorList := ExpressionBA(uReduced);
  if codify then begin
    CodifyFunction(FBlocks[0], fs[0],  FBlockLengths[0], actorList);
    ff[0,0,0]:= fs[0];
    f:= fs[0];
  end;
  if withPDs then begin
    BuildPDs;
    tirar:= 0;
    for i:= 0 to  High(FBlocks) do
      tirar:= tirar + FBlockLengths[i];
  end;
end;



PROCEDURE TFunctionAndPartDerivs_64.PopFromAppStack;
VAR
  m : byte;
  stub: ByteArray;
BEGIN
  m := c_LastReg + 1 - emptyRegisters; {ST(m) is the first empty register}
  if m = 1 then begin
    IBAppend(fCode,  b_FLD_rec);
    SetLength(stub,4);
    stub:= Int32Str(c_sfr- 8*lastPreserved);
    IBAppend(fCode, stub);
    IBAppend(fCode, $D9);
    inc(fCode.h);
    fCode.arr[fCode.h]:= $C8 + m;
  end
  else begin
    IBAppend(fCode, $DD);
    inc(fCode.h);
    fCode.arr[fCode.h]:= $D8 + m;
    IBAppend(fCode, b_FLD_rec);
    stub:= Int32Str(c_sfr- 8*lastPreserved);
    IBAppend(fCode, stub);
    IBAppend(fCode, $D9);
    inc(fCode.h);
    fCode.arr[fCode.h]:= $C8 + m;
  end;
  dec(lastPreserved);
  dec(emptyregisters);
  stub:= nil;
END;

procedure TFunctionAndPartDerivs_64.ReduceConstants(var str: ansiString);
{Apply only to an actorList}
var
  i, len: Integer;
  a: Double;
  ch:AnsiChar;
  good: Boolean;
begin
  if not reduceConsts then
    Exit;
  len:= length(str);
  good:= True;
  if (len > 0) then begin
    i:= 0;
    while i < len do begin
      inc(i);
      ch:= str[i];
      if (ch = kLastChar) and (i<len) and (ord(str[i+1]) < kInitIntegers) then
        exit;
    end;
    if ((len = 2) and (str[1] = kLastChar)) or (len = 1) then
      exit;
    a:=EvaluateActorList(str);
    if (a = 1) then
      str:= '1'
    else if a= 0 then
      str:= '0'
    else begin
      i:= kInitIntegers-1;
      good:= False;
      while i < endFD do begin
        inc(i);
        good:= (a = data[i]);
        if good then
          break
      end;
      if good then
        str:= kLastChar + chr(i)
      else  begin
         inc(endFD);
         data[endFD]:= a;
         str:= kLastChar + chr(endFD)
      end;
    end;
  end;
end;



procedure  TFunctionAndPartDerivs_64.ReduceSymbols(source: AnsiString; var dest: AnsiString);
var
  hotSpot, formulaLength, k: Integer;
  currentChar, ch: AnsiChar;
begin {ReduceSymbols}
  dest:= '';
  formulaLength := Length(source);
  hotSpot := 0;
  repeat
    inc(hotSpot);
    currentChar := source[hotSpot];
    k:= Pos(currentChar, allVars);
    if k > 0 then begin
      if k <= argNum then
        dest:= dest +  kLastChar + chr(k)
      else
        dest:= dest + kLastChar + chr(kInitParams -1 + k-argNum)
    end
    else if currentChar in ['p', '0', '1', '#','+', '-', '*', '/','^', '(', ')']  then
      dest:= dest + currentChar
    else if currentChar = kFE then begin
      inc(hotSpot);
      ch:= source[hotSpot];
      if ch = 'S' then
        raise Exception.Create(eNoIntegrals)
      else
        dest:= dest + ch;
    end
    else if currentChar = kLastChar then begin
      ch:= source[hotSpot+1];
      dest:= dest + kLastChar + ch;
      inc(hotSpot);
    end
    else
       raise Exception.Create(eBadCharacters);
  until hotSpot >= formulaLength;
end; {ReduceSymbols}

function TFunctionAndPartDerivs_64.TermBA(theCode: AnsiString) : AnsiString;
const
  times =['*'];
var
  aux, anF: AnsiString;
  lst, last: Integer;
  symbol: AnsiChar;
begin
  result:= '';
  last:= Length(theCode);
  if last > 0 then begin
    aux:= PeelExpression(theCode);
    if aux <> '' then
      result:= ExpressionBA(aux)
    else begin
      lst := PieceEndStr(1, theCode, last, times, symbol);
      if symbol = '?' then
        result:= FactorBA(theCode)
      else begin
        aux:= Copy(theCode, 1, lst);
        result:= FactorBA(aux);
        if result <> '0' then begin
          aux:= Copy(theCode, lst+2, last-lst-1);
          anF:= ExpressionBA(aux);
          if result = '1' then
            result:=  anF
          else if anF= '0' then
            result:= '0'
          else if anF <> '1' then
            result:= result + anF + '*'
        end
      end;
    end;
  end;
  if result = '' then
    result:= '0'
  else
    ReduceConstants(result);
end;


procedure TFunctionAndPartDerivs_64.UnifyChars;
var
  place, theLength: Integer;
  currentChar, precChar, toAdd : AnsiChar;
  pst: string2;
begin {UnifyChars}
  rawText := AnsiUpperCase(rawText);
  rawText:= DropBlanksEtc(rawText);
  theLength := Length(rawText);
  uFormula := '';
  if theLength = 0 then
    raise Exception.Create(eNothingWritten);
  precChar := '(';
  toAdd := ' ';
  place := 0;
  depth := 0;
  repeat
    inc(place);
    currentChar := rawText[place];
    case currentChar of
      ' ':  toAdd := ' ';
      ';': begin
             toAdd:= currentChar;
             precChar := toAdd;
           end;
      '#': if (place < theLength) and (rawtext[place+1] in digits) then begin
               uFormula := uFormula + InNumber(rawText, currentChar, place);
               precChar:= currentChar;
               toAdd:= ' ';
             end
             else begin
                toAdd:= '#';
                precChar := '#';
             end;
      'A'..'Z' : begin
           if not (precChar in (['(','#']+operatorschar)) then
             uFormula := uFormula + '*' + InText(rawText, currentChar, place)
           else
             uFormula := uFormula + InText(rawText, currentChar, place);
           precChar:= currentChar;
           toAdd:= ' ';
         end;

      '0'..'9': begin
           if not (precChar in (['(','#']+operatorschar)) then
             uFormula := uFormula + '*';
           if (currentChar = '1') and (place < theLength) and
                 not (rawText[place+1] in (['.',',','''']+digits)) then begin
             toAdd:= '1';
             precChar:= toAdd;
           end
           else begin
             uFormula := uFormula + InNumber(rawText, currentChar, place);
             precChar:= currentChar;
             toAdd:= ' ';
           end;
         end;
      '(' :
        begin
          if not (precChar in (['(', '#']+ operatorsChar) ) then
            uFormula := Concat(uFormula, '*');
          toAdd := currentChar;
          precChar:= toAdd;
          depth := depth + 1;
        end;

      ')':
        begin
          if precChar in (['('] + operatorsChar) then
            raise Exception.Create(eSyntaxError);
          toAdd := currentChar;
          precChar := toAdd;
          depth := depth - 1;
        end;
      ',', '.', '''' :{if user writes  7,56, or 7.56 or 7'56 it will be
                       converted to 7.56. Thus, we are not enforcing totally
                       the law that the decimal separator must be the dot.}
           if (place = 1) or (place >= theLength)  or  not (rawText[place+1] in digits) or not
                 (precChar in digits) then
              raise Exception.Create(eBadNumber)
           else begin
             toAdd := '.';
             precChar := toAdd;
           end;
      '+', '-', '*', '/', '^':
          if (place < theLength) and (rawText[place+1] in properOpChars) then begin
            if (place < theLength-1) and (rawText[place+2] in properOpChars) then
              raise Exception.Create(ebadOpSeq);
            pst:= CorrectPair(string2(currentChar + rawText[place+1]));
            if pst <>'' then begin
              uFormula:= uFormula + pst;
              precChar:= uFormula[length(uFormula)];
              inc(place);
              toAdd:= ' ';
            end
            else raise Exception.Create(ebadOpSeq);
         end
         else case currentChar of
           '+': if (precChar in (['(']+operatorschar)) then
                toAdd:= ' '
              else begin
                toAdd:= '+';
                precChar:= '+';
              end;
           '-': begin
                 if precChar = '(' then
                   toAdd:= '#'
                 else
                   toAdd := '-';
                 precChar:= toAdd;
               end;
           '*' :
             begin
               if precChar in (['(', ';']) then
                 raise Exception.Create(eSyntaxError);
              toAdd := '*';
              precChar := toAdd;
            end;
           '/':  if precChar in (['(',';']) then
                    raise Exception.Create(eSyntaxError)
                else begin
                   toAdd := currentChar; {Thus,  'cos(x)\sin(x)' will be converted to 'COS(X)/SIN(X)'}
                   precChar:= toAdd;
                end;
           '^':  if (place = theLength) or (precChar in ['(',';']) then
                 raise Exception.Create(eSyntaxError)
               else begin
                 toAdd:= '^';
                 precChar := toAdd;
              end;
            else
              ;
         end;
      kLastChar: begin
        inc(place);
        toAdd:= rawText[place];
        uFormula:= uFormula +kLastChar;
      end;
    end;
    if toAdd <> ' ' then
      uFormula := uFormula + toAdd;
  until (place = theLength) or (depth < 0);
  theLength:= Length(uFormula);
  if (uFormula[theLength] in operatorsChar) and ((theLength=1) or (uFormula[theLength-1]<> kLastChar)) then
    Delete(uFormula, theLength, 1);
  if depth <> 0 then
    raise Exception.Create(eBadParentheses);
end;

procedure TFunctionAndPartDerivs_64.UpdateParameters(parameters:
          array of Double);
{Changes the value of the parameters stored in 'data', so that they
 will have the new values in the following evaluations}
  var
    h,i: Integer;
begin
  h:= high(parameters);
  if (h >-1) and (h < parNum) then for i:=0 to high(parameters) do
    data[i+kInitParams]:= parameters[i];
end;


function TFunctionAndPartDerivs_64.UseTilde(t: AnsiString; var xp: Double) : Boolean;
{It is supposed that ' t ' is an exponent. It seeks whether the 't' is a constant integer
 In that case the result is True, and the value of 'xp' is that integer}
var
  i, h: Integer;
  ch: AnsiChar;
  str, aux: AnsiString;
  oldRedConsts: Boolean;
begin
  result:= False;
  aux:= PeelExpression(t);
  if aux = '' then
    aux:= t;
  h:= length(aux);
  if (aux = '0') then
    xp:= 0
  else if aux= '1' then
    xp:= 1;
  if (aux = '0') or (aux= '1') then
    Exit;
  if (length(aux) = 2) then begin
    i:= ord(aux[2]);
    if (aux[1] = kLastChar) and (i >= kInitIntegers) and (data[i] = trunc(data[i])) then begin
      result:= True;
      xp:= data[i];
    end
  end
  else begin
    i:= 0;
    while i < h-1 do begin
      inc(i);
      ch:= aux[i];
      if ch = kLastChar then begin
        inc(i);
        ch:= aux[i];
        if (ord(ch) < kInitIntegers) then
          Exit;
      end;
    end;
    ChangeMinusAndDiv(aux);
    oldRedConsts:= reduceConsts;
    reduceConsts:= False;
    str := ExpressionBA(aux);
    reduceConsts:= oldRedConsts;
    xp:= EvaluateActorList(str);
    result:=  ( xp= trunc(xp) );
  end;
end;

function TFunctionAndPartDerivs_64.APD(t: AnsiString;  ind: Integer):  AnsiString;
var
  i: Integer;
  anArg: AnsiString;
begin
  result:= '';
  anArg:= PeelExpression(t);
  if anArg <> '' then
    result:= XPD(anArg, ind)
  else begin
    if t = '0' then
      result:= '0'
    else begin
      if (t[1] = kFE) then
        raise exception.create(eNoIntegrals)
      else if t[1] in upCaseLetters then begin
        anArg:= Copy(t, 3, length(t)-3);
        result:= FunctionPD(AnArg, t[1], ind);
      end
      else if (t[1] = kLastChar) then begin
        i:= ord(t[2]);
       if (i = ind) then
          result:= '1'
        else
          result:= '0';
      end
    end;
  end;
  if result= '' then
    result:= '0';
end;


function TFunctionAndPartDerivs_64.TPD(t: AnsiString;  ind: Integer):  AnsiString;
var
  first, last, lst, key: Integer;
  aF, afPD, str: AnsiString;
  ch: AnsiChar;
begin
  if t = '' then
    raise Exception.Create(eSyntaxError);
  result:= '';
  first:= 1;
  last:=  Length(t);
  aF:= PeelExpression(t);
  if aF <> '' then
    result:= XPD(aF, ind)
  else begin
    lst:= PieceEndStr(first, t, last, ['*'], ch);
    if (ch = '?') then
      result:= FPD(t, ind)
    else begin
      aF:= Copy(t, 1, lst);
      aFPD:= FPD(aF, ind);
      result:= aFPD;
      str:= aF;
      if lst < last then
        repeat
          first:= lst+2;
          lst:=  PieceEndStr(first, t, last, ['*'], ch);
          aF:= Copy(t, first, lst-first+1);
          aFPD:= FPD(aF, ind);

          if (str = '0') or (aFPD = '0') then begin  {result:=  result +'*' + aF}
            if (result <> '0') and (aF <> '0') then begin
              if (result= '1') and (aF= '1') then
                result:= '1'
              else if result = '1' then
                result:= aF
              else if aF <> '1' then
                result:= '('+result +')*(' + aF + ')'
            end
            else
              result:= '0';
          end
          else if (result = '0') or (aF = '0') then begin {result:=  str + '*(' +aFPD +')'  }
            { (str <> '0') and (aFPD <> '0')  is necessary}
            if (str = '1') and (aFPD= '1') then
              result:= '1'
            else if (str = '1') then
              result:= aFPD
            else if aFPD = '1' then
              result:= str
            else
              result:= '(' +str +')*(' +aFPD +')';
          end
          else begin

            {  Now, all of result, str, aF and AfPD must be non-zero. Thus, they are 1 or
               another non-zero number.  Remember that now
                   result:= '('+result +')*' + aF + '+(' +str +')*(' +aFPD +')';            }

            key:= ord(aFPD <> '1') + 2* ord(str<> '1') + 4* ord(aF<>'1')+ 8*ord(result <> '1');
            case key of
              0 : result:= klastChar + chr(24);
              1 : result:= '1+' + aFPD;
              2 : result:='1+' + str;
              3 : result:= '1+(' +str +')*(' +aFPD +')';
              4 : result:= aF + '+1';
              5:  result:= aF +'+'+ aFPD;
              6:  result:= aF  +'+'+ str;
              7:  result:= aF   + '+(' +str +')*(' +aFPD +')';
              8:  result:= result + '+1';
              9:  result:= result  +'+'+ afPD;
              10: result:= result +'+'+ str;
              11: result:= result +  str + '*(' +aFPD +')';
              12: result:= '('+result +')*(' + aF + ')+1';
              13: result:= '('+result +')*(' + aF + ')+' +aFPD ;
              14: result:= '('+result +')*(' + aF + ')+' +str ;
              15: result:= '('+result +')*(' + aF + ')+(' +str +')*(' +aFPD +')';
            end;
          end;
          if aF = '0' then
            str:= '0'
          else if aF <> '1' then
            str:= '(' + aF + ')*(' +str +')';
        until lst >= last;
      end;
  end;
  if result = '' then
    result:= '0';
end;


function TFunctionAndPartDerivs_64.FPD(t: AnsiString;  ind: Integer):  AnsiString;
var
  first, last, i: Integer;
  x, xp, y, yp: AnsiString;
  ch: AnsiChar;
begin
  result:= '';
  first:= 1;
  last:=  Length(t);
  x:= PeelExpression(t);
  if x <>'' then
    result:= XPD(x, ind)
  else begin
    i:= PieceEndStr(1, t, last, ['^', kTilde], ch);
    if ch = '?' then
      result:= APD(t, ind)
    else begin
      x:= Copy(t, first, i-first+1);
      y:= Copy(t, i+2, last-i-1);
      xp:= APD(x, ind);
      yp := XPD(y, ind);
      if (x='0') then begin  {result:= '((' + yp + ')*F(' + x + ')*' + t +
                 '+(' + y + ')*(' + xp + ')*'+ x + ch+ '(' + y+  '+V(1))'; }

      end;
      if xp = '0' then begin
        if yp = '0' then
          result:= '0'
        else if yp ='1' then
          result:= 'F(' + x + ')*'+t
        else
          result:= '(' + yp + ')*F(' + x + ')*'+t ;
      end
      else if xp = '1' then begin
        if yp = '0' then
          result:= '(' + y + ')*'+ x+ ch+ '(' + y+  '+V(1))'
        else if yp = '1' then
          result:= 'F(' + x + ')*' + t + '+(' + y + ')*' + x + ch+ '(' + y + '+V(1))'
        else
          result:= '(' + yp + ')*F(' + x + ')*' + t + '+(' + y + ')*' + x +  ch + '(' + y + '+V(1))'
      end
      else begin
        if yp = '0' then
          result:=  '(' + y + ')*('+ xp + ')*'+ x + ch+ '(' + y + '+V(1))'
        else if yp = '1' then
          result:= 'F(' + x + ')*' + t +
                 '+(' + y + ')*(' + xp + ')*'+ x+ ch+ '(' + y+  '+V(1))'
        else
          result:= '(' + yp + ')*F(' + x + ')*' + t +
                 '+(' + y + ')*(' + xp + ')*'+ x+ ch+ '(' + y+  '+V(1))';
      end;
    end;
  end;
  if result = '' then
    result:= '0';
end;

function TFunctionAndPartDerivs_64.FunctionPD(t: AnsiString; b: AnsiChar;  ind: Integer):  AnsiString;
var
  arg, argPD: AnsiString;
begin
  result:='';
  arg:= PeelExpression(t);
  if arg = '' then
    arg:= t;
  argPD:= XPD(arg, ind) ;
  if argPD = '0' then begin
    result:= '0';
    Exit;
  end;
  if argPD <> '1' then
    case b of
      'A': result:= 'V(B(' + arg + '))*(' + argPD +')';
      'B': result:= 'A(' + arg + ')*(' + argPD +')';
      'C': result:= 'C(' + arg + ')*(' + argPD +')';
      'D': result:= kLastChar + char(24) + '*(' + arg + ')*(' + argPD +')';
      'E': result:= kLastChar + char(33) + '*U(E(' + arg + '))' + '*(' + argPD +')';
      'F': result:= '(' + argPD +')*U(' + arg + ')';
      'G': result:= '(1+D(G(' + arg + ')))*(' + argPD +')';
      'H': result:=  'I(' + arg + ')*(' + argPD +')';
      'I': result:= 'H(' + arg + ')*(' + argPD +')';
      'J': result:= '(1+ V(D(J(' + arg + '))))*(' + argPD +')';
      'K': result:= '(' + argPD + ')*U(E(1+V(D(' + arg + '))))';
      'L': result:= 'V(' + argPD + ')*U(E(1+V(D(' + arg+ '))))';
      'M': result:= '(' + argPD + ')*U(1+D(' + arg + '))';
      'N': result:= 'T('+ arg+ ')*(' +  argPd +')';
      'O': result:= 'V(' + argPD + ')*(1+D(O(' + arg + ')))';
      'P': result:= '(' + argPD + ')*U(E(V(1)+D(' + arg + ')))';
      'Q': result:= '(' + argPD + ')*U(1+ V(D(' + arg + ')))';
      'R': result:= '(' + argPD + ')*U(E(1+D(' + arg + ')))';
      'S': result:= '0';
      'T': result:=  '0';
      'U': result:=  'V(' + argPD + ')*U(D(' + arg + '))';
      'V': result:=  'V(' + argPD+ ')';
      'W': result:=  kLastChar + char(25) + '*D(' + arg + ')*(' + argPD +')';
      'X': result:=  kLastChar + char(26) + '*W(' + arg + ')*(' + argPD +')';
      else
        ;
    end
  else
    case b of
      'A': result:= 'V(B(' + arg + '))';
      'B': result:= 'A(' + arg + ')';
      'C': result:= 'C(' + arg + ')';
      'D': result:= kLastChar + char(24) + '*(' + arg + ')';
      'E': result:= kLastChar + char(33) + '*U(E(' + arg + '))';
      'F': result:= 'U(' + arg + ')';
      'G': result:= '(1+D(G(' + arg + ')))';
      'H': result:= 'I(' + arg + ')';
      'I': result:= 'H(' + arg + ')';
      'J': result:= '(1+V(D(J(' + arg + '))))';
      'K': result:= 'U(E(1+V(D(' + arg + '))))';
      'L': result:= 'V(U(E(1+V(D(' + arg+ ')))))';
      'M': result:= 'U(1+D(' + arg + '))';
      'N': result:= 'T('+ arg+ ')';
      'O': result:= 'V(1+D(O(' + arg + ')))';
      'P': result:= 'U(E(V(1)+D(' + arg + ')))';
      'Q': result:= 'U(1+V(D(' + arg + ')))';
      'R': result:= 'U(E(1+D(' + arg + ')))';
      'S': result:= '0';
      'T': result:= '0';
      'U': result:= 'V(U(D(' + arg + ')))';
      'V': result:= 'V(1)';
      'W': result:=  kLastChar + char(25) + '*D(' + arg + ')';
      'X': result:=  kLastChar + char(26) + '*W(' + arg + ')';
      else
        ;
    end;
  if result = '' then
    result:= '0';
end;

function TFunctionAndPartDerivs_64.XPD(t: AnsiString; ind: Integer):  AnsiString;
const
  plus =['+'];
var
  first, last, initLast, lst: Integer;
  aT, atPD: AnsiString;
  ch: AnsiChar;
begin
  result:= '';
  if t <> '' then begin
    first:= 1;
    last:= Length(t);
    initLast:= last;
    aT:= PeelExpression(t);
    if aT <> '' then
      result:= XPD(aT, ind)
    else begin
      lst := PieceEndStr(first, t, last, plus, ch);
      if lst = last then
        Result := TPD(t, ind)
      else begin
        aT := Copy(t, first, lst-first+1);
        first:= 1;
        last:= Length(aT);
        result := TPD(aT, ind);
        if result = '0' then
          result:= '';
        if lst < initLast then
          repeat
            first:=lst+2;
            last:= initLast;
            lst := PieceEndStr(first, t, last, plus, ch);
            aT:= Copy(t, first, lst-first+1);
            first:= 1;
            last:= Length(aT);
            aTPD:=  TPD(aT, ind);
            if (aTPD <> '0') then begin
              if result <> '' then
                result:= result + '+' + aTPD
              else
                result:= aTPD;
            end;
          until lst >= initLast;
      end;
    end;
  end;
  if result = '' then
    result:= '0';
end;




initialization;

  InitializeParser;


end.

