unit FunctionParserBytes_64_XE2;

{.$mode delphi}
{
Unit Name: FunctionParser_64
parser  for real functions of several variables of Double type. For X86-64 machines

Author:    Angel Montesinos Amilibia
 November 11, 2011 -
 Purpose: Class TFunctionParser_64. Parses a text for a "Realfunction" type.
    Encodes as a string an rpn notation of the function. That rpn string may be
    passed to object code (for x68-64) by means of the accompanying class
    TEvaluator_32. Three object codes may so be obtained: for evaluationg the
    function, its first and its second derivatives. This unit has an evaluator
    for constant expressions, that is without independent variables, for instance
    like  'cos(3 pi/5)/ sqrt(22)'.

 History: I am indebted to Dmitry Kaschenko and Vladimir Safin for the general
    idea of translating an evaluation tree to direct X86 opcode that I learned
    by studying their component "FormulaCompiler".

Copyright (C) 2006-2011 Angel Montesinos Amilibia
Parametricas4, workshop for the geometry of surfaces in R^4.


MIT License

Copyright (c) [2003 - 2018] [Angel Montesinos-Amilibia], email montesin@uv.es

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


}




interface
uses
  InfByteArrays;

const
  cN1 = 512; {The number of steps that make the grid for the
    interpolation of a function given by an integral depending on one
    variable}
  cN2 = 128; {The number of steps that make the grid for the
    interpolation of a function given by an integral depending on two
    variables}
  cQ = 15; {The greater this integer, the better will be the
    approximation of integrals by B-Splines, but with more CPU time}

  cNodes1 = cN1 + 2*cQ + 4;
  cNodes2 = cN2 + 2*cQ + 4;

  {Functions defined by integrals are allowed. Such a function may
  involve at least one and at most two independent variables
  chosen between the allowed variables "allowedvars" of the main formula.
  These variables may appear in the integrand or in the extremes of
  the integral. Then "var1" is the position in "allowedvars" (counting
  from zero) of the first variable, and "var2" the position of the second,
  if there is one (if not, var2 = -1.)

  An integral in the variables, say 'x' and 'y' must be written in the follo-
  wing form

      int( f(x,y,r) ; m(x,y) ; M(x,y) ),

  where 'f(x,y,r)' is a valid formula for a function in the variables 'x','y',
  'r'. This formula stands for the integrand, and 'r' is always the variable of
  integration;

  where 'm(x,y)' and  'M(x,y)' are valid formulas for two functions in the
  variables 'x', 'y'. They represents the inferior and superior limits of the
  integration, respectively.

  Warning: It is not alllowed to write integrals inside integrals, that is
  in the formulas for the integrand of the integration limits. Nor are allowed
  parameters inside integrals.}


  kcvNum = 5; {number of special constants that the evaluator may
               insert in the object code}
  fpTypeSize = SizeOf(Double);
  eMaxNum: Double = 1.0e100;
  eMinNum: Double = 1.0e-100; {must be equal to 1/eMaxNum}
  eMaxExp: Double = 230;

    {error messages}
  eThereIsNoError = 'There is no error';
  eBadNumber = 'A number is miswritten.';
  eBadCharacters = 'You have used an illegal character.';
  eBadFunctionOrVariables = 'You mispelled a function or variable';
  eBadParentheses = 'Check the parentheses, please.';
  eTooMuchConstants = 'I cannot digest that much constants.';
  eSyntaxError = 'There is a sintax fault in the formula.';
  eErrorInParsing = 'I did fail in analysing the function.';
  eExpressionTooLarge = 'The expression is too long.';
  eNothingWritten = 'Some field may be empty';
  eBadVar = 'The allowed variables in this formula are: ';
  eBadVarSet = 'The allowed variables must be in the range ''p'' to ''z''.';
  eBadParamSet = 'Bad parameter set';
  eNoIntegralsInsideIntegrals = 'Integrals inside integrals are not allowed';
  eBadIntegralSyntax = 'Bad integral sintax';
  eAtMostTwoVarsInInt = 'One or at most two vars are allowed in integrals, besides ''r''.';
  eNaughtIntegral = 'The integral is naught because it doesn''t contain variables.';
  eTooMuchParts = 'Implicit function too long';
  eCouldnotCreateBlock= 'Could not create code block';
  eEmptyFormula= 'Formula reduces to nothing';


type
  BArray2= array[0..1] of Byte;
  BArray3= array[0..2] of Byte; 
  BArray4= array[0..3] of Byte;
  BArray6= array[0..5] of Byte;
  BArray8= array[0..7] of Byte;
  BArray10= array[0..9] of Byte;
  Array255D = array[0..255] of Double;
  PArray255D = ^array255D;
  ArrayNodes1 = array [0..cNodes1] of Double;
  ArrayNodes2 = array [0..cNodes2] of Double;
  ArrayArrayNodes2 = array[0..cNodes2, 0..cNodes2] of Double;
  PArrayNodes1 = ^ArrayNodes1;
  PArrayArrayNodes2 = ^ArrayarrayNodes2;
  RealFunction = function (const X: array of Double): Double of object;

  firstDeriv = record  //used for the following: given a function in R^n, a
    fa0: Double;        //curve a:R --> R^n, f0 will be the value of f(a(t))
    dfa0: Double;       //as Double at t=0, dfa0 will be f(a(t))'(0) (that is the
  end;                  //value at t=0 of the derivative of f(a(t))

  secondDeriv = record //used for the following: given a function in R^n, a
    fa0: Double;        //curve a:R --> R^n, f0 will be the value as Double of
    dfa0 : Double;      //f(a(t)) at t=0, dfa0 will be f(a(t))'(0) (that is the
    ddfa0 : Double;     //value at t=0 of the derivative of f(a(t)); and ddfa0
  end;                  //will be f(a(t))''(0).

  IntegralData1 = record
    var1: Integer;
    data: array[0..2] of ArrayNodes1;
  end;

  IntegralData2= record
    var1, var2: Integer; {'var1' and 'var2' code the independent variables from
                  'allowedVars' on which the integrand or its upper and lower
                  limits may depend.}
    data: array[0..2,0..2] of arrayarrayNodes2; {A grid of data that is computed
           when parsing the function. These data serve for interpolating by
           B-splines the value of the integral as function of the vars
           whose position in 'allowedVars' is 'var1' and 'var2'}
  end;
  IntegralDataList1= array of IntegralData1;
       {For all integrals with one variable that may appear in the formula}
  IntegralDataList2= array of IntegralData2;
       {For all integrals with two variable that may appear in the formula}

  TFunctionParserBytes_64 = class(TObject)
  private
    rawText : String; {the text of the function}
    uFormula : String;
    uReduced : InfBArray;
    uReducedBuffer: ByteArray;
    allowedVars, allowedPars, allVars: String; {variable names that may appear in 'rawText'.
                 must be a subset of [p,q,r,s,t,u,v,w,x,y,z]}

    lastPreserved: Byte; {last preserved item from the overflowing FPU stack}
    {------------------------------------------------------}

    emptyRegisters: Byte;
    
    fCode: InfBArray; { 'fCode.arr' shall be equal to FBlock}
    FBlock: PByte; {executable DEP block where the object code of the function is
                    stored}
    FBlockLength: Integer;
                                                     
    endParam: Word; {end of indexes in 'data' containing, resp.,
        parameters, function constants, intermediate velocities}

    procedure AdjustIntegralDataForBSplines(prov1: PArrayNodes1;
        prov2: PArrayArrayNodes2; dim: Integer);
    function BuildIntegralDataList(tripleArgument: String): Byte;
    function CheckFunction : String;
    procedure CodifyFunction; {builds 'fC' from 'actorList'}
    function ComputeCodeLength: Integer;
    function ComputeExponent(var n : Integer; ini, fini: Integer): Boolean;
    procedure ComputeIntegralData(integrand, inferior,superior:
        TFunctionParserBytes_64); {used when integrals appear in formula}
    function ConvertIntegral(var hotSpot: Integer): Byte;
    procedure ConvertPowers;
    function FactorizeText(targetString : String; integralsAllowed: Boolean;
        var mayEndInFunction : Boolean): ByteArray;
    procedure GetRoomAndCoeffsFirstDeriv;
    procedure GetRoomAndCoeffsSecondDeriv;
    procedure LeaveRoomInFPUStack(neededPlaces: Integer; var cod: InfBArray);
    function  LengthLeaveRoom(neededPlaces: Integer): Integer;
    function LengthPopFromStack: Integer;
    procedure LoadCVPar(const j: Byte; var iba: InfBArray);
    procedure LoadData(const j: Word; var yarr: InfBArray);
    procedure LoadX(i: Word; var yarr: InfBArray); 
    procedure OCStr(stub: ByteArray; var yarr: InfBArray); 
    function PieceBeginning(const last : Integer; soughtChars : ByteSet) :
            Integer;
    procedure PopFromAppStack;
    procedure ReduceSymbols(integralsAllowed: Boolean);
    procedure SetUWLength;
    procedure UnifyChars;
  //protected

  public
    actorList : ByteArray; {the RPN string that represents the function}
    argNum, parNum : Integer; {Length[vars], Length[params] as called in "ParseNewText"}
    intList1: IntegralDataList1;
    intList2: IntegralDataList2;
    lastIntegral1, lastIntegral2, maxJ: Integer;
    minCorner, maxCorner, jump1, jump2, invJump1, invJump2: array of double;
    endFD, endVD: Word;
    vaNum: Integer; {number of entries that occupy the intermediate values of velocities,
       which is equal to that occupied by accelerations}
    data : array of Double; {stores the values of general constants, independent variables,
       parameters, constants that appear in the formula, and the
       intermediate values of velocities and accelerations as follows:
          useful consts:    data[1] := eMaxNum;
                            data[2] := -eMaxNum;
                            data[3] := eMinNum;
                            data[4] := -eMinNum;
                            data[5] := eMaxExp;
          parameters:       data[6 .. endParam],
          formula consts:   data[endParam+1 .. endFD]
          velocities:       data[endFD+1 .. endVD]
          accelerations:    data[endVD+1 .. endA],
       where
          endParam := 5 + ParNum,
          endFD := endParam + constNum,
          endVD :=  endFD + vaNum;
          endA:=   endVD + vaNum;

       After any call to Fv, Dv and DDv, @data[0] is immediately loaded in EAX}

    {---------used when integrals appear in formula--------}
    usedVarsAndParams: String;

    Fv : RealFunction;//function(const X : array of Double) : Double;
     {'X' must be an array of Double whose length is 'argNum'.
      Gives as result the value of the function at ''.}

    constructor Create;
    destructor Destroy; override;
    procedure ParseNewText(t, vars, pars: String;
        integralsAllowed, codify: Boolean; minC, maxC: array of double);
    {'t' is the text of the function (capitalization is unimportant); 'vars' is
     a string with some of the characters in 'PQRSTUVWXYZ'. ' pars '
     is a string with some of the characters in 'PQRSTUVWXYZ' that do not
     appear in 'vars'. The characters in 'vars' will be treated as independent
     variables, that is, if vars= 'xyw', the function will be called as
     f(3, 7.21, 2.4) and this will mean that the value of the variable x is 3,
     that of y is 7.21 and that of w is 2.4. However, the function f may have
     a text as  t = 'cos(p x) + q y^2 w'. Then 'pars' would be 'pq'. The value
     of the parameters p and q will be passed to the function using the
     procedure UpdateParameters.}
    {function Garbage(const X: array of Double): Double;}

  end;

  var
    theParser: TFunctionParserBytes_64;

function FormatFloatString(x: Double; totalStrLength: Integer): string;
procedure ReadConstantField(theStr: String; var val: double); overload;
function ReadConstantField(theStr: String): double; overload;

implementation
uses
  SysUtils,
  {$IFDEF MSWINDOWS}
  Windows,
  {$ELSE IF LINUX}
  BaseUnix,
  {$ENDIF}
  Forms;
type

  pieceClass = (openParent, closeParent, operationSymb,
    negation, constantSymb, varSymb, functionSymb, badThing);

  WrittenFunction = record
    t : String;
    by : Byte;
  end; 

const
  c_sfr: Integer= $01A0; {room for storing locals and overflowings of fpu stack}
  c_LastReg = 7; {last register in the FPU}

  {Don't change these constants! They are needed to convert a grid of
   points to another grid that, being used for evaluation of B-splines,
   makes the initial grid the interpolation points, that is the curve
   or surface defined by the B-splines passes by the initial points
   with great approximation that may be increased as one likes.
   For a detailed explanation of the use of B-splines, read the
   attached paper "AproxFuncBSplinesEnglish.pdf"}
  cA= -0.0755861459167170;
  cB= 2.28831641118603;
  cU= -0.01372542929733973;
  cV= -0.36134122590022017;


  kMaxNumOfNames = 44; {The number of strings of length less}
                       {than 8 that the parser recognizes as}
                       {function names. For instance, asin  }
                       {and arcsin are both recognized.     }
  upCaseLetters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
  variables = 'PQRSTUVWXYZ';
  digits = '0123456789';

  {AnsiChar constants}
  kReturn = AnsiChar(13);
  kLineFeed = AnsiChar(10);
  kTab = AnsiChar($09);
  kLastChar = #$FF;

  bA= $41 {'A'};
  bB= $42 {'B'}; 
  bC= $43 {'C'};
  bD= $44 {'D'};
  bE= $45 {'E'};
  bF= $46 {'F'};
  bG= $47 {'G'};
  bH= $48 {'H'};
  bI= $49 {'I'};
  bJ= $4A {'J'};
  bK= $4B {'K'};
  bL= $4C {'L'};
  bM= $4D {'M'};
  bN= $4E {'N'};
  bO= $4F {'O'};
  bP= $50 {'P'};
  bQ= $51 {'Q'};
  bR= $52 {'R'};
  bS= $53 {'S'};
  bT= $54 {'T'};
  bU= $55 {'U'};
  bV= $56 {'V'};
  bW= $57 {'W'};
  bX= $58 {'X'};
  bY= $59 {'Y'};
  bZ= $5A {'Z'};
  
  bPi = $70  {'p'};
  bSemiCol = $3B {';'};
  bQLow= $71{'q'};
  bZlow = $7A {'z'};
  
  b0= $30;
  b1= $31;
  b2= $32;
  b3= $33;
  b4= $34;
  b5= $35;
  b6= $36;
  b7= $37;
  b8= $38;
  b9= $39;
  
  bPlus= $2B;
  bMinus= $2D;
  bTimes= $2A;
  bDiv= $2F;
  bPow= $5E;
  bOpPar= $28;
  bClPar= $29;
  bNeg= $23;

  
  operators: ByteSet = [bPlus, bMinus, bTimes, bDiv, bPow];
  
  bLastByte= $FF;
    
  bQuestion= $3F;
   
  knownFuncs : array[1..kMaxNumOfNames] of
      WrittenFunction= 
            ((t:'COS'; by: bA {A}),
            (t:'SIN'; by: bB ),
            (t:'SEN'; by: bB {B}),  
            (t:'EXP'; by: bC {C}),
            (t:'SQR'; by: bD {D}),
            (t:'SQRT'; by: bE {E}),
            (t:'LN'; by: bF {F}), 
            (t:'LOG'; by: bF {F}),
            (t:'TG'; by: bG {G}),
 {10}       (t:'TAN'; by: bG {G}),
            (t:'SH'; by: bH {H}),
            (t:'SINH'; by: bH {H}), 
            (t:'CH'; by:  bI {I}),
            (t:'COSH'; by: bI {I}),
            (t:'TH'; by: bJ {J}),
            (t:'TANH'; by: bJ {J}),
            (t:'ARCSEN'; by: bK {K}),
            (t:'ASEN';  by: bK {K}),
            (t:'ARCSIN';  by: bK {K}),
 {20}       (t:'ASIN';  by: bK {K}),
            (t:'ARCCOS'; by: bL {L}),
            (t:'ACOS'; by: bL {L}),
            (t:'ARCOS'; by: bL {L}),
            (t:'ARCTG'; by: bM {M}),
            (t:'ARCTAN'; by: bM {M}),
            (t:'ATAN'; by: bM {M}),
            (t:'ABS'; by: bN {N}),
            (t:'COTG'; by: bO {O}),
            (t:'COTAN'; by: bO {O}),
 {30}       (t:'COT'; by: bO {O}),
            (t:'ARGCH'; by: bP {P}),
            (t:'ACOSH'; by: bP {P}),
            (t:'ARGCOSH'; by: bP {P}),
            (t:'ARCCOSH'; by: bP {P}),
            (t:'ARCTANH'; by: bQ {Q}),
            (t:'ARGTANH'; by: bQ {Q}),
            (t:'ARGTH'; by: bQ {Q}),
            (t:'ATANH'; by: bQ {Q}),
            (t:'ARCSINH'; by: bR {R}),
            (t:'ASINH'; by: bR {R}),
 {40}       (t:'ARGSH'; by: bR {R}),
            (t:'ARGSINH'; by: bR {R}),
            (t: 'INT'; by: bS {S}),
            (t: 'INTEGRAL'; by: bS {S}));

var
  uAttribution : array[bNeg..$9F] of pieceClass;
  qq : array[-cQ..cQ] of Double; {used in the adjusting of the control points
                                  for B-splines}
  {Code chunks. Don't change them when you change "Double"
   Complete information about the instruction sets x86
   x87, etc. may be downloaded from the sites of Intel and AMD. We use
   the instruction names that appear there and the format of hexadecimal
   characters. A byte is shown here as part of a string, with a syntax
   as #$D7, where #$ tells that the two symbols that follow make a character,
   writen as a byte or two hexadecimal digits}

  b_FtimesC: BArray3 = (bF, bTimes, bC);
  b_DOpPar: BArray2 = (bD, bOpPar);
  b_LoadPi: BArray2= ($D9, $EB);
  b_LoadHalfPi: array[0..9] of byte =($D9, $E8, //FLD1
                                       $D9 ,$E0, //FCHS
                                       $D9, $EB, //FLDPI
                                       $D9, $FD,  //FSCALE
                                       $DD, $D9); //FSTP(1)
  b_Load1: BArray2 = ($D9,$E8);
  b_Load0: BArray2 = ($D9,$EE);
  b_zero_comp: BArray4= ($D9, $EE, $DF, $F1);
    {4 bytes. ST(cLastReg) must be empty. Comparison of 0 with x=ST(0). After it,
     JA jumps if 0>x; JNA jumps if 0 <= x; JB, if 0 < x; JNB, if 0 >= x; JNE if 0<>x}
  b_1_comp: BArray4= ($D9, $E8, $DF, $F1);
    {4 bytes.  ST(cLastReg) must be empty. Comparison of 1 with x=ST(0). After it,
     JA jumps if 1>x; JNA jumps if 1 <= x; JB, if 1 < x; JNB, if 1 >= x; JNE if 1<>x}
  b_mod_pi: BArray8= ($D9, $EB, $D9, $C9, $D9, $F8, $DD, $D9);
    {8 bytes. Last ST must be empty}
  b_mod_2pi: array[0..9] of Byte= ($D9, $EB, $D8, $C0, $D9, $C9, $D9, $F8, $DD, $D9);
    {10 bytes. Last ST must be empty}
  b_ln: BArray6= ($D9, $ED, $D9, $C9, $D9, $F1);
    {6 bytes. Requires ST(0) > 0 and last ST empty}
  b_2toX: array[0..17] of Byte= ($D9, $C0, $D9, $FC, $D9, $C9, $D8, $E1, $D9,
                                 $F0, $D9, $E8, $DE, $C1, $D9, $FD, $DD, $D9);
    {18 bytes. Requires two last ST empty}
  b_div2: BArray6= ($D9, $E8, $D8, $C0, $DE, $F9);
    {6 bytes. Requires ST(cLastReg) empty}
  b_div_S: array[0..15] of Byte= ($D9, $EE, $DF, $F1, {zero compare}
             $75, $08, //if r_j <> 0, jump 8 bytes
             $DD, $D8,
             $DD, $D8, // nothing in FST
             $D9, $EE, $EB, $02, $DE, $F9);
    {Safe division. 16 bytes. Gives 0 if divisor is zero}
    
  b_times_Log2_e: BArray4= ($D9, $EA, $DE, $C9);
  b_FSTP_rec: BArray2 = ($DD,$9D); {FSTP fp64bit [RBP] plus 32 bit displacement}
  b_FLD_rec: BArray2 = ($DD,$85); {FLD fp64bit [RBP] plus 32 bit displacement}
  b_Ld_X1: BArray2 = ($DD, $02);  {FLD fp64bits [EDX]} {used only for fld the first indep var}
  b_Ld_X: BArray2  = ($DD, $42);  {FLD fp64bits [EDX] + disp8} {used for remaining
          indep vars and initial velocities and accelerations. Requires to fill it with
          its third entry, a byte, }
  b_Ld_Dt: BArray2 =  ($DD, $80);  {FLD from [EAX] + disp32}{used for
      fld formula consts, intermediate velocities and accelerations}
  b_Ld_Dt8: BArray2 = ($DD, $40);  {FLD from [EAX] + disp8}{used for freq const and params}  

  b_ln_S: Array[0..20] of Byte = ({b_zero_comp:}$D9, $EE, $DF, $F1, $72, $09, $DD, $D8,
      {LoadCVPar(5): next three bytes. Fill place 10} $DD, $40, 0, 
      $D9, $E0, $EB, $06, {b_ln:} $D9, $ED, $D9, $C9, $D9, $F1);
  
  b_2toX_S: Array[0..48] of Byte = ($D9, $C0, $D9, $E1, {LoadCVPar(5), next 3 bytes} $DD, $40, 0,
      $DF, $F1, $DD, $D8, $72, $14, {b_2toX:} $D9, $C0, $D9, $FC, $D9, $C9, $D8, $E1, $D9,
      $F0, $D9, $E8, $DE, $C1, $D9, $FD, $DD, $D9 {b_2toX ended}, $EB, $10,
      {b_zero_comp:}$D9, $EE, $DF, $F1,  $DD, $D8, $72, $05,
      {LoadCVPar(3):} $DD, $40,0 , $EB, $03, {LoadCVPar(1):} $DD, $40, 0);
  
  {These are the prolog and epilog when target is x86_64}
  b_Prolog: array[0..10] of Byte= ({push RBP: }$55,
     {RSP  <--  RSP - $01A0: }$48, $81, $EC, {Int32Str(c_sfr):} $A0,$01,$00,$00,
     {RBP  <-- RSP:} $48, $89, $E5);


  b_Epilog: array[0..16] of Byte= ({fstp  [RBP-$08]. This clears ST(0):}  $DD, $5D, $10,
           {movsd  xmm0, [RBP-$08]. Result goes to xmm0: }$F2,$0F,$10,$45,$10,
           $48,$8D,$A5, $A0,$01,$00,$00, {lea rsp,[rbp+c_sfr]. Recover initial value of RSP}
           {pop RBP:}  $5D,{ret:} $C3);
  
  b_plus_Code: BArray2= ($DE, $C1);
  b_minus_Code: BArray2= ($DE, $E9);
  b_times_Code: BArray2= ($DE, $C9);
  b_neg_code: BArray2= ($D9, $E0);
  b_cos_code: array[0..11] of Byte=  ({c_mod_2pi:} $D9, $EB, $D8, $C0, $D9, $C9, $D9, $F8, $DD, $D9,
                                      {cos:} $D9, $FF);
  b_sin_code: array[0..11] of Byte=  ({c_mod_2pi:} $D9, $EB, $D8, $C0, $D9, $C9, $D9, $F8, $DD,$D9,
                                      {sin:} $D9, $FE);
  b_Sqr: BArray2 = ($D8, $C8);
  b_Sqrt: array[0..13] of Byte= ($D9, $EE, $DF, $F1, $72, $06,
                    $DD, $D8, $D9, $EE, $EB, $02, $D9, $FA);
  b_Tan: array[0..11] of Byte = ($D9, $EB, $D9, $C9, $D9, $F8, $DD, $D9, $D9, $F2, $DD, $D8);


function FormatFloatString(x: Double; totalStrLength: Integer): string;
  {Gives the string of 'x' with 'totalStrLength' characters unless this
   number be greater than 15}
    var
      sigDig: Integer;
  begin
    result:= FloatToStrF(x, ffFixed, totalStrLength+2,15);
    sigDig:= Length(result);
    if sigDig > totalStrLength  then
      sigDig:= totalStrLength;
    result:= Copy(result,1,sigDig);
  end;

procedure AddInt32ToCode(n: Integer; var cod: InfBArray);
{Adds to 'cod' the code of the integer 'n', in the format that is
required by the CPU and increments the index of the last meanngful entry
of 'cod'. It is supposed that the length of 'cod' is not surpassed}
var
  c: PByte;
  theInt, i: Integer;
begin
  if cod.h+4 > cod.bigH then
    raise exception.create(eWrongAttemptToWrite);
  theInt:= n;
  c := @theInt;
  for i := 0 to 3 do begin
    inc(cod.h);
    cod.arr[cod.h] := c[i];
  end;
end;

function Int32Str(n: Integer): ByteArray;
var
  c: PByte;
  theInt, i: Integer;
begin
  theInt:= n;
  c := @theInt;
  SetLength(result, 4);
  for i := 0 to 3 do
    result[i] := c[i];
end;

function ConvertPowerSymbol(x : ByteArray; n : Integer) : ByteArray;
  function ConvertPowerSymbolProv(xx : ByteArray; nn : Integer) : ByteArray;
  var
    zz : ByteArray;
    m, np : Integer;
  begin
    zz := xx;
    np := abs(nn);
    m := np div 2;
    if np = 0 then begin
      SetLength(result, 1);
      Result[0] := b1;
    end
    else if nn = 1 then
      Result:= xx
    else if m > 0 then
      begin
        result := ConvertPowerSymbolProv(zz, m);        
        PrependBA(result, b_DOpPar);
        AppendBA(result, bClPar);
        if odd(np) then begin
          AppendBA(result, bTimes);
          AppendBA(result, zz);
        end;
      end
    else
      Result := zz;
  end;
begin
  Result := ConvertPowerSymbolProv(x, n);
  
  PrependBA(result, bOpPar);
  AppendBA(result, bClPar);
  if n < 0 then begin
   PrependBA(Result, bDiv);
   PrependBA(result, b1);
   PrependBA(result, bOpPar);
   AppendBA(result, bClPar);
  end;
end;

function IdentifyFunction(theFunctionText : String) : Byte;
{If 'theFunctionText' is among the known functions, IdentifyFunction }
{returns the capital letter that corresponds to it. else, it    }
{returns '?'.                                                        }
var
  i : Integer;
  foundFunction : Boolean;
begin
  i := 0;
  repeat
    i := i + 1;
    foundFunction := (theFunctionText = knownFuncs[i].t);
  until (i = kMaxNumOfNames) or foundFunction;
  if foundFunction then
    IdentifyFunction := knownFuncs[i].by
  else
    IdentifyFunction := bQuestion;
end;

procedure InitializeParser;
var
  u, v: Double;
  character : Byte;
  i: Integer;
begin
  for character := bNeg to bZlow do
    case character of
      bOpPar:
        uAttribution[character] := openParent;
      bNeg :
        uAttribution[character] := negation;
      bClPar :
        uAttribution[character] := closeParent;
      bPlus, bMinus, bTimes, bDiv, bPow, bSemicol:
        uAttribution[character] := operationSymb;
      b1, bpi :
        uAttribution[character] := constantSymb;
      bQLow..bZLow :
        uAttribution[character] := varSymb;
      bA..bZ :
        uAttribution[character] := functionSymb;
    else
      uAttribution[character] := badThing;
    end; {of case}

  u:= cA;
  v:= cB;
  qq[0]:= u + v;
  for i:= 1 to cQ do begin {initialize the coefficients that will serve to
      convert a grid of interpolating points to a grid of control points
      for a B-spline}
    u:= u*cU;
    v:= v*cV;
    qq[i]:= u+v;
    qq[-i]:= qq[i];
  end;
end; {InitializeParser}

{function MyFloor(x: Double): Integer;
begin
  result:= Trunc(x);
  if (result<>x) and (x < 0) then
    dec(result);
end;}

function inNumber(targetStr : String;
  var startChar : Char;
  var startPos : Integer) : String;
var
  valid, alreadyBlank, alreadyDot: boolean;
  charPos, stringLength : Integer;
  theChar: Char;
begin
  stringLength := length(targetStr);
  alreadyDot := false;
  valid := true;
  alreadyBlank := false;
  result := Concat('', startChar);
  charPos := startPos;
  theChar := startChar;
  while (charPos < stringLength) and valid and not alreadyBlank do
    begin
      inc(charPos);
      theChar := targetStr[charPos];
      case theChar of
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' :
          result := result + theChar;
        '.' :
          if alreadyDot then
            valid := False
          else begin
            alreadyDot:= True;
            result := result + theChar;
          end;
        ' ' :
          if result[length(result)] = '.' then
            valid := False
          else
            alreadyBlank := true;
        else begin
          alreadyBlank := true;
          charPos := charPos - 1;
          theChar := targetStr[charPos];
        end;
      end; {de CASE}
    end;
  if valid then begin
    startPos := charPos;
    startChar := theChar;
  end
  else
    startChar := '?';
end;



procedure FillMargins(prov1: PArrayNodes1; prov2: PArrayArrayNodes2; dim: Integer);
var
  i, j: Integer;
begin
  if dim = 1 then begin
    for i:= 1 to cQ+2 do
      prov1^[cQ+2-i]:= 2*prov1^[cQ+2] - prov1^[cQ+2+i];
    for i:= 1 to cQ+3 do
      prov1^[cN1+cQ+1+i]:= 2*prov1^[cN1+cQ+1] - prov1^[cN1+cQ+1-i];
  end
  else begin
    for j:= cQ+2 to cN2+cQ+1 do begin
      for i:= 1 to cQ+2 do
        prov2^[j,cQ+2-i]:= 2*prov2^[j,cQ+2] - prov2^[j,cQ+2+i];
      for i:= 1 to cQ+3 do
        prov2^[j, cN2+cQ+1+i]:= 2*prov2^[j,cN2+cQ+1] - prov2^[j, cN2+cQ+1-i];
    end;
    for j:= 0 to cN2+2*cQ+4 do begin
      for i:= 1 to cQ+2 do
        prov2^[cQ+2-i,j]:= 2*prov2^[cQ+2,j] - prov2^[cQ+2+i,j];
      for i:= 1 to cQ+3 do
        prov2^[cN2+cQ+1+i,j]:= 2*prov2^[cN2+cQ+1,j] - prov2^[cN2+cQ+1-i,j];
    end;
  end;
end;

function GetEffectiveVars(allowedVars, list1, list2, list3: String;
               var var1, var2: Integer): integer;
{Used only for the functions that define an integral after they have been
 converted to a codified RPN notation}
  var
    charSet: Set of AnsiChar;
    i, ll: Integer;
    c: Char;
begin
  charSet:= [];
  if length(list1) > 0 then for i:= 1 to length(list1) do begin
    c:= list1[i];
    if c<>'R' then {letter 'R' denotes the integration variable and cannot be
                    considered a variable of the integrated function}
      charSet:= charSet + [c];
  end;
  if length(list2) >0 then for i:= 1 to length(list2) do
    charSet:= charSet + [ list2[i] ];
  if length(list3) >0 then for i:= 1 to length(list3) do
    charSet:= charSet + [ list3[i] ];
  {Now in 'charSet' there is the set of all characters appearing in the three
   functions that make the integral other than 'R'}
  ll:= Length(allowedVars);
  result:=0;
  var1:= -1;
  var2:= -1;
  if ll > 0 then for i:= 1 to ll do
    if allowedVars[i] in charSet then begin
      inc(result);
      if result=1 then var1:= i-1  {take note of the position of the first var
           in the list of allowed variables}
      else if result = 2 then var2:= i-1; {id for the second var}
      {if result = 2 we do not stop the loop, because the caller may be
       interested in the number 'result' of vars that effectively appear in
       the integral.}
    end;
end;


function IntegrateRationalExtr(anF: RealFunction; var args: array of Double;
   room: Integer; tMin, tMax: Double): Double;
{Implementation of a method for numerical integration described in the
 book "Introduction to Numerical Analysis", by J. Stoer and room. Bulirsch,
 Springer, 1991}
const
  cMaxM = 24;
  cNum = 10;
  cTol = 1.0e-8;
var
  i,j,n,m: Integer;
  t0, h, a,b, err: Double;
  finished: Boolean;
  S: Array [0..1] of Double;
  den: array[0..cMaxM] of Integer;
  T: array[0..cMaxM,-1..cMaxM] of double;
  jump: Double;
begin
  jump:= (tMax-tMin)/cNum;
  result:= 0;
  for i:= 0 to cMaxM do
    T[i,-1]:= 0;
  for n:= 1 to cNum do begin
    t0:= tMin + (n-1)*jump;
    args[room]:= t0;
    S[0]:= anF(args);
    args[room]:= t0+ jump;
    S[0]:= (S[0] + anF(args))/2;
    S[1]:= S[0];
    den[0]:= 1;
    T[0,0]:= jump*S[0];
    den[1]:= 2;
    h:= jump/den[1];
    args[room]:= t0 + h;
    S[1]:= S[1] + anF(args);
    T[1,0]:= h*S[1];
    den[2]:= 3;
    h:= jump/den[2];
    args[room]:= t0 + h;
    S[0]:= S[0] + anF(args);
    args[room]:= t0 + 2*h;
    S[0]:= S[0] + anF(args);
    T[2,0]:= h*S[0];
    a:= T[1,0]-T[0,-1];
    b:= sqr(den[1]/den[0])*(T[0,0]-T[0,-1])-a;
    if b<>0 then
      T[1,1]:= T[1,0] + a*(T[1,0]-T[0,0])/b
    else
      T[1,1]:= T[1,0];

    a:= T[2,0]-T[1,-1];
    b:= sqr(den[2]/den[1])*(T[1,0]-T[1,-1])-a;
    if (b<>0) then
      T[2,1]:= T[2,0] + a*(T[2,0]-T[1,0])/b
    else
      T[2,1]:= T[2,0];

    a:= T[2,1]-T[1,0];
    b:= sqr(den[2]/den[0])*(T[1,1]-T[1,0])-a;
    if b<>0 then
      T[2,2]:= T[2,1] + a*(T[2,1]-T[1,1])/b
    else
      T[2,2]:= T[2,1];
    err:= abs(T[2,2]*cTol);
    m:= 2;
    finished:= False;
    while not finished do begin
      inc(m);
      den[m]:= 2*den[m-2];
      h:= jump/den[m];
      j:= m mod 2;
      for i:= 0 to (den[m] div 2 -1) do begin
        args[room]:= t0 + + (2*i+1)*h;
        S[j]:= S[j] + anF(args);
      end;
      T[m,0]:= h*S[j];
      for i:= 1 to m-1 do begin
        a:= T[m,i-1]-T[m-1,i-2];
        b:= sqr(den[m]/den[m-i])*(T[m-1,i-1]-T[m-1,i-2])-a;
        if b<>0 then begin
          T[m,i]:= T[m,i-1]+ a*(T[m,i-1]-T[m-1,i-1])/b;
          finished:= (m>6) and (abs(T[m,i]-T[m-1,i]) < err);
          if finished then
            result:= result+ T[m,i];
        end
        else if a=0 then begin
          finished:= True;
          result:= result+ T[m,i-1];
        end
        else
          T[m,i]:= T[m,i-1];
        if finished then
          break;
      end;
      if not finished then begin
        a:= T[m,m-1]-T[m-1,m-2];
        b:= sqr(den[m]/den[0])*(T[m-1,m-1]-T[m-1,m-2])-a;
        if b<>0 then begin
          T[m,m]:= T[m,m-1]+ a*(T[m,m-1]-T[m-1,m-1])/b;
          finished:= (m >= cMaxM);
          if finished then
            result:= result+ T[m,m];
        end
        else begin
          finished:= True;
          result:= result+ T[m,m-1];
        end;
      end;
    end;
  end;
end;

function PieceEnd(const first : Integer; const theCode: PByte; h: Integer;
   soughtBytes: ByteSet; var symbol : Byte) : Integer;
{This function acts upon 'theCode^', from which only the bytes 'theCode[0]..
theCode[h]' are meaningful. It assumes that 'theCode[first] >= 0'' and that 
it is not in 'soughtBytes'. It scans forward 'theCode' until in reaches one
of the bytes in 'soughtBytes' while being at non-negative parentheses depth.
Result is the index previous to that byte, or 'h' otherwise . In the first
case, 'symbol' is the found byte in 'soughbytes', and it is 'bQuestion' in
the other case}
var
  i, depth : Integer;
  byt : Byte;
  found : Boolean;
begin
  symbol := bQuestion;
  i := first-1;
  depth := 0;
  found := False;
  while (i < h) and not found do
    begin
      inc(i);
      byt := theCode[i];
      if (byt = bOpPar) then
        inc(depth)
      else if (byt = bClPar) then
        dec(depth);
      if byt = bLastByte then
        i:= i+3
      else if ((byt in soughtBytes) and (depth = 0)) or (depth < 0) then
        found := true;
    end;
  if found then
    begin
      result := i - 1;
      symbol := byt;
    end
  else
    result := h;
end;

function ExpressionBA(theCode: PByte; h: Integer) : ByteArray;
  FORWARD;

function AtomBA(theCode: PByte; h: Integer) : ByteArray;
const
  lowVar: ByteSet = [bQ+$20, bR+$20, bS+20, bT+$20, bU+$20, bV+$20, bW+$20,
      bX+$20, bY+$20, bZ+$20];
var
  base : ByteArray;
  i : Integer;
  byt : Byte;
begin
  result:= nil;
  if h>= 0 then begin
    byt := theCode[0];
    if byt = bOpPar then begin
      SetLength(base, h-1);
      for i:= 0 to h-2 do 
        base[i] := theCode[1+i];
      result := ExpressionBA(@base[0], High(base));
    end
    else begin
      Assert(not (byt in lowVar));
      case byt of
        b1, bPi :  AppendBA(Result, byt);
        bLastByte : begin
            SetLength(result, h+1);
            for i:= 0 to h do
              result[i] := theCode[i];
          end;
        bA..bZ : 
          begin
            SetLength(base, h-2);            
            for i:= 0 to h-3 do
              base[i] := theCode[2+i];
            Assert(byt <>  bS, 'AtomString');
            result:= ExpressionBA(@base[0], high(base));            
            AppendBA(result, byt);
          end
        else
            ;
        end;
      end;
    end;
  base:= nil;
end; {AtomString}

function FactorBA(theCode: PByte; h: Integer) : ByteArray;
var
  neg : Boolean;
  redArr, anAtom : ByteArray;
  first, last, i : Integer;
  symbol, prevSymbol : byte;
begin
  if h < 0  then
    result := nil
  else begin
    neg := (theCode[0] = bNeg);
    if neg then begin 
      SetLength(redArr, h);
      for i:= 0 to h-1 do
        redArr[i]:= theCode[i+1];
    end
    else begin 
      SetLength(redArr, h+1);
      for i:= 0 to h do
        redArr[i]:= theCode[i];
    end;   
  
    if High(redArr) = -1 then
      Result := nil
    else begin
      last := PieceEnd(0, @redArr[0], High(redArr), [bPow], symbol);
      if last = High(redArr) then
        result := AtomBA(@redArr[0],last)
      else begin
        anAtom := ExtractBA(@redArr[0], 0, last+1);//Copy(redArr, 0, last+1);
        prevSymbol := symbol;
        result := AtomBA(@anAtom[0], High(anAtom));
        repeat
          first := last + 2;
          last := PieceEnd(first, @redArr[0], High(redArr), [bPow], symbol);
          anAtom := ExtractBA(@redArr[0], first, last - first + 1);
          if (prevSymbol = bPow) then begin
            PrependBA(result, AtomBA(@anAtom[0], High(anAtom)));
            AppendBA(result, b_FtimesC);
            //result := Atom(@anAtom, High(anAtom)) + Result + b_FtimesC
          end
          else begin
            PrependBA(result, AtomBA(@anAtom[0], High(anAtom)));
            AppendBA(result, prevSymbol); 
            //result :=  AtomBA(@anAtom, High(anAtom))+ Result + prevSymbol;
          end;
          prevSymbol := symbol;
        until last >= h;
      end;
      if neg then AppendBA(Result, bNeg);
    end;
  end;
  redArr:= nil;
  anAtom:= nil;
end; {of FactorString}  

function TermBA(theCode: PByte; h: Integer) : ByteArray;
const
  timesdiv: ByteSet=[bTimes, bDiv];
var
  aFactor : ByteArray;
  first, last, i : Integer;
  symbol, prevSymbol : Byte;
begin
  if h < 0 then
    Result := nil
  else begin
    last := PieceEnd(0, theCode, h, timesdiv, symbol);
    if last = h then
      Result := FactorBA(theCode, h)
    else begin
      SetLength(aFactor, last + 1);
      for i:= 0 to last do
        aFactor[i]:= theCode[i];
      prevSymbol := symbol;
      result := FactorBA(@aFactor[0], last);
      repeat
        first := last + 2;
        last := PieceEnd(first, @theCode[0], h, timesdiv, symbol);
        SetLength(aFactor, last - first + 1);
        for i:= 0 to last-first do
          aFactor[i]:= theCode[first + i];
        AppendBA(result, FactorBA(@aFactor[0], High(aFactor)) );
        AppendBA(result, prevSymbol);
        prevSymbol := symbol;
      until last >= h;
    end;
  end;
  aFactor:= nil;
end; {of TermBA}


function ExpressionBA(theCode : PByte; h: Integer) : ByteArray;
const
  plusminus: ByteSet =[bPlus, bMinus];
var
  aTerm : ByteArray;
  first, last : Integer;
  symbol, prevSymbol : Byte;
begin
  if h < 0 then
    Result := nil
  else begin
    last := PieceEnd(0, theCode, h, plusminus, symbol);
    if last = h then
      Result := TermBA(theCode, h)
    else begin
      aTerm := ExtractBA(theCode, 0, last+1);
      prevSymbol := symbol;
      result := TermBA(@aTerm[0], High(aTerm));
      repeat
        first := last + 2;
        last := PieceEnd(first, @theCode[0], h, plusminus, symbol);
        aTerm := ExtractBA(theCode, first, last - first + 1);
        AppendBA(result,TermBA(@aTerm[0], High(aTerm)));
        AppendBA(result, prevSymbol);
        prevSymbol := symbol;
        until last = h;
      end;
    end;
  aTerm:= nil;
end; {of ExpressionBA}

function disp32(n: Integer): BArray4;
{32 bit displacement from @data[0] to @data[n] written as ByteArray}
var
  bytes, i: Integer;
  c: PByte;
begin
  bytes:= fpTypeSize * n;
  c:= @bytes;
  for i := 0 to 3 do begin
    Result[i] := c[i];
  end;
end;


procedure sn(n: Pointer; var stub: array of Byte);
var
  c: PByte;
  i, sz: Integer;
begin
  c:= PByte(@n);
  sz:= SizeOf(Pointer);
  for i := 0 to sz - 1 do
    stub[i]:= c[i];
end;

procedure TFunctionParserBytes_64.AdjustIntegralDataForBSplines(prov1: PArrayNodes1;
    prov2: PArrayArrayNodes2; dim: Integer);
{From data stored in 'provData' that consists of values at equal intervals of the
 function that is to be interpolated, the procdedure builds the control values
'intData' that will be directly used in the evaluation by de Boor algorithm.
 See the attached paper "AproxFuncBSplinesEnglish.pdf"}
var
    i, j, k: Integer;
    dd: PArrayArrayNodes2;
begin
  if dim=1 then
    for i:= cQ-2 to cN1+cQ+2 do begin
      intList1[lastIntegral1].Data[0,i]:= 0;
      for k:=-cQ to cQ do
        intList1[lastIntegral1].Data[0,i]:= intList1[lastIntegral1].Data[0,i]+qq[k]*prov1^[i+k+2];
    end
  else begin
    New(dd);
    try
      for j:= 0 to cN2 + 2*cQ+ 4 do
        for i:= cQ-2 to cN2 + cQ+ 2 do begin
          dd^[i,j]:= 0;
          for k:= -cQ to cQ do
            dd^[i,j]:= dd^[i,j]+ qq[k]*prov2^[i+k+2,j];
        end;
      for i:= cQ-2 to cN2+ cQ+2 do
        for j:= cQ-2 to cN2+cQ+2 do begin
          intList2[lastIntegral2].Data[0,0,i,j] := 0;
          for k:= -cQ to cQ do
            intList2[lastIntegral2].Data[0,0,i,j]:=
                   intList2[lastIntegral2].Data[0,0,i,j]+ qq[k]*dd^[i,j+k+2];
        end;
    finally
      Dispose(dd);
    end;
  end;
end;


function TFunctionParserBytes_64.BuildIntegralDataList(tripleArgument: String): Byte;
{An integral is build by three functions. The integrand and the limits. Here,
 we separate the strings containing those functions and parse them but now
 specifying that no integrals are allowed}
  var
    len, i, i0: Integer;
    c: Char;
    funcText: String;
    integrand, inferior, superior: TFunctionParserBytes_64;
    mc: array of double;
begin
  result:= $FF;
  len:= Length(tripleArgument);
  i:= 0;
  mc:= nil;
  inferior := TFunctionParserBytes_64.create;
  superior := TFunctionParserBytes_64.create;
  integrand := TFunctionParserBytes_64.create;
  try
    repeat  {Find the semicolon}
      inc(i);
      c:= tripleArgument[i];
    until (c= ';') or (i = len);
    if c = ';' then begin
      funcText:= Copy(tripleArgument,1,i-1);
      {Parse integrand}
      integrand.ParseNewText(funcText, allowedVars+'r','', False, True, mc, mc);
    end;
    if (i > len-3) then
      raise exception.Create(eBadIntegralSyntax);
    i0:= i+1;
    repeat  {Find the semicolon}
      inc(i);
      c:= tripleArgument[i];
    until (c=';') or (i = len);
    if c=';' then begin
      funcText:= Copy(tripleArgument,i0,i-i0);
      {Parse inferior limit}
      inferior.ParseNewText(funcText, allowedVars, '', False, True, mc, mc);
    end;
    if (i > len-1) then
      raise exception.Create(eBadIntegralSyntax);
    i0:= i+1;
    funcText:= Copy(tripleArgument,i0,len-i0+1);
    {Parse superior limit}
    superior.ParseNewText(funcText, allowedVars,'', False, True, mc, mc);

    ComputeIntegralData(integrand, inferior, superior);

    {Once the integral data have been computed, these functions are
     useless}
    finally
      integrand.Free;
      inferior.Free;
      superior.Free;
  end;
  result:= lastIntegral1+1 + (lastIntegral2+1) shl 4;
end;



function TFunctionParserBytes_64.CheckFunction : String;
{Checks the syntax of the function}
var
  lastRead : integer;
  species, speciesAnt : pieceClass;
  tok : Byte;
  stopHere : boolean;

  function hasError : boolean;
  var
    foundError, usual : boolean;
  begin
    foundError := false;
    if species = badThing then
      foundError := true
    else
      begin
        usual := (species = operationSymb) or (species = closeParent);
        case speciesAnt of
          openParent :
            foundError := usual;
          closeParent, constantSymb, varSymb :
            foundError := not usual;
          negation, operationSymb :
            foundError := usual or (species = negation);
          functionSymb :
            foundError := (species <> openParent);
        end; {of case}
      end; {of else}
    hasError := foundError;
  end; {of hasError}

begin {CheckFunction}
  if uReduced.h = -1 then begin
    result:= eEmptyFormula;
    Exit;
  end;
  result := eThereIsNoError;
  lastRead := -1;
  speciesAnt := openParent;
  repeat
    inc(lastRead);
    tok := uReduced.arr[lastRead];
    if tok = bLastByte then begin
      if uReduced.arr[lastRead+1] = $01 then
        species:= varSymb
      else
        species := constantSymb;
      lastRead:= lastRead + 3;
    end
    else
      species := uAttribution[tok];
    stopHere := hasError;
    speciesAnt := species;
  until (lastRead = uReduced.h) or stopHere;
  if (not stopHere) then
    begin
      species := closeParent;
      stopHere := hasError;
    end;
  if stopHere then
    result := eSyntaxError;
end; {CheckFunction}

procedure TFunctionParserBytes_64.CodifyFunction;
var
  i: Integer;
  ch: Byte;
  stub: ByteArray;
  {$IFDEF MSWINDOWS}
    prot: Word;
  {$ELSE IF LINUX}
  prot: cInt;
  {$ENDIF}
begin
  lastPreserved:= 1;
  {$IFDEF MSWINDOWS}
  FBlockLength:= ComputeCodeLength;
  if FBlock = nil then begin
    //FBlockLength:= ComputeCodeLength;
    prot:= PAGE_EXECUTE_READWRITE;
    FBlock:= VirtualAlloc(nil, FBlockLength, MEM_COMMIT, prot);
  end
  else begin
    prot:= PAGE_EXECUTE_READ;
    VirtualProtect(FBlock, FBlockLength, PAGE_EXECUTE_READWRITE, @prot);
  end;
  {$ELSE IF LINUX}
   if FBlock <> nil then
    if Fpmunmap(FBlock, FBlockLength) <> 0 then
      raise exception.Create(eBadIntegralSyntax);
   prot:= PROT_EXEC OR PROT_READ OR PROT_WRITE;
   FBlockLength:= ComputeCodeLength;
   FBlock:= FPmMAP(nil, FBlockLength, prot, MAP_PRIVATE or MAP_ANONYMOUS, 0, 0);
  {$ENDIF}
  fCode.arr:= FBlock;
  fCode.h:= -1;
  fCode.bigH:= FBlockLength-1;
  IBAppend(fCode, b_Prolog);{14 bytes. initiate 'FBlock' with the prologue. }
  
  {Save the address of 'data[0]' to rax}
  IBAppend(fCode, [$48, $B8]); {2 bytes}
  SetLength(stub, SizeOf(Pointer));
  sn(@data[0], stub);
  IBAppend(fCode, stub); {SizeOf(Pointer)}

  emptyRegisters := c_LastReg + 1;
  i:= -1;
  while i < High(ActorList) do
    begin
      inc(i);
      ch := actorList[i];
      if ch in [bLastByte, bPi, b1] then begin
        if ch <> bLastByte then
          LeaveRoomInFPUStack(1, fCode);
        case ch of
          bLastByte:  begin
              LeaveRoomInFPUStack(1, fCode);
              stub:= ExtractBA(@actorList[0], i, 4);
              OCStr(stub, fCode);
              i:= i+3;
              dec(emptyRegisters);
            end;
          bPi: begin
                 IBAppend(fCode, b_LoadPi);
                 dec(emptyRegisters);
               end;
          b1: begin
                 IBAppend(fCode, b_Load1);
                 dec(emptyRegisters);
               end
          else
            ;
        end;
      end
      else
        begin
          if (ch in [bPlus, bMinus, bTimes, bDiv, bPow]) then
            begin
              if (emptyRegisters = c_LastReg) then
                PopFromAppStack;
              case ch of
                bPlus : IBAppend(fCode, b_plus_code) ;  //fCode:= fCode + b_plus_code ;
                bMinus : IBAppend(fCode, b_minus_code) ;//fCode:= fCode + b_minus_code;
                bTimes : IBAppend(fCode, b_times_code) ;//fCode:= fCode + b_times_code;
                bDiv :
                  begin
                    LeaveRoomInFPUStack(1, fCode);
                    IBAppend(fCode, b_div_S) ;//fCode:= fCode + b_div_S;
                  end;
                bPow :  ;{This shall not be used. Used exp(#2 * ln#1) <-- #1^#2}
              end;
              inc(emptyRegisters);
            end
          else {opCode of a function}
            begin
              if (ch in [bA, bB, bC, bE, bF, bG, bM]) then
                LeaveRoomInFPUStack(1, fCode)
              else if (ch in [bH, bI, bJ, bK, bL, bO, bP, bQ, bR]) then
                LeaveRoomInFPUStack(2, fCode);
              case ch of
                bNeg : IBAppend(fCode, b_neg_code);
                bA : IBAppend(fCode, b_cos_code);
                bB : IBAppend(fCode, b_sin_code);
                bC : begin
                      IBAppend(fCode, b_times_Log2_e );
                      IBAppend(fCode,  b_2ToX_S);
                     end;
                bD : IBAppend(fCode, b_Sqr); 
                bE : IBAppend(fCode, b_Sqrt);
                bF : IBAppend(fCode, b_ln_S);
                bG : IBAppend(fCode, b_tan);
                bH : begin
                       IBAppend(fCode, [$D9, $EA, $DE, $C9]);
                       IBAppend(fCode, b_2ToX_S);
                       IBAppend(fCode, [$D9, $C0, $D9, $E8, $DE, $F1, $DE, $E9]);
                       IBAppend(fCode, b_div2);
                     end;                      
                bI : begin
                       IBAppend(fCode, [$D9, $EA, $DE, $C9]);
                       IBAppend(fCode, b_2ToX_S);
                       IBAppend(fCode, [$D9, $C0, $D9, $E8, $DE, $F1, $DE, $C1]); 
                       IBAppend(fCode, b_div2);
                     end;
                bJ : begin
                       IBAppend(fCode, [$D9, $EA, $DE, $C9]);
                       IBAppend(fCode, b_2ToX_S);
                       IBAppend(fCode, [$D8, $C8, $D9, $E8, $D8, $E9, $D9, $C9,
                                        $D9, $E8, $DE, $C1, $DE, $F9]);
                     end;
                bK : begin
                       IBAppend(fCode, [$D9, $C0, $D9, $E1]); 
                       IBAppend(fCode, b_1_comp);
                       IBAppend(fCode, [$DD, $D8, $72, $0E, $D9, $C0, $D8, $C8, $D9, $E8, $DE,
                               $E1, $D9, $FA, $D9, $F3, $EB, $14]);
                       IBAppend(fCode, b_zero_comp);
                       IBAppend(fCode, [$DD, $D8]);
                       IBAppend(fCode, b_LoadHalfPi);
                       IBAppend(fCode, [$72, $02, $D9, $E0]); 
                     end;
                bL : begin
                       IBAppend(fCode, [$D9, $C0, $D9, $E1]);
                       IBAppend(fCode, b_1_comp);
                       IBAppend(fCode, [$DD, $D8, $72, $10, $D9, $C0, $D8, $C8, $D9, $E8, $DE,
                               $E1, $D9, $FA, $D9, $C9, $D9, $F3, $EB, $0E]);
                       IBAppend(fCode, b_zero_comp);
                       IBAppend(fCode, [$DD, $D8, $72, $04, $D9, $EB, $EB, $02, $D9, $EE]); 
                     end;
                bM : IBAppend(fCode, [$D9, $E8, $D9, $F3]);
                bN : IBAppend(fCode, [$D9, $E1]);  
                bO : begin
                       IBAppend(fCode, b_mod_pi);
                       IBAppend(fCode, [$D9, $FB, $D9, $C9]);
                       IBAppend(fCode, b_div_S);
                     end;
                bP : begin
                       IBAppend(fCode, b_1_comp);
                       IBAppend(fCode,[$77, $14, $D9, $C0, $D8, $C8, $D9, $E8, $DE, $E9,
                               $D9, $FA, $DE, $C1]);
                       IBAppend(fCode, b_ln);
                       IBAppend(fCode,[$EB, $04, $DD, $D8, $D9, $EE]);
                     end;
                bQ : begin
                       IBAppend(fCode,[$D9, $C0, {fld(0), r in ST(0) and in ST(1)}
                           $D9, $E1 {abs, |r| in ST(0), r in ST(1)}]);
                       IBAppend(fCode, b_1_comp);
                       IBAppend(fCode,[$DD, $D8,{pop(0), r in ST(0)}
                           $76, $1C, {if |r| >=1, jump $1C= 28 bytes}
                           $D9, $C0, $D9, $E8, $DE, $C1, $D9, $C9,
                           $D9, $E8, $DE, $E1, $DE, $F9 {(1+r)/(1-r) in ST(0)}]);
                       IBAppend(fCode, b_ln {2 atanh(r) in ST(0)});
                       IBAppend(fCode, b_div2 {atanh(r) in ST(0)});
                       IBAppend(fCode,[$EB, $0F {jump 15 bytes}]);
                       IBAppend(fCode, b_zero_comp);
                       IBAppend(fCode,[$DD, $D8,  {pop(0), nothing in ST(0)}
                              $77, $05] {if r < 0, jump 5 bytes} );
                       LoadCVPar(5, fCode);
                       IBAppend(fCode,[$EB, $02 {jump 2 bytes}]);
                       LoadCVPar(0, fCode);
                     end;
                bR : begin
                       IBAppend(fCode, b_zero_comp);
                       IBAppend(fCode,[$D9, $E1, $D9, $C0, $D8, $C8, $D9, $E8, $DE, $C1,
                                $D9, $FA, $DE, $C1]); 
                       IBAppend(fCode, b_ln);
                       IBAppend(fCode,[$76, $02, $D9, $E0]);
                     end;
                else
                  raise Exception.Create(eErrorInParsing);
                end;
            end;
        end;
    end; {del while}
  IBAppend(fCode, b_Epilog);
  @FV := @fCode.arr[0];
  stub:= nil;
end;

function TFunctionParserBytes_64.ComputeCodeLength: Integer;
var
  i: Integer;
  ch: Byte;
begin
  result:= 0;
  inc(result, 11); {b_Prolog, 11 bytes}
  inc(result, 2+SizeOf(Pointer));  {Save the address of 'data'}
  emptyRegisters := c_LastReg + 1;
  i:= -1;
  while i < High(ActorList) do
    begin
      inc(i);
      ch := actorList[i];
      if ch in [bLastByte, bPi, b1] then begin
        if ch <> bLastByte then
          inc(result, LengthLeaveRoom(1));
        case ch of
          bLastByte:  begin
              inc(result, LengthLeaveRoom(1));
              inc(result, 6); {OCRBytes adds at most 6 bytes}
              i:= i+3;
              dec(emptyRegisters);
            end;
          bPi: begin
                 inc(result, 2); {b_LoadPi};
                 dec(emptyRegisters);
               end;
          b1: begin
                 inc(result, 2); {b_Load1};
                 dec(emptyRegisters);
               end
          else
            ;
        end;
      end
      else
        begin
          if (ch in [bPlus, bMinus, bTimes, bDiv, bPow]) then
            begin
              if (emptyRegisters = c_LastReg) then
                inc(result, LengthPopFromStack);
              case ch of
                bPlus, bMinus, bTimes : inc(result, 2);
                bDiv :
                  begin
                    inc(result, LengthLeaveRoom(1));
                    inc(result, 16); { b_div_S};
                  end;
                bPow :  ;{This shall not be used. Used exp(#2 * ln#1) <-- #1^#2}
              end;
              inc(emptyRegisters);
            end
          else {opCode of a function}
            begin
              if (ch in [bA, bB, bC, bE, bF, bG, bM]) then
                inc(result, LengthLeaveRoom(1))
              else if (ch in [bH, bI, bJ, bK, bL, bO, bP, bQ, bR]) then
                inc(result, LengthLeaveRoom(2));
              case ch of
                bNeg : inc(result, 2);
                bA, bB: inc(result, 12); {b_cos_code, b_sin_code}
                bC : begin
                      inc(result, 4); {b_times_Log2_e}
                      inc(result, 49);  {b_2ToX_S}
                     end;
                bD : inc(result, 2);  {b_Sqr}
                bE : inc(result, 14); {b_Sqrt}
                bF : inc(result, 21);  {b_ln_S}
                bG : inc(result, 12);  {b_tan}
                bH, bI : begin
                       inc(result, 4);
                       inc(result, 49);  {b_2ToX_S}
                       inc(result, 8);
                       inc(result, 6);  {b_div2}
                     end;
                bJ : begin
                       inc(result, 4);
                       inc(result, 49);  {b_2ToX_S}
                       inc(result, 14);
                     end;
                bK : begin
                       inc(result, 4);
                       inc(result, 4); {b_1_comp}
                       inc(result, 18);
                       inc(result, 4); {b_zero_comp}
                       inc(result, 2);
                       inc(result, 10); {b_LoadHalfPi}
                       inc(result, 4);
                     end;
                bL : begin
                       inc(result, 4);
                       inc(result, 4); {b_1_comp}
                       inc(result, 20);
                       inc(result, 4); {b_zero_comp}
                       inc(result, 10);
                     end;
                bM : inc(result, 4);
                bN : inc(result, 2);
                bO : begin
                       inc(result, 8); {b_mod_pi}
                       inc(result, 4);
                       inc(result, 16); {b_div_S}
                     end;
                bP : begin
                       inc(result, 4); {b_1_comp}
                       inc(result, 14);
                       inc(result, 6); {b_ln}
                       inc(result, 6);
                     end;
                bQ : begin
                       inc(result, 4);
                       inc(result, 4); {b_1_comp}
                       inc(result, 18); {(1+r)/(1-r)]}
                       inc(result, 6); {b_ln}
                       inc(result, 6);  {b_div2}
                       inc(result, 2);
                       inc(result, 4); {b_zero_comp}
                       inc(result, 4);
                       inc(result, 3); {LoadCVPar(5, fCode)}
                       inc(result, 2);
                       inc(result, 2); {LoadCVPar(0, fCode)}
                     end;
                bR : begin
                       inc(result, 4); {b_zero_comp}
                       inc(result, 14);
                       inc(result, 6); {b_ln}
                       inc(result, 4);
                     end;
                else
                  raise Exception.Create(eErrorInParsing);
                end;
            end;
        end;
    end; {del while}
  inc(result, 17); {b_Epilog};
end;

function TFunctionParserBytes_64.ComputeExponent(var n : Integer; ini, fini: Integer): Boolean;
var
  c1, c2 : Byte;
  k: word;
  negate : boolean;
  firstExp, lastExp: Integer;
  exponent : Double;
begin
  firstExp := ini;
  lastExp := fini;
  result:= lastExp >= firstExp;
  c1 := uReduced.arr[firstExp];
  c2 := uReduced.arr[lastExp];
  if result then begin
    {peal off parentheses and compute resulting exponent}
    if (c1 = bOpPar) then begin
      result := (c2 = bClPar);
      if result then begin
        inc(firstExp);
        dec(lastExp);
        result:= ComputeExponent(n, firstExp, lastExp);
        Exit;
      end;
    end;

    negate := (c1 = bNeg);
    if negate then  begin
      inc(firstExp);
      c1 := uReduced.arr[firstExp];
    end;
    result := (c1 = b1) or (c1 = bLastByte);
    if result then begin
      if (lastExp = firstExp) and (c1 = b1) then
        n := 1
      else if (c1 = bLastByte) and (lastExp - firstExp = 3) and 
             (uReduced.arr[firstExp+1] = $03) then begin
        k:= ord(uReduced.arr[firstExp+2]) shl 8 + ord(uReduced.arr[firstExp+3]);
        if k > endParam  then begin
          exponent := data[k];
          result:= (Frac(exponent) = 0);
          if result then
            n:= round(exponent);
        end
        else
          result:= False;
      end
      else
        result:= False;
    end;
    if result and negate then
      n := -n;
  end;
end;

procedure TFunctionParserBytes_64.ComputeIntegralData(
             integrand, inferior,superior: TFunctionParserBytes_64);
{Computes the integral values at a grid of points. That grid with the
 computed values will be then the interpolating points for a curve or
 surface of B-splines. Approximate evaluation of the integral will be
 done afterwards by using the de Boor algorithm. See the attached paper
 "AproxFuncBSplinesEnglish.pdf"}
var
  provData1: PArrayNodes1;
  provData2: PArrayArrayNodes2;
  i, j, var1, var2: Integer;
  x0, infE, supE: Double;
  argum: array of double;
begin
  {Check the number of effective independent variables that appear in the
  functions defining the integral. That number must be 1 or 2}
  i:= GetEffectiveVars(integrand.allowedVars, integrand.usedVarsAndParams, inferior.usedVarsAndParams, superior.usedVarsAndParams,
    var1, var2);
  if i>2 then
    raise Exception.Create(eAtMostTwoVarsInInt);
  if i=0 then
    raise Exception.Create(eNaughtIntegral); {the integral is constant. The user
                                              should compute it by other means}

  if var2 < 0 then begin
    inc(lastIntegral1);
    if High(intList1) < lastIntegral1 then
      SetLength(intList1, lastIntegral1+1);
    IntList1[lastIntegral1].var1:= var1;
  end
  else begin
    inc(lastIntegral2);
    if High(intList2) < lastIntegral2 then
      SetLength(intList2, lastIntegral2+1);
    IntList2[lastIntegral2].var1:= var1;
    IntList2[lastIntegral2].var2:= var2;
  end;

  SetLength(argum, argNum+1);
  for i:= 0 to argNum do
    argum[i]:=0;

  if var2 = -1 then begin {there is only and independent variable. The domain of
                           the function defined by the integral is a segment}
    New(provData1);

    {Calculate values of the integral at cN1+1 evenly spaced points, such that
     the distance between minCorner[var1] and the first point is half the
      separation between the points of the grid, and the same at the end}
    try
      argum[var1]:= minCorner[var1]-jump1[var1]/2;
      for i:= cQ+2 to cN1+cQ+1 do begin
        argum[var1]:= argum[var1] + jump1[var1];
        infE:= inferior.fV(argum);
        supE:= superior.fV(argum);
        provData1^[i]:=IntegrateRationalExtr(integrand.fV, argum, argNum, infE, supE);
      end;
      FillMargins(provData1, nil, 1);
      AdjustIntegralDataForBSplines(provData1, nil, 1);
    finally
      dispose(provData1)
    end;
  end
  else begin
    New(provData2);
    try
      x0:= minCorner[var1]-jump2[var1]/2;
      argum[var2]:= minCorner[var2]-jump2[var2]/2;
      for j:= cQ+2 to cN2+cQ+1 do begin
        argum[var2]:= argum[var2] + jump2[var2];
        argum[var1]:= x0 ;
        for i:= cQ+2 to cN2+cQ+1 do begin
          argum[var1]:= argum[var1] + jump2[var1];
          infE:= inferior.fV(argum);
          supE:= superior.fV(argum);
          provData2^[i,j]:=IntegrateRationalExtr(integrand.fV, argum, argNum, infE, supE);
        end;
      end;
      FillMargins(nil, provData2, 2);
      AdjustIntegralDataForBSplines(nil, provData2, 2);
    finally
      dispose(provData2);
    end;
  end;
  {Prepare the B-splines control points that make the calculted
   values interpolation values with good approximation}
  argum:= nil;
end;

function TFunctionParserBytes_64.ConvertIntegral(var hotSpot: Integer): Byte;
{Extract the string containing the definition of the integral and convert the
 integral to a grid of control points of a B-Spline, ready to be evaluated by the
 de Boor algorithm. The result is the place of the integral in the list of
 integrals that appear in the function}
var
  tripleArgument: String;
  reducedLength, depth: Integer;
  c: Char;
begin
  result:= 0;
  reducedLength := Length(uFormula);
  if reducedLength < 8 then
    Exit;
  depth:= 1;
  tripleArgument:= '';
  repeat
    inc(hotSpot);
    c:= uFormula[hotSpot];
    if c='(' then Inc(depth);
    if c= ')' then Dec(depth);
    if depth > 0 then
      tripleArgument := tripleArgument+c;
  until (depth = 0) or (hotSpot=reducedLength);
  if (depth =0) and (tripleArgument <> '') then begin
    {the triple argumen string has been extracted. Build the B-splines}
    result:= BuildIntegralDataList(tripleArgument);
  end;
end;



procedure TFunctionParserBytes_64.ConvertPowers;
var
  newBA : ByteArray;
  firstBase, lastBase, firstExp, lastExp, i, n, aux: Integer;
  c : Byte;
  good : Boolean;
begin
  if uReduced.h < 2 then
    Exit;
  i := 0;
  n:= 0;
  repeat
    inc(i);
    c := uReduced.arr[i];
    if (c = bPow) then begin
      firstExp := i + 1;
      lastExp := PieceEnd(firstExp, uReduced.arr, uReduced.h, operators, c);
      good := computeExponent(n, firstExp, lastExp);
      if good then begin
        lastBase := i - 1;
        firstBase := PieceBeginning(lastBase, operators + [bNeg]);
        newBA := ExtractBA(uReduced.arr, firstBase, lastBase - firstBase + 1);
        newBA := ConvertPowerSymbol(newBA, n);
        DeleteBA(uReduced.arr, uReduced.h, firstBase, lastExp);
        aux:= High(newBA) + uReduced.h + 1;
        if aux > uReduced.bigH then begin
          SetLength(uReducedBuffer, aux + 50);
          uReduced.bigH:= High(uReducedBuffer);
        end;
        IBInsert(newBA, uReduced, firstBase);
        i := firstBase + Length(newBA) - 1;
      end
    end;
  until (i >= uReduced.h-2);
  newBA:= nil;
end;

constructor TFunctionParserBytes_64.Create;
begin
  inherited Create;
  rawText := '';
  SetLength(data, 32);
  data[0]:= -eMaxExp;
  data[1] := eMaxNum;
  data[2] := -eMaxNum;
  data[3] := eMinNum;
  data[4] := -eMinNum;
  data[5] := eMaxExp;
  b_ln_s[10]:= fpTypeSize * 5;
  b_2toX_S[6]:= fpTypeSize * 5;
  b_2toX_S[43]:= fpTypeSize * 3;
  b_2toX_S[48]:= fpTypeSize;
  FBlockLength:= 0;
end;

destructor TFunctionParserBytes_64.Destroy;
begin
  rawText := '';
  data := nil;
  minCorner:= nil;
  maxCorner:= nil;
  jump1:= nil;
  invJump1:= nil;
  jump2:= nil;
  invJump2:= nil;
  intList1:= nil;
  intList2:= nil;
  if FBlock <> nil then begin
  {$IFDEF MSWINDOWS}
    VirtualFree(FBlock, 0, MEM_RELEASE);
  {$ELSE IF LINUX}
    fpmunmap(FBlock, FBlockLength);
  {$ENDIF}
    FBlock:= nil;
  END;
  uReduced.arr:= nil;
  fCode.arr:= nil;
  actorList:=nil;
  inherited;
end;



function TFunctionParserBytes_64.FactorizeText(targetString : String; integralsAllowed: Boolean;
        var mayEndInFunction : Boolean): ByteArray;
var
  numberOfChars, hotSpot, count, position, i: Integer;
  currentByte : Byte;
  ch: Char;
  presumedText : String;
begin {FactorizeText}
  SetLength(result, 0);
  numberOfChars := Length(targetString);
  hotSpot := numberOfChars;
  if mayEndInFunction and (numberOfChars > 1) then begin
    {we try to recognize a standard function first in the whole string,
     then in it minus the first AnsiChar, then it minus the two first
     chars, etc.}
    count := numberOfChars;
    hotSpot := 1;
    repeat
      presumedText := Copy(targetString, hotSpot, count);
      currentByte := IdentifyFunction(presumedText);
      hotSpot := hotSpot + 1;
      count := count - 1;
    until (currentByte <> bQuestion) or (count = 0);
    if (currentByte <> bQuestion) then begin
      hotSpot := hotSpot - 2;
      if (currentByte= bS) and not integralsAllowed then
      {user has submitted a text with the subexpression 'INT(' that
      can only be interpreted as meaning the beginning of
      the expression of an integral. If integrals are not allowed,
      then interrupt the parsing by raising an exception}
        raise Exception.Create(eNoIntegralsInsideIntegrals);
      mayEndInFunction := TRUE;
      if (hotSpot > 0) then begin
        {hotSpot = 0 would mean that the function was recognized in
         the first pass and the routine would end here. Thus, here we
         are in a situation on which the initial part of 'targetString'
         must be the product of variables with perhaps the pair 'pi'}
        SetLength(result, 2);
        result[0] := bTimes;
        result[1]:= currentByte;
        {and leave the parsing of the first part of 'targetString'
         to the code following the line below that begins by
         'if hotSpot > 0'    }
      end
      else begin
        SetLength(result, 1);
        result[0] := currentByte;
        Exit;
      end;
    end
    else begin {No function was recognnized. Hence, it cannot
     end in a function, and we are yet to factorize all the string}
      hotSpot := numberOfChars;
      mayEndInFunction := FALSE;
    end;
  end {of IF mayEndInFunction ...}
  else
    mayEndInFunction := FALSE;
  if hotSpot > 0 then begin{we will go reading characters backwards,
     in order to discover an 'I' that cannot be a variable nor a
     parameter and in consequence should be preceded by a 'p'. Here
     we take the first step}
    ch := targetString[hotSpot];
    position := pos(ch, allVars);
    if (position <> 0) then begin
      {we encode the variable or param by its position in the list of 'allVars'
       preceded by kLastchar}
      if pos(ch, usedVarsAndParams)=0 then
          {add 'currentByte' to the list of vars and params that in fact appear
           in the parsed expression}
        usedVarsAndParams:= usedVarsAndParams + ch;
      SetLength(result, High(result)+5);
      if High(result) > 3 then
        for i:= high(result)-4 downto 0 do
            result[i+4]:= result[i];
 
      if position <= argNum then begin{it is a indep variable. First byte after 'kLastChar'
              is $01, second byte  is zero and third byte means the position in the list
              of params passed to the routine when evaluation is called}
        result[0]:= bLastByte;
        result[1]:= $01;
        result[2]:= 0;
        result[3]:= position;
      end
      else begin{it is a parameter. First byte after 'kLastChar'
              is $02, second byte is zero and third byte means the position in 'data'
              where the actual value of the parameter will be stored}
        result[0]:= bLastByte;
        result[1]:= $02;
        result[2]:= 0;
        result[3]:= kcvNum + position - argNum;
      end  
    end
    else  begin {check if maybe it is a 'pi'}
      if (ch = 'I') and (hotSpot > 1) and (targetString[hotSpot - 1]
            = 'P') then begin
          hotSpot := hotSpot - 1;
          SetLength(result, High(result)+2);
          if high(result) > 0 then
            for i:= High(result)-1 downto 0 do
              result[i+1]:= result[i];
          result[0] := bPi;
          {Thus, the appearance of a 'p' will mean from now on that a
           value equal to pi is present}
      end
      else {fatal error}
        raise Exception.Create(eBadFunctionOrVariables);
    end;
      hotSpot := hotSpot - 1;
  end;
  while hotSpot > 0 do begin
    ch := targetString[hotSpot];
    position := pos(ch, allVars);
    if (position <> 0) then begin
      SetLength(result, High(result)+6);
      if High(result) > 4 then
        for i:= high(result)-5 downto 0 do
            result[i+5]:= result[i];
      result[0]:= bLastByte;
      if pos(ch, usedVarsAndParams)=0 then
        usedVarsAndParams:= usedVarsAndParams + ch;
      if position <= argNum then begin{it is a independent variable}
        result[1] := $01;
        result[2] := $00;
        result[3] := position;
        result[4] := bTimes;
      end
      else begin{it is a parameter}
        result[1] := $02;
        result[2] := $00;
        result[3] := kcvNum+position-argNum;
        result[4] := bTimes;
      end;
    end
    else begin {check if maybe it is a 'pi'}
      if (ch = 'I') and (hotSpot > 1) and (targetString[hotSpot - 1]
            = 'P') then begin
        hotSpot := hotSpot - 1;
        SetLength(result, High(result)+3);
        if high(result) > 2 then
          for i:= high(result)-2 to 0 do
            result[i+2]:= result[i];
        result[0]:= bPi;
        result[1]:= bTimes;
      end
      else {fatal error}
        raise Exception.Create(eSyntaxError);
    end;
    hotSpot := hotSpot - 1;
  end; {of WHILE...}
end; {FactorizeText}


procedure TFunctionParserBytes_64.GetRoomAndCoeffsFirstDeriv;
{Computes and stores the interpolating points needed for calculating
 the first derivatives of the integrals involved in the function}
  var
    i, a, b, v1, v2: Integer;
begin
  if lastIntegral1 >= 0 then
    for i:= 0 to lastIntegral1 do begin
      v1:= IntList1[i].var1;
      for a:= cQ-1 to cNodes1 do
        IntList1[i].data[1,a]:=
             (IntList1[i].data[0,a]-IntList1[i].data[0,a-1])*invJump1[v1];
    end;
  if lastIntegral2 >= 0 then
    for i := 0 to lastIntegral2 do begin
      v1:= IntList2[i].var1;
      v2:= IntList2[i].var2;
      for a:= cQ-1 to cNodes2 do
        for b:= cQ-2 to cNodes2 do
          IntList2[i].data[1,0,a,b]:=
             (IntList2[i].data[0,0,a,b]-IntList2[i].data[0,0,a-1,b])*invJump2[v1];
      for a:= cQ-2 to cNodes2 do
        for b:= cQ-1 to cNodes2 do
          IntList2[i].data[0,1,a,b]:=
             (IntList2[i].data[0,0,a,b]-IntList2[i].data[0,0,a,b-1])*invJump2[v2];
    end;
end;

procedure TFunctionParserBytes_64.GetRoomAndCoeffsSecondDeriv;
{Reserves the memory necessary for storing the interpolating points needed for
 calculating the second derivatives of the integrals involved in the function}
  var
    i, a, b, v1, v2: Integer;
begin
  If lastIntegral1 >= 0 then
    for i:= 0 to lastIntegral1 do begin
      v1:= IntList1[i].var1;
      for a:= cQ to cNodes1 do
        IntList1[i].data[2,a]:=
             (IntList1[i].data[1,a]-IntList1[i].data[1,a-1])*invJump1[v1];
    end;
  if lastIntegral2 >= 0 then
    for i:= 0 to lastIntegral2 do begin
      v1:= IntList2[i].var1;
      v2:= IntList2[i].var2;
      for a:= cQ to cNodes2 do
        for b:= cQ-2 to cNodes2 do
          IntList2[i].data[2,0,a,b]:=
              (IntList2[i].data[1,0,a,b]-IntList2[i].data[1,0,a-1,b])*invJump2[v1];
      for a:= cQ-2 to cNodes2 do
        for b:= cQ to cNodes2 do
          IntList2[i].data[0,2,a,b]:=
              (IntList2[i].data[0,1,a,b]-IntList2[i].data[0,1,a,b-1])*invJump2[v2];
      for a:= cQ-1 to cNodes2 do
        for b:= cQ-1 to cNodes2 do
          IntList2[i].data[1,1,a,b]:=
              (IntList2[i].data[1,0,a,b]-IntList2[i].data[1,0,a,b-1])*invJump2[v2];
    end;
end;


PROCEDURE TFunctionParserBytes_64.LeaveRoomInFPUStack(neededPlaces: Integer; var cod: InfBArray);
var
  stub: ByteArray;
BEGIN
  IF emptyRegisters >= neededPlaces THEN
    Exit;
  SetLength(stub, 4);
  IF emptyRegisters = 0 THEN BEGIN
    inc(lastPreserved);
    inc(cod.h);
    cod.arr[cod.h] := $D9;
    inc(cod.h);
    cod.arr[cod.h] := $F6;  {fdecstp: ST(cLastReg] becomes ST(0)}
    IBAppend(cod, b_FSTP_rec);
    stub:= Int32Str(c_sfr-8*lastPreserved);  {fstp [RBP - 8*lastPreserved]}
    IBAppend(cod, stub);
    inc(emptyRegisters);
  END;
  IF emptyRegisters < neededPlaces then begin
    if emptyRegisters = c_LastReg then begin
      inc(lastPreserved);
      IBAppend(cod, b_FSTP_rec);
      stub:= Int32Str(c_sfr-8*lastPreserved);  {fstp [RBP - 8*lastPreserved]}
      IBAppend(cod, stub);
      inc(emptyRegisters);
    end
    else REPEAT
      inc(lastPreserved);
      inc(cod.H);
      cod.arr[cod.H]:= $D9;
      inc(cod.H);
      cod.arr[cod.H]:= $C0 + c_LastReg - emptyRegisters; {push last occupied ST to ST(0)}
      IBAppend(cod, b_FSTP_rec);
      stub:= Int32Str(c_sfr-8*lastPreserved);  {fstp [RBP - 8*lastPreserved]}
      IBAppend(cod, stub);
      inc(cod.H);
      cod.arr[cod.H]:= $DD;
      inc(cod.H);
      cod.arr[cod.H]:= $C0 + c_LastReg - emptyRegisters;
         {last occupied place contains the former value: declare it free}
      inc(emptyRegisters);
    UNTIL emptyRegisters = neededPlaces;
  end;
  stub:= nil;
END;

function  TFunctionParserBytes_64.LengthLeaveRoom(neededPlaces: Integer): Integer;
begin
  result:= 0;
  if emptyRegisters >= neededPlaces then
    Exit;
  if emptyRegisters = 0 then begin
    (*inc(result, 2); {fdecstp: ST(cLastReg] becomes ST(0)}
    inc(result, 6); {b_FSTP_rec + disp32} *)
    inc(result, 8);
    inc(emptyRegisters);
  end;
  if emptyRegisters < neededPlaces then begin
    if emptyRegisters = c_LastReg then begin
      inc(result, 6); {b_FSTP_rec + disp32}
      inc(emptyRegisters);
    end
    else repeat
      (*inc(result, 2); {push last occupied ST to ST(0)}
      inc(result, 6); {b_FSTP_rec + disp32}
      inc(result, 2); {declare free last occupied place} *)
      inc(result, 10);
      inc(emptyRegisters);
    until emptyRegisters = neededPlaces;
  end;
end;


function  TFunctionParserBytes_64.LengthPopFromStack: Integer;
VAR
  m : byte;
BEGIN
  result:= 0;
  m := c_LastReg + 1 - emptyRegisters; {ST(m) is the first empty register}
  if m = 1 then
    inc(result, 8)
  else
    inc(result, 10);
  dec(emptyregisters);
END;

procedure TFunctionParserBytes_64.LoadCVPar(const j: Byte; var iba: InfBArray);
begin
  if j=0 then
    IBAppend(iba, [$DD,$00])
  else begin
    IBAppend(iba, b_Ld_Dt8);
    inc(iba.h);
    iba.arr[iba.h]:= fpTypeSize*j;
  end;
end;

procedure TFunctionParserBytes_64.LoadData(const j: word; var yarr: InfBArray);
{Reads from the address of the var data[j] as a Double and stores it in ST(0)}
var 
  bytes, i: Integer;
  c: PByte;
begin
  if j= 0 then
    IBAppend(yarr, [$DD,$00])
  else begin
    IBAppend(yarr, b_Ld_Dt);
    bytes:= fpTypeSize * j;
    c:= @bytes;
    for i := 0 to 3 do begin
      inc(yarr.h);
      yarr.arr[yarr.h] := c[i];
    end;
  end;
end;


procedure TFunctionParserBytes_64.LoadX(i: Word; var yarr: InfBArray);
{Reads from the function stack the independent variable 'i' of FPType
type and stores it in ST(0)}
begin
  if i = 1 then
    IBAppend(yarr, b_Ld_X1)
  else begin
    IBAppend(yarr, b_Ld_X);
    inc(yarr.h);
    yarr.arr[yarr.h]:= fpTypeSize * (i-1);
  end;
end;

procedure TFunctionParserBytes_64.OCStr(stub: ByteArray; var yarr: InfBArray);
{'stub' is a string of four bytes beginning with $FF that represent the loading
 of an argument, a parameter or a constant. Gives the bytes to be added to an
 object code string}
begin
  case stub[1] of
    $01: LoadX(stub[3], yarr);
    $02: LoadCVPar(stub[3], yarr);
    $03: LoadData(stub[2] shl 8 + stub[3], yarr)
    else
    ;
  end;
end;

procedure TFunctionParserBytes_64.ParseNewText(t, vars, pars: String;
        integralsAllowed, codify: Boolean; minC, maxC: array of Double);
{Parses and builds the function given in the string 't'. Each
 character in the string 'vars' identifies one of the possible arguments
 on which the function may depend. Each character in the string 'pars'
 identifies one of the possible parameters that may appear in the function. If
 'integralsAllowed' is True, 't' may contain integrals and then the arguments
 'minC' and 'maxC' describe the domain where the variables can move. If
 'integralsAllowed' is False, then 'minC' and 'maxC' are useless, and the
 appearance of an integral in 't' will raise an exception.

 Warning: don't repeat chars in 'vars' or 'pars'.}
var
  i, len : Integer;
  ch : Char;
begin
  rawText := t;
  allowedVars:= '';
  lastIntegral1:= -1;
  lastIntegral2:= -1;
  usedVarsAndParams:= '';
  if vars = '' then begin
    allowedVars := '';
    argNum := 0;
  end
  else begin
    allowedVars := AnsiUpperCase(vars);    
    argNum := Length(vars);
    {Check that the variable names are among those allowed}
    for i := 1 to argNum do begin
      ch := allowedVars[i];
      if pos(ch, variables)=0 then
        raise Exception.Create(eBadVarSet);
    end;
  end;

  allowedPars := pars;
  parNum := Length(allowedPars);

  {Check that the parameter names are among those allowed and
   that none of them has been included also as a variable}
  if pars <> '' then begin
    allowedPars:= AnsiUpperCase(allowedPars);
    for i := 1 to parNum do begin
      ch := allowedPars[i];
      if (pos(ch,allowedVars)<>0) or (pos(ch, variables) = 0) then
        raise Exception.Create(eBadParamSet)
    end;
  end;
  allVars:= allowedVars + allowedPars;
  endParam:= kcvNum + parNum;
  if high(data) < endParam then
    SetLength(data, endParam + 20);

  for i:= kcvNum + 1 to endParam do
    data[i]:= 0;
  endFD:= endParam;
  if integralsAllowed then begin  {get room for several variables}
    len:= Length(minC);
    SetLength(minCorner,len);
    SetLength(maxCorner,len);
    SetLength(jump1,len);
    SetLength(invJump1,len);
    SetLength(jump2,len);
    SetLength(invJump2,len);
    if len >0 then
      for i:= 0 to len-1 do begin
        minCorner[i]:= minC[i];
        maxCorner[i]:= maxC[i];
        jump1[i]:= (maxCorner[i]-minCorner[i])/cN1;
        invJump1[i]:= 1/jump1[i];
        jump2[i]:= (maxCorner[i]-minCorner[i])/cN2;
        invJump2[i]:= 1/jump2[i];
      end;
  end;

  UnifyChars;
  SetLength(uReducedBuffer, Length(uFormula)*6);
  uReduced.h:= -1;
  uReduced.bigH:= High(uReducedBuffer);
  uReduced.arr:= @uReducedBuffer[0];
  try
    ReduceSymbols(integralsAllowed);
    uFormula:= '';
    CheckFunction;
    ConvertPowers;
    actorList := ExpressionBA(uReduced.arr, uReduced.h);
    SetUWLength;
    SetLength(data, endFD + 2*vaNum + 1);
    endVD :=  endFD + vaNum;
    if codify then
      CodifyFunction;
    if (lastIntegral1 >= 0) or (lastIntegral2 >= 0) then begin
      GetRoomAndCoeffsFirstDeriv;
      GetRoomAndCoeffsSecondDeriv;
    end;
  finally
    uReduced.arr:= nil;
    uReducedBuffer:= nil;
  end;
end;

function TFunctionParserBytes_64.PieceBeginning(const last : Integer; soughtChars : ByteSet) :
  Integer;
{This function acts upon 'uReduced'. It is assumed that 'uReduced.arr[last]'
is not in the string 'soughtChars'. Then, it scans backwards
'uReduced' until in reaches one of the symbols in 'soughtChars' while being at
<=zero parentheses depth. Result is 1 (if not found) or the index following that symbol.}
var
  i, depth : Integer;
  byt : Byte;
  found : Boolean;
begin
  i := last + 1;
  depth := 0;
  found := False;
  while (i > 0) and not found do
    begin
      dec(i);
      byt := uReduced.arr[i];
      if (byt = bClPar) then
        inc(depth)
      else if (byt = bOpPar) then
        dec(depth)
      else if (i > 2) and (uReduced.arr[i-3] = bLastByte) then
        i:= i-3;
      if ((byt in soughtChars) and (depth = 0)) or (depth < 0) then
        found := true;
    end;
  if found then
    result := i + 1
  else
    result := 0;
end;


PROCEDURE TFunctionParserBytes_64.PopFromAppStack;
VAR
  m : byte;
  stub: ByteArray;
BEGIN
  m := c_LastReg + 1 - emptyRegisters; {ST(m) is the first empty register}
  if m = 1 then begin
    IBAppend(fCode,  b_FLD_rec);
    SetLength(stub,4);
    stub:= Int32Str(c_sfr- 8*lastPreserved);
    IBAppend(fCode, stub);
    IBAppend(fCode, $D9);
    inc(fCode.h);
    fCode.arr[fCode.h]:= $C8 + m;
  end
  else begin
    IBAppend(fCode, $DD);
    inc(fCode.h);
    fCode.arr[fCode.h]:= $D8 + m;
    IBAppend(fCode, b_FLD_rec);
    stub:= Int32Str(c_sfr- 8*lastPreserved);
    IBAppend(fCode, stub);
    IBAppend(fCode, $D9);
    inc(fCode.h);
    fCode.arr[fCode.h]:= $C8 + m;
  end;
  dec(lastPreserved);
  dec(emptyregisters);
  stub:= nil;
END;



procedure TFunctionParserBytes_64.ReduceSymbols(integralsAllowed: Boolean);
{Substitutes the name of each function by a single character; any number, independent
variable or parameter by an "escape character", 'kLastChar', followed by three bytes,
that signal the nature of the symbol and the place where the actual value must be stored
in the array 'data'. Also, each integral is processed to build an array of data
that will define the B-splines that will be used in evaluating it, and it is
substituted by 'S('+kLastChar+ n+')', where 'n' means the byte that signals
the place of that integral in the list of integrals that appear in the function}
var
  byt: Byte;
  auxBA: ByteArray;
  numberOfChars, hotSpot, formulaLength: Integer;
  currentChar : Char;
  nm: Double;
  letterStr, numStr : String;
  numberBegins, inLetter, mayHaveFunction: Boolean;
begin {ReduceSymbols}  
  uReduced.h:= -1;
  formulaLength := Length(uFormula);
  numberOfChars := formulaLength;
  hotSpot := 0;
  repeat
    inc(hotSpot);
    currentChar := uFormula[hotSpot];
    inLetter := (pos(currentChar, upCaseLetters) <> 0);
    numberBegins := (pos(currentChar, digits) <> 0) or (currentChar = '.');
    if (not inLetter) and (not numberBegins) then begin{it is a symbol as '+' or ')'}
      inc(ureduced.h);
      uReduced.arr[uReduced.h] := ord(currentChar);
    end;
    if numberBegins then begin
      numStr := '0';
      if currentChar = '.' then
        currentChar := '?'
      else
        numStr := inNumber(uFormula, currentChar, hotSpot);
      if (currentChar <> '?') and (currentChar <> '.') then begin
        nm:= StrToFloat(numStr);
        if (nm = 1) then 
          IBAppend(uReduced, b1)
        else begin
          inc(endFD);
          if high(data) < endFD then
            SetLength(data, endFD + 16);
          data[endFD]:= nm;

          {codification of the constant. Byte $03 means that it is a constant,
           and 'endFD' is codified as a word. So, we admit more than 65000
           constants in the formula}
          uReduced.arr[uReduced.h+1] :=  bLastByte;
          uReduced.arr[uReduced.h+2] :=  $03;
          uReduced.arr[uReduced.h+3] :=  (endFD shr 8);
          uReduced.arr[uReduced.h+4] := (endFD and $FF);
          inc(uReduced.h, 4);
        end;
      end
      else
        raise Exception.Create(eBadNumber);
    end;
    if inLetter then begin
      letterStr := currentChar;
      while (hotSpot < numberOfChars) and inLetter do begin
        inc(hotSpot);
        currentChar := uFormula[hotSpot];
        inLetter := (pos(currentChar, upCaseLetters) <> 0);
        if inLetter then
          letterStr := Concat(letterStr, currentChar);
      end;
      mayHaveFunction := (currentChar = '('); {mayHaveFunction = True says that
          after the letters comes an opening parenthesis, that is these letters
          may be the identifier of a known function}
      if not inLetter then
        hotSpot := hotSpot - 1;
      {Check if the final part of letterStr is a known function, substitute it
       by its corresponding letter, and factorize the initial part}
      auxBA := FactorizeText(letterStr, integralsAllowed, mayHaveFunction);
      IBAppend(uReduced, auxBA);
      if (currentChar = '(') and (not mayHaveFunction) then {the parenthesis
         initiates a Factor that multiplies the last Factor}
        IBAppend(uReduced, bTimes);
      if (high(auxBA) = 0) and (auxBA[0] = bS) and (currentChar = '(') and mayHaveFunction then begin
        inc(hotSpot);
        {prepare B-splines data}
        dec(uReduced.h); {the 'bS' byte must be overriden by the following four ones}
        byt:= ConvertIntegral(hotSpot);
        if byt = $FF then
           raise exception.create(eBadIntegralSyntax);
        //IBAppend(uReduced, bLastByte);
        IBAppend(uReduced, [bLastByte, $04, $00, byt]);
        //IBAppend(uReduced, byt);
      end;
      auxBA:= nil;
    end; {of inLetter}
  until (hotSpot >= numberOfChars);
end; {ReduceSymbols}

procedure TFunctionParserBytes_64.SetUWLength;
var
  i, j: Integer;
  c : Byte;
  escape : Boolean;
begin
  i := -1;
  j:= 0;
  vaNum:=0;
  while i < High(ActorList) do begin
    inc(i);
    c := actorList[i];
    escape := (c = bLastByte);
    if escape or (c = bPi) or (c = b1) then begin
      if escape then
        i:= i+3;
      inc(j);
      if j > vaNum then
        vaNum:= j;
    end
    else if (c in operators) then
      dec(j);
  end; {del while}
end;

procedure TFunctionParserBytes_64.UnifyChars;
var
  place, depth, theLength : Integer;
  currentChar, precChar, lastAdded, toAdd : Char;
begin {UnifyChars}
  theLength := Length(rawText);
  rawText := AnsiUpperCase(rawText);
  uFormula := '';
  if theLength = 0 then
    raise Exception.Create(eNothingWritten);
  lastAdded := '(';
  currentChar := '(';
  toAdd := ' ';
  place := 0;
  depth := 0;
  repeat
    precChar := currentChar;
    inc(place);
    currentChar := rawText[place];
    case currentChar of
      'A'..'Z' :
        begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) or
            ((pos(lastAdded, upCaseLetters) <> 0) and (precChar = ' ')) then
            uFormula := Concat(uFormula, '*'); {only a product can be
             accepted as separator from the previous part of the text. Add
             '*'}
          toAdd := CurrentChar; {save this Char to add it after the end of "case"}
        end;
      '0'..'9' :
        begin
          if (lastAdded = ')') or (pos(lastAdded, upCaseLetters) <> 0) or
            ((pos(lastAdded, digits) <> 0) and (precChar = ' ')) then
            uFormula := Concat(uFormula, '*');
          toAdd := CurrentChar;
        end;
      '(', '[', '{' :
        begin
          if (lastAdded = ')') or (pos(lastAdded, digits) <> 0) then
            uFormula := Concat(uFormula, '*');
          toAdd := '(';
          depth := depth + 1;
        end;
      '*' :
        if (lastAdded = '*') then begin
          {the user want to build a power expression. So, replace
           the previous '*' with a caret}
          uFormula[Length(uFormula)]:= '^';
          lastAdded := '^';
          toAdd := ' ';
        end
        else
          toAdd := '*';
      ')', ']', '}' :
        begin
          toAdd := ')';
          depth := depth - 1;
        end;
      ',', '.', '''' : {if user writes  7,56, or 7.56 or 7'56 it will be
                       converted to 7.56. Thus, we are not enforcing totally
                       the law that the decimal separator must be the dot.}
        toAdd := '.';
      '-' :
        if (lastAdded = '(') then
          toAdd := '#' {we distinguish the operator "change sign" that we
                        denote by '#', from the subtract operator. The
                        first one demands one argument; the other, demands two.}
        else
          toAdd := '-';
      '/', '\' :
        toAdd := '/'; {Thus,  'cos(x)\sin(x)' will be converted to 'COS(X)/SIN(X)'}
      ' ', '^', ';' :
        toAdd := currentChar; {the semicolon ';' is used only in expressions for
                               integrals}
      '+' :
        if (lastAdded = '(') then
          toAdd := ' '
        else
          toAdd := '+';
          {Note that a string as '+cos(x)- sin(y)' should be converted to
           'COS(X)-SIN(Y)'. This is the reason for having initialized 'lastAdded'
           as '('   }
      kReturn, kTab, kLineFeed :  {Eat all these characters}
          toAdd := ' ';
      else
        raise Exception.Create(eBadCharacters);
    end; {of CASE}

    if toAdd <> ' ' then
      begin
        uFormula := uFormula + toAdd;
        lastAdded := toAdd;
      end;
  until (place = theLength) or (depth < 0);
  if depth <> 0 then
    raise Exception.Create(eBadParentheses);
end; {of UnifyChars}


procedure ReadConstantField(theStr: String; var val: double);
var
  forConstFunct: TFunctionParserBytes_64;
  mc: array of Double;
begin
  val := 0;
  mc:= nil;
  forConstFunct:= TFunctionParserBytes_64.create;
  try
    forConstFunct.ParseNewText(theStr,'','', False, True, mc, mc);
    val := forConstFunct.fV(mc);
  finally
    forConstFunct.Free;
  end;
end;

function ReadConstantField(theStr: String): double; overload;
var
  forConstFunct: TFunctionParserBytes_64;
  mc: array of Double;
begin
  result := 0;
  mc:= nil;
  forConstFunct:= TFunctionParserBytes_64.create;
  try
    forConstFunct.ParseNewText(theStr,'','', False, True, mc, mc);
    try
      result := forConstFunct.fV(mc);
    except on Exception do
      SysUtils.beep;
    end;
  finally
    forConstFunct.Free;
  end;
end;




initialization;

  InitializeParser;


end.
 

