
{-----------------------------------------------------------------------------
 Unit Name: HigherDegreeLines
 Project Name : Parametricas4
 Author:    Angel Montesinos Amilibia
 Dates: 2003-2009
 Purpose: Computation and drawing of integral curves of extremal deviation
    lines, and Strong principal lines (Montaldi) in the sense of Mara Garca Monera, "Desarrollo
    de Taylor de la Aplicacin Exponencial y Aplicaciones Geomtricas", Tesis
    de Master, Facultad de Matemticas, Universidad de Valencia, 2008.

    These lines are of three types:
       - Strong principal lines
       - Lines of extremal normal deviation
       - Lines of extremal lateral deviation

    Computation of these types involve the numerical solution of polynomial
    equations of degrees four or five.

-----------------------------------------------------------------------------}
{

Parametricas4, workshop for the geometry of surfaces in R^4.


MIT License

Copyright (c) [2003 - 2018] [Angel Montesinos-Amilibia], email montesin@uv.es

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


}

unit HigherDegreeLines;

interface
uses
  Graphics,
  FParam4, VectorsAndUtil;

type
  RealPolynomial = record { for real polynomials of degrees 0 to 4}
    degree: Integer; {for the actual degree of the polynomial}
    coefs: array[0..5] of Double; {the polynomial must be understood as
    coefs[0]*x^degree + coefs[1]*x^(degree-1)+ ... + coefs[degree]}
  end;

  Roots = record
    number: Integer;
    values: array[1..5] of Vector3;
    {values[i,1] and values[i,2] for giving the direction
  that is values[i,1]:= cos(t), values[i,2]:= sin(t). The third field, values[i,3] receives
  other data. For instance, here it denotes, for the case of extremal deviation
  lines the value of the length of  H + B cos(2t) + C sin(2t)}
  end;

procedure DrawHigherDegreeLines(cl: Integer; can: TCanvas;
  start: Vector2; foliations: Byte);
procedure GetHigherDegreeDirections(cl: Integer; var dirs: Roots; param:
  Vector2);
function ComputeStrongPrincipalPolynomial(var p5: RealPolynomial; p: Vector2): Boolean;

implementation

uses
  Forms, Windows, Math, SysUtils,
  FParametrizationR2R4;

type
  SturmSequence = record
    numPolyn: Integer;
    polynomials: array[0..5] of RealPolynomial;
  end;

const
  eMinScale = 1.0E-12;
  eMinRootsSep = 1.0E-5;
  eSmallShift = 1.0E-16;

var
  uSmallValue: Double;
  theRoots: Roots;
  co, si: Double;
  pol: RealPolynomial;
  stm: SturmSequence;
  rk1, rk2, rk3: Vector2;

procedure SortRoots;
// Quick-Sort for sorting rawData points in the lexicographic order.
  procedure qs(left, right: integer);
  var
    i, j: integer;
    pivot: Double;
    t: Vector3;
  begin
    if left < right then begin
      i:= left;
      j:= right;
      pivot:= theRoots.values[(left + right) shr 1, 3];
      repeat
        while theRoots.values[i, 3] < pivot do
          inc(i);
        while pivot < theRoots.values[j, 3] do
          dec(j);
        if i <= j then begin
          if i < j then begin
            t:= theRoots.values[i];
            theRoots.values[i]:= theRoots.values[j];
            theRoots.values[j]:= t;
          end;
          inc(i);
          dec(j);
        end;
      until i > j;
      qs(left, j);
      qs(i, right);
    end;
  end;
begin
  qs(1, theRoots.number);
end;

function ComputeStrongPrincipalPolynomial(var p5: RealPolynomial; p: Vector2): Boolean;
{Gives to the variable 'pol' the coefficients of the fifth-degree
 polynomial whose roots will give the strong principal directions at a
 point of the surface in R^4. See the memory
    "Asymp3orderInR5_StrongPpalDirsInR4.pdf"
 in this same folder. It it fails, results in False}
var
  raa, rab, rbb: Vector4;
  m: matrix44;
  E, F, G, d, uudu, uudv, uvdu, uvdv, vvdu, vvdv, duuuv, duuvv, duvvv, puu, quu,
    puv, quv, pvv, qvv: Double;
  i: Integer;
begin
  with MapR2R4 do begin
    PosFirstSecondThirdDerivativesFR2R4(p[1], p[2]);
    E:= EuclEnergy(xu);
    F:= EuclInnerProd(xu,xv);
    G:= EuclEnergy(xv);
    d:= sqrt(abs(E*G-sqr(F)));
    result:= d > 0;
    if result then begin
      E:= E*d;
      F:= F*d;
      G:= G*d;
      for i:= 1 to 4 do begin
        raa[i]:= -F*xuu[i] + E*xuv[i];
        rab[i]:= -G*xuu[i]+ E*xvv[i];
        rbb[i]:= F*xvv[i] - G*xuv[i];
      end;
      m[1]:= xu;
      m[2]:= xv;
      m[3]:= xuu;
      m[4]:= xuv;
      duuuv:= determinant44(m);
      m[4]:= xvv;
      duuvv:= determinant44(m);
      m[3]:= xuv;
      duvvv:= determinant44(m);
      uudu:= EuclInnerProd(xuu,xu);
      uudv:= EuclInnerProd(xuu,xv);
      uvdu:= EuclInnerProd(xuv,xu);
      uvdv:= EuclInnerProd(xuv,xv);
      vvdu:= EuclInnerProd(xvv,xu);
      vvdv:= EuclInnerProd(xvv,xv);
      puu:= d*(G*uudu - F*uudv);
      quu:= d*(E*uudv - F*uudu);
      puv:= d*(G*uvdu - F*uvdv);
      quv:= d*(E*uvdv - F*uvdu);
      pvv:= d*(G*vvdu - F*vvdv);
      qvv:= d*(E*vvdv - F*vvdu);
      p5.degree:= 5;

      m[3]:= raa;
      m[4]:= xuuu;
      p5.coefs[5]:= determinant44(m) + 3*(F*quu +E*puu)*duuuv;

      m[4]:= xuuv;
      p5.coefs[4]:= 3*determinant44(m) + 3*(F*(2*quv+puu)+2*E*puv + G*quu)*duuuv +
             3*(E*puu+F*quu)*duuvv;
      m[3]:= rab;
      m[4]:= xuuu;
      p5.coefs[4]:= p5.coefs[4] + determinant44(m);

      m[4]:= xuuv;
      p5.coefs[3]:= 3*determinant44(m) + 3*(F*(2*puv+qvv)+2*G*quv + E*pvv)*duuuv +
             3*(2*(E*puv + F*quv)+F*puu + G*quu)*duuvv+
             3*(F*quu + E*puu)*duvvv;
      m[3]:= rbb;
      m[4]:= xuuu;
      p5.coefs[3]:= p5.coefs[3] + determinant44(m);
      m[3]:= raa;
      m[4]:= xuvv;
      p5.coefs[3]:= p5.coefs[3] + 3*determinant44(m);

      m[3]:= rbb;
      m[4]:= xuuv;
      p5.coefs[2]:= 3*determinant44(m) + 3*(F*(2*quv+puu)+2*E*puv + G*quu)*duvvv +
             3*(2*(G*quv + F*puv)+F*qvv + E*pvv)*duuvv+
             3*(F*pvv + G*qvv)*duuuv;
      m[3]:= rab;
      m[4]:= xuvv;
      p5.coefs[2]:= p5.coefs[2] + 3*determinant44(m);
      m[3]:= raa;
      m[4]:= xvvv;
      p5.coefs[2]:= p5.coefs[2] + determinant44(m);

      m[3]:= rab;
      p5.coefs[1]:= determinant44(m) + 3*(F*(2*puv+qvv)+2*G*quv + E*pvv)*duvvv +
             3*(G*qvv+F*pvv)*duuvv;
      m[3]:= rbb;
      m[4]:= xuvv;
      p5.coefs[1]:= p5.coefs[1] + 3*determinant44(m);

      m[3]:= rbb;
      m[4]:= xvvv;
      p5.coefs[0]:= determinant44(m) + 3*(F*pvv +G*qvv)*duvvv;
    end
  end;
end;

function ComputeCoefficientsExtrDevDirs(cl: Integer; p: Vector2): Boolean;
{Gives to the variable 'pol' the coefficients of the fourth-degree
 polynomial whose roots will give the extremal deviations directions at a
 point of the surface in R^4. See the memory
    "MasterTesisGarciaMonera.pdf"
 in this same folder.}
var
  a: Double;
  i: Integer;
begin
  result:= MapR2R4.ComputeHBCandProducts(p);
  if result then with MapR2R4 do begin
      case cl of
        kExtrNormDevLines: begin
            pol.degree:= 4;
            pol.coefs[0]:= bc - hc;
            pol.coefs[1]:= 2 * (bb - hb - cc);
            pol.coefs[2]:= -6 * bc;
            pol.coefs[3]:= 2 * (cc - hb - bb);
            pol.coefs[4]:= bc + hc;
          end;
        kExtrLatDevLines: begin
            pol.degree:= 4;
            pol.coefs[0]:= hb + cc - bb;
            pol.coefs[1]:= 2 * (4 * bc - hc);
            pol.coefs[2]:= 6 * (bb - cc);
            pol.coefs[3]:= -2 * (4 * bc + hc);
            pol.coefs[4]:= cc - bb - hb;
          end;
        kStrongPrincipalLines:
          ComputeStrongPrincipalPolynomial(pol, p);
      else
        ;
      end;
      a:= eMinScale * 0.001;
      for i:= 0 to 4 do
        a:= a + abs(pol.coefs[i]);
      uSmallValue:= a * eMinScale;
      {a measure of the absolute value below which some variables
       shall be taken as zero}
    end;
end;

function ReduceDegree(var p: RealPolynomial): Integer;
{suppress all the coefficients starting from coefs[0] until we reach
 one with absolute value greater than 'uSmallValue'. Build then the
 polynomial so that the new coef[0] will be the coefficient of
 greatest degree. Returns the order num of the first coefficient not
 suppressed}
var
  i, j: Integer;
begin
  i:= 0;
  while (i <= p.degree) and (abs(p.coefs[i]) <= uSmallValue) do
    inc(i);
  result:= i;
  if (i > 0) and (i <= p.degree) then
    for j:= i to p.degree do
      p.coefs[j - i]:= p.coefs[j];
  p.degree:= p.degree - i;
end;

procedure MinusDerivative(p: RealPolynomial; var deriv: RealPolynomial);
{gives in 'deriv' minus the derivative of the polynomial 'p'}
var
  i: Integer;
begin
  if p.degree > 0 then
    for i:= 0 to p.degree - 1 do
      deriv.coefs[i]:= -(p.degree - i) * p.coefs[i];
  deriv.degree:= p.degree - 1;
end;

procedure MinusRemainder(num, denom: RealPolynomial; var remd:
  RealPolynomial);
{it is assumed that num and denom have been previously reduced, so
 that abs(num.coefs[0]) and abs(denom.coefs[0]) are greater than
 'uSmallValue'. Returns in 'remd' minus the remainder of dividing 'num'
 by 'denom'}
var
  a: Double;
  i: Integer;
begin
  remd:= num;
  while (remd.degree >= denom.degree) do begin
    a:= remd.coefs[0] / denom.coefs[0];
    if denom.degree > 0 then
      for i:= 1 to denom.degree do
        remd.coefs[i - 1]:= remd.coefs[i] - a * denom.coefs[i];
    if remd.degree > denom.degree then
      for i:= denom.degree + 1 to remd.degree do
        remd.coefs[i - 1]:= remd.coefs[i];
    dec(remd.degree);
    ReduceDegree(remd);
  end;
  if remd.degree >= 0 then
    for i:= 0 to remd.degree do
      remd.coefs[i]:= -remd.coefs[i];
end;

procedure Quotient(var num, quot: RealPolynomial; denom: RealPolynomial);
{Gives in 'quot' the quotient of dividing 'num' by 'denom'. Assumes that
 'quot' has been reduced so that 'abs(quot.coefs[0]) > uSmallValue'. Gives
 in 'num' the remainder of the division after reducing it}
var
  i, j, shift: Integer;
  a: Double;
begin
  i:= 0;
  if denom.degree < 0 then
    quot.degree:= -1
  else
    quot.degree:= num.degree - denom.degree;
  if quot.degree < 0 then Exit;
  while (num.degree >= denom.degree) do begin
    a:= num.coefs[0] / denom.coefs[0];
    quot.coefs[i]:= a;
    if denom.degree > 0 then
      for j:= 1 to denom.degree do
        num.coefs[j - 1]:= num.coefs[j] - a * denom.coefs[j];
    if num.degree > denom.degree then
      for j:= denom.degree + 1 to num.degree do
        num.coefs[j - 1]:= num.coefs[j];
    dec(num.degree);
    shift:= ReduceDegree(num);
    if shift > 0 then
      for j:= 1 to shift do begin
        inc(i);
        quot.coefs[i]:= 0;
      end;
    inc(i);
  end;
end;

procedure MakeSturmSequence;
{It is assumed that 'stm.polynomials[0] = pol' has been previously created, initialized and reduced,
and that 'pol.degree >= 3'. Otherwise, we should compute directly the roots of
'pol'}
var
  i: Integer;
  provPol: RealPolynomial;
begin
  MinusDerivative(stm.polynomials[0], stm.polynomials[1]);
  i:= 1;
  while (stm.polynomials[i].degree > 0) do begin
    inc(i);
    MinusRemainder(stm.polynomials[i - 2],
      stm.polynomials[i - 1], stm.polynomials[i]);
  end;
  stm.numPolyn:= i + 1;
  if stm.polynomials[i].degree < 0 then begin
    provPol:= stm.polynomials[0];
    Quotient(provPol, stm.polynomials[0], stm.polynomials[i - 1]);
    MakeSturmSequence;
  end;
end;

function EvalPolU(p: RealPolynomial; u: Double): Double;
{Evaluates 'p' at 'u'}
var
  i: Integer;
begin
  if p.degree < 0 then
    Result:= 0
  else begin
    result:= p.coefs[0];
    if p.degree > 0 then
      for i:= 1 to p.degree do
        result:= result * u + p.coefs[i];
  end;
end;

function Variations(a: Double): integer;
{Gives the variations of the Sturm sequence at a}
var
  i: integer;
  precSign, curSign: boolean;
begin
  result:= 0;
  curSign:= (evalPolU(stm.polynomials[0], a) > 0);
  if stm.numPolyn > 1 then
    for i:= 1 to stm.numPolyn - 1 do begin
      precSign:= curSign;
      curSign:= (evalPolU(stm.polynomials[i], a) > 0);
      if precSign <> curSign then
        inc(result);
    end;
end;

function MidPointRule(a, b: Double; fa: Boolean): Double;
var
  a1, b1, vc: Double;
  fa1, fc: Boolean;
  counter: integer;
begin
  a1:= a;
  b1:= b;
  fa1:= fa;
  counter:= 0;
  repeat
    result:= 0.5 * (a1 + b1);
    vc:= EvalPolU(pol, result);
    if abs(vc) < uSmallValue then
      Exit;
    fc:= vc > 0;
    if fa1 = fc then begin
      a1:= result;
      fa1:= fc;
    end else
      b1:= result;
    inc(counter);
  until counter > 50;
end;

procedure AddRoot(cl: Integer; r: Double);
{given the root 'r', which is the tangent of the angle that the corresponding
 direction makes with the first axis, compute and save the cosine and sine, and
 fill the third component of values[i] with the value of the function that these
 directions make stationary.}
var
  sine, cosine, sine2, cosine2: Double;
begin
  inc(theRoots.number);
  cosine:= 1 / sqrt(1 + sqr(r));
  sine:= r * cosine;
  theRoots.values[theRoots.number, 1]:= cosine;
  theRoots.values[theRoots.number, 2]:= sine;
  if cl = kExtrNormDevLines then begin
    sine2:= 2 * sine * cosine;
    cosine2:= sqr(cosine) - sqr(sine);
    with MapR2R4 do {save the value of |H + B cosine2 + C sine2|^2}
      theRoots.values[theRoots.number, 3]:= hh + bb * sqr(cosine2) + cc *
          sqr(sine2) +
          2 * hb * cosine2 + 2 * hc * sine2 + 2 * bc * sine2 * cosine2;
  end
  else if cl = kExtrLatDevLines then begin
    sine2:= 2 * sine * cosine;
    cosine2:= sqr(cosine) - sqr(sine);
    with MapR2R4 do {save (H + B cosine2 + C sine2).(-B sine2 + C cosine2)}
      theRoots.values[theRoots.number, 3]:= -hb * sine2 + hc * cosine2 + (cc -
          bb) * cosine2 * sine2 +
          bc * (sqr(cosine2) - sqr(sine2));
  end
end;

procedure GetRootsInInterval(cl: Integer; a, b: Double; va, vb:
  Integer);
{It is assumed that the number of variations in the interval a is va, and in b
is vb. The procedure computes all those roots and stores them in gRoots.}
var
  c, a0, fa, fb, fc: Double;
  v0, vc: Integer;
  sa, sb: Boolean;
begin
  if va = vb then Exit;
  if abs(va - vb) = 1 then begin
    fa:= EvalPolU(pol, a);
    fb:= EvalPolU(pol, b);
    if (abs(fa) = 0) then begin
      fc:= EvalPolU(pol, a + eSmallShift);
      if (fc * fb < 0) then begin
        sa:= (fc > 0);
        c:= MidPointRule(a + eSmallShift, b, sa);
        AddRoot(cl, c);
        Exit;
      end else
        AddRoot(cl, a);
    end else if (abs(fb) = 0) then begin
      fc:= EvalPolU(pol, b - eSmallShift);
      if (fa * fc < 0) then begin
        sb:= (fc > 0);
        c:= MidPointRule(a, b - eSmallShift, not sb);
        AddRoot(cl, c);
        Exit;
      end else
        AddRoot(cl, b);
    end else begin
      sa:= (fa > 0);
      sb:= (fb > 0);
      if (sa <> sb) then begin
        c:= MidPointRule(a, b, sa);
        AddRoot(cl, c);
      end;
    end;
  end else begin
    if abs(a - b) < eMinRootsSep then
      Exit;
    c:= (a + b) / 2;
    vc:= Variations(c);
    a0:= a;
    v0:= va;
    GetRootsInInterval(cl, a0, c, v0, vc);
    a0:= b;
    v0:= vb;
    GetRootsInInterval(cl, c, a0, vc, v0);
  end;
end;

procedure GetRoots(cl: Integer);
{ 'pol'  must have already its values set}
var
  rMax, aux: Double;
  i, va, vb: Integer;
begin
  try
    theRoots.number:= 0;
    if uSmallValue < 1.0E-30 then
      Exit;
    if (abs(Pol.coefs[0]) < uSmallValue) then
    {angle is equal to pi/2}begin {add the root directly with the obvious values}
      inc(theRoots.number);
      theRoots.values[1, 1]:= 0;
      theRoots.values[1, 2]:= 1;
      if cl = kExtrNormDevLines then
        theRoots.values[1, 3]:= MapR2R4.hh - 2 * MapR2R4.hb + MapR2R4.bb
      else if cl = kExtrLatDevLines then
        theRoots.values[1, 3]:= -MapR2R4.hc + MapR2R4.bc;
      for i:= 1 to pol.degree do
        pol.coefs[i - 1]:= pol.coefs[i];
      dec(pol.degree);
      if abs(pol.coefs[0]) < uSmallValue then
        ReduceDegree(pol);
    end;
    if pol.degree = 1 then begin
      Assert(abs(pol.coefs[0]) > uSmallValue);
      addRoot(cl, -pol.coefs[1] / pol.coefs[0]);
    end else if pol.degree = 2 then
    {compute the roots directly and save them}begin
      aux:= sqr(pol.coefs[1]) - 4 * pol.coefs[0] * pol.coefs[2];
      if aux > 0 then begin
        aux:= sqrt(aux);
        AddRoot(cl, (-pol.coefs[1] + aux) / (2 * pol.coefs[0]));
        AddRoot(cl, (-pol.coefs[1] - aux) / (2 * pol.coefs[0]));
      end else if aux = 0 then
        AddRoot(cl, -pol.coefs[1] / (2 * pol.coefs[0]));
    end else if pol.degree > 2 then {apply the Sturm sequence technique} begin
      stm.polynomials[0]:= pol; {we know already that 'pol.coefs[0] <> 0'}
      rMax:= 1; {will serve to bound the roots from above and from below}
      for i:= 1 to stm.polynomials[0].degree do begin
        stm.polynomials[0].coefs[i]:= stm.polynomials[0].coefs[i]
          /
          stm.polynomials[0].coefs[0];
        if abs(stm.polynomials[0].coefs[i]) + 1 > rMax then
          rMax:= abs(stm.polynomials[0].coefs[i]) + 1;
      end;
      uSmallValue:= uSmallValue / abs(stm.polynomials[0].coefs[0]);
      stm.polynomials[0].coefs[0]:= 1;
      MakeSturmSequence;
      va:= Variations(-rMax);
      vb:= Variations(rMax);
      GetRootsInInterval(cl, -rMax, rMax, va, vb);
    end;
  except on exception do
      {beep}; {eventually used for debugging}
  end;
end;

function InitialDirections(axis: Integer): Vector2;
{Called only for Axymptotic lines.
'axis' can have any value between 1 and two times the number of roots. This
 gives the correspondign oriented field direction. In fact, when 'axis' goes
 from 1 to, say, 10, '(axis + 1) div 2'  takes the following values: 1,1,2,2,
 3,3,4,4,5,5, and the orientation changes for instance from axis= 3 to axis=4,
 etc.}
begin
  result[1]:= theRoots.values[(axis + 1) div 2, 1];
  result[2]:= theRoots.values[(axis + 1) div 2, 2];
  if not odd(axis) then begin
    result[1]:= -result[1];
    result[2]:= -result[2];
  end;
end;

procedure TrigDerivative(var pp: RealPolynomial);
{It is assumed here that we associate to the polynomial pp a function in 't',
given by
     p[0] cos(t)^d + p[1] cos(t)^(d-1) sin(t) + ...+ p[k] sin(t)^k,
where for brevity we have put  p[i]:= pp.coefs[i], and k:= pp.degree.
This routine gives the same degree polynomial whose associated function
is the derivative of the given associated function with respect to 't'}
var
  j: integer;
begin
  pp.degree:= pol.degree;
  pp.coefs[0]:= -pol.coefs[1];
  pp.coefs[pp.degree]:= pol.coefs[pp.degree - 1];
  for j:= 1 to pp.degree - 1 do
    pp.coefs[j]:= -(j + 1) * pol.coefs[j + 1] + (pp.degree + 1 - j) *
      pol.coefs[j - 1];
end;

function EvalPolTrig(p: RealPolynomial; alpha: Double): Double;
{It is assumed here that we associate to the polynomial pp a function in 't',
 given by
     p[0] cos(t)^d + p[1] cos(t)^(d-1) sin(t) + ...+ p[k] sin(t)^k.
 This function gives the value of this function when 't=alpha'.}
var
  sinus, cosinus, vp: Double;
  i: Integer;
begin
  if p.degree < 0 then
    result:= 0
  else begin
    sinus:= sin(alpha);
    cosinus:= cos(alpha);
    result:= p.coefs[0];
    vp:= 1;
    if p.degree > 0 then
      for i:= 1 to p.degree do begin
        vp:= vp * cosinus;
        result:= result * sinus + p.coefs[i] * vp;
      end;
  end;
end;

function ClassifyRoot(cl: Integer; y: Vector2; sine, cosine: Double): Integer;
{Each root gives an extremal value of some function. This function is applied
 to the lines of extremal deviation, so that at each point there are exactly two,
 four or infinite directions. If that numnber is two, we number the roots as
 1 or 4 according to them giving the maximum or minimum of the function values;
 if that number is 4, then we give them their order number in the increasing
 sequence of their function values. Now, we are given a direction at 'y' by
 its sine and cosine. We compute the roots at 'y' and the value of the function
 at the direction (cosine,sine), and assign to that direction the classification
 number of the root whose function value is nearest to that of (cosine,sine)}
var
  sine2, cosine2, value, dif, curDif: Double;
  i: Integer;
begin
  {compute and sort roots at 'y'}
  theRoots.number:= 0;
  if ComputeCoefficientsExtrDevDirs(cl, y) then begin
    GetRoots(cl);
    if cl <> kStrongPrincipalLines then
      SortRoots;

    {compute function value at (cosine,sine)}
    sine2:= 2 * sine * cosine;
    cosine2:= sqr(cosine) - sqr(sine);
    with MapR2R4 do
      if cl = kExtrNormDevLines then
        value:= hh + bb * sqr(cosine2) + cc * sqr(sine2) +
          2 * hb * cosine2 + 2 * hc * sine2 + 2 * bc * sine2 * cosine2
      else if cl = kExtrLatDevLines then
        value:= -hb * sine2 + hc * cosine2 + (cc - bb) * cosine2 * sine2 +
          bc * (sqr(cosine2) - sqr(sine2))
      else
        value:= 0;
    result:= 0;
    if theRoots.number > 0 then begin
      dif:= 1.0e20;
      for i:= 1 to theRoots.number do begin
        curDif:= abs(value - theRoots.values[i, 3]);
        if curDif < dif then begin
          dif:= curDif;
          result:= i; {assign the number of the root with nearest function value}
        end
      end
    end;

    if (theRoots.number = 2) and (result = 2) then {the case of two roots. So
      the meaning of the classification numbers of 1 for the minimum and 4 for the
      maximum is preserved}
      result:= 4;
  end else
    result:= 0;
end;

function Direction(cl: Integer; var alpha: Double; var dir: Vector2; param:
  Vector2)
  : Byte;
{computes the direction at 'param' nearest to the given direction 'prevDir'.
 If a direction is found, result > 0, else result is equal to 1 or two according
 with the function derivative at that direction being negative or positive.}
const
  cShift = 0.001;
var
  f, fp, a, b, c, aux: Double;
  counter: integer;
  pp: RealPolynomial;
begin
  result:= 0;
  if ComputeCoefficientsExtrDevDirs(cl, param) then begin
    TrigDerivative(pp);
    {the derivative of the function with respect to the angle}
    counter:= 0;
    repeat {find by Newton-Cotes method the root nearest to 'alpha'}
      inc(counter);
      f:= EvalPolTrig(pol, alpha);
      fp:= EvalPolTrig(pp, alpha);
      if abs(f) < uSmallValue then
        break;
      if abs(fp) < abs(f) then
        Exit;
      alpha:= alpha - f / fp;
    until counter > 50;
    if counter > 50 then
      Exit; {root not found}
    result:= 1;
    sincos(alpha, si, co);
    if cl = kStrongPrincipalLines then begin
      dir[1]:= co;
      dir[2]:= si;
    end
    else begin
      aux:= sqrt(abs(MapR2R4.guu * MapR2R4.gvv - sqr(MapR2R4.guv)));
      if (aux > 0) and (MapR2R4.guu > 0) then begin
        a:= 1 / sqrt(MapR2R4.guu);
        aux:= a / aux;
        b:= -MapR2R4.guv * aux;
        c:= MapR2R4.guu * aux;
        dir[1]:= a * co + b * si;
        dir[2]:= c * si;
        if fp > 0 then {add one if derivative of function is positive}
          result:= result + 1;
      end else
        result:= 0;
    end;
  end
end;

function FoliationClass(cl: Integer; dir, param: Vector2): Byte;
{give two or one according with the derivative of the function that
 the the direction 'dir' makes extremal being positive or not.}
var
  pp: RealPolynomial;
  alpha, fp: Double;
begin
  Result:= 0;
  if ComputeCoefficientsExtrDevDirs(cl, param) then begin
    TrigDerivative(pp);
    alpha:= myArctan2(dir[2], dir[1]);
    fp:= EvalPolTrig(pp, alpha);
    result:= 1 + ord(fp > 0);
  end;
end;

function ComputeElem(cl: Integer; var alpha: Double; var error: Double; u:
  Vector2; nextStep: Double): byte;
var
  y: Vector2;
begin
  error:= 1.0E20;
  result:= Direction(cl, alpha, rk1, u);
  if result > 0 then begin
    y[1]:= u[1] + nextStep * rk1[1];
    y[2]:= u[2] + nextStep * rk1[2];
    result:= Direction(cl, alpha, rk2, y);
    if result > 0 then begin
      y[1]:= u[1] + 0.25 * nextStep * (rk1[1] + rk2[1]);
      y[2]:= u[2] + 0.25 * nextStep * (rk1[2] + rk2[2]);
      result:= Direction(cl, alpha, rk3, y);
      if result > 0 then begin
        error:= abs(nextStep * (2.0 * rk3[1] - rk1[1] -
          rk2[1]));
        error:= error + abs(nextStep * (2.0 * rk3[2] - rk1[2] -
          rk2[2]));
        error:= error + 1.3e-12;
      end;
    end;
  end;
end;

function GetInitialStep(cl: Integer; alpha: Double; u: Vector2): Double;
const
  tol: Double = 1.0E-4;
var
  err, coef: Double;
  v: Vector2;
begin
  result:= kTheError;
  if (Direction(cl, alpha, v, u) > 0) then begin
    result:= param4.radius2D / hypot(v[1], v[2]);
    ComputeElem(cl, alpha, err, u, result);
    coef:= 0.8 * exp(ln(tol / err) / 3.0);
    result:= result * Max(0.1, Min(5, coef));
  end;
end;

function RunKut(cl: Integer; var step, jump, guessStep, alpha: Double;
  var x: Vector2; foliations: Byte): boolean;
{For this and the following related routines see comments more detailed in the
 unit  "FRunKut.pas" of the project Superficies by same author, available at
 the same site}
const
  tol: Double = 1.0E-4;
var
  err, coef: Double;
  advance: Vector2;
  goOn: boolean;
begin
  repeat
    goOn:= (ComputeElem(cl, alpha, err, x, guessStep) and foliations) <> 0;
    if goOn then begin
      coef:= 0.8 * exp(ln(tol / err) / 3.0);
      if coef < 0.5 then
        coef:= 0.5;
      if 2.0 < coef then
        coef:= 2.0;
      step:= guessStep;
      guessStep:= coef * step;
    end
  until (not goOn) or (err < tol);
  if goOn then begin
    advance[1]:= step * (rk1[1] + rk2[1] + 4.0 * rk3[1]) / 6.0;
    x[1]:= x[1] + advance[1];
    advance[2]:= step * (rk1[2] + rk2[2] + 4.0 * rk3[2]) / 6.0;
    x[2]:= x[2] + advance[2];
    jump:= abs(advance[1]) + abs(advance[2]);
  end;
  RunKut:= goOn;
end;

function Interpolate(x, rk1, rk2, rk3: Vector2; h, theta: Double): Vector2;
const
  threeOverTwo: Double = 1.5;
  twoOverThree: Double = 0.666666666666666666;
  fourOverThree: Double = 1.33333333333333333;
var
  theta2, b1, b2, b3: Double;
begin
  theta2:= sqr(theta);
  b1:= theta + (-threeOverTwo + twoOverThree * theta) * theta2;
  b2:= (-0.5 + twoOverThree * theta) * theta2;
  b3:= (2.0 - fourOverThree * theta) * theta2;
  Interpolate[1]:= x[1] + h * (b1 * rk1[1] + b2 * rk2[1] + b3 * rk3[1]);
  Interpolate[2]:= x[2] + h * (b1 * rk1[2] + b2 * rk2[2] + b3 * rk3[2]);
end;

procedure GetHigherDegreeDirections(cl: Integer; var dirs: Roots; param:
  Vector2);
{return in 'dirs' the directions that make extremal the corresponding function}
begin
  theRoots.Number:= 0;
  if ComputeCoefficientsExtrDevDirs(cl, param) then begin
    GetRoots(cl);
    if (cl <> kStrongPrincipalLines) then
      SortRoots;
    dirs:= theRoots;
  end;
end;

procedure DrawHigherDegreeLines(cl: Integer; can: TCanvas;
  start: Vector2; foliations: Byte);
{compute and draw the curves of extremal deviation with initial contidion 'start'.
 See "FRunKut.pas" of the project Superficies}
var
  initialRoots: Roots;
  noFaraway, goNext: boolean;
  x, y, xInter, xPrevious: Vector2;
  inside, step, guessStep, jump, theta, alpha: Double;
  pieces, i, axis, counter, colorNo, newColor: integer;
begin
  try
    if ComputeCoefficientsExtrDevDirs(cl, start) then begin
      GetRoots(cl);
      if cl <> kStrongPrincipalLines then
        SortRoots;
      initialRoots:= theRoots;
      if theRoots.number > 0 then begin
        if cl = kStrongPrincipalLines then
          colorNo:= 5;
        for axis:= 1 to 2 * initialRoots.number do begin
          counter:= 1;
          for i:= 1 to 2 do
            rk3[i]:= initialRoots.values[(axis + 1) div 2, i];
          if not odd(axis) then for i:= 1 to 2 do
              rk3[i]:= -rk3[i];
          alpha:= myArctan2(rk3[2], rk3[1]); {the angle of 'rk3'}
          {draw only the curves corresponding to the allowed foliations}
          if cl <> kStrongPrincipalLines then
            goNext:= (foliations and FoliationClass(cl, rk3, start)) = 0
          else
            goNext:= False;
          if goNext then
            Continue;
          step:= GetInitialStep(cl, alpha, start);
          guessStep:= step;
          x:= start;

          {select color for the next segment}
          if (cl <> kStrongPrincipalLines) then begin
            if (initialRoots.number = 2) then
              case axis of {truly, only options 1 and 4 will appear}
                1, 2: colorNo:= 1;
                3, 4: colorNo:= 4;
              else
                colorNo:= 4;
              end
            else
              colorNo:= (axis + 1) div 2;
          end;
          param4.AddAndDrawLineSegmentExtremals(x, x, 1, colorNo);
          repeat
            xPrevious:= x;
            repeat
              counter:= counter + 1;
              if ((counter mod 30) = 0) then begin
                Application.ProcessMessages;
                {give user the chance of click the stop button}
                if param4.stop then begin
                  {stop buttom was pressed}
                  param4.stop:= False;
                  Exit;
                end
              end;
              x:= xPrevious;
              noFaraway:= RunKut(cl, step, jump, guessStep, alpha, x,
                foliations);
              inside:= min(param4.halfEdges2D[1] - abs(x[1] -
                param4.center2D[1]),
                param4.halfEdges2D[2] - abs(x[2] - param4.center2D[2]));
              if not (noFaraway and (inside > 0)) then begin
                guessStep:= 0.5 * step;
                step:= guessStep;
                jump:= abs(step) * hypot(rk3[1], rk3[2]);
              end;
              if (jump < 0.02 * param4.radius2D) then
                goNext:= True;
            until (noFaraway and (inside > 0)) or goNext;
            if goNext then
              break;
            pieces:= round(1 + jump / param4.maxSegLen);
            theta:= 1.0 / pieces;
            xInter:= xPrevious;
            for i:= 1 to pieces do begin
              y:= Interpolate(xPrevious, rk1, rk2, rk3, step, i * theta);
              {draw the segment (xInter,y) and store the point 'y'}
              param4.AddAndDrawLineSegmentExtremals(xInter, y, 0, colorNo);
              xInter:= y;
            end;
            {select color, draw and store the point 'y'}
            if (cl <> kStrongPrincipalLines) then begin
              newColor:= ClassifyRoot(cl, y, si, co);
              if newColor <> colorNo then begin
                colorNo:= newColor;
                param4.AddAndDrawLineSegmentExtremals(y, y, 1, colorNo);
              end;
            end;
          until (not (inside > 0)) or (jump < 0.01 * param4.radius2D) or
            (counter > 3000);
          if goNext then
            continue;
        end;
      end;
    end;
  except on exception do
      {beep}; {eventually used for debugging}
  end;
end;

end.

